// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "tree_controller.h"

#include "kit.h"

#if 1
#include "uml/uml.h"
#include "uml/foundation/core/class.h"
#include "uml/foundation/core/generalization.h"
#include "uml/foundation/core/structural_feature.h"
#include "uml/foundation/core/attribute.h"
#include "uml/foundation/core/operation.h"
#include "uml/foundation/core/association.h"
#include "uml/foundation/core/parameter.h"
#include "uml/foundation/data_types/name.h"
#include "uml/model_management/package.h"
#include "uml/model_management/model.h"
#include "uml/diagram/diagram.h"
#endif

#include "reflective/changes.h"
#include "reflective/context.h"
#include "reflective/attribute.h"
#include "reflective/iterator.h"

#include "toolkit/mouse_event.h"
#include "toolkit/factory.h"
#include "toolkit/tree.h"
#include "toolkit/menu.h"

#include "drawing/factory.h"

#include "foundation/utility.h"
#include "foundation/undo_manager.h"
#include "foundation/macros.h"
#include "foundation/command_callback.h"

#include "python/python.h"
#include "python/toolkit/tree.h"
#include "python/reflective/context.h"


#include <sstream>
using namespace std;

// the stuff in "#if 1 ... #endif" enclosements has to go... I have
// left it in for now because the functionality that it provides
// is useful for testing. But it isn't built around the reflective interface.



#ifdef USING_PYTHON

//================================================================================

TreeController::TreeController(const char * package, Toolkit::Tree * tree) : Object() {
  Python::Module module = Python::Module("forest");
  PyObject * args = Py_BuildValue("(sO)", package, Python::tree_reference_new(tree));
  mObject = Python::Callable(module, "factory").call(args);
  Py_DECREF(args);
}

TreeController::~TreeController() {
  Py_XDECREF(mObject);
}

void
TreeController::set_context(REF::Context * context) {
  PyObject * ref_context = Python::context_reference_new(context);
  PyObject * result = PyObject_CallMethod(mObject, "set_context", "(O)", ref_context);
  Py_DECREF(ref_context);
  if (result == NULL) {
    ERROR("Error returned by 'set_context'");
    if (PyErr_Occurred())
      PyErr_Print();
  }
  else {
    Py_DECREF(result);
  }
}

//================================================================================



#else
const TreeController::Rule TreeController::RULES[] = {
    {
	"Namespace",
	"owned_element",
	"ModelElement",
	true
    },

    {
	"Classifier",
	"feature",
	"Feature",
	true
    },

    {
	"BehavioralFeature",
	"parameter",
	"Parameter",
	true
    },

    {
	"ModelElement",
	"tagged_value",
	"TaggedValue",
	true
    },

    {
	"Stereotype",
	"stereotype_constraint",
	"Constraint",
	true
    },

    {
	"Stereotype",
	"defined_tag",
	"TagDefinition",
	true
    },

    {
	"Association",
	"connection",
	"AssociationEnd",
	false
    },
};

const int TreeController::NUM_RULES = ( sizeof(TreeController::RULES) /
					sizeof(TreeController::RULES[0]) );

const TreeController::Icon TreeController::ICONS[] = {
    {
	"Package",
	"palette/package.png"
    },

    {
	"Model",
	"palette/model.png"
    },

    {
	"Class",
	"palette/class.png"
    },

    {
	"Generalization",
	"palette/generalization.png"
    },

    {
	"Dependency",
	"palette/dependency.png"
    },

    {
	"Operation",
	"palette/operation-public.png"
    },

    {
	"Attribute",
	"palette/attribute-public.png"
    },

    {
	"Parameter",
	"palette/parameter-in.png"
    },

    {
	"Diagram",
	"palette/class-diagram.png"
    },

    {
	"Association",
	"palette/aggregation.png"
    },

    {
	"AssociationEnd",
	"palette/association_end.png"
    },
};

const int TreeController::NUM_ICONS = ( sizeof(TreeController::ICONS) /
					sizeof(TreeController::ICONS[0]) );


TreeController::TreeController(Toolkit::Tree * tree) : Object(),
    mTree(tree),
    mContext(0) {
    ASSERT(mTree);

    UndoManager * undo_manager = UndoManager::instance();
    undo_manager->add_change_listener(static_cast<ChangeListener *>(this));

    mTree->set_maintainer(static_cast<Toolkit::TreeMaintainer *>(this));
    mTree->add(static_cast<Toolkit::MouseListener *>(this));
}

TreeController::~TreeController() {

    mTree->remove(static_cast<Toolkit::MouseListener *>(this));
    mTree->set_maintainer(static_cast<Toolkit::TreeMaintainer *>(0));

    UndoManager * undo_manager = UndoManager::instance();
    undo_manager->remove_change_listener(static_cast<ChangeListener *>(this));
}

void
TreeController::set_context(REF::Context * context) {

    if(mContext != context) {
	if(mContext) {
	    mTree->clear();
	}

	mContext = context;

	if(mContext) {
	    REF::Class * root = mContext->ref_get_root();

	    ASSERT(root);

	    if(root) {
		recursive_build(0, root, 0, true);
	    }
	}
    }
}

//


void
TreeController::changes_occurred(const list<ChangeEvent *> & changes) throw() {
    if(!mContext) {
	return;
    }

    for(list<ChangeEvent *>::const_iterator iter = changes.begin(); iter != changes.end(); ++iter) {
	ChangeEvent * change = *iter;

	if(typeid(*change) == typeid(REF::ChangeAttributeChanged)) {
	    //
	    // Attribute changed
	    //

	    REF::ChangeAttributeChanged * ac = known_cast<REF::ChangeAttributeChanged *>(change);
	    REF::Class * object = ac->get_object();
	    REF::Attribute * old_value = ac->get_old_value();
	    REF::Attribute * new_value = ac->get_new_value();

	    ASSERT(object);
	    ASSERT(old_value || new_value);

	    if(object->ref_is_a_type_of("ModelElement") && ac->get_attribute_name() == "name") {
		if(mTree->contains(object)) {
		    string value;

		    if(new_value) {
			new_value->ref_get_value(value);
		    }
		    else {
			value = "<unnamed>";
		    }

		    mTree->set_label(object, value);
		}
	    }
	}
	else if(typeid(*change) == typeid(REF::ChangeRoleChanged)) {
	    //
	    // Role changed
	    //

	    REF::ChangeRoleChanged * crc = known_cast<REF::ChangeRoleChanged *>(change);
	    REF::Class * object = crc->get_object();
	    REF::Class * old_role = crc->get_old_value();
	    REF::Class * new_role = crc->get_new_value();

	    ASSERT(object);
	    ASSERT(old_role || new_role);
	}
	else if(typeid(*change) == typeid(REF::ChangeRoleAdded)) {
	    //
	    // Role added
	    //

	    REF::ChangeRoleAdded * cra = known_cast<REF::ChangeRoleAdded *>(change);
	    REF::Class * object = cra->get_object();
	    REF::Class * role = cra->get_value();

	    ASSERT(object);
	    ASSERT(role);

	    for(int i = 0; i < NUM_RULES; ++i) {
		if(object->ref_is_a_type_of(RULES[i].type) &&
		   cra->get_role_name() == RULES[i].role_name) {
		    recursive_build(object, role, 0, RULES[i].expand);
		}
	    }
	}
	else if(typeid(*change) == typeid(REF::ChangeRoleRemoved)) {
	    //
	    // Role removed
	    //

	    REF::ChangeRoleRemoved * crr = known_cast<REF::ChangeRoleRemoved *>(change);
	    REF::Class * object = crr->get_object();
	    REF::Class * role = crr->get_value();

	    ASSERT(object);
	    ASSERT(role);

	    for(int i = 0; i < NUM_RULES; ++i) {
		if(object->ref_is_a_type_of(RULES[i].type) &&
		   crr->get_role_name() == RULES[i].role_name) {
		    mTree->remove(role);
		}
	    }
	}
	else {
	    WARNING("Unknown change type");
	}
    }
}

//

bool
TreeController::can_relocate(Toolkit::Tree * tree,
			     Object * node, Object * old_parent, Object * new_parent) const {
    REF::Class * klass = known_cast<REF::Class *>(node);
    REF::Class * old_p = known_cast<REF::Class *>(old_parent);
    REF::Class * new_p = known_cast<REF::Class *>(new_parent);

    NYI();

    return true;
}

void
TreeController::relocated(Toolkit::Tree * tree, Object * node, Object * new_parent) {
    REF::Class * klass = known_cast<REF::Class *>(node);
    REF::Class * new_p = known_cast<REF::Class *>(new_parent);

    NYI();
}

void
TreeController::renamed(Toolkit::Tree * tree, Object * node, const string & n_name) {
    REF::Class * klass = known_cast<REF::Class *>(node);

    const REF::Attribute * old_name = klass->ref_get_attribute("name");
    string o_name;

    if(old_name) {
	old_name->ref_get_value(o_name);
    }

    REF::Attribute * new_name = klass->ref_create_attribute("name");
    new_name->ref_set_value(n_name);

    if(o_name != n_name) {
	begin_transaction();

	klass->ref_set_attribute("name", new_name);

	end_transaction();
    }

    new_name->unref();
}

void
TreeController::selection(const set<Object *> & nodes) {
    NYI();
}

//

bool
TreeController::mouse_entered(const Toolkit::MouseEvent * mouse_event) {
    return false;
}

bool
TreeController::mouse_exited(const Toolkit::MouseEvent * mouse_event) {
    return false;
}

bool
TreeController::mouse_pressed(const Toolkit::MouseEvent * mouse_event) {
    //cerr << "mouse_pressed" << endl;

#if 0
    if(mouse_event->get_button() == 1) {
	if(mouse_event->get_click_count() == 2) {
	    // double-click

	    // if it is a diagram, then we edit it,
	    // for anything else, show the properties of it (to be like rose)
	    // or seek to it in the document

	    Object * object = mTree->point_to_node(mouse_event->get_point());

	    if(object) {
		REF::Class * klass = known_cast<REF::Class *>(object);

		if(klass->ref_get_type() == "Diagram") {
		    //cerr << "Double clicked diagram!!!" << endl;

		    Kit * kit = Kit::instance();
		    kit->set_diagram(known_cast<UML::Diagram *>(klass));
		}
	    }
	}
    }
    if(mouse_event->get_button() == 3) {
	//
	// *** this code is bad because it doesn't operate at the reflective level
	//

	Object * object = mTree->point_to_node(mouse_event->get_point());

	if(object) {
	    UML::Element * element = known_cast<UML::Element *>(object);

	    Toolkit::Factory * factory = Toolkit::Factory::instance();

	    Toolkit::Menu * menu = factory->create_menu();

	    menu->add_push(command_callback(this, element, &TreeController::cut), "Cut");
	    menu->add_push(0, "Copy");
	    menu->add_push(0, "Paste");

	    if(typeid(*element) == typeid(UML::Class)) {
		UML::Class * klass = known_cast<UML::Class *>(element);

		menu->add_push(command_callback(this, klass, &TreeController::add_sub_class),
			       "New Sub-Class...");

		menu->add_push(command_callback(this, klass, &TreeController::add_operation),
			       "New Operation...");

		menu->add_push(command_callback(this, klass, &TreeController::add_attribute),
			       "New Attribute...");
	    }
	    else if(typeid(*element) == typeid(UML::Operation)) {
		UML::Operation * operation = known_cast<UML::Operation *>(element);

		menu->add_push(command_callback(this, operation, &TreeController::add_parameter),
			       "New Parameter...");
	    }
	    else if(typeid(*element) == typeid(UML::Package)) {
		UML::Package * package = known_cast<UML::Package *>(element);

		menu->add_push(command_callback(this, package, &TreeController::add_class),
			       "New Class");

		menu->add_push(command_callback(this, package, &TreeController::add_sub_package),
			       "New Sub Package");

		menu->add_push(command_callback(this, package, &TreeController::add_class_diagram),
			       "New Class Diagram");

		menu->add_push(command_callback(this, package, &TreeController::add_use_case_diagram),
			       "New Use Case Diagram");
	    }
	    else if(typeid(*element) == typeid(UML::Diagram)) {
		UML::Diagram * diagram = known_cast<UML::Diagram *>(element);

		menu->add_push(command_callback(this, diagram, &TreeController::edit_diagram),
			       "Edit Diagram");
	    }

#if 0
	    menu->add_push(0, "Properties...");
#endif

	    menu->popup(mouse_event);
	}
    }
#endif

    return false;
}

bool
TreeController::mouse_released(const Toolkit::MouseEvent * mouse_event) {
    return false;
}

bool
TreeController::mouse_moved(const Toolkit::MouseEvent * mouse_event) {
    return false;
}

bool
TreeController::mouse_dragged(const Toolkit::MouseEvent * mouse_event) {
    return false;
}

//

#if 0
void
TreeController::cut(UML::Element * element) {
    ASSERT(element);

    //cerr << "Deleting element" << endl;
    element->ref_delete();
    //cerr << "Deleted" << endl;
}

void
TreeController::add_operation(UML::Class * klass) {
    ASSERT(klass);

    begin_transaction();

    UML::Operation * op = new UML::Operation();

    UML::Name * name = new UML::Name();
    name->ref_set_value("new operation");
    op->set_name(name);
    name->unref();

    klass->add_feature(op);

    op->unref();

    end_transaction();
}

void
TreeController::add_attribute(UML::Class * klass) {
    ASSERT(klass);

    begin_transaction();

    UML::Attribute * at = new UML::Attribute();

    UML::Name * name = new UML::Name();
    name->ref_set_value("new attribute");
    at->set_name(name);
    name->unref();

    klass->add_feature(at);

    at->unref();

    end_transaction();
}

void
TreeController::add_parameter(UML::Operation * operation) {
    ASSERT(operation);

    begin_transaction();

    UML::Parameter * p = new UML::Parameter();

    UML::Name * name = new UML::Name();
    name->ref_set_value("new parameter");
    p->set_name(name);
    name->unref();

    operation->add_parameter(p);

    p->unref();

    end_transaction();
}

void
TreeController::add_class(UML::Package * package) {
    ASSERT(package);

    begin_transaction();

    UML::Class * kl = new UML::Class();

    UML::Name * name = new UML::Name();
    name->ref_set_value("New Class");
    kl->set_name(name);
    name->unref();

    package->add_owned_element(kl);

    kl->unref();

    end_transaction();
}

void
TreeController::add_sub_class(UML::Class * klass) {
    ASSERT(klass);

    begin_transaction();

    UML::Class * kl = new UML::Class();

    UML::Name * name = new UML::Name();
    name->ref_set_value("New Class");
    kl->set_name(name);
    name->unref();

    klass->add_owned_element(kl);
    kl->unref();

    end_transaction();
}

void
TreeController::add_sub_package(UML::Package * package) {
    ASSERT(package);

    begin_transaction();

    UML::Package * pack = new UML::Package();

    UML::Name * name = new UML::Name();
    name->ref_set_value("New Package");
    pack->set_name(name);
    name->unref();

    package->add_owned_element(pack);
    pack->unref();

    end_transaction();
}

void
TreeController::add_class_diagram(UML::Package * package) {
    ASSERT(package);

    begin_transaction();

    UML::Diagram * diagram = new UML::Diagram();

    UML::Name * name = new UML::Name();
    name->ref_set_value("New Class Diagram");
    diagram->set_name(name);
    name->unref();

    package->add_owned_element(diagram);
    diagram->unref();

    end_transaction();
}

void
TreeController::add_use_case_diagram(UML::Package * package) {
    ASSERT(package);

    begin_transaction();

    UML::Diagram * diagram = new UML::Diagram();

    UML::Name * name = new UML::Name();
    name->ref_set_value("New Use Case Diagram");
    diagram->set_name(name);
    name->unref();

    package->add_owned_element(diagram);
    diagram->unref();

    end_transaction();
}

//

void
TreeController::edit_diagram(UML::Diagram * diagram) {
    NYI();
}
#endif

//

void
TreeController::recursive_build(REF::Class * parent, REF::Class * child,
                                REF::Class * pos, bool expand) {
    if(mTree->contains(child)) {
	// not sure if this is the best workaround...

	return;
    }

    ASSERT(child);

    Drawing::Factory * factory = Drawing::Factory::instance();

    //

    const REF::Attribute * name = child->ref_get_attribute("name");

    Drawing::Image * image = 0;

    for(int i = 0; i < NUM_ICONS; ++i) {
	if(ICONS[i].type == child->ref_get_type()) {
	    image = factory->create_image(ICONS[i].icon);
	    break;
	}
    }

    if(!image) {
	WARNING("No image for child of type: " << child->ref_get_type() << ", using default");
	image = factory->create_image("???");
    }

    string value;

    if(name) {
	name->ref_get_value(value);
    }
    else {
	value = "<unnamed>";
    }

    mTree->add(child, image, value, expand, parent);

    //

    for(int i = 0; i < NUM_RULES; ++i) {
	if(child->ref_is_a_type_of(RULES[i].type)) {
	    REF::Iterator * iterator = child->ref_get_role_iterator(RULES[i].role_name);

	    while(iterator->is_valid()) {
		recursive_build(child, iterator->get(), 0, RULES[i].expand);
		iterator->increment();
	    }

	    iterator->unref();
	}

    }
}
#endif
