// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "cpp_stream.h"

#include "foundation/macros.h"

#include <iterator>
#include <algorithm>
#include <exception>
#include <iostream>
using namespace std;

namespace Common {

    CppStream::CppStream(ostream & ost) : mOs(ost) {
	ASSERT(ost);

	mIndentLevel = 0;
	mIndentWidth = DEFAULT_INDENT_WIDTH;
    }

    CppStream::~CppStream() {
	if(!uncaught_exception()) {
	    if(mIndentLevel != 0) {
		error("Warning indent level not zero");
	    }
	    else if(!mBuffer.empty()) {
		error("Warning buffer not empty");
	    }
	}
    }

    void
    CppStream::set_indent_width(int new_value) {
	ASSERT(new_value >= 0);
	mIndentWidth = new_value;
    }

    int
    CppStream::get_indent_width() const {
	return mIndentWidth;
    }

    void
    CppStream::write(const string & s) {
	// take a copy of the buffer to work on
	string buffer = mBuffer + s;

	string::size_type new_line_offset = buffer.find('\n');

	while(new_line_offset != string::npos) {
	    string::size_type white_space_offset = buffer.find_first_not_of(" \t");
	    string::size_type substring_size = new_line_offset + 1 - white_space_offset;

	    const string line = buffer.substr(white_space_offset, substring_size);

	    if(line[0] == '}') {
		if(mStack.empty()) {
		    error("Excess number of \'}\'s");
		}
		else if(mStack.top().first != BK_CURLY) {
		    error("Bad closing bracket, expected \'}\'");
		}
		else {
		    mStack.pop();
		    --mIndentLevel;

		    ASSERT(mIndentLevel >= 0);
		}
	    }

	    int columns;

	    if(line[0] == '#') {
		columns = 0;
	    }
	    else if(mStack.empty() || mStack.top().first != BK_ROUND) {
		columns = mIndentLevel * mIndentWidth;

		if(line.substr(0, 7) == "public:" ||
		   line.substr(0, 10) == "protected:" ||
		   line.substr(0, 8) == "private:") {
		    if(mStack.empty() || mStack.top().first != BK_CURLY) {
			error("Line \"" + line + "\" can only occur inside \'{\'");
		    }
		    else {
			columns -= mIndentWidth;
		    }
		}
	    }
	    else {
		columns = mStack.top().second + 1;
	    }

	    int counter = 0;

	    for(string::const_iterator iter = line.begin(); iter != line.end(); ++iter) {
		char c = *iter;

		if(c == '{') {
		    if(!mStack.empty() && mStack.top().first != BK_CURLY) {
			ERROR("Bracket \'{\' cannot occur here");
		    }

		    mStack.push(make_pair(BK_CURLY, 0));
		    ++mIndentLevel;
		}
		else if(c == '(') {
		    mStack.push(make_pair(BK_ROUND, columns + counter));
		}
		else if(c == '[') {
		    mStack.push(make_pair(BK_SQUARE, 0));
		}
		else if(iter != line.begin() && c == '}') {
		    // this case is handled for the first character above

		    if(mStack.empty() || mStack.top().first != BK_CURLY) {
			ERROR("Unexpected \'}\'");
		    }
		    else {
			mStack.pop();
			--mIndentLevel;
			ASSERT(mIndentLevel >= 0);
		    }
		}
		else if(c == ')') {
		    if(mStack.empty() || mStack.top().first != BK_ROUND) {
			ERROR("Unexpected \')\'");
		    }
		    else {
			mStack.pop();
		    }
		}
		else if(c == ']') {
		    if(mStack.empty() || mStack.top().first != BK_SQUARE) {
			ERROR("Unexpected \']\'");
		    }
		    else {
			mStack.pop();
		    }
		}

		++counter;
	    }

	    // write to the stream

	    fill_n(ostream_iterator<char>(mOs), columns / 8, '\t');
	    fill_n(ostream_iterator<char>(mOs), columns % 8, ' ');
	    mOs << line;

	    // the buffer becomes the remainder

	    buffer = buffer.substr(new_line_offset + 1);

	    new_line_offset = buffer.find('\n');
	}

	// the buffer becomes what is left after processing
	mBuffer = buffer;
    }

    void
    CppStream::checkpoint(const string & m) {
	mLastCheckpoint = m;
    }

    void
    CppStream::error(const string & m) const {
	cerr << "CppStream error: " << m << endl;

	if(!mLastCheckpoint.empty()) {
	    cerr << "Last checkpoint: " << mLastCheckpoint << endl;
	}
	else {
	    cerr << "No checkpoint available" << endl;
	}
    }

} // namespace Common
