// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "editor.h"

#include "foundation/macros.h"
#include "foundation/functional.h"

#include "drawing/rectangle.h"

#include "toolkit/canvas.h"
#include "toolkit/cursor.h"
#include "toolkit/mouse_event.h"
#include "toolkit/canvas_event.h"
#include "toolkit/key_event.h"

#include "editor_change_event.h"
#include "editor_change_listener.h"
#include "mode_manager.h"
#include "mode_change_event.h"
#include "mode.h"
#include "selection_manager.h"
#include "selection.h"
#include "layer_manager.h"
#include "layer_perspective.h"
#include "layer_grid.h"         // FIXME
#include "fig.h"

#include <algorithm>
using namespace std;

namespace Diagram {

    const int Editor::HIT_SIZE = 8;

    Editor * Editor::mCurrent = 0;

    list<EditorChangeListener *> Editor::mEditorChangeListeners;

    void
    Editor::add(EditorChangeListener * editor_change_listener) {
	mEditorChangeListeners.push_back(editor_change_listener);
    }

    void
    Editor::remove(EditorChangeListener * editor_change_listener) {
	mEditorChangeListeners.remove(editor_change_listener);
    }

    Editor::Editor(Toolkit::Canvas * canvas) : Object(),
	mActive(false),
	mCanvas(canvas),
	mCurrentFig(0),
	mCurrentSelection(0) {
	ASSERT(canvas);

	mModeManager = new ModeManager();
	mModeManager->add(static_cast<ModeChangeListener *>(this));

	mLayerManager = new LayerManager(this);

	// FIXME editor shouldn't need to know about a concrete subclass
	// of LayerGuide
	mLayerGuide = new LayerGrid();
	mLayerGuide->set_hidden(true);
	mLayerManager->add(mLayerGuide);

	mCanvas->ref();
	mCanvas->add(static_cast<Toolkit::CanvasListener *>(this));
	mCanvas->add(static_cast<Toolkit::MouseListener *>(this));
	mCanvas->add(static_cast<Toolkit::KeyListener *>(this));


	// if there is no current editor then we make ourself the
	// current editor

	if(!current()) {
	    set_current(this);
	}
    }

    Editor::~Editor() {
	// if we were the current editor then clear the current editor

	if(current() == this) {
	    set_current(0);
	}

	mCanvas->remove(static_cast<Toolkit::CanvasListener *>(this));
	mCanvas->remove(static_cast<Toolkit::MouseListener *>(this));
	mCanvas->remove(static_cast<Toolkit::KeyListener *>(this));
	mCanvas->unref();

	mLayerGuide->unref();
	mLayerManager->unref();

	mModeManager->remove(static_cast<ModeChangeListener *>(this));
	mModeManager->unref();
    }

    void
    Editor::set_active(bool active) {
	if(mActive != active) {
	    mActive = active;

	    mLayerGuide->set_hidden(!active);
	}
    }

    void
    Editor::add(Layer * layer) {
	RETURN_IF_FAIL(layer);

	mLayerManager->add(layer);
    }

    void
    Editor::remove(Layer * layer) {
	RETURN_IF_FAIL(layer);

	// clear the selection manager in case it contains selections from
	// this layer

	SelectionManager * selection_manager = SelectionManager::instance();
	selection_manager->damage();
	selection_manager->clear();

	mLayerManager->remove(layer);
    }

    void
    Editor::add(Fig * fig) {
	RETURN_IF_FAIL(fig);
	RETURN_IF_FAIL(mActive);

	mLayerManager->add(fig);
    }

    void
    Editor::remove(Fig * fig) {
	RETURN_IF_FAIL(fig);
	RETURN_IF_FAIL(mActive);

	mLayerManager->remove(fig);
    }

    void
    Editor::push(Mode * mode) {
	RETURN_IF_FAIL(mode);

	mModeManager->push(mode);
    }

    void
    Editor::pop() {
	mModeManager->pop();
    }

    Fig *
    Editor::hit(const Drawing::Rectangle & rect) const {
	if(mLayerManager) {
	    return mLayerManager->hit(rect);
	}
	else {
	    return 0;
	}
    }

    Fig *
    Editor::hit(const Drawing::Point & point) const {
	Drawing::Rectangle hit_rect(point - Drawing::Vector(HIT_SIZE / 2, HIT_SIZE / 2),
				    Drawing::Vector(HIT_SIZE + 1, HIT_SIZE + 1));

	return mLayerManager->hit(hit_rect);
    }

    void
    Editor::damage(const Mode * mode) {
	ASSERT(mode);

	damage(mode->get_bounds());
    }

    void
    Editor::damage(const Selection * selection) {
	ASSERT(selection);

	damage(selection->get_bounds());
    }

    void
    Editor::damage(const Fig * fig) {
	ASSERT(fig);

	damage(fig->get_bounds());
    }

    void
    Editor::damage(const Drawing::Rectangle & rect) {
	if(mCanvas) {
	    mCanvas->damage(rect);
	}
    }

    Graph::Model *
    Editor::get_graph_model() {
	if(mLayerManager) {
	    Layer * layer = mLayerManager->get_active_layer();

	    if(layer) {
		LayerPerspective * layer_perspective = dynamic_cast<LayerPerspective *>(layer);

		if(layer_perspective) {
		    return layer_perspective->get_graph_model();
		}
	    }
	}

	return 0;
    }

    Renderer *
    Editor::get_renderer() {
	if(mLayerManager) {
	    Layer * layer = mLayerManager->get_active_layer();

	    if(layer) {
		LayerPerspective * layer_perspective = dynamic_cast<LayerPerspective *>(layer);

		if(layer_perspective) {
		    return layer_perspective->get_renderer();
		}
	    }
	}

	return 0;
    }

    void
    Editor::snap(Drawing::Point & snapped_point) {
	if(mLayerGuide) {
	    mLayerGuide->snap(snapped_point);
	}
    }

    // invoked when the mouse enters the editor
    bool
    Editor::mouse_entered(const Toolkit::MouseEvent * mouse_event) {
	if(mActive) {
	    set_current(this);

	    TRACE(10, "enter: " << mouse_event->get_point());

	    set_under_mouse(mouse_event);

	    return mModeManager->mouse_entered(mouse_event);
	}
	else {
	    return false;
	}
    }

    bool
    Editor::mouse_exited(const Toolkit::MouseEvent * mouse_event) {
	if(mActive) {
	    // note: we don't say "set_current(0)"

	    TRACE(10, "leave: " << mouse_event->get_point());

	    set_under_mouse(mouse_event);

	    return mModeManager->mouse_exited(mouse_event);
	}
	else {
	    return false;
	}
    }

    bool
    Editor::mouse_pressed(const Toolkit::MouseEvent * mouse_event) {
	if(mActive) {
	    if(mCurrentFig) {
		MouseListener * mouse_listener = dynamic_cast<MouseListener *>(mCurrentFig);

		if(mouse_listener) {
		    mouse_listener->mouse_pressed(mouse_event);
		}
	    }

	    SelectionManager * selection_manager = SelectionManager::instance();

	    return (selection_manager->mouse_pressed(mouse_event) ||
		    mModeManager->mouse_pressed(mouse_event));
	}
	else {
	    return false;
	}
    }

    bool
    Editor::mouse_released(const Toolkit::MouseEvent * mouse_event) {
	if(mActive) {
	    if(mCurrentFig) {
		MouseListener * mouse_listener = dynamic_cast<MouseListener *>(mCurrentFig);

		if(mouse_listener) {
		    mouse_listener->mouse_released(mouse_event);
		}
	    }

	    SelectionManager * selection_manager = SelectionManager::instance();

	    return (selection_manager->mouse_released(mouse_event) ||
		    mModeManager->mouse_released(mouse_event));
	}
	else {
	    return false;
	}
    }

    bool
    Editor::mouse_moved(const Toolkit::MouseEvent * mouse_event) {
	if(mActive) {
	    set_under_mouse(mouse_event);

	    if(mCurrentFig) {
		// show tool tip
	    }

	    SelectionManager * selection_manager = SelectionManager::instance();

	    return (selection_manager->mouse_moved(mouse_event) ||
		    mModeManager->mouse_moved(mouse_event));
	}
	else {
	    return false;
	}
    }

    bool
    Editor::mouse_dragged(const Toolkit::MouseEvent * mouse_event) {
	if(mActive) {
	    set_under_mouse(mouse_event);

	    SelectionManager * selection_manager = SelectionManager::instance();

	    return (selection_manager->mouse_dragged(mouse_event) ||
		    mModeManager->mouse_dragged(mouse_event));
	}
	else {
	    return false;
	}
    }

    bool
    Editor::key_pressed(const Toolkit::KeyEvent * key_event) {
	if(mActive) {
	    SelectionManager * selection_manager = SelectionManager::instance();

	    if(selection_manager->key_pressed(key_event) ||
	       mModeManager->key_pressed(key_event)) {
		return true;
	    }
	    else {
		// if the selection manager and mode manager didn't handle
		// the key press, then maybe we will...

		if(key_event->get_key() == 'a' && key_event->is_ctrl_down()) {
		    // i.e.: ^A

		    selection_manager->damage();
		    selection_manager->clear();

		    for_each(get_figs().begin(),
			     get_figs().end(),
			     bind1st(mem_fun(&SelectionManager::add), selection_manager));

		    selection_manager->damage();

		    return true;
		}
		else {
		    return false;
		}
	    }
	}
	else {
	    return false;
	}
    }

    bool
    Editor::key_released(const Toolkit::KeyEvent * key_event) {
	if(mActive) {
	    SelectionManager * selection_manager = SelectionManager::instance();
	    return (selection_manager->key_released(key_event) ||
		    mModeManager->key_released(key_event));
	}
	else {
	    return false;
	}
    }

    void
    Editor::add(ModeChangeListener * mode_change_listener) {
	if(mModeManager) {
	    mModeManager->add(mode_change_listener);
	}
    }

    void
    Editor::remove(ModeChangeListener * mode_change_listener) {
	if(mModeManager) {
	    mModeManager->remove(mode_change_listener);
	}
    }

    void
    Editor::set_under_mouse(const Toolkit::MouseEvent * mouse_event) {
	ASSERT(mActive);

	// FIXME If the user is modifying the selection, for example
	// dragging a handle of an edge, then the under mouse oscillates
	// between the edge and the objects that are being passed over...

	Fig * fig = hit(mouse_event->get_point());

	if(fig != mCurrentFig) {
	    MouseListener * mouse_listener;
	    
	    if(mCurrentFig) {
		mouse_listener = dynamic_cast<MouseListener *>(mCurrentFig);

		if(mouse_listener) {
		    mouse_listener->mouse_exited(mouse_event);
		}
	    }

	    if(fig) {
		mouse_listener = dynamic_cast<MouseListener *>(fig);

		if(mouse_listener) {
		    mouse_listener->mouse_entered(mouse_event);
		}
	    }

	    mCurrentFig = fig;
	}

	SelectionManager * selection_manager = SelectionManager::instance();
	Selection * selection = selection_manager->find_selection_at(mouse_event->get_point());

	if(selection != mCurrentSelection) {
	    if(mCurrentSelection) {
		mCurrentSelection->mouse_exited(mouse_event);
	    }

	    if(selection) {
		selection->mouse_entered(mouse_event);
	    }
	}

	mCurrentSelection = selection;
    }

    void
    Editor::set_current(Editor * editor) {
	if(mCurrent != editor) {
	    mCurrent = editor;

	    EditorChangeEvent editor_change_event(mCurrent);

	    for_each(mEditorChangeListeners.begin(), mEditorChangeListeners.end(),
		     bind2nd(mem_fun(&EditorChangeListener::editor_changed), &editor_change_event));
	}
    }

    bool
    Editor::canvas_exposed(const Toolkit::CanvasEvent * canvas_event) {
	Toolkit::Canvas * canvas = canvas_event->get_canvas();
	ASSERT(canvas == mCanvas);

	if(mLayerManager) {
	    mLayerManager->paint(canvas);
	}

	SelectionManager * selection_manager = SelectionManager::instance();
	selection_manager->paint(canvas);

	if(mModeManager) {
	    mModeManager->paint(canvas);
	}

	return true;
    }

    void
    Editor::mode_changed(const ModeChangeEvent * mode_change_event) {
	Mode * mode = mode_change_event->get_mode();

	if(mode) {
	    mCanvas->set_cursor(mode->get_cursor());
	}
    }

    void
    Editor::paint(Drawing::Drawable * drawable) const {
	ASSERT(mLayerManager);
	mLayerManager->paint(drawable);

	SelectionManager * selection_manager = SelectionManager::instance();
	selection_manager->paint(drawable);

	ASSERT(mModeManager);
	mModeManager->paint(drawable);
    }

    void
    Editor::print(Drawing::Drawable * drawable) const {
	ASSERT(mLayerManager);

	// only the layers are printed
	mLayerManager->paint(drawable);
    }

    const list<Fig *> &
    Editor::get_figs() {
	ASSERT(mLayerManager);

	return mLayerManager->get_figs();
    }

} // namespace Diagram
