// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef DIAGRAM_EDITOR_H
#define DIAGRAM_EDITOR_H

#include "foundation/object.h"
#include "toolkit/mouse_listener.h"
#include "toolkit/canvas_listener.h"
#include "toolkit/key_listener.h"
#include "mode_change_listener.h"

#include <list>
using namespace std;

namespace Toolkit {

    class Canvas;

} // namespace Toolkit

namespace Drawing {

    class Point;
    class Drawable;
    class Rectangle;

} // namespace Drawing

namespace Graph {

    class Model;

} // namespace Graph

namespace Diagram {

    class Fig;
    class Mode;
    class ModeManager;
    class ModeChangeListener;
    class EditorChangeListener;
    class LayerManager;
    class Selection;
    class LayerGuide;
    class Renderer;
    class Layer;

    class Editor : public Object,
		   protected Toolkit::MouseListener,
		   protected Toolkit::CanvasListener,
		   protected Toolkit::KeyListener,
		   protected ModeChangeListener {
    public:
	static Editor * current() { return mCurrent; }
	static void add(EditorChangeListener * editor_change_listener);
	static void remove(EditorChangeListener * editor_change_listener);

	explicit Editor(Toolkit::Canvas * canvas);

	void set_active(bool active);

	// proxied to the LayerManager
	void add(Layer * layer);
	void remove(Layer * layer);
	Fig * hit(const Drawing::Point & point) const;
	Fig * hit(const Drawing::Rectangle & rect) const;
	const list<Fig *> & get_figs();
	void add(Fig * fig);
	void remove(Fig * fig);

	// proxied to the ModeManager
	void push(Mode * mode);
	void pop();

	// damage the canvas
	void damage(const Mode * mode);
	void damage(const Selection * selection);
	void damage(const Fig * fig);
	void damage(const Drawing::Rectangle & rect);

	// these are proxied to the ModeManager
	void add(ModeChangeListener * mode_change_listener);
	void remove(ModeChangeListener * mode_change_listener);

	Graph::Model * get_graph_model();
	Renderer * get_renderer();

	// the point will be snapped according to the guide
	void snap(Drawing::Point & point);

	void paint(Drawing::Drawable * drawable) const;
	void print(Drawing::Drawable * drawable) const;

    protected:
	// MouseListener interface

	bool mouse_entered(const Toolkit::MouseEvent * mouse_event);
	bool mouse_exited(const Toolkit::MouseEvent * mouse_event);
	bool mouse_pressed(const Toolkit::MouseEvent * mouse_event);
	bool mouse_released(const Toolkit::MouseEvent * mouse_event);
	bool mouse_moved(const Toolkit::MouseEvent * mouse_event);
	bool mouse_dragged(const Toolkit::MouseEvent * mouse_event);

	// KeyListener interface

	bool key_pressed(const Toolkit::KeyEvent * key_event);
	bool key_released(const Toolkit::KeyEvent * key_event);

	// CanvasListener interface

	bool canvas_exposed(const Toolkit::CanvasEvent * canvas_event);

	// ModeChangeListener interface

	void mode_changed(const ModeChangeEvent * mode_change_event);

	~Editor();

	void set_under_mouse(const Toolkit::MouseEvent * mouse_event);

	static void set_current(Editor * editor);

    private:
	static const int HIT_SIZE;

	bool mActive;

	Toolkit::Canvas * mCanvas;
	ModeManager * mModeManager;
	LayerManager * mLayerManager;

	Fig * mCurrentFig;		// the fig that the mouse is over
	Selection * mCurrentSelection;
	LayerGuide * mLayerGuide;

	static Editor * mCurrent;
	static list<EditorChangeListener *> mEditorChangeListeners;
    };

} // namespace Diagram

#endif // DIAGRAM_EDITOR_H
