// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "fig.h"

#include "drawing/drawable.h"
#include "drawing/point.h"

#include "foundation/macros.h"
#include "foundation/functional.h"

#include "layer.h"
#include "handle.h"

#include <algorithm>
using namespace std;

namespace Diagram {

    const string Fig::DEFAULT_LINE_COLOR = "#dd0000";
    const string Fig::DEFAULT_FILL_COLOR = "#eebb00";
    const Drawing::LineStyle Fig::DEFAULT_LINE_STYLE = Drawing::LS_SOLID;

    const Drawing::Vector Fig::DEFAULT_MINIMUM_SIZE(17, 17);

    Fig::Fig(Object * owner) : Object(),
	mMinimumSize(DEFAULT_MINIMUM_SIZE),
	mLayer(0),
	mFilled(false),
	mLineColor(DEFAULT_LINE_COLOR),
	mFillColor(DEFAULT_FILL_COLOR),
	mLineStyle(DEFAULT_LINE_STYLE),
	mOwner(owner) {
	TRACE(10, "Fig created: " << this);
    }

    Fig::~Fig() {
	TRACE(10, "Fig destroyed: " << this);
    }

    //
    //
    //

    void
    Fig::print(Drawing::Drawable * drawable) const {
	// by default printing figs is the same as painting them
	// but subclasses can change this behaviour
	paint(drawable);
    }

    void
    Fig::relocate(const Drawing::Point & location) {
	damage();
	mBounds.relocate(location.get_x(), location.get_y());
	damage();
    }

    void
    Fig::resize(const Drawing::Vector & size) {
	damage();
	mBounds.resize(size.get_x(), size.get_y());
	damage();
    }

    Selection *
    Fig::create_selection() {
	return 0;
    }

    void
    Fig::create_drag(const Drawing::Point & anchor, const Drawing::Point & location,
		     const Drawing::Point & snapped) {
	Drawing::Rectangle new_bounds(anchor, snapped);

	if(new_bounds.get_w() < mMinimumSize.get_x()) {
	    new_bounds.set_w(mMinimumSize.get_x());
	}

	if(new_bounds.get_h() < mMinimumSize.get_y()) {
	    new_bounds.set_h(mMinimumSize.get_y());
	}

        relocate(Drawing::Point(new_bounds.get_x(), new_bounds.get_y()));

        resize(Drawing::Point(new_bounds.get_xw(), new_bounds.get_yh()) -
               Drawing::Point(new_bounds.get_x(), new_bounds.get_y()));
    }

    void
    Fig::begin_interaction() {
    }

    void
    Fig::end_interaction() {
    }

    void
    Fig::update_to_owner() {
    }

    void
    Fig::cleanup() {
    }

#if 0
    Point
    Fig::get_connection_point(const Point & point) {
	return get_closest_point_to(point);
    }

    Point
    Fig::get_closest_point_to(const Point & point) {
	return Point(point.get_x() < mBounds.get_cx() ? mBounds.get_x() : mBounds.get_xw(),
		     point.get_y() < mBounds.get_cy() ? mBounds.get_y() : mBounds.get_yh());
    }
#endif

    //
    //
    //

    void
    Fig::get_center(Drawing::Point & point) const {
	point.set(mBounds.get_cx(), mBounds.get_cy());
    }

    void
    Fig::set_filled(bool filled) {
	if(mFilled != filled) {
	    mFilled = filled;
	    damage();
	}
    }

    void
    Fig::set_line_color(const string & color) {
	if(mLineColor != color) {
	    mLineColor = color;
	    damage();
	}
    }

    void
    Fig::set_fill_color(const string & color) {
	if(mFillColor != color) {
	    mFillColor = color;
	    damage();
	}
    }

    void
    Fig::set_line_style(Drawing::LineStyle line_style) {
	if(mLineStyle != line_style) {
	    mLineStyle = line_style;
	    damage();
	}
    }

    bool
    Fig::hit(const Drawing::Rectangle & rect) const {
	if(contained_by(rect)) {
	    return true;
	}
	else if(intersects(rect)) {
	    // we only do the involved math if there is an intersection

	    int corners_contained = count_corners_contained(rect);

	    return is_filled() ? corners_contained > 0 : corners_contained > 0 && corners_contained < 4;
	}
	else {
	    return false;
	}
    }

    bool
    Fig::intersects(const Drawing::Rectangle & rect) const {
	return mBounds & rect;
    }

    bool
    Fig::contains(const Drawing::Rectangle & rect) const {
	return (mBounds | rect) == mBounds;
    }

    bool
    Fig::contained_by(const Drawing::Rectangle & rect) const {
	return (rect | mBounds) == rect;
    }

    bool
    Fig::contains(const Drawing::Point & point) const {
	return mBounds.contains_point(point);
    }

    void
    Fig::damage() {
	if(mLayer) {
	    mLayer->damage(this);
	}
    }

    void
    Fig::set_layer(Layer * layer) {
	damage();

	mLayer = layer;

	damage();
    }

    void
    Fig::rebound(const Drawing::Rectangle & bounds) {
        relocate(Drawing::Point(bounds.get_x(), bounds.get_y()));
        resize(Drawing::Vector(bounds.get_xw() - bounds.get_x() + 1,
                               bounds.get_yh() - bounds.get_y() + 1));
    }

    void
    Fig::translate(const Drawing::Vector & offset) {
        relocate(Drawing::Point(mBounds.get_x(), mBounds.get_y()) + offset);
    }

    int
    Fig::count_corners_contained(const Drawing::Rectangle & rect) const {
	int counter = 0;

	if(contains(Drawing::Point(rect.get_x(), rect.get_y()))) {
	    ++counter;
	}

	if(contains(Drawing::Point(rect.get_x() + rect.get_w(), rect.get_y()))) {
	    ++counter;
	}

	if(contains(Drawing::Point(rect.get_x() + rect.get_w(), rect.get_y() + rect.get_h()))) {
	    ++counter;
	}

	if(contains(Drawing::Point(rect.get_x(), rect.get_y() + rect.get_h()))) {
	    ++counter;
	}

	return counter;
    }

} // namespace Diagram
