// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef DIAGRAM_FIG_H
#define DIAGRAM_FIG_H

#include "foundation/object.h"

#include <string>
using namespace std;

#include "drawing/line_style.h"
#include "drawing/rectangle.h"
#include "drawing/point.h"
#include "drawing/vector.h"

namespace Drawing {

    class Drawable;

} // namespace Drawing

namespace Diagram {

    class Layer;
    class Selection;

    class Fig : public Object {
    public:
	Fig(Object * owner = 0);
	~Fig();

	//
	//
	//

	// each fig class is responsible for drawing itself
	virtual void paint(Drawing::Drawable * drawable) const = 0;

	// just calls paint by default, but
	// some figs may look different on screen from on paper...
	virtual void print(Drawing::Drawable * drawable) const;

	// figs must create their own selection object
	virtual Selection * create_selection();

	// sometimes it is advantageous for subclasses to handle
	// translation/versus resizing, etc... otherwise the change
	// just shows up as a change in the bounds
	virtual void relocate(const Drawing::Point & location);
	virtual void resize(const Drawing::Vector & size);

	// this gets called while the user is dragging the mouse after
	// first creating the object with a button press
	virtual void create_drag(const Drawing::Point & anchor, const Drawing::Point & location,
				 const Drawing::Point & snapped);

	// allows the figs to review their state before an interaction proceeds.
	// could be useful for undo support with Mementos.
	virtual void begin_interaction();

	// calls cleanup() and
	// allows figs to compare their current state with the state when
	// the transaction started and pass it to the undo manager
	virtual void end_interaction();

	// invoked if the owner of this object has changed
	virtual void update_to_owner();

	// allows the fig to any internal reduction after an interaction
	// has completed. for example, poly-lines get rid of any redundant
	// points.
	virtual void cleanup();

#if 0
	virtual Drawing::Point get_connection_point(const Drawing::Point & point);
	virtual Drawing::Point get_closest_point_to(const Drawing::Point & point);
#endif

	// does the rectangle intersect the fig, the default
	// handler just returns true if the rect intersects
	// the bounds of the fig
	virtual bool hit(const Drawing::Rectangle & rect) const;

	//
	//
	//

	virtual void set_filled(bool filled);
	virtual void set_line_color(const string & line_color);
	virtual void set_fill_color(const string & fill_color);
	virtual void set_line_style(Drawing::LineStyle line_style);
	virtual void set_layer(Layer * layer);

        // built on top of relocate and resize

        void rebound(const Drawing::Rectangle & bounds);
        void translate(const Drawing::Vector & offset);

	// tell our layer that we are damaged

	void damage();

	// geometric queries
	bool intersects(const Drawing::Rectangle & rect) const;
	bool contains(const Drawing::Rectangle & rect) const;
	bool contained_by(const Drawing::Rectangle & rect) const;
	bool contains(const Drawing::Point & point) const;

	bool is_layer(const Layer * layer) { return mLayer == layer; }
	Layer * get_layer() { return mLayer; }

	bool is_owner(const Object * owner) { return mOwner == owner; }
	Object * get_owner() const { return mOwner; }

	void get_center(Drawing::Point & point) const;

	const Drawing::Rectangle & get_bounds() const { return mBounds; }
	const Drawing::Vector & get_minimum_size() const { return mMinimumSize; }

	bool is_filled() const { return mFilled; }
	const string & get_line_color() const { return mLineColor; }
	const string & get_fill_color() const { return mFillColor; }
	Drawing::LineStyle get_line_style() const { return mLineStyle; }

    protected:
	virtual int count_corners_contained(const Drawing::Rectangle & rect) const;

	static const string DEFAULT_LINE_COLOR;
	static const string DEFAULT_FILL_COLOR;
	static const Drawing::Vector DEFAULT_MINIMUM_SIZE;
	static const Drawing::LineStyle DEFAULT_LINE_STYLE;

	// subclasses have access to these so that they can keep them up-to-date
	Drawing::Rectangle mBounds;
	Drawing::Vector mMinimumSize;		// only applies to resizable

    private:
	Fig(const Fig & fig) { }		// don't allow figs to be copied

	// the layer that this fig belongs to
	Layer * mLayer;

	bool mFilled;

	string mLineColor, mFillColor;
	Drawing::LineStyle mLineStyle;

	Object * mOwner;
    };

} // namespace Diagram

#endif // DIAGRAM_FIG_H
