// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "fig_edge.h"

#include "fig_node.h"
#include "fig_edge_text.h"
#include "layer.h"
#include "selection_edge.h"
#include "handle.h"

#include "foundation/macros.h"
#include "drawing/segment.h"

#include <algorithm>
#include <functional>
using namespace std;

namespace Diagram {

    FigEdge::FigEdge(Object * owner, FigNode * source_fig,
		     FigNode * destination_fig) : FigPoly(owner),
	mUseNearest(false),
	mHighlight(false),
	mSourceFig(source_fig),
	mDestinationFig(destination_fig) {
	mSourceFig->add_edge(this);
	mDestinationFig->add_edge(this);
    }

    FigEdge::~FigEdge() {
	if(mSourceFig) {
	    mSourceFig->remove_edge(this);
	}

	if(mDestinationFig) {
	    mDestinationFig->remove_edge(this);
	}

	for_each(mAnnotations.begin(), mAnnotations.end(), mem_fun(&Object::unref));
    }

    void
    FigEdge::compute_route() {
	ASSERT(mSourceFig);
	ASSERT(mDestinationFig);
	ASSERT(get_points().size() >= 2);

	const Drawing::Rectangle & source_bounds = mSourceFig->get_bounds();
	const Drawing::Rectangle & destin_bounds = mDestinationFig->get_bounds();

	bool changed = false;

	if(mPoints.size() == 2) {
	    if(source_bounds & destin_bounds) {
		Drawing::Point pa(source_bounds.get_cx(), source_bounds.get_cy());
		Drawing::Point pz(destin_bounds.get_cx(), destin_bounds.get_cy());

		// the source fig and the destination fig overlap
		// the best we can do is draw from the center of one to
		// the center of the other

		if(mPoints.front() != pa) {
		    changed = true;
		    mPoints.front() = pa;
		}

		if(mPoints.back() != pz) {
		    changed = true;
		    mPoints.back() = pz;
		}
	    }
	    else {
		// the bounds of the two figs are completely separated

		if(source_bounds.get_xw() >= destin_bounds.get_x() &&
		   destin_bounds.get_xw() >= source_bounds.get_x()) {
		    // vertical line between the two

		    int a = source_bounds.get_w();
		    int b = destin_bounds.get_w();
		    int c = destin_bounds.get_cx() - source_bounds.get_cx();

		    int x = source_bounds.get_cx() + (c * a) / (a + b);
		    int ya;
		    int yz;

		    if(source_bounds.get_y() > destin_bounds.get_yh()) {
			ya = source_bounds.get_y();
			yz = destin_bounds.get_yh();
		    }
		    else if(destin_bounds.get_y() > source_bounds.get_yh()) {
			ya = source_bounds.get_yh();
			yz = destin_bounds.get_y();
		    }
		    else {
			ASSERT_NOT_REACHED();
		    }

		    Drawing::Point pa(x, ya);
		    Drawing::Point pz(x, yz);

		    if(mPoints.front() != pa) {
			changed = true;
			mPoints.front() = pa;
		    }

		    if(mPoints.back() != pz) {
			changed = true;
			mPoints.back() = pz;
		    }
		}
		else if(source_bounds.get_yh() >= destin_bounds.get_y() &&
			destin_bounds.get_yh() >= source_bounds.get_y()) {
		    // horizontal line between the two

		    int a = source_bounds.get_h();
		    int b = destin_bounds.get_h();
		    int c = destin_bounds.get_cy() - source_bounds.get_cy();

		    int y = source_bounds.get_cy() + (c * a) / (a + b);
		    int xa;
		    int xz;

		    if(source_bounds.get_x() > destin_bounds.get_xw()) {
			xa = source_bounds.get_x();
			xz = destin_bounds.get_xw();
		    }
		    else if(destin_bounds.get_x() > source_bounds.get_xw()) {
			xa = source_bounds.get_xw();
			xz = destin_bounds.get_x();
		    }
		    else {
			ASSERT_NOT_REACHED();
		    }

		    Drawing::Point pa(xa, y);
		    Drawing::Point pz(xz, y);

		    if(mPoints.front() != pa) {
			changed = true;
			mPoints.front() = pa;
		    }

		    if(mPoints.back() != pz) {
			changed = true;
			mPoints.back() = pz;
		    }
		}
		else {
		    // diagonal line between the two
		    Drawing::Point p1(source_bounds.get_cx(), source_bounds.get_cy());
		    Drawing::Point p2(destin_bounds.get_cx(), destin_bounds.get_cy());

		    Drawing::Segment segment(p1, p2);

		    Drawing::Point pa, pz;

		    ASSERT(source_bounds.get_intercept(segment, pa));
		    ASSERT(destin_bounds.get_intercept(segment, pz));

		    if(mPoints.front() != pa) {
			changed = true;
			mPoints.front() = pa;
		    }

		    if(mPoints.back() != pz) {
			changed = true;
			mPoints.back() = pz;
		    }
		}
	    }
	}
	else {
#if 1
	    Drawing::Point source_result;
	    solve(mPoints[1], source_bounds, source_result);

	    if(mPoints[0] != source_result) {
		changed = true;
		mPoints[0] = source_result;
	    }

	    Drawing::Point destin_result;
	    solve(mPoints[mPoints.size() - 2], destin_bounds, destin_result);

	    if(mPoints[mPoints.size() - 1] != destin_result) {
		changed = true;
		mPoints[mPoints.size() - 1] = destin_result;
	    }
#else
	    Drawing::Point pa(source_bounds.get_cx(), source_bounds.get_cy());	// first point
	    Drawing::Point pb(mPoints[1]);					// second point
	    //
	    // c d e ... v w x
	    //
	    Drawing::Point py(mPoints[mPoints.size() - 2]);		// second to last point
	    Drawing::Point pz(destin_bounds.get_cx(), destin_bounds.get_cy());	// last point

	    Drawing::Segment source_segment(pa, pb);
	    Drawing::Segment destination_segment(pz, py);

	    Drawing::Point source_point;
	    if(!source_bounds.get_intercept(source_segment, source_point)) {
		source_point.set(source_bounds.get_cx(), source_bounds.get_cy());
	    }

	    if(mPoints.front() != source_point) {
		changed = true;
		mPoints.front() = source_point;
	    }

	    Drawing::Point destination_point;
	    if(!destin_bounds.get_intercept(destination_segment, destination_point)) {
		destination_point.set(destin_bounds.get_cx(), destin_bounds.get_cy());
	    }

	    if(mPoints.back() != destination_point) {
		changed = true;
		mPoints.back() = destination_point;
	    }
#endif
	}

	if(changed) {
	    update_bounds_to_points();
	}
    }

    void
    FigEdge::compute_text_positions() {
	for_each(mAnnotations.begin(), mAnnotations.end(), mem_fun(&FigEdgeText::compute_position));
    }

    void
    FigEdge::solve(const Drawing::Point & point, const Drawing::Rectangle & rectangle,
		   Drawing::Point & result) {
	if(rectangle.contains_point(point)) {
	    result.set(rectangle.get_cx(), rectangle.get_cy());
	}
	else {
	    if(point.get_x() >= rectangle.get_x() && point.get_x() <= rectangle.get_xw()) {
		// vertical line between the two

		int x = point.get_x();
		int y;

		if(point.get_y() < rectangle.get_y()) {
		    y = rectangle.get_y();
		}
		else if(point.get_y() > rectangle.get_yh()) {
		    y = rectangle.get_yh();
		}
		else {
		    ASSERT_NOT_REACHED();
		}

		result.set(x, y);
	    }
	    else if(point.get_y() >= rectangle.get_y() && point.get_y() <= rectangle.get_yh()) {
		// horizontal line between the two

		int x;
		int y = point.get_y();

		if(point.get_x() < rectangle.get_x()) {
		    x = rectangle.get_x();
		}
		else if(point.get_x() > rectangle.get_xw()) {
		    x = rectangle.get_xw();
		}
		else {
		    ASSERT_NOT_REACHED();
		}

		result.set(x, y);
	    }
	    else {
		Drawing::Segment segment(Drawing::Point(rectangle.get_cx(), rectangle.get_cy()),
					 point);

		ASSERT(rectangle.get_intercept(segment, result));
	    }
	}
    }

    void
    FigEdge::set_source_fig(FigNode * source_fig_node) {
	if(mSourceFig != source_fig_node) {
	    if(mSourceFig) {
		mSourceFig->remove_edge(this);
	    }

	    mSourceFig = source_fig_node;

	    if(mSourceFig) {
		mSourceFig->add_edge(this);
	    }

	    cleanup();
	}
    }

    void
    FigEdge::set_destination_fig(FigNode * destination_fig_node) {
	if(mDestinationFig != destination_fig_node) {
	    if(mDestinationFig) {
		mDestinationFig->remove_edge(this);
	    }

	    mDestinationFig = destination_fig_node;

	    if(mDestinationFig) {
		mDestinationFig->add_edge(this);
	    }

	    cleanup();
	}
    }

    void
    FigEdge::remove_node(FigNode * node) {
	ASSERT(!( (mSourceFig == node) && (mDestinationFig == node)));	// we are screwed

	if(mSourceFig == node) {
	    set_source_fig(0);
	}
	else if(mDestinationFig == node) {
	    set_destination_fig(0);
	}
	else {
	    ERROR("remove_node called, but node we are not connected to that node");
	}
    }

    void
    FigEdge::update_bounds_to_points() {
	FigPoly::update_bounds_to_points();
    }

    bool
    FigEdge::mouse_entered(const Toolkit::MouseEvent * mouse_event) {
#if 0
	set_highlight(true);
#endif

	return true;
    }

    bool
    FigEdge::mouse_exited(const Toolkit::MouseEvent * mouse_event) {
#if 0
	set_highlight(false);
#endif

	return true;
    }

    bool
    FigEdge::mouse_pressed(const Toolkit::MouseEvent * mouse_event) {
	return true;
    }

    bool
    FigEdge::mouse_released(const Toolkit::MouseEvent * mouse_event) {
	return true;
    }

    bool
    FigEdge::mouse_moved(const Toolkit::MouseEvent * mouse_event) {
	return true;
    }

    bool
    FigEdge::mouse_dragged(const Toolkit::MouseEvent * mouse_event) {
	return true;
    }

    void
    FigEdge::cleanup() {
	FigPoly::cleanup();
	compute_route();
	compute_text_positions();
    }

    void
    FigEdge::add_point(const Drawing::Point & point) {
	FigPoly::add_point(point);
	compute_route();
	compute_text_positions();
    }

    void
    FigEdge::set_point(Handle & handle, const Drawing::Point & point) {
	FigPoly::set_point(handle, point);

	// we only compute the route if the point that has been moved is
	// a "second-from-the-end" point, because these are the only points
	// that affect the route

	int index = handle.get_index();

	if(mPoints.size() > 2 && (index == 1 || index == mPoints.size() - 2)) {
	    compute_route();
	}

	compute_text_positions();
    }

    void
    FigEdge::remove_point(int index) {
	// the first and last point should never be removed

	ASSERT(index != 0);
	ASSERT(index != mPoints.size() - 1);

	FigPoly::remove_point(index);

	// we only compute the route if the point is
	// a "second-from-the-end" point

	if(index == 1 || index == mPoints.size() - 2) {
	    compute_route();
	}

	compute_text_positions();
    }

    void
    FigEdge::insert_point(int index, const Drawing::Point & point) {
	// can't insert a point at the beginning or the end

	ASSERT(index != 0);
	ASSERT(index != mPoints.size());

	FigPoly::insert_point(index, point);

	// we only compute the route if the new point will be
	// a "second-from-the-end" point

	if(index == 1 || index == mPoints.size() - 1) {
	    compute_route();
	}

	compute_text_positions();
    }

    Selection *
    FigEdge::create_selection() {
	return new SelectionEdge(this);
    }

    void
    FigEdge::resize(const Drawing::Vector & size) {
	ERROR("Don't resize FigEdges");
    }

    void
    FigEdge::set_layer(Layer * layer) {
	Fig::set_layer(layer);

	for_each(mAnnotations.begin(), mAnnotations.end(), bind1st(mem_fun(&Layer::add), layer));
    }

    void
    FigEdge::update_edge_to_nodes() {
	compute_route();
	compute_text_positions();
    }

} // namespace Diagram
