// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "fig_group.h"

#include "foundation/functional.h"
#include "foundation/macros.h"

#include "drawing/drawable.h"

#include <algorithm>
#include <functional>
#include <numeric>
using namespace std;

namespace Diagram {

    FigGroup::FigGroup(Object * owner) : Fig(owner) {
    }

    FigGroup::~FigGroup() {
	for_each(mChildren.begin(), mChildren.end(), mem_fun(&Fig::unref));
    }

    void
    FigGroup::paint(Drawing::Drawable * drawable) const {
	// we push a clip mask that is equal to our bounds so
	// that children do not draw outside of our bounds

	drawable->push_clip(get_bounds());
	for_each(mChildren.rbegin(), mChildren.rend(), bind2nd(mem_fun(&Fig::paint), drawable));
	drawable->pop_clip();
    }

    void
    FigGroup::relocate(const Drawing::Point & location) {
        Drawing::Vector offset = location - Drawing::Point(mBounds.get_x(), mBounds.get_y());

	Fig::relocate(location);
	for_each(mChildren.begin(), mChildren.end(), bind2nd_vr(mem_fun(&Fig::translate), offset));
    }

    void
    FigGroup::resize(const Drawing::Vector & size) {
	Fig::resize(size);
	layout();
    }

#if 0
    void
    FigGroup::resize(int w, int h) {
	Fig::resize(w, h);
	NYI();
    }

    void
    FigGroup::set_bounds(const Drawing::Rectangle & bounds) {
	int old_x = mBounds.get_x();
	int old_y = mBounds.get_y();
	int old_w = mBounds.get_w();
	int old_h = mBounds.get_h();

	Fig::set_bounds(bounds);

	int new_x = mBounds.get_x();
	int new_y = mBounds.get_y();
	int new_w = mBounds.get_w();
	int new_h = mBounds.get_h();

	for(list<Fig *>::iterator iter = mChildren.begin(); iter != mChildren.end(); ++iter) {
	    Fig * fig = *iter;

	    const Drawing::Rectangle & child_bounds = fig->get_bounds();

	    int x = child_bounds.get_x();
	    int y = child_bounds.get_y();
	    int xw = child_bounds.get_xw();
	    int yh = child_bounds.get_yh();

	    Drawing::Point p0(new_x + ((x - old_x) * new_w) / old_w,
			      new_y + ((y - old_y) * new_h) / old_h);

	    Drawing::Point p1(new_x + ((xw - old_x) * new_w) / old_w,
			      new_y + ((yh - old_y) * new_h) / old_h);

	    fig->set_bounds(Drawing::Rectangle(p0, p1));
	}
    }
#endif

    bool
    FigGroup::hit(const Drawing::Rectangle & rect) const {
#if 1
	if(mChildren.empty()) {
	    return rect & mBounds;
	}
	else if(rect & mBounds) {
	    // we check that it is in our outer bounds before asking each child fig
	    return find_if(mChildren.begin(), mChildren.end(),
			   bind2nd_vr(mem_fun(&Fig::hit), rect)) != mChildren.end();
	}
	else {
	    return false;
	}
#else
	return rect & mBounds;
#endif
    }

    void
    FigGroup::set_filled(bool filled) {
	Fig::set_filled(filled);

	for_each(mChildren.begin(), mChildren.end(), bind2nd(mem_fun(&Fig::set_filled), filled));
    }

    void
    FigGroup::set_line_color(const string & line_color) {
	Fig::set_line_color(line_color);

	for_each(mChildren.begin(), mChildren.end(),
		 bind2nd_vr(mem_fun(&Fig::set_line_color), line_color));
    }

    void
    FigGroup::set_fill_color(const string & fill_color) {
	Fig::set_fill_color(fill_color);

	for_each(mChildren.begin(), mChildren.end(),
		 bind2nd_vr(mem_fun(&Fig::set_fill_color), fill_color));
    }

    void
    FigGroup::set_line_style(Drawing::LineStyle line_style) {
	Fig::set_line_style(line_style);

	for_each(mChildren.begin(), mChildren.end(),
		 bind2nd(mem_fun(&Fig::set_line_style), line_style));
    }

    void
    FigGroup::set_layer(Layer * layer) {
	Fig::set_layer(layer);

	for_each(mChildren.begin(), mChildren.end(), bind2nd(mem_fun(&Fig::set_layer), layer));
    }

#if 0
    Fig *
    FigGroup::hit_fig(const Drawing::Rectangle & rect) {
	if(rect & mBounds) {
	    if(mChildren.empty()) {
		return this;
	    }
	    else {
		list<Fig *>::iterator iter = find_if(mChildren.begin(), mChildren.end(),
						     bind2nd_vr(mem_fun(&Fig::hit), rect));

		return iter == mChildren.end() ? 0 : *iter;
	    }
	}
	else {
	    return 0;
	}
    }
#endif

} // namespace Diagram
