// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "fig_poly.h"

#include "foundation/macros.h"
#include "foundation/functional.h"
#include "drawing/point.h"
#include "drawing/rectangle.h"
#include "drawing/segment.h"

#include "handle.h"
#include "selection_poly.h"

#include <algorithm>
#include <cmath>
using namespace std;

namespace Diagram {

    FigPoly::FigPoly(Object * owner) : Fig(owner) {
    }

    FigPoly::~FigPoly() {
    }

    //
    //
    //

    void
    FigPoly::relocate(const Drawing::Point & location) {
        Drawing::Vector v = location - Drawing::Point(mBounds.get_x(), mBounds.get_y());

	Fig::relocate(location);

#if 1
	for(int i = 0; i < mPoints.size(); ++i) {
	    mPoints[i].translate(v);
	}
#else
	// FIXME this is broken. I think it's because somewhere the reference
	// is being passed by value, so the change to it is lost
	for_each(mPoints.begin(), mPoints.end(),
		 bind2nd_rr(mem_fun_ref(&Drawing::Point::translate), v));
#endif
    }

    void
    FigPoly::resize(const Drawing::Vector & size) {
	int old_x = mBounds.get_x();
	int old_y = mBounds.get_y();
	int old_w = mBounds.get_w();
	int old_h = mBounds.get_h();

	Fig::resize(size);

	int new_x = mBounds.get_x();
	int new_y = mBounds.get_y();
	int new_w = mBounds.get_w();
	int new_h = mBounds.get_h();

	for(unsigned int i = 0; i < mPoints.size(); ++i) {
	    Drawing::Point & point = mPoints[i];

	    point.set_x(new_x + ((point.get_x() - old_x) * new_w) / old_w);
	    point.set_y(new_y + ((point.get_y() - old_y) * new_h) / old_h);
	}
    }

    void
    FigPoly::set_points(const vector<Drawing::Point> & points) {
	mPoints.resize(points.size());
	copy(points.begin(), points.end(), mPoints.begin());
	update_bounds_to_points();
    }

    void
    FigPoly::update_bounds_to_points() {
	damage();

	if(mPoints.empty()) {
	    mBounds.clear();
	}
	else {
	    int xmin = mPoints.front().get_x();
	    int xmax = xmin;
	    int ymin = mPoints.front().get_y();
	    int ymax = ymin;

	    for(unsigned int i = 1; i < mPoints.size(); ++i) {
		const Drawing::Point & point = mPoints[i];

		int x = point.get_x();
		int y = point.get_y();

		if(x < xmin) { xmin = x; }
		else if(x > xmax) { xmax = x; }

		if(y < ymin) { ymin = y; }
		else if(y > ymax) { ymax = y; }
	    }

	    // pixels have a width and height of 1
	    ++xmax;
	    ++ymax;

	    mBounds.set(xmin, ymin, xmax - xmin, ymax - ymin);
	}

	damage();
    }

    void
    FigPoly::add_point(const Drawing::Point & point) {
	mPoints.push_back(point);
	update_bounds_to_points();
    }

    void
    FigPoly::set_point(Handle & handle, const Drawing::Point & point) {
	if(!handle.get_exists()) {
	    // the trick for postponing the insertion of points when the
	    // user drags part of a poly
	    insert_point(handle.get_index(), point);
	    handle.set_exists(true);
	}
	else {
	    mPoints[handle.get_index()] = point;
	}

	update_bounds_to_points();
    }

    void
    FigPoly::prepend_two_points() {
	mPoints.resize(mPoints.size() + 2);
	copy_backward(mPoints.begin(), mPoints.end() - 2, mPoints.end());
	mPoints[0] = mPoints[2];
	mPoints[1] = mPoints[2];
	update_bounds_to_points();
    }

    void
    FigPoly::append_two_points() {
	mPoints.resize(mPoints.size() + 2);
	mPoints[mPoints.size() - 2] = mPoints[mPoints.size() - 3];
	mPoints[mPoints.size() - 1] = mPoints[mPoints.size() - 3];
	update_bounds_to_points();
    }

    void
    FigPoly::remove_point(int index) {
	if(mPoints.size() > 2 && index < static_cast<int>(mPoints.size())) {
	    copy(mPoints.begin() + index + 1, mPoints.end(), mPoints.begin() + index);
	    mPoints.pop_back();
	    update_bounds_to_points();
	}
    }

    void
    FigPoly::insert_point(int index, const Drawing::Point & point) {
	ASSERT(index <= static_cast<int>(mPoints.size()));

	mPoints.push_back(Drawing::Point());
	copy_backward(mPoints.begin() + index, mPoints.end() - 1, mPoints.end());
	mPoints[index] = point;
	update_bounds_to_points();
    }

    void
    FigPoly::set_end_points(const Drawing::Point & point_start, const Drawing::Point & point_end) {
	while(mPoints.size() < 2) {
	    add_point(point_start);
	}

	Handle handle;

	handle.set_index(0);
	set_point(handle, point_start);

	handle.set_index(mPoints.size() - 1);
	set_point(handle, point_end);

	update_bounds_to_points();
    }

    void
    FigPoly::cleanup() {
	Fig::cleanup();

	if(mPoints.size() > 2) {
	    // if there are more than to points in the poly then
	    // remove points that occupy the same position

	    for(int i = 0; i < static_cast<int>(mPoints.size()) - 1; ++i) {
		const Drawing::Point & a = mPoints[i];
		const Drawing::Point & b = mPoints[i + 1];

		if(a == b) {
		    remove_point(i--);
		}
	    }
	}

	bool clean;

	do {
	    // if three points form a nearly straight line then remove
	    // the middle point

	    clean = true;

	    for(int i = 1; i < static_cast<int>(mPoints.size()) - 1; ++i) {
		const Drawing::Point & a = mPoints[i - 1];
		const Drawing::Point & b = mPoints[i];

		const Drawing::Point & c = mPoints[i + 1]; 
		double ab = atan2(static_cast<double>(b.get_x() - a.get_x()),
				  static_cast<double>(b.get_y() - a.get_y()));
		double bc = atan2(static_cast<double>(c.get_x() - b.get_x()),
				  static_cast<double>(c.get_y() - b.get_y()));

		while(ab - bc > M_PI) {
		    ab -= 2.0 * M_PI;
		}

		while(bc - ab > M_PI) {
		    bc -= 2.0 * M_PI;
		}

		if(abs(ab - bc) < MINIMUM_ANGLE) {
		    remove_point(i);
		    clean = false;
		    break;
		}
	    }
	}
	while(!clean);
    }

    Selection *
    FigPoly::create_selection() {
	return new SelectionPoly(this);
    }

    bool
    FigPoly::hit(const Drawing::Rectangle & rect) const {
	if(contained_by(rect)) {
	    return true;
	}
	else if(intersects(rect)) {
	    // we only do the involved math if there is an intersection

	    Drawing::Point dummy;

	    for(unsigned int i = 0; i < mPoints.size() - 1; ++i) {
		Drawing::Segment segment(mPoints[i], mPoints[i + 1]);

		if(rect.contains_segment(segment) || rect.get_intercept(segment, dummy)) {
		    return true;
		}
	    }

	    return false;
	}
	else {
	    return false;
	}
    }

} // namespace Diagram
