// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "fig_spline.h"

#include "drawing/drawable.h"
#include "drawing/point.h"
#include "drawing/segment.h"

#include "foundation/macros.h"

#include <cmath>
using namespace std;

namespace Diagram {

    FigSpline::FigSpline() : FigPoly() {
    }

    FigSpline::~FigSpline() {
    }

    void
    FigSpline::update_bounds_to_points() {
	FigPoly::update_bounds_to_points();
	update_curve();
    }

    void
    FigSpline::make_bezier(const Drawing::Point & p1, const Drawing::Point & p2,
			   const Drawing::Point & p3) {
	Drawing::Point pa((p1.get_x() + p2.get_x()) / 2, (p1.get_y() + p2.get_y()) / 2);
	Drawing::Point pc((p2.get_x() + p3.get_x()) / 2, (p2.get_y() + p3.get_y()) / 2);
	Drawing::Point pb((pa.get_x() + pc.get_x()) / 2, (pa.get_y() + pc.get_y()) / 2);
	Drawing::Point pp((p1.get_x() + pb.get_x()) / 2, (p1.get_y() + pb.get_y()) / 2);

	if(abs(pa.get_x() - pp.get_x()) + abs(pa.get_y() - pp.get_y()) > SPLINE_THRESHOLD) {
	    make_bezier(p1, pa, pb);
	}
	else {
	    mCurve.push_back(pb);
	}

	pp.set_x((p3.get_x() + pb.get_x()) / 2);
	pp.set_y((p3.get_y() + pb.get_y()) / 2);

	if(abs(pc.get_x() - pp.get_x()) + abs(pc.get_y() - pp.get_y()) > SPLINE_THRESHOLD) {
	    make_bezier(pb, pc, p3);
	}
	else {
	    mCurve.push_back(p3);
	}
    }

    void
    FigSpline::paint(Drawing::Drawable * drawable) const {
	drawable->set_line_style(get_line_style());
        drawable->set_color(get_line_color());
        drawable->draw_lines(mCurve);
    }

    bool
    FigSpline::hit(const Drawing::Rectangle & rect) const {
	// check that is intersects our outer bounds before doing the more involved math
	if(rect & mBounds && !mCurve.empty()) {
	    Drawing::Point dummy;

	    for(unsigned int i = 0; i < mCurve.size() - 1; ++i) {
		if(rect.get_intercept(Drawing::Segment(mCurve[i], mCurve[i + 1]), dummy)) {
		    return true;
		}
	    }

	    return false;
	}
	else {
	    return false;
	}
    }

    void
    FigSpline::set_junction_point(const Drawing::Point & p1, const Drawing::Point & p2,
				  const Drawing::Point & p3) {
	double delta0 = (p2 - p1).get_length();
	double delta1 = (p3 - p2).get_length();
	double denom = delta0 + delta1;
	double junction_t = denom <= OTHER_THRESHOLD ? 0.0 : delta1 / denom;

	mJunction.set_x(static_cast<int>(junction_t * p2.get_x() +
					(1.0 - junction_t) * p3.get_x() + 0.5));
	mJunction.set_y(static_cast<int>(junction_t * p2.get_y() +
					(1.0 - junction_t) * p3.get_y() + 0.5));
    }

    void
    FigSpline::update_curve() {
	mKnots.resize(mPoints.size() > 3 ? 2 * mPoints.size() - 3 : mPoints.size());

	if(mPoints.size() > 3) {
	    mKnots.front() = mPoints.front();
	    mKnots.back() = mPoints.back();

	    for(unsigned int i = 0; i < mPoints.size() - 3; ++i) {
		set_junction_point(mPoints[i], mPoints[i + 1], mPoints[i + 2]);

		mKnots[2 * (i + 1)] = mJunction;
	    }

	    for(unsigned int i = 1; i < mPoints.size() - 1; ++i) {
		mKnots[2 * i - 1] = mPoints[i];
	    }
	}
	else {
	    for(unsigned int i = 0; i < mPoints.size(); ++i) {
		mKnots[i] = mPoints[i];
	    }
	}

        // update the curve

	mCurve.clear();

        if(mPoints.size() > 2) {
            int segments = mPoints.size() - 2;

            mCurve.push_back(mKnots[0]);

            for(int i = 0; i < segments; ++i) {
                make_bezier(mKnots[2 * i], mKnots[2 * i + 1], mKnots[2 * i + 2]);
            }
        }
        else {
            ASSERT(mPoints.size() == 2);

            mCurve.resize(2);

            mCurve[0] = mPoints[0];
            mCurve[1] = mPoints[1];
        }
    }

    void
    FigSpline::relocate(const Drawing::Point & location) {
	FigPoly::relocate(location);
	update_curve();
    }

    void
    FigSpline::resize(const Drawing::Vector & size) {
	FigPoly::resize(size);
	update_curve();
    }

} // namespace Diagram
