// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "layer_manager.h"

#include "layer.h"
#include "fig.h"
#include "editor.h"

#include "foundation/macros.h"
#include "foundation/algorithm.h"

#include "drawing/drawable.h"
#include "drawing/rectangle.h"

#include <algorithm>
#include <functional>
using namespace std;

namespace Diagram {

    LayerManager::LayerManager(Editor * editor) : Object(),
	mEditor(editor),
	mActiveLayer(0) {
	ASSERT(editor);
    }

    LayerManager::~LayerManager() {
	for_each(mLayers.begin(), mLayers.end(), bind2nd(mem_fun(&Layer::remove_editor), mEditor));
	for_each(mLayers.begin(), mLayers.end(), mem_fun(&Layer::unref));
    }

    void
    LayerManager::add(Layer * layer) {
	RETURN_IF_FAIL(layer);

	layer->ref();
	mLayers.push_front(layer);
	layer->add_editor(mEditor);

	mActiveLayer = layer;
    }

    void
    LayerManager::remove(Layer * layer) {
	RETURN_IF_FAIL(layer);

	list<Layer *>::iterator iter = find(mLayers.begin(), mLayers.end(), layer);

	if(iter == mLayers.end()) {
	    ERROR("layer not found");
	}
	else {
	    ASSERT(*iter == layer);

	    layer->remove_editor(mEditor);
	    mLayers.erase(iter);
	    layer->unref();

	    if(layer == mActiveLayer) {
		mActiveLayer = mLayers.empty() ? 0 : mLayers.front();
	    }
	}
    }

    void
    LayerManager::add(Fig * fig) {
	if(mActiveLayer) {
	    mActiveLayer->add(fig);
	}
	else {
	    ERROR("No active layer");
	}
    }

    void
    LayerManager::remove(Fig * fig) {
	if(mActiveLayer) {
	    mActiveLayer->remove(fig);
	}
	else {
	    ERROR("No active layer");
	}
    }

    void
    LayerManager::clear() {
	if(mActiveLayer) {
	    mActiveLayer->clear();
	}
	else {
	    ERROR("No active layer");
	}
    }

    Fig *
    LayerManager::hit(const Drawing::Rectangle & rect) const {
	if(mActiveLayer) {
	    return mActiveLayer->hit(rect);
	}
	else {
	    WARNING("No active layer");

	    return 0;
	}
    }

    Fig *
    LayerManager::hit_not(const Drawing::Rectangle & rect, Fig * disregard) const {
	if(mActiveLayer) {
	    return mActiveLayer->hit_not(rect, disregard);
	}
	else {
	    WARNING("No active layer");

	    return 0;
	}
    }

    void
    LayerManager::paint(Drawing::Drawable * drawable) const {
	for_each_if(mLayers.rbegin(), mLayers.rend(),
		    not1(mem_fun(&Layer::is_hidden)),
		    bind2nd(mem_fun(&Layer::paint), drawable));
    }

    void
    LayerManager::print(Drawing::Drawable * drawable) const {
	for_each_if(mLayers.rbegin(), mLayers.rend(),
		    not1(mem_fun(&Layer::is_hidden)),
		    bind2nd(mem_fun(&Layer::print), drawable));
    }

    const list<Fig *> &
    LayerManager::get_figs() {
	if(mActiveLayer) {
	    return mActiveLayer->get_figs();
	}
	else {
	    static const list<Fig *> dummy;
	    return dummy;
	}
    }

    void
    LayerManager::set_active_layer(Layer * active_layer) {
	ASSERT(active_layer);

	mActiveLayer = active_layer;
    }

} // namespace Diagram
