// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "layer_perspective.h"

#include "graph/model.h"
#include "graph/event.h"

#include "foundation/macros.h"

#include "selection_manager.h"
#include "renderer.h"
#include "fig_node.h"
#include "fig_edge.h"

namespace Diagram {

    LayerPerspective::LayerPerspective(Graph::Model * graph_model,
				       Renderer * renderer) : LayerDiagram(),
	mModel(graph_model),
	mRenderer(renderer) {
	ASSERT(graph_model);
	ASSERT(renderer);

	mModel->ref();
	mRenderer->ref();

	mModel->add(static_cast<Listener *>(this));
    }

    LayerPerspective::~LayerPerspective() {
	mModel->remove(static_cast<Listener *>(this));

	mRenderer->unref();
	mModel->unref();
    }

    void
    LayerPerspective::set_graph_model(Graph::Model * graph_model) {
	RETURN_IF_FAIL(graph_model);

	if(mModel != graph_model) {
	    mModel->remove(static_cast<Listener *>(this));
	    mModel->unref();

	    mModel = graph_model;

	    mModel->ref();
	    mModel->add(static_cast<Graph::Listener *>(this));
	}
    }

    void
    LayerPerspective::set_renderer(Renderer * renderer) {
	RETURN_IF_FAIL(renderer);

	if(mRenderer != renderer) {
	    mRenderer->unref();

	    mRenderer = renderer;

	    mRenderer->ref();
	}
    }

    void
    LayerPerspective::node_added(const Graph::Event * graph_event) {
	TRACE(5, "LayerPerspective::node_added");

	ASSERT(graph_event);
	ASSERT(graph_event->get_model());
	ASSERT(graph_event->get_object());

	SelectionManager * selection_manager = SelectionManager::instance();
	selection_manager->damage();
	selection_manager->clear();

	Object * node = graph_event->get_object();
	FigNode * fig_node = mRenderer->create_fig_node_for(this, node);
        fig_node->update_to_owner();

	ASSERT(fig_node);

	add(fig_node);

        selection_manager->set(fig_node);
        selection_manager->damage();
    }

    void
    LayerPerspective::node_changed(const Graph::Event * graph_event) {
	TRACE(5, "LayerPerspective::node_changed");

	ASSERT(graph_event);
	ASSERT(graph_event->get_model());
	ASSERT(graph_event->get_object());

#if 0
	SelectionManager * selection_manager = SelectionManager::instance();
	selection_manager->damage();
	selection_manager->clear();
#endif

	Fig * fig = get_presentation_for(graph_event->get_object());

	// FIXME when a node is added to the diagram, the graph model
	// sometimes configures the node (for example, the name of a uml class is set)
	// the problem is that these changes occur in the same transaction that
	// adding to the diagram occurs, so a node_changed is emitted before
	// the node_added... NOTE, I think this is fixed now
	RETURN_IF_FAIL(fig);

	fig->damage();
	fig->update_to_owner();
	fig->damage();
    }

    void
    LayerPerspective::node_removed(const Graph::Event * graph_event) {
	TRACE(5, "LayerPerspective::node_removed");

	ASSERT(graph_event);
	ASSERT(graph_event->get_model());
	ASSERT(graph_event->get_object());

	SelectionManager * selection_manager = SelectionManager::instance();

	Fig * fig = get_presentation_for(graph_event->get_object());

	RETURN_IF_FAIL(fig);

        if(selection_manager->contains(fig)) {
            selection_manager->remove(fig);
            selection_manager->damage();
        }

	remove(fig);
    }

    void
    LayerPerspective::edge_added(const Graph::Event * graph_event) {
	TRACE(5, "LayerPerspective::edge_added");

	ASSERT(graph_event);
	ASSERT(graph_event->get_model());
	ASSERT(graph_event->get_object());

	SelectionManager * selection_manager = SelectionManager::instance();
	selection_manager->damage();
	selection_manager->clear();

	Object * edge = graph_event->get_object();
	FigEdge * fig_edge = mRenderer->create_fig_edge_for(this, edge);
        fig_edge->update_to_owner();

	ASSERT(fig_edge);

	add(fig_edge);

        selection_manager->set(fig_edge);
        selection_manager->damage();
    }

    void
    LayerPerspective::edge_changed(const Graph::Event * graph_event) {
	TRACE(5, "LayerPerspective::edge_changed");

	ASSERT(graph_event);
	ASSERT(graph_event->get_model());
	ASSERT(graph_event->get_object());

#if 0
	SelectionManager * selection_manager = SelectionManager::instance();
	selection_manager->damage();
	selection_manager->clear();
#endif

	Fig * fig = get_presentation_for(graph_event->get_object());

	// FIXME when a node is added to the diagram, the graph model
	// sometimes configures the node (for example, the name of a uml class is set)
	// the problem is that these changes occur in the same transaction that
	// adding to the diagram occurs, so a node_changed is emitted before
	// the node_added... NOTE, I think this is fixed now
	RETURN_IF_FAIL(fig);

	fig->damage();
	fig->update_to_owner();
	fig->damage();
    }

    void
    LayerPerspective::edge_removed(const Graph::Event * graph_event) {
	TRACE(5, "LayerPerspective::edge_removed");

	ASSERT(graph_event);
	ASSERT(graph_event->get_model());
	ASSERT(graph_event->get_object());

	SelectionManager * selection_manager = SelectionManager::instance();

	Fig * fig = get_presentation_for(graph_event->get_object());

	ASSERT(fig);

        if(selection_manager->contains(fig)) {
            selection_manager->remove(fig);
            selection_manager->damage();
        }

	remove(fig);
    }

} // namespace Diagram
