// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "mode_create_fig_poly_edge.h"

#include "editor.h"
#include "fig_node.h"
#include "fig_edge.h"
#include "fig_poly_line.h"

#include "foundation/macros.h"
#include "foundation/utility.h"
#include "foundation/undo_manager.h"

#include "toolkit/mouse_event.h"
#include "toolkit/factory.h"
#include "toolkit/message_box.h"

#include "graph/exception.h"
#include "graph/model.h"

namespace Diagram {

    ModeCreateFigPolyEdge::ModeCreateFigPolyEdge() : ModeCreateFigPoly(),
	mStartNode(0),
	mFinishNode(0) {
    }

    ModeCreateFigPolyEdge::~ModeCreateFigPolyEdge() {
    }

    Fig *
    ModeCreateFigPolyEdge::create_new_item(const Toolkit::MouseEvent * mouse_event,
					   const Drawing::Point & point) {
	FigPoly * fig_poly = new FigPolyLine();
	fig_poly->add_point(point);
	fig_poly->add_point(point);
	fig_poly->set_line_style(Drawing::LS_DOTTED);

	mLast = point;

	return fig_poly;
    }

    bool
    ModeCreateFigPolyEdge::can_start(const Toolkit::MouseEvent * mouse_event) {
	Editor * editor = Editor::current();

	ASSERT(editor);

	Fig * fig = editor->hit(mouse_event->get_point());

	if(fig) {
	    FigNode * fig_node = dynamic_cast<FigNode *>(fig);

	    if(fig_node) {
		mStartNode = fig_node->get_owner();

		ASSERT(mStartNode);

		return true;
	    }
	}

	return false;
    }

    bool
    ModeCreateFigPolyEdge::can_finish(const Toolkit::MouseEvent * mouse_event) {
	ASSERT(mStartNode);

	Editor * editor = Editor::current();

	ASSERT(editor);

	Fig * fig = editor->hit(mouse_event->get_point());

	if(fig) {
	    FigNode * fig_node = dynamic_cast<FigNode *>(fig);

	    if(fig_node) {
		mFinishNode = fig_node->get_owner();

		ASSERT(mFinishNode);

		return true;
	    }
	}

	return false;
    }

    void
    ModeCreateFigPolyEdge::finish(const Toolkit::MouseEvent * mouse_event) {
	ASSERT(mNewItem);
	ASSERT(mStartNode);
	ASSERT(mFinishNode);

	FigPoly * fig_poly = known_cast<FigPoly *>(mNewItem);
	ASSERT(fig_poly->get_points().size() >= 2);

	begin_transaction();

	// create the edge

	Object * edge = create_new_edge(fig_poly->get_points());

	// remove the facade item from the editor and unref it

	Editor * editor = Editor::current();

	ASSERT(editor);

	editor->remove(mNewItem);

	mNewItem->unref();

	mNewItem = 0;

	// add the new edge to the graph model

	Graph::Model * graph_model = editor->get_graph_model();

	ASSERT(graph_model);

	try {
	    graph_model->add_edge(edge, mStartNode, mFinishNode);
	}
	catch(Graph::Exception & ex) {
	    end_transaction();

	    UndoManager * undo_manager = UndoManager::instance();

	    undo_manager->undo();
	    undo_manager->clear_redo();

	    edge->unref();

            Toolkit::Factory * factory = Toolkit::Factory::instance();

            Toolkit::MessageBox * message_box =
                factory->create_message_box(ex.get_details(), Toolkit::MBK_ERROR,
                                            mouse_event->get_widget());

            message_box->go();
            message_box->unref();

            ModeCreateFigPoly::finish(mouse_event);

            return;
	}

	edge->unref();

	end_transaction();

	ModeCreateFigPoly::finish(mouse_event);
    }

} // namespace Diagram
