// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "mode_manager.h"

#include "foundation/macros.h"
#include "drawing/drawable.h"
#include "toolkit/mouse_event.h"

#include "mode.h"
#include "mode_change_listener.h"
#include "mode_change_event.h"

#include <algorithm>
#include <functional>
#include <typeinfo>
using namespace std;

// different modes should show different cursors

namespace Diagram {

    ModeManager::ModeManager() : Object(),
        mGrabbedMode(0) {
    }

    ModeManager::~ModeManager() {
	while(!mModes.empty()) {
	    mModes.top()->unref();
	    mModes.pop();
	}
    }

    void
    ModeManager::push(Mode * mode) {
	ASSERT(mode);

	if(mModes.empty() || mModes.top() != mode) {
	    while(!mModes.empty() && mModes.top()->is_replaceable()) {
		if(mModes.top()->is_running()) {
		    mModes.top()->cancel();
		}

		mModes.top()->unref();
		mModes.pop();
	    }

	    mModes.push(mode);
	    mModes.top()->ref();

	    ModeChangeEvent mode_change_event(mModes.top());

	    for_each(mModeChangeListeners.begin(), mModeChangeListeners.end(),
		     bind2nd(mem_fun(&ModeChangeListener::mode_changed), &mode_change_event));
	}

	TRACE(1, mModes.size());
    }

    void
    ModeManager::replace(Mode * mode) {
	ASSERT(mode);
	ASSERT(mModes.size() > 1);

	if(mode != mModes.top()) {
	    if(mModes.top()->is_running()) {
		mModes.top()->cancel();
	    }

	    mModes.top()->unref();
	    mModes.top() = mode;
	    mModes.top()->ref();

	    ModeChangeEvent mode_change_event(mModes.top());

	    for_each(mModeChangeListeners.begin(), mModeChangeListeners.end(),
		     bind2nd(mem_fun(&ModeChangeListener::mode_changed), &mode_change_event));
	}

	TRACE(1, mModes.size());
    }

    void
    ModeManager::pop() {
	if(mModes.size() > 1 && !mModes.top()->is_sticky()) {
	    if(mModes.top()->is_running()) {
		mModes.top()->cancel();
	    }

	    mModes.top()->unref();
	    mModes.pop();

	    ModeChangeEvent mode_change_event(mModes.top());

	    for_each(mModeChangeListeners.begin(), mModeChangeListeners.end(),
		     bind2nd(mem_fun(&ModeChangeListener::mode_changed), &mode_change_event));
	}

	TRACE(1, mModes.size());
    }

    bool
    ModeManager::mouse_moved(const Toolkit::MouseEvent * mouse_event) {
	if(mModes.empty()) {
	    return false;
	}
	else {
	    return mModes.top()->mouse_moved(mouse_event);
	}
    }

    bool
    ModeManager::mouse_dragged(const Toolkit::MouseEvent * mouse_event) {
	if(mModes.empty()) {
	    return false;
	}
	else {
	    return mModes.top()->mouse_dragged(mouse_event);
	}
    }

    bool
    ModeManager::mouse_pressed(const Toolkit::MouseEvent * mouse_event) {
	if(mModes.empty()) {
	    return false;
	}
	else {
	    return mModes.top()->mouse_pressed(mouse_event);
	}
    }

    bool
    ModeManager::mouse_released(const Toolkit::MouseEvent * mouse_event) {
	if(mModes.empty()) {
	    return false;
	}
	else {
	    return mModes.top()->mouse_released(mouse_event);
	}
    }

    bool
    ModeManager::mouse_exited(const Toolkit::MouseEvent * mouse_event) {
	if(mModes.empty()) {
	    return false;
	}
	else {
	    return mModes.top()->mouse_exited(mouse_event);
	}
    }

    bool
    ModeManager::mouse_entered(const Toolkit::MouseEvent * mouse_event) {
	if(mModes.empty()) {
	    return false;
	}
	else {
	    return mModes.top()->mouse_entered(mouse_event);
	}
    }

    //

    bool
    ModeManager::key_pressed(const Toolkit::KeyEvent * key_event) {
	if(mModes.empty()) {
	    return false;
	}
	else {
	    return mModes.top()->key_pressed(key_event);
	}
    }

    bool
    ModeManager::key_released(const Toolkit::KeyEvent * key_event) {
	if(mModes.empty()) {
	    return false;
	}
	else {
	    return mModes.top()->key_released(key_event);
	}
    }

    //

    void
    ModeManager::add(ModeChangeListener * mode_change_listener) {
	mModeChangeListeners.push_front(mode_change_listener);
    }

    void
    ModeManager::remove(ModeChangeListener * mode_change_listener) {
	list<ModeChangeListener *>::iterator iter =
	    find(mModeChangeListeners.begin(), mModeChangeListeners.end(), mode_change_listener);

	if(iter == mModeChangeListeners.end()) {
	    WARNING("Tried to remove a mode change listener that couldn't be found");
	}
	else {
	    mModeChangeListeners.erase(iter);
	}
    }

    void
    ModeManager::paint(Drawing::Drawable * drawable) const {
	if(!mModes.empty()) {
	    const Drawing::Rectangle & clip = drawable->get_clip();
	    Mode * mode = mModes.top();

	    if(clip) {
		if(mode->intersects(clip)) {
		    mode->paint(drawable);
		}
	    }
	    else {
		mode->paint(drawable);
	    }
	}
    }

} // namespace Diagram
