// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "mode_modify.h"

#include "foundation/macros.h"
#include "foundation/utility.h"
#include "drawing/rectangle.h"
#include "drawing/vector.h"
#include "drawing/drawable.h"
#include "toolkit/mouse_event.h"
#include "toolkit/widget.h"

#include "editor.h"
#include "selection_manager.h"
#include "selection.h"

#include <algorithm>
#include <iterator>
#include <functional>
#include <cmath>
using namespace std;

namespace Diagram {

    const int ModeModify::THRESHOLD = 8;

    ModeModify::ModeModify() : Mode(),
	mThresholdAchieved(false) {
	mCursor = Toolkit::C_CROSSHAIR;
    }

    ModeModify::~ModeModify() {
    }

    string
    ModeModify::get_instructions() const {
	return "Modify selected objects";
    }

    bool
    ModeModify::mouse_pressed(const Toolkit::MouseEvent * mouse_event) {
	start(mouse_event);

	Editor * editor = Editor::current();

	ASSERT(editor);

	const Drawing::Point & point = mouse_event->get_point();

	mAnchor = point;

	SelectionManager * selection_manager = SelectionManager::instance();

	Drawing::Rectangle rect(mAnchor - Drawing::Vector(4, 4), mAnchor + Drawing::Vector(4, 4));

	selection_manager->hit_handle(rect, mHandle);

	Drawing::Point snapped_point = point;

	editor->snap(snapped_point);

	mLast = snapped_point;

	return true;
    }

    bool
    ModeModify::mouse_dragged(const Toolkit::MouseEvent * mouse_event) {
	SelectionManager * selection_manager = SelectionManager::instance();

	if(!mThresholdAchieved) {
	    mThresholdAchieved = (mouse_event->get_point() - mAnchor).get_length() > THRESHOLD;

	    if(mThresholdAchieved) {
		selection_manager->begin_interaction();
	    }
	}

	if(mThresholdAchieved) {
	    Editor * editor = Editor::current();

	    ASSERT(editor);

	    const Drawing::Point & point = mouse_event->get_point();

	    Drawing::Point snapped_point = point;

	    if(mouse_event->is_ctrl_down()) {
		if(abs(point.get_x() - mAnchor.get_x()) > abs(point.get_y() - mAnchor.get_y())) {
		    snapped_point.set_y(mAnchor.get_y());
		}
		else {
		    snapped_point.set_x(mAnchor.get_x());
		}
	    }

	    editor->snap(snapped_point);

	    //

	    Drawing::Drawable * drawable =
		known_cast<Drawing::Drawable *>(mouse_event->get_widget());

	    Drawing::Vector size = drawable->get_size();

	    size.set_x(size.get_x() - (size.get_x() % 8));
	    size.set_y(size.get_y() - (size.get_y() % 8));

	    Drawing::Vector correction(0, 0);

	    if(mHandle.get_index() == -1) {
		Drawing::Rectangle bounds;

		selection_manager->get_contents_bounds(bounds);

		Drawing::Vector delta = snapped_point - mLast;

		Drawing::Point top_left =
                    Drawing::Point(bounds.get_x(), bounds.get_y()) + delta;

		Drawing::Point bottom_right =
                    Drawing::Point(bounds.get_xw(), bounds.get_yh()) + delta;

		if(bottom_right.get_x() >= size.get_x()) {
		    correction.set_x(bottom_right.get_x() - size.get_x());
		}
		else if(top_left.get_x() < 0) {
		    correction.set_x(top_left.get_x());
		}

		if(bottom_right.get_y() >= size.get_y()) {
		    correction.set_y(bottom_right.get_y() - size.get_y());
		}
		else if(top_left.get_y() < 0) {
		    correction.set_y(top_left.get_y());
		}

		if(delta - correction) {
		    selection_manager->damage();
		    selection_manager->translate(delta - correction);
		    selection_manager->damage();
		}
	    }
	    else {
		if(snapped_point.get_x() >= size.get_x()) {
		    correction.set_x(snapped_point.get_x() - size.get_x());
		}
		else if(snapped_point.get_x() < 0) {
		    correction.set_x(snapped_point.get_x());
		}

		if(snapped_point.get_y() >= size.get_y()) {
		    correction.set_y(snapped_point.get_y() - size.get_y());
		}
		else if(snapped_point.get_y() < 0) {
		    correction.set_y(snapped_point.get_y());
		}

		if(snapped_point - correction != mLast) {
		    selection_manager->damage();
		    selection_manager->drag_handle(snapped_point - correction, mAnchor, mHandle);
		    selection_manager->damage();
		}
	    }

	    mLast = snapped_point - correction;
	}

	return true;
    }

    bool
    ModeModify::mouse_released(const Toolkit::MouseEvent * mouse_event) {
	if(mThresholdAchieved) {
	    SelectionManager * selection_manager = SelectionManager::instance();

	    selection_manager->end_interaction();
	}

	finish(mouse_event);

	return true;
    }

    void
    ModeModify::start(const Toolkit::MouseEvent * mouse_event) {
	Mode::start(mouse_event);

	mThresholdAchieved = false;
    }

    void
    ModeModify::finish(const Toolkit::MouseEvent * mouse_event) {
	Editor * editor = Editor::current();

	ASSERT(editor);

	SelectionManager * selection_manager = SelectionManager::instance();

	if(selection_manager) {
	    selection_manager->damage();
	    selection_manager->cleanup();
	    selection_manager->damage();
	}

	Mode::finish(mouse_event);
    }

} // namespace Diagram
