// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "mode_select.h"

#include "foundation/iterator.h"
#include "foundation/algorithm.h"
#include "foundation/functional.h"
#include "foundation/macros.h"
#include "drawing/drawable.h"
#include "toolkit/mouse_event.h"

#include "handle.h"
#include "editor.h"
#include "selection_manager.h"
#include "fig.h"
#include "selection.h"
#include "mode_modify.h"
//#include "mode_broom.h"
#include "mode_pan.h"
#include "mode_popup.h"

#include <algorithm>
#include <functional>
#include <set>
#include <iterator>
using namespace std;

namespace Diagram {

    const int ModeSelect::HIT_SIZE = 8;

    ModeSelect::ModeSelect() : Mode() {
	mCursor = Toolkit::C_DEFAULT;
    }

    ModeSelect::~ModeSelect() {
    }

    string
    ModeSelect::get_instructions() const {
	return "Select";
    }

    bool
    ModeSelect::mouse_pressed(const Toolkit::MouseEvent * mouse_event) {
	start(mouse_event);

	Editor * editor = Editor::current();

	ASSERT(editor);

	SelectionManager * selection_manager = SelectionManager::instance();

	if(mouse_event->get_button() == 1) {
	    mAnchor = mouse_event->get_point();
	    mBounds.set(mAnchor.get_x(), mAnchor.get_y(), 0, 0);

	    Drawing::Rectangle hit_rect(mAnchor - Drawing::Vector(HIT_SIZE / 2, HIT_SIZE / 2),
					Drawing::Vector(HIT_SIZE + 1, HIT_SIZE + 1));

	    Handle handle;

	    selection_manager->hit_handle(hit_rect, handle);

	    bool toggle_selection = mouse_event->is_shift_down();

	    if(handle) {
		Mode * next_mode = new ModeModify();

		// the mode will start
		// and push itself
		// onto the mode_manager stack

		next_mode->mouse_pressed(mouse_event);
		next_mode->unref();
	    }
	    else {
		Fig * hit_fig = editor->hit(mAnchor);

		if(hit_fig) {
		    if(toggle_selection) {
			selection_manager->damage();
			selection_manager->toggle(hit_fig);
			selection_manager->damage();
		    }
		    else {
			if(!selection_manager->contains(hit_fig)) {
			    selection_manager->damage();
			    selection_manager->set(hit_fig);
			    selection_manager->damage();
			}

			Mode * next_mode = new ModeModify();

			// the mode will start
			// and push itself
			// onto the mode_manager stack

			next_mode->mouse_pressed(mouse_event);
			next_mode->unref();
		    }
		}
		else {
		    selection_manager->damage();
		    selection_manager->clear();
		    selection_manager->damage();
		}
	    }

	    return true;
	}
	else if(mouse_event->get_button() == 2) {
	    selection_manager->damage();
	    selection_manager->clear();
	    selection_manager->damage();

	    //Mode * next_mode = new ModePan();
	    //Mode * next_mode = new ModeBroom();

	    // the mode will start
	    // and push itself
	    // onto the mode_manager stack

	    //bool rval = next_mode->mouse_pressed(mouse_event);

	    //next_mode->unref();

	    //return rval;

            return false;
	}
	else if(mouse_event->get_button() == 3) {
	    Fig * hit_fig = editor->hit(mouse_event->get_point());

	    if(hit_fig) {
		// the user right clicked on a fig
		if(!selection_manager->contains(hit_fig)) {
		    // and that fig is not part of the current selection, so we selet it
		    selection_manager->damage();
		    selection_manager->set(hit_fig);
		    selection_manager->damage();
		}
	    }
	    else {
		// the user right clicked in space, clear the selection
		selection_manager->damage();
		selection_manager->clear();
	    }

	    // now we want to post a popup menu

	    Mode * next_mode = new ModePopup();

	    // the mode will start
	    // and push itself
	    // onto the mode_manager stack

	    bool rval = next_mode->mouse_pressed(mouse_event);

	    next_mode->unref();

	    return rval;
	}
	else {
	    return false;
	}
    }

    bool
    ModeSelect::mouse_dragged(const Toolkit::MouseEvent * mouse_event) {
	if(!mouse_event->is_button1_down()) {
	    // We are only interested in left mouse presses
	    return false;
	}

	const Drawing::Point & point = mouse_event->get_point();

	damage();
	mBounds.set(mAnchor, point);
	damage();

	update_selection();

	return true;
    }

    bool
    ModeSelect::mouse_released(const Toolkit::MouseEvent * mouse_event) {
	if(mouse_event->get_button() != 1) {
	    // We are only interested in left mouse presses
	    return false;
	}

	damage();
	mBounds.clear();

	finish(mouse_event);

	return true;
    }

    void
    ModeSelect::update_selection() {
	Editor * editor = Editor::current();

	ASSERT(editor);

	const list<Fig *> & all_figs = editor->get_figs();

	SelectionManager * selection_manager = SelectionManager::instance();

	ASSERT(selection_manager);

	const list<Selection *> & selections = selection_manager->get_selections();

	set<Fig *> currently_are_selected;
	set<Fig *> should_be_selected;

	copy_if(all_figs.begin(), all_figs.end(),
		associative_inserter(should_be_selected),
		bind2nd_vr(mem_fun(&Fig::contained_by), mBounds));

	transform(selections.begin(), selections.end(),
		  associative_inserter(currently_are_selected),
		  mem_fun(&Selection::get_content));

	vector<Fig *> add_these;
	vector<Fig *> remove_these;

	set_difference(currently_are_selected.begin(), currently_are_selected.end(),
		       should_be_selected.begin(), should_be_selected.end(),
		       back_inserter(remove_these));

	set_difference(should_be_selected.begin(), should_be_selected.end(),
		       currently_are_selected.begin(), currently_are_selected.end(),
		       back_inserter(add_these));

	if(!remove_these.empty()) {
	    selection_manager->damage();
	}

	for_each(remove_these.begin(), remove_these.end(),
		 bind1st(mem_fun(&SelectionManager::remove), selection_manager));

	for_each(add_these.begin(), add_these.end(),
		 bind1st(mem_fun(&SelectionManager::add), selection_manager));

	if(!add_these.empty()) {
	    selection_manager->damage();
	}
    }

    void
    ModeSelect::paint(Drawing::Drawable * drawable) const {
	if(mBounds) {
	    drawable->set_color(get_line_color());
	    drawable->draw_rectangle(mBounds, false);
	}
    }

} // namespace Diagram
