// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "rectangle.h"

#include "foundation/macros.h"

#include "point.h"
#include "vector.h"
#include "segment.h"

#include <algorithm>
#include <iostream>
using namespace std;

namespace Drawing {

    Rectangle::Rectangle() {
	mX = mY = mW = mH = 0;
    }

    Rectangle::Rectangle(int x, int y, int w, int h) :
	mX(x),
	mY(y),
	mW(w),
	mH(h) {
    }

    Rectangle::Rectangle(const Point & p1, const Point & p2) {
	set(p1, p2);
    }

    Rectangle::Rectangle(const Point & point, const Vector & vec) {
	mX = point.get_x();
	mY = point.get_y();
	mW = vec.get_x();
	mH = vec.get_y();
    }

    int
    Rectangle::get_area() const {
	return mW * mH;
    }

    Rectangle &
    Rectangle::translate(const Vector & v) {
	mX += v.get_x();
	mY += v.get_y();
	return *this;
    }

    bool
    Rectangle::contains_point(const Point & p) const {
	return p.get_x() >= mX && p.get_x() < mX + mW && p.get_y() >= mY && p.get_y() < mY + mH;
    }

    bool
    Rectangle::contains_segment(const Segment & s) const {
	return contains_point(s.get_point1()) && contains_point(s.get_point2());
    }

    void
    Rectangle::set(const Point & p1, const Point & p2) {
	if(p1.get_x() < p2.get_x()) {
	    mX = p1.get_x();
	    mW = p2.get_x() - p1.get_x() + 1;
	}
	else {
	    mX = p2.get_x();
	    mW = p1.get_x() - p2.get_x() + 1;
	}

	if(p1.get_y() < p2.get_y()) {
	    mY = p1.get_y();
	    mH = p2.get_y() - p1.get_y() + 1;
	}
	else {
	    mY = p2.get_y();
	    mH = p1.get_y() - p2.get_y() + 1;
	}
    }

    Rectangle
    operator & (const Rectangle & r1, const Rectangle & r2) {
	if(!r1 || !r2) {
	    return Rectangle(0, 0, 0, 0);
	}
	else {
	    int left = max(r1.get_x(), r2.get_x());
	    int right = min(r1.get_xw(), r2.get_xw());
	    int bottom = max(r1.get_y(), r2.get_y());
	    int top = min(r1.get_yh(), r2.get_yh());

	    if(right >= left && top >= bottom) {
		return Rectangle(left, bottom, right - left + 1, top - bottom + 1);
	    }
	    else {
		return Rectangle(0, 0, 0, 0);
	    }
	}
    }

    Rectangle
    operator | (const Rectangle & r1, const Rectangle & r2) {
	if(!r1) {
	    return r2;
	}
	else if(!r2) {
	    return r1;
	}
	else {
	    int left = min(r1.get_x(), r2.get_x());
	    int right = max(r1.get_xw(), r2.get_xw());
	    int bottom = min(r1.get_y(), r2.get_y());
	    int top = max(r1.get_yh(), r2.get_yh());

	    return Rectangle(left, bottom, right - left + 1, top - bottom + 1);
	}
    }

    Rectangle &
    Rectangle::operator &= (const Rectangle & rect) {
	return *this = *this & rect;
    }

    Rectangle &
    Rectangle::operator |= (const Rectangle & rect) {
	return *this = *this | rect;
    }

    bool
    Rectangle::get_intercept(const Segment & segment, Point & point) const {
	return (segment.get_intercept(Segment(Point(mX, mY),
					    Point(mX + mW - 1, mY)), point) ||
		segment.get_intercept(Segment(Point(mX, mY),
					    Point(mX, mY + mH - 1)), point) ||
		segment.get_intercept(Segment(Point(mX, mY + mH - 1),
					    Point(mX + mW - 1, mY + mH - 1)), point) ||
		segment.get_intercept(Segment(Point(mX + mW - 1, mY),
					    Point(mX + mW - 1, mY + mH - 1)), point));
    }

    bool
    operator == (const Rectangle & r1, const Rectangle & r2) {
	return (!r1 && !r2 ||
		r1.get_x() == r2.get_x() &&
		r1.get_y() == r2.get_y() &&
		r1.get_w() == r2.get_w() &&
		r1.get_h() == r2.get_h());
    }

    bool
    operator != (const Rectangle & r1, const Rectangle & r2) {
	return !(r1 == r2);
    }

    ostream &
    operator << (ostream & ost, const Rectangle & rect) {
	ost << "("
	    << rect.get_x() << ", "
	    << rect.get_y() << ", "
	    << rect.get_w() << ", "
	    << rect.get_h()
	    << ")";

	return ost;
    }

    istream &
    operator >> (istream & ist, Rectangle & rect) {
	char c;

        //                                    '('

	ist >> c;

        if(!ist.good()) {
            // probably just an eof
            return ist;
        }

        if(c != '(') {
            ERROR("Expected '('");
            ist.clear(ios::failbit);
            return ist;
        }

        //                                     x

	int x;

	ist >> x;

        if(!ist.good()) {
            ERROR("Expected an int");
            return ist;
        }

        //                                    ','

	ist >> c;

        if(!ist.good()) {
            ERROR("Expected a char");
            return ist;
        }

        if(c != ',') {
            ERROR("Expected ','");
            ist.clear(ios::failbit);
            return ist;
        }

        //                                     y

	int y;

	ist >> y;

        if(!ist.good()) {
            ERROR("Expected an int");
            return ist;
        }

        //                                    ','

	ist >> c;

        if(!ist.good()) {
            ERROR("Expected a char");
            return ist;
        }

        if(c != ',') {
            ERROR("Expected ','");
            ist.clear(ios::failbit);
            return ist;
        }

        //                                     w

	int w;

	ist >> w;

        if(!ist.good()) {
            ERROR("Expected an int");
            return ist;
        }


        //                                    ','

	ist >> c;

        if(!ist.good()) {
            ERROR("Expected a char");
            return ist;
        }

        if(c != ',') {
            ERROR("Expected ','");
            ist.clear(ios::failbit);
            return ist;
        }

        //                                     h

	int h;

	ist >> h;

        if(!ist.good()) {
            ERROR("Expected an int");
            return ist;
        }

        //                                    ')'

	ist >> c;

        if(!ist.good()) {
            ERROR("Expected a char");
            return ist;
        }

        if(c != ')') {
            ERROR("Expected ')'");
            ist.clear(ios::failbit);
            return ist;
        }

        // everything went correctly

	rect.set(x, y, w, h);

	return ist;
    }

} // namespace Drawing
