// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef DRAWING_RECTANGLE_H
#define DRAWING_RECTANGLE_H

// This class of Rectangle embraces the concept of _pixels_.
// As such, the width and height represent the number of pixels
// in that dimension. c.f. the way X does it.
//
//   "01234"	width = 5
//   0+---+
//   1| + |		+ -> the center (2,1)
//   2+---+
//
//   height = 3
//
//   if the left edge is at location x, then the right edge is
//   at location xw, where xw = x + w - 1
//

#include <iosfwd>
using namespace std;

namespace Drawing {

    class Point;
    class Vector;
    class Line;
    class Segment;

    class Rectangle {
    public:
	Rectangle();
	Rectangle(int x, int y, int w, int h);
	Rectangle(const Point & p1, const Point & p2);
	Rectangle(const Point & p, const Vector & v);

	int get_x() const { return mX; }			// left edge
	int get_y() const { return mY; }			// top edge
	int get_w() const { return mW; }			// width
	int get_h() const { return mH; }			// height
	int get_xw() const { return mX + mW - 1; }		// right edge
	int get_yh() const { return mY + mH - 1; }		// bottom edge

	int get_cx() const { return mX + mW / 2; }		// the center in x-axis
	int get_cy() const { return mY + mH / 2; }		// the center in y-axis

	bool contains_point(const Point & p) const;		// point must be within or on the edge
	bool contains_segment(const Segment & s) const;	// both end points must be
							    // within or on the edge

	void set(int x, int y, int w, int h) { mX = x; mY = y; mW = w; mH = h; }
	void set(const Point & p1, const Point & p2);
	void clear() { mX = mY = mW = mH = 0; }

	void set_x(int x) { mX = x; }
	void set_y(int y) { mY = y; }
	void set_w(int w) { mW = w; }
	void set_h(int h) { mH = h; }

	Rectangle & translate(const Vector & v);
	Rectangle & relocate(int x, int y) { mX = x; mY = y; return *this; }
	Rectangle & resize(int w, int h) { mW = w; mH = h; return *this; }
	Rectangle & grow(int dw, int dh) { mW += dw; mH += dh; return *this; }
	Rectangle & shrink(int dw, int dh) { mW -= dw; mH -= dh; return *this; }

	operator bool () const { return get_area() != 0; }
	bool operator ! () const { return !operator bool(); }

	Rectangle & operator &= (const Rectangle & rect);
	Rectangle & operator |= (const Rectangle & rect);

	int get_area() const;

	bool get_intercept(const Segment & segment, Point & point) const;

    private:
	int mX, mY;
	int mW, mH;
    };

    bool operator == (const Rectangle & r1, const Rectangle & r2);
    bool operator != (const Rectangle & r1, const Rectangle & r2);

    Rectangle operator & (const Rectangle & r1, const Rectangle & r2);	// intersection
    Rectangle operator | (const Rectangle & r1, const Rectangle & r2);	// union

    ostream & operator << (ostream & ost, const Rectangle & rect);
    istream & operator >> (istream & ist, Rectangle & rect);


    template<class T>
    class AccumulateAreaUnion {
    public:
	AccumulateAreaUnion(T t) : mT(t) { }

	template<class U>
	Rectangle operator() (const Rectangle & rect, U u) const { return mT(u) | rect; }

    private:
	T mT;
    };

    template<class T> AccumulateAreaUnion<T> unite_area(T t) {
	return AccumulateAreaUnion<T>(t);
    }

    template<class T>
    class AccumulateAreaIntersect {
    public:
	AccumulateAreaIntersect(T t) : mT(t) { }

	template<class U>
	Rectangle operator() (const Rectangle & rect, U u) const { return mT(u) & rect; }

    private:
	T mT;
    };

    template<class T> AccumulateAreaIntersect<T> intersect_area(T t) {
	return AccumulateAreaIntersect<T>(t);
    }

} // namespace Drawing

#endif // DRAWING_RECTANGLE_H
