// Dachshund
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef FOUNDATION_COMMANDS_H
#define FOUNDATION_COMMANDS_H

#include "command.h"

#include "macros.h"

#include <functional>
#include <algorithm>
#include <list>
#include <set>
using namespace std;

template<class Obj, class Arg>
class CommandFlipArgument_s : public Command {
public:
    CommandFlipArgument_s(Obj * object, void (Obj::*method)(const set<Arg *> &),
			  const set<Arg *> & arg1, const set<Arg *> & arg2) : Command(),
	mObject(object),
	mMethod(method),
	mArg1(arg1),
	mArg2(arg2) {
	ASSERT(mObject);
	ASSERT(mMethod);
	mObject->ref();

	for_each(mArg1.begin(), mArg1.end(), mem_fun(&Object::ref));
	for_each(mArg2.begin(), mArg2.end(), mem_fun(&Object::ref));
    }

    void
    execute() {
	(mObject->*mMethod)(mArg1);
    }

    void
    unexecute() {
	(mObject->*mMethod)(mArg2);
    }

protected:
    ~CommandFlipArgument_s() {
	for_each(mArg1.begin(), mArg1.end(), mem_fun(&Object::unref));
	for_each(mArg2.begin(), mArg2.end(), mem_fun(&Object::unref));

	mObject->unref();
    }

private:
    Obj * mObject;
    void (Obj::*mMethod)(const set<Arg *> &);
    set<Arg *> mArg1;
    set<Arg *> mArg2;
};

template<class Obj, class Arg>
CommandFlipArgument_s<Obj, Arg> * command_flip_argument(Obj * object,
							void (Obj::*method)(const set<Arg *> &),
							const set<Arg *> & arg1,
							const set<Arg *> & arg2) {
    return new CommandFlipArgument_s<Obj, Arg>(object, method, arg1, arg2);
}

//
//
//

template<class Obj, class Arg>
class CommandFlipArgument_r : public Command {
public:
    CommandFlipArgument_r(Obj * object, void (Obj::*method)(const Arg &),
			 const Arg & arg1, const Arg & arg2) : Command(),
	mObject(object),
	mMethod(method),
	mArg1(arg1),
	mArg2(arg2) {
	ASSERT(mObject);
	ASSERT(mMethod);
	mObject->ref();
    }

    void
    execute() {
	(mObject->*mMethod)(mArg1);
    }

    void
    unexecute() {
	(mObject->*mMethod)(mArg2);
    }

protected:
    ~CommandFlipArgument_r() {
	mObject->unref();
    }

private:
    Obj * mObject;
    void (Obj::*mMethod)(const Arg &);
    Arg mArg1;
    Arg mArg2;
};

template<class Obj, class Arg>
CommandFlipArgument_r<Obj, Arg> * command_flip_argument(Obj * object,
						       void (Obj::*method)(const Arg &),
						       const Arg & arg1, const Arg & arg2) {
    return new CommandFlipArgument_r<Obj, Arg>(object, method, arg1, arg2);
}

//
//
//

template<class Obj, class Arg>
class CommandFlipArgument_v : public Command {
public:
    CommandFlipArgument_v(Obj * object, void (Obj::*method)(Arg),
			 Arg arg1, Arg arg2) : Command(),
	mObject(object),
	mMethod(method),
	mArg1(arg1),
	mArg2(arg2) {
	ASSERT(mObject);
	ASSERT(mMethod);
	mObject->ref();
    }

    void
    execute() {
	(mObject->*mMethod)(mArg1);
    }

    void
    unexecute() {
	(mObject->*mMethod)(mArg2);
    }

protected:
    ~CommandFlipArgument_v() {
	mObject->unref();
    }

private:
    Obj * mObject;
    void (Obj::*mMethod)(Arg);
    Arg mArg1;
    Arg mArg2;
};

template<class Obj, class Arg>
CommandFlipArgument_v<Obj, Arg> * command_flip_argument(Obj * object, void (Obj::*method)(Arg),
						       Arg arg1, Arg arg2) {
    return new CommandFlipArgument_v<Obj, Arg>(object, method, arg1, arg2);
}

//
//
//

template<class Obj, class Arg>
class CommandFlipArgument_l : public Command {
public:
    CommandFlipArgument_l(Obj * object, void (Obj::*method)(const list<Arg *> &),
			  const list<Arg *> & arg1, const list<Arg *> & arg2) : Command(),
	mObject(object),
	mMethod(method),
	mArg1(arg1),
	mArg2(arg2) {
	ASSERT(mObject);
	ASSERT(mMethod);
	mObject->ref();

	for_each(mArg1.begin(), mArg1.end(), mem_fun(&Object::ref));
	for_each(mArg2.begin(), mArg2.end(), mem_fun(&Object::ref));
    }

    void
    execute() {
	(mObject->*mMethod)(mArg1);
    }

    void
    unexecute() {
	(mObject->*mMethod)(mArg2);
    }

protected:
    ~CommandFlipArgument_l() {
	for_each(mArg1.begin(), mArg1.end(), mem_fun(&Object::unref));
	for_each(mArg2.begin(), mArg2.end(), mem_fun(&Object::unref));

	mObject->unref();
    }

private:
    Obj * mObject;
    void (Obj::*mMethod)(const list<Arg *> &);
    list<Arg *> mArg1;
    list<Arg *> mArg2;
};

template<class Obj, class Arg>
CommandFlipArgument_l<Obj, Arg> * command_flip_argument(Obj * object,
							void (Obj::*method)(const list<Arg *> &),
							const list<Arg *> & arg1,
							const list<Arg *> & arg2) {
    return new CommandFlipArgument_l<Obj, Arg>(object, method, arg1, arg2);
}

//
//
//

template<class Obj, class Arg>
class CommandFlipArgument_p : public Command {
public:
    CommandFlipArgument_p(Obj * object, void (Obj::*method)(Arg *),
			    Arg * arg1, Arg * arg2) : Command(),
	mObject(object),
	mMethod(method),
	mArg1(arg1),
	mArg2(arg2) {
	ASSERT(mObject);
	mObject->ref();

	if(mArg1) {
	    mArg1->ref();
	}

	if(mArg2) {
	    mArg2->ref();
	}
    }

    void
    execute() {
	(mObject->*mMethod)(mArg1);
    }

    void
    unexecute() {
	(mObject->*mMethod)(mArg2);
    }

protected:
    ~CommandFlipArgument_p() {
	if(mArg1) {
	    mArg1->unref();
	}

	if(mArg2) {
	    mArg2->unref();
	}

	mObject->unref();
    }

private:
    Obj * mObject;
    void (Obj::*mMethod)(Arg *);
    Arg * mArg1;
    Arg * mArg2;
};

template<class Obj, class Arg>
CommandFlipArgument_p<Obj, Arg> * command_flip_argument(Obj * object, void (Obj::*method)(Arg *),
						       Arg * arg1, Arg * arg2) {
    return new CommandFlipArgument_p<Obj, Arg>(object, method, arg1, arg2);
}

//
//
//

template<class Obj, class Arg>
class CommandFlipArgument_pc : public Command {
public:
    CommandFlipArgument_pc(Obj * object, void (Obj::*method)(const Arg *),
			 Arg * arg1, Arg * arg2) : Command(),
	mObject(object),
	mMethod(method),
	mArg1(arg1),
	mArg2(arg2) {
	ASSERT(mObject);
	mObject->ref();

	if(mArg1) {
	    mArg1->ref();
	}

	if(mArg2) {
	    mArg2->ref();
	}
    }

    void
    execute() {
	(mObject->*mMethod)(mArg1);
    }

    void
    unexecute() {
	(mObject->*mMethod)(mArg2);
    }

protected:
    ~CommandFlipArgument_pc() {
	if(mArg1) {
	    mArg1->unref();
	}

	if(mArg2) {
	    mArg2->unref();
	}

	mObject->unref();
    }

private:
    Obj * mObject;
    void (Obj::*mMethod)(const Arg *);
    Arg * mArg1;
    Arg * mArg2;
};

template<class Obj, class Arg>
CommandFlipArgument_pc<Obj, Arg> * command_flip_argument(Obj * object,
						       void (Obj::*method)(const Arg *),
						       Arg * arg1, Arg * arg2) {
    return new CommandFlipArgument_pc<Obj, Arg>(object, method, arg1, arg2);
}

//
//
//

template<class Obj, class Arg>
class CommandFlipMethod_r : public Command {
public:
    CommandFlipMethod_r(Obj * object,
		       void (Obj::*method1)(const Arg &),
		       void (Obj::*method2)(const Arg &),
		       const Arg & arg) : Command(),
	mObject(object),
	mMethod1(method1),
	mMethod2(method2),
	mArg(arg) {
	ASSERT(mMethod1);
	ASSERT(mMethod2);
	ASSERT(mObject);
	mObject->ref();
    }

    void
    execute() {
	(mObject->mMethod1)(arg);
    }

    void
    unexecute() {
	(mObject->mMethod2)(arg);
    }

protected:
    ~CommandFlipMethod_r() {
	mObject->unref();
    }

private:
    Obj * mObject;
    void (Obj::*mMethod1)(const Arg &);
    void (Obj::*mMethod2)(const Arg &);
    Arg mArg;
};

template<class Obj, class Arg>
CommandFlipMethod_r<Obj, Arg> * command_flip_method(Obj * object,
						   void (Obj::*method1)(const Arg &),
						   void (Obj::*method2)(const Arg &),
						   const Arg & arg) {
    return new CommandFlipMethod_r<Obj, Arg>(object, method1, method2, arg);
}

//
//
//

template<class Obj, class Arg>
class CommandFlipMethod_p1 : public Command {
public:
    CommandFlipMethod_p1(Obj * object,
			void (Obj::*method1)(Arg *),
			void (Obj::*method2)(Arg *),
			Arg * arg) : Command(),
	mObject(object),
	mMethod1(method1),
	mMethod2(method2),
	mArg(arg) {
	ASSERT(mArg);
	mArg->ref();
	mObject->ref();
    }

    void
    execute() {
	(mObject->*mMethod1)(mArg);
    }

    void
    unexecute() {
	(mObject->*mMethod2)(mArg);
    }

protected:
    ~CommandFlipMethod_p1() {
	mObject->unref();
	mArg->unref();
    }

private:
    Obj * mObject;
    void (Obj::*mMethod1)(Arg *);
    void (Obj::*mMethod2)(Arg *);
    Arg * mArg;
};


template<class Obj, class Arg1, class Arg2>
class CommandFlipMethod_p2 : public Command {
public:
    CommandFlipMethod_p2(Obj * object,
			void (Obj::*method1)(Arg1 *, Arg2),
			void (Obj::*method2)(Arg1 *),
			Arg1 * arg1, Arg2 arg2) : Command(),
	mObject(object),
	mMethod1(method1),
	mMethod2(method2),
	mArg1(arg1),
	mArg2(arg2) {
	ASSERT(mArg1);
	ASSERT(mMethod1);
	ASSERT(mMethod2);
	mObject->ref();
	mArg1->ref();
    }

    void
    execute() {
	(mObject->*mMethod1)(mArg1, mArg2);
    }

    void
    unexecute() {
	(mObject->*mMethod2)(mArg1);
    }

protected:
    ~CommandFlipMethod_p2() {
	mArg1->unref();
	mObject->unref();
    }

private:
    Obj * mObject;
    void (Obj::*mMethod1)(Arg1 *, Arg2);
    void (Obj::*mMethod2)(Arg1 *);
    Arg1 * mArg1;
    Arg2 mArg2;
};

template<class Obj, class Arg1, class Arg2>
class CommandFlipMethod_p3 : public Command {
public:
    CommandFlipMethod_p3(Obj * object,
			void (Obj::*method1)(Arg1 *),
			void (Obj::*method2)(Arg1 *, Arg2), Arg1 * arg1, Arg2 arg2) : Command(),
	mObject(object),
	mMethod1(method1),
	mMethod2(method2),
	mArg1(arg1),
	mArg2(arg2) {
	ASSERT(mArg1);
	ASSERT(mMethod1);
	ASSERT(mMethod2);
	mObject->ref();
	mArg1->ref();
    }

    void
    execute() {
	(mObject->*mMethod1)(mArg1);
    }

    void
    unexecute() {
	(mObject->*mMethod2)(mArg1, mArg2);
    }

protected:
    ~CommandFlipMethod_p3() {
	mArg1->unref();
	mObject->unref();
    }

private:
    Obj * mObject;
    void (Obj::*mMethod1)(Arg1 *);
    void (Obj::*mMethod2)(Arg1 *, Arg2);
    Arg1 * mArg1;
    Arg2 mArg2;
};

template<class Obj, class Arg1, class Arg2>
class CommandFlipMethod_p4 : public Command {
public:
    CommandFlipMethod_p4(Obj * object,
			void (Obj::*method1)(Arg1 *, Arg2),
			void (Obj::*method2)(Arg1 *, Arg2),
			Arg1 * arg1, Arg2 arg2) : Command(),
	mObject(object),
	mMethod1(method1),
	mMethod2(method2),
	mArg1(arg1),
	mArg2(arg2) {
	ASSERT(mArg1);
	ASSERT(mMethod1);
	ASSERT(mMethod2);
	mObject->ref();
	mArg1->ref();
    }

    void
    execute() {
	(mObject->*mMethod1)(mArg1, mArg2);
    }

    void
    unexecute() {
	(mObject->*mMethod2)(mArg1, mArg2);
    }

protected:
    ~CommandFlipMethod_p4() {
	mArg1->unref();
	mObject->unref();
    }

private:
    Obj * mObject;
    void (Obj::*mMethod1)(Arg1 *, Arg2);
    void (Obj::*mMethod2)(Arg1 *, Arg2);
    Arg1 * mArg1;
    Arg2 mArg2;
};

template<class Obj, class Arg>
CommandFlipMethod_p1<Obj, Arg> * command_flip_method(Obj * object,
						    void (Obj::*method1)(Arg *),
						    void (Obj::*method2)(Arg *),
						    Arg * arg) {
    return new CommandFlipMethod_p1<Obj, Arg>(object, method1, method2, arg);
}

template<class Obj, class Arg1, class Arg2>
CommandFlipMethod_p2<Obj, Arg1, Arg2> * command_flip_method(Obj * object,
							   void (Obj::*method1)(Arg1 *, Arg2),
							   void (Obj::*method2)(Arg1 *),
							   Arg1 * arg1, Arg2 arg2) {
    return new CommandFlipMethod_p2<Obj, Arg1, Arg2>(object, method1, method2, arg1, arg2);
}

template<class Obj, class Arg1, class Arg2>
CommandFlipMethod_p3<Obj, Arg1, Arg2> * command_flip_method(Obj * object,
							   void (Obj::*method1)(Arg1 *),
							   void (Obj::*method2)(Arg1 *, Arg2),
							   Arg1 * arg1, Arg2 arg2) {
    return new CommandFlipMethod_p3<Obj, Arg1, Arg2>(object, method1, method2, arg1, arg2);
}

template<class Obj, class Arg1, class Arg2>
CommandFlipMethod_p4<Obj, Arg1, Arg2> * command_flip_method(Obj * object,
							   void (Obj::*method1)(Arg1 *, Arg2),
							   void (Obj::*method2)(Arg1 *, Arg2),
							   Arg1 * arg1, Arg2 arg2) {
    return new CommandFlipMethod_p4<Obj, Arg1, Arg2>(object, method1, method2, arg1, arg2);
}

//
//
//

template<class Obj, class Arg>
class CommandReverseArgument_p : public Command {
public:
    CommandReverseArgument_p(Obj * object,
			    void (Obj::*method)(Arg *, Arg *),
			    Arg * arg1, Arg * arg2) : Command(),
    mObject(object),
    mMethod(method),
    mArg1(arg1),
    mArg2(arg2) {
	ASSERT(mObject);
	mObject->ref();

	if(mArg1) {
	    mArg1->ref();
	}

	if(mArg2) {
	    mArg2->ref();
	}
    }

    void
    execute() {
	(mObject->*mMethod)(mArg1, mArg2);
    }

    void
    unexecute() {
	(mObject->*mMethod)(mArg2, mArg1);
    }

protected:
    ~CommandReverseArgument_p() {
	if(mArg1) {
	    mArg1->unref();
	}

	if(mArg2) {
	    mArg2->unref();
	}

	mObject->unref();
    }

private:
    Obj * mObject;
    void (Obj::*mMethod)(Arg *, Arg *);
    Arg * mArg1;
    Arg * mArg2;
};

template<class Obj, class Arg>
CommandReverseArgument_p<Obj, Arg> * command_reverse_argument(Obj * object,
							     void (Obj::*method)(Arg *, Arg *),
							     Arg * arg1, Arg * arg2) {
    return new CommandReverseArgument_p<Obj, Arg>(object, method, arg1, arg2);
}

#endif // FOUNDATION_COMMANDS_H
