// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "object.h"

#include "macros.h"
#include "undo_manager.h"

#include <iostream>
#include <typeinfo>
using namespace std;

#ifdef DEBUG

#include <algorithm>
#include <functional>

#include <signal.h>
#include <sys/types.h>

set<Object *> Object::mLiveObjects;
set<string> Object::mWatchedObjects;

Object::Object() : mCounter(1) {
    mLiveObjects.insert(this);
    TRACE(10, "Object created");
}

Object::Object(const Object & object) : mCounter(1) {
    mLiveObjects.insert(this);
    TRACE(10, "Object copied");
}

Object::~Object() {
    TRACE(10, "Object destroyed");
    mLiveObjects.erase(this);
}

void
Object::watch_type(const type_info & type) {
    mWatchedObjects.insert(type.name());
}

void
Object::ref() throw() {
    ASSERT(this);

    ++mCounter;

    if(!mWatchedObjects.empty()) {
	const string & name = typeid(*this).name();

	if(mWatchedObjects.find(name) != mWatchedObjects.end()) {
	    cout
		<< " ++\t" << name << "\t" << static_cast<void *>(this)
		<< "\t" << mCounter << endl;
	}
    }
}

void
Object::unref() {
    ASSERT(this);

    if(!mWatchedObjects.empty()) {
	const string & name = typeid(*this).name();

	if(mWatchedObjects.find(name) != mWatchedObjects.end()) {
	    cout
		<< " --\t" << name << "\t" << static_cast<void *>(this)
		<< "\t" << mCounter << endl;
	    // kill(0, SIGTRAP);
	}
    }

#  if 0
    if(mCounter < 1) {
	cerr
	    << "Unreferencing unreferenced object: " << flush
	    << typeid(*this).name() << endl;

	abort();
    }
#  else
    if(mLiveObjects.find(this) == mLiveObjects.end()) {
	cerr
	    << "Unreferencing unreferenced object: "
	    << this << ", " << typeid(*this).name() << endl;

	abort();
    }
#  endif

    if(--mCounter == 0) {
#  if 0
	this->~Object();
#  else
	delete this;
#  endif
    }
}
#else	// DEBUG
Object::~Object() {
}
#endif // DEBUG

void
Object::begin_transaction() {
    UndoManager::instance()->begin_transaction();
}

void
Object::end_transaction() {
    UndoManager::instance()->end_transaction();
}

void
Object::log_command(Command * command) {
    UndoManager::instance()->log_command(command);
}

void
Object::log_change(ChangeEvent * change_event) {
    UndoManager::instance()->log_change(change_event);
}

void
Object::show_details() {
    cout
	<< "Reference count: " << mCounter << ",    Type name: " << flush
	<< typeid(*this).name() << endl;
}

void
Object::show_remaining() {
#ifdef DEBUG
    cout << "Remaining object count: " << mLiveObjects.size() << endl;
    for_each(mLiveObjects.begin(), mLiveObjects.end(), mem_fun(&Object::show_details));
#else
    cout << "Statistics only available in debug mode" << endl;
#endif
}
