// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "undo_manager.h"

#include "macros.h"
#include "composite_command.h"
#include "change_event.h"
#include "change_listener.h"
#include "undo_event.h"
#include "undo_listener.h"
#include "vetoable_change_listener.h"

#include <algorithm>
#include <functional>
#include <iostream>
using namespace std;

#include "functional.h"

UndoManager * UndoManager::mInstance = 0;

const unsigned int UndoManager::DEFAULT_UNDO_LIMIT = 0;

UndoManager *
UndoManager::instance() {
    if(!mInstance) {
	mInstance = new UndoManager();
    }

    return mInstance;
}

void
UndoManager::shutdown() {
    ASSERT(mInstance);

    delete mInstance;

    mInstance = 0;
}

UndoManager::UndoManager() : Object(),
    mUndoLimit(DEFAULT_UNDO_LIMIT),
    mTransactionDepth(0),
    mTransactionsForbidden(false),
    mUndoingOrRedoing(false) {
}

UndoManager::~UndoManager() {
    mTransactionsForbidden = true;

    for_each(mPast.begin(), mPast.end(), mem_fun(&Object::unref));
    for_each(mFuture.begin(), mFuture.end(), mem_fun(&Object::unref));
    for_each(mChangeEvents.begin(), mChangeEvents.end(), mem_fun(&Object::unref));
}

void
UndoManager::begin_transaction() {
    ASSERT(!mTransactionsForbidden);

    if(mTransactionDepth++ == 0) {
	if(!mUndoingOrRedoing) {
	    // If there was an "old future" then it is now invalid.
	    // This happens when the user does some stuff, undoes it and then
	    // does something different
	    for_each(mFuture.begin(), mFuture.end(), mem_fun(&Object::unref));
	    mFuture.clear();

	    mPast.push_back(new CompositeCommand());

	    if(mUndoLimit > 0 && mPast.size() > mUndoLimit) {
		MESSAGE("Undo limit exceeded (" << mPast.size() << "), " << "popping!");
		mPast.front()->unref();
		mPast.pop_front();
	    }
	}
    }
}

void
UndoManager::end_transaction() {
    ASSERT(!mTransactionsForbidden);

    if(--mTransactionDepth == 0) {
	// ensure consistency between changes and commands
	//ASSERT(mChangeEvents.empty() == mPast.back()->is_empty());	// FIXME this fails

	if(mChangeEvents.empty()) {
	    // no changes occurred, delete the composite command

	    mPast.back()->unref();
	    mPast.pop_back();
	}
	else {
	    flush_changes();
	}
    }
}

void
UndoManager::log_command(Command * command) {
    RETURN_IF_FAIL(command);
    RETURN_IF_FAIL(mTransactionDepth > 0);

    if(mUndoingOrRedoing) {
	// ignore the commands
    }
    else {
	mPast.back()->add(command);
    }
}

void
UndoManager::log_change(ChangeEvent * change_event) {
    RETURN_IF_FAIL(change_event);
    RETURN_IF_FAIL(mTransactionDepth > 0);

    change_event->ref();
    mChangeEvents.push_back(change_event);
}

void
UndoManager::undo() {
    RETURN_IF_FAIL(mTransactionDepth == 0);
    RETURN_IF_FAIL(!mPast.empty());

    mUndoingOrRedoing = true;
    begin_transaction();
    mPast.back()->unexecute();
    end_transaction();
    mUndoingOrRedoing = false;

    mFuture.push_front(mPast.back());
    mPast.pop_back();

    //

    UndoEvent undo_event(mPast.size(), mFuture.size());

    for_each(mUndoListeners.begin(), mUndoListeners.end(),
	     bind2nd(mem_fun(&UndoListener::undo_event), &undo_event));
}

void
UndoManager::redo() {
    RETURN_IF_FAIL(mTransactionDepth == 0);
    RETURN_IF_FAIL(!mFuture.empty());

    mUndoingOrRedoing = true;
    begin_transaction();
    mFuture.front()->execute();
    end_transaction();
    mUndoingOrRedoing = false;

    mPast.push_back(mFuture.front());
    mFuture.pop_front();

    if(mUndoLimit > 0 && mPast.size() > mUndoLimit) {
	mPast.front()->unref();
	mPast.pop_front();
    }

    //

    UndoEvent undo_event(mPast.size(), mFuture.size());

    for_each(mUndoListeners.begin(), mUndoListeners.end(),
	     bind2nd(mem_fun(&UndoListener::undo_event), &undo_event));
}

bool
UndoManager::can_undo() {
    return mTransactionDepth == 0 && !mPast.empty();
}

bool
UndoManager::can_redo() {
    return mTransactionDepth == 0 && !mFuture.empty();
}

void
UndoManager::clear() {
    RETURN_IF_FAIL(mTransactionDepth == 0);

    MESSAGE("Clearing the undo past (" << mPast.size() << ") " <<
	    "and future (" << mFuture.size() << ")");

    mTransactionsForbidden = true;

    for_each(mPast.begin(), mPast.end(), mem_fun(&Object::unref));
    mPast.clear();

    for_each(mFuture.begin(), mFuture.end(), mem_fun(&Object::unref));
    mFuture.clear();

    //

    UndoEvent undo_event(mPast.size(), mFuture.size());

    for_each(mUndoListeners.begin(), mUndoListeners.end(),
	     bind2nd(mem_fun(&UndoListener::undo_event), &undo_event));

    //

    mTransactionsForbidden = false;
}

void
UndoManager::clear_undo() {
    RETURN_IF_FAIL(mTransactionDepth == 0);

    MESSAGE("Clearing the undo past (" << mPast.size() << ")");

    mTransactionsForbidden = true;

    for_each(mPast.begin(), mPast.end(), mem_fun(&Object::unref));
    mPast.clear();

    //

    UndoEvent undo_event(mPast.size(), mFuture.size());

    for_each(mUndoListeners.begin(), mUndoListeners.end(),
	     bind2nd(mem_fun(&UndoListener::undo_event), &undo_event));

    //

    mTransactionsForbidden = false;
}

void
UndoManager::clear_redo() {
    RETURN_IF_FAIL(mTransactionDepth == 0);

    MESSAGE("Clearing the redo future (" << mFuture.size() << ")");

    mTransactionsForbidden = true;

    for_each(mFuture.begin(), mFuture.end(), mem_fun(&Object::unref));
    mFuture.clear();

    //

    UndoEvent undo_event(mPast.size(), mFuture.size());

    for_each(mUndoListeners.begin(), mUndoListeners.end(),
	     bind2nd(mem_fun(&UndoListener::undo_event), &undo_event));

    //

    mTransactionsForbidden = false;
}

void
UndoManager::set_undo_limit(unsigned int i) {
    mUndoLimit = i;

    if(mUndoLimit > 0) {
	mTransactionsForbidden = true;

	while(mPast.size() > mUndoLimit) {
	    mPast.front()->unref();
	    mPast.pop_front();
	}

	mTransactionsForbidden = false;
    }

    //

    UndoEvent undo_event(mPast.size(), mFuture.size());

    for_each(mUndoListeners.begin(), mUndoListeners.end(),
	     bind2nd(mem_fun(&UndoListener::undo_event), &undo_event));
}

void
UndoManager::add(UndoListener * undo_listener) {
    mUndoListeners.push_front(undo_listener);
}

void
UndoManager::remove(UndoListener * undo_listener) {
    mUndoListeners.remove(undo_listener);
}

void
UndoManager::add_change_listener(ChangeListener * change_listener) {
    mChangeListeners.push_front(change_listener);
}

void
UndoManager::remove_change_listener(ChangeListener * change_listener) {
    mChangeListeners.remove(change_listener);
}

void
UndoManager::add_vetoable_change_listener(VetoableChangeListener * change_listener) {
    mVetoableChangeListeners.push_front(change_listener);
}

void
UndoManager::remove_vetoable_change_listener(VetoableChangeListener * change_listener) {
    mVetoableChangeListeners.remove(change_listener);
}

void
UndoManager::flush_changes() {
    RETURN_IF_FAIL(mTransactionDepth == 0);

    // we only tell the Vetoable listeners of the changes if we are not undoing/redoing
    if(!mUndoingOrRedoing) {
	try {
	    for_each(mVetoableChangeListeners.begin(), mVetoableChangeListeners.end(),
		     bind2nd_vr(mem_fun(&VetoableChangeListener::changes_occurred), mChangeEvents));
	}
	catch(Veto &) {
	    // if we are here then one of the vetoable listeners didn't like
	    // the change, so we back it out
	    for_each(mChangeEvents.begin(), mChangeEvents.end(), mem_fun(&Object::unref));
	    mChangeEvents.clear();

	    mUndoingOrRedoing = true;
	    mPast.back()->unexecute();
	    mUndoingOrRedoing = false;
	    mPast.back()->unref();
	    mPast.pop_back();

	    throw;		// now the user code can handle it
	}
    }

    // if we get here then all of the vetoable listeners were happy with
    // the changes, so we proceed to tell the non-vetoable listeners

    mTransactionsForbidden = true;

    for_each(mChangeListeners.begin(), mChangeListeners.end(),
	     bind2nd_vr(mem_fun(&ChangeListener::changes_occurred), mChangeEvents));

    for_each(mChangeEvents.begin(), mChangeEvents.end(), mem_fun(&Object::unref));
    mChangeEvents.clear();

    //

    UndoEvent undo_event(mPast.size(), mFuture.size());

    for_each(mUndoListeners.begin(), mUndoListeners.end(),
	     bind2nd(mem_fun(&UndoListener::undo_event), &undo_event));

    //

    mTransactionsForbidden = false;
}
