// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "class.h"

#include "mof/class.h"
#include "mof/reference.h"
#include "mof/attribute.h"
#include "mof/data_type.h"
#include "mof/association_end.h"
#include "mof/association.h"

#include "common/file_utilities.h"
#include "common/diff_stream.h"
#include "common/cpp_stream.h"

#include "foundation/macros.h"
#include "foundation/utility.h"

#include "utility.h"

#include <typeinfo>
#include <iostream>
using namespace std;

namespace Generator {

    Class::Class(MOF::Class * klass, const string & model_name) : Object(),
        mClass(klass),
	mModelName(model_name),
	mDetailsLoaded(false) {
	ASSERT(mClass);
    }

    Class::~Class() {
    }

    void
    Class::generate(const string & output_dir, const string & include_dir) {
	if(!mDetailsLoaded) {
	    load_details();
	    mDetailsLoaded = true;
	}

	string full_path = output_dir + "/" + mClassDetails.qualified_path;

	Common::DiffStream hpp_stream(full_path + ".h");
	ASSERT(hpp_stream);
	Common::CppStream hpp(hpp_stream);

	Common::DiffStream cpp_stream(full_path + ".cc");
	ASSERT(cpp_stream);
	Common::CppStream cpp(cpp_stream);

	generate_header(hpp);
	generate_source(cpp);
    }

    void
    Class::generate_header(Common::CppStream & hpp) {
	string guard = path_to_guard(mClassDetails.qualified_path);

	hpp << get_prologue()
	    << "\n#ifndef " << guard
	    << "\n#define " << guard << "\n";

	generate_header_includes(hpp);

	hpp << "\nnamespace " << mModelName << " {\n";

	generate_header_forward_declarations(hpp);

	hpp << "\nclass " << mClassDetails.name << " : ";

	if(mSupertypeDetails.empty()) {
	    hpp << "public REF::Class";
	}
	else {
	    vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();

	    for(;;) {
		hpp << (iter->is_virtual ? "virtual " : "") <<  "public " << iter->name;

		++iter;

		if(iter == mSupertypeDetails.end()) {
		    break;
		}
		else {
		    hpp << ", ";
		}
	    }
	}

	hpp << " {";

	hpp << "\npublic:";
	generate_header_public_members(hpp);

	hpp << "\nprotected:";
	generate_header_protected_members(hpp);

	hpp << "\nprivate:";
	generate_header_private_members(hpp);

	hpp << "\n};\n";

	hpp << "\n} // namespace " << mModelName << "\n";

	hpp << "\n#endif // " << guard << "\n";
    }

    void
    Class::generate_header_includes(Common::CppStream & hpp) {
	set<string> hpp_includes;
	set<string> std_includes;

	for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
	    iter != mReferenceDetails.end();
	    ++iter) {
	    if(iter->other_cardinality == MULTI) {
		if(iter->other_ordering == ORDERED) {
		    std_includes.insert("<list>");
		}
		else if(iter->other_ordering == UNORDERED) {
		    std_includes.insert("<set>");
		}
		else {
		    ASSERT_NOT_REACHED();
		}
	    }
	}

	if(mSupertypeDetails.empty()) {
	    hpp_includes.insert("\"reflective/class.h\"");
	}
	else {
	    for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
		iter != mSupertypeDetails.end();
		++iter) {
		if(iter->klass->get_container() == mClassDetails.klass->get_container()) {
		    // the are in the same package
		    hpp_includes.insert("\"" + iter->role + ".h\"");
		}
		else {
		    hpp_includes.insert("\"" + iter->qualified_path + ".h\"");
		}
	    }
	}

	if(!hpp_includes.empty()) {
	    for(set<string>::iterator iter = hpp_includes.begin();
		iter != hpp_includes.end();
		++iter) {
		const string & include = *iter;
		hpp << "\n#include " << include;
	    }

	    hpp << "\n";
	}

	if(!std_includes.empty()) {
	    for(set<string>::iterator iter = std_includes.begin();
		iter != std_includes.end();
		++iter) {
		const string & include = *iter;
		hpp << "\n#include " << include;
	    }

	    hpp << "\nusing namespace std;\n";
	}
    }

    void
    Class::generate_header_forward_declarations(Common::CppStream & hpp) {
	set<string> forward_declarations;

	for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
	    iter != mReferenceDetails.end();
	    ++iter) {
	    forward_declarations.insert(iter->other_type_name);
	}

	for(vector<AttributeDetails>::iterator iter = mAttributeDetails.begin();
	    iter != mAttributeDetails.end();
	    ++iter) {
	    forward_declarations.insert(iter->type_name);
	}

	if(!forward_declarations.empty()) {
	    for(set<string>::iterator iter = forward_declarations.begin();
		iter != forward_declarations.end();
		++iter) {
		hpp << "\nclass " << *iter << ";";
	    }

	    hpp << "\n";
	}
    }

    void
    Class::generate_header_public_members(Common::CppStream & hpp) {
	// constructor/destructor

	hpp << "\n//"
	    << "\n// constructor/destructor"
	    << "\n//\n";

	hpp << "\n" << mClassDetails.name << "();"
	    << "\n~" << mClassDetails.name << "();\n";

	// access to roles and role containers

	hpp << "\n//"
	    << "\n// access to roles and role containers"
	    << "\n//\n";

	for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
	    iter != mReferenceDetails.end();
	    ++iter) {

	    if(iter->other_cardinality == UNARY) {
		if(iter->other_end_navigable) {
		    hpp << "\n" << iter->other_type_name << " * get_"
			<< iter->other_end_role << "() const "
			<< "{ return " << iter->other_end_member << "; }"
			<< "\nvoid set_" << iter->other_end_role
			<< "(" << iter->other_type_name << " * " << iter->other_end_variable << ");"
			<< "\n";
		}

		if(iter->this_end_navigable) {
		    hpp << "\nvoid internal_set_" << iter->other_end_role
			<< "(" << iter->other_type_name
			<< " * " << iter->other_end_variable << ");\n";
		}
	    }
	    else if(iter->other_cardinality == MULTI) {
		if(iter->other_ordering == UNORDERED) {
		    if(iter->other_end_navigable) {
			hpp << "\nconst set<" << iter->other_type_name << " *> & get_"
			    << iter->other_end_role << "() const "
			    << "{ return " << iter->other_end_member << "; }"
			    << "\nvoid set_" << iter->other_end_role
			    << "(const set<" << iter->other_type_name << " *> & "
			    << iter->other_end_variable
			    << ");"
			    << "\nvoid add_" << iter->other_end_role << "("
			    << iter->other_type_name
			    << " * " << iter->other_end_variable << ");"
			    << "\nvoid remove_" << iter->other_end_role << "("
			    << iter->other_type_name
			    << " * " << iter->other_end_variable << ");\n";
		    }

		    if(iter->this_end_navigable) {
			hpp << "\nvoid internal_add_" << iter->other_end_role
			    << "(" << iter->other_type_name
			    << " * " << iter->other_end_variable << ");";
			hpp << "\nvoid internal_remove_" << iter->other_end_role
			    << "(" << iter->other_type_name
			    << " * " << iter->other_end_variable << ");\n";
		    }
		}
		else if(iter->other_ordering == ORDERED) {
		    if(iter->other_end_navigable) {
			hpp << "\nconst list<" << iter->other_type_name << " *> & get_"
			    << iter->other_end_role << "() const "
			    << "{ return " << iter->other_end_member << "; }"
			    << "\nvoid set_" << iter->other_end_role
			    << "(const list<" << iter->other_type_name << " *> & "
			    << iter->other_end_variable
			    << ");"
			    << "\nvoid add_" << iter->other_end_role << "("
			    << iter->other_type_name
			    << " * " << iter->other_end_variable << ", int index = -1);"
			    << "\nvoid remove_" << iter->other_end_role
			    << "(" << iter->other_type_name
			    << " * " << iter->other_end_variable << ");\n";
		    }

		    if(iter->this_end_navigable) {
			hpp << "\nvoid internal_add_" << iter->other_end_role
			    << "(" << iter->other_type_name
			    << " * " << iter->other_end_variable << ");";
			hpp << "\nvoid internal_remove_" << iter->other_end_role
			    << "(" << iter->other_type_name
			    << " * " << iter->other_end_variable << ");\n";
		    }
		}
		else {
		    ASSERT_NOT_REACHED();
		}
	    }
	    else {
		ASSERT_NOT_REACHED();
	    }
	}

	// access to attributes

	hpp << "\n//"
	    << "\n// access to attributes"
	    << "\n//\n";

	for(vector<AttributeDetails>::iterator iter = mAttributeDetails.begin();
	    iter != mAttributeDetails.end();
	    ++iter) {

	    if(iter->cardinality == UNARY) {
		hpp << "\nconst " << iter->type_name << " * get_" << iter->role << "() const {"
		    << " return " << iter->member << "; }";
		hpp << "\nvoid set_" << iter->role << "(const " << iter->type_name << " * "
		    << iter->variable << ");\n";
	    }
	    else if(iter->cardinality == MULTI) {
		hpp << "\nconst vector<const " << iter->type_name
		    << " *> & get_" << iter->role << "() const;";

		hpp << "\nvoid set_" << iter->role
		    << "(const vector<const " << iter->type_name << " *> & "
		    << iter->variable << ");\n";
	    }
	    else {
		ASSERT_NOT_REACHED();
	    }
	}

	// ref_get_type

	hpp << "\nconst string & ref_get_type() const;\n";
    }

    void
    Class::generate_header_protected_members(Common::CppStream & hpp) {
	// reflective stuff

	hpp << "\n//"
	    << "\n// reflective interface"
	    << "\n//"
	    << "\n"
	    << "\nvoid ref_set_context_impl(set<string> & visited, REF::Context * context);"
	    << "\nvoid ref_implant_impl(set<string> & visited) throw();"
	    << "\nvoid ref_uproot_impl(set<string> & visited) throw();"
	    << "\n"
	    << "\n//"
	    << "\n"
	    << "\nbool ref_is_a_type_of_impl(set<string> & visited,"
	    << "\nconst string & class_name) const;"
	    << "\n"
	    << "\nvoid ref_delete_impl(set<string> & visited);"
	    << "\n"
	    << "\n//"
	    << "\n"
	    << "\nbool ref_add_role_impl(set<string> & visited,"
	    << "\nconst string & role_name,"
	    << "\nREF::Class * role,"
	    << "\nREF::Class * position);"
	    << "\n"
	    << "\nbool ref_remove_role_impl(set<string> & visited,"
	    << "\nconst string & role_name,"
	    << "\nREF::Class * role);"
	    << "\n"
	    << "\nbool ref_get_role_iterator_impl(set<string> & visited,"
	    << "\nREF::Iterator * & iterator,"
	    << "\nconst string & role_name);"
	    << "\n"
	    << "\nbool ref_get_role_impl(set<string> & visited,"
	    << "\nREF::Class * & role,"
	    << "\nconst string & role_name);"
	    << "\n"
	    << "\nbool ref_set_role_impl(set<string> & visited,"
	    << "\nconst string & role_name,"
	    << "\nREF::Class * role);"
	    << "\n"
	    << "\nbool ref_create_class_impl(set<string> & visited,"
	    << "\nREF::Class * & role,"
	    << "\nconst string & role_name) const;"
	    << "\n"
	    << "\n//"
	    << "\n"
	    << "\nbool ref_get_attribute_impl(set<string> & visited,"
	    << "\nconst REF::Attribute * & attribute,"
	    << "\nconst string & attribute_name) const;"
	    << "\n"
	    << "\nbool ref_set_attribute_impl(set<string> & visited,"
	    << "\nconst string & attribute_name,"
	    << "\nconst REF::Attribute * attribute);"
	    << "\n"
	    << "\nbool ref_get_attribute_set_impl(set<string> & visited,"
	    << "\nconst vector<const REF::Attribute *> * & attribute_set,"
	    << "\nconst string & attribute_name) const;"
	    << "\n"
	    << "\nbool ref_set_attribute_set_impl(set<string> & visited,"
	    << "\nconst string & attribute_name,"
	    << "\nconst vector<const REF::Attribute *> * attribute_set);"
	    << "\n"
	    << "\nbool ref_create_attribute_impl(set<string> & visited,"
	    << "\nREF::Attribute * & attribute,"
	    << "\nconst string & attribute_name) const;"
	    << "\n";
    }

    void
    Class::generate_header_private_members(Common::CppStream & hpp) {
	// copy constructor is private

	hpp << "\n//"
	    << "\n// copy constructor (copying is forbidden)"
	    << "\n//\n";

	hpp << "\n" << mClassDetails.name << "(const " << mClassDetails.name << " & "
	    << mClassDetails.variable << ");\n";

	// roles storage

	if(!mReferenceDetails.empty()) {
	    hpp << "\n//"
		<< "\n// roles storage"
		<< "\n//\n";

	    for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
		iter != mReferenceDetails.end();
		++iter) {
		if(iter->other_cardinality == UNARY) {
		    hpp << "\n" << iter->other_type_name << " * " << iter->other_end_member << ";";
		}
		else if(iter->other_cardinality == MULTI) {
		    if(iter->other_ordering == ORDERED) {
			hpp << "\nlist<" << iter->other_type_name << " *> "
			    << iter->other_end_member << ";";
		    }
		    else if(iter->other_ordering == UNORDERED) {
			hpp << "\nset<" << iter->other_type_name << " *> "
			    << iter->other_end_member << ";";
		    }
		    else {
			ASSERT_NOT_REACHED();
		    }
		}
		else {
		    ASSERT_NOT_REACHED();
		}
	    }

	    hpp << "\n";
	}

	// attributes storage

	if(!mAttributeDetails.empty()) {
	    hpp << "\n//"
		<< "\n// attributes storage"
		<< "\n//\n";

	    for(vector<AttributeDetails>::iterator iter = mAttributeDetails.begin();
		iter != mAttributeDetails.end();
		++iter) {
		if(iter->cardinality == UNARY) {
		    hpp << "\n" << iter->type_name << " * " << iter->member << ";";
		}
		else if(iter->cardinality == MULTI) {
		    hpp << "\nvector<" << iter->type_name << " *> " << iter->member << ";";
		}
		else {
		    ASSERT_NOT_REACHED();
		}
	    }

	    hpp << "\n";
	}

	// reflective stuff

	hpp << "\n//"
	    << "\n// our type"
	    << "\n//\n";

	hpp << "\nstatic const string TYPE;";
    }

    //
    //
    //

    void
    Class::generate_source(Common::CppStream & cpp) {
	cpp << get_prologue()
	    << "\n#ifdef HAVE_CONFIG_H"
	    << "\n#include \"config.h\""
	    << "\n#endif // HAVE_CONFIG_H\n";

	generate_source_includes(cpp);

	cpp << "\nnamespace " << mModelName << " {\n";

	generate_source_statics(cpp);
	generate_source_constructors(cpp);
	generate_source_destructor(cpp);
	generate_source_body(cpp);

	cpp << "\n} // namespace " << mModelName << "\n";
    }

    void
    Class::generate_source_includes(Common::CppStream & cpp) {
	cpp << "\n#include \"" << mClassDetails.role << ".h\"\n";

	set<string> cpp_includes;

	for(vector<AttributeDetails>::iterator iter = mAttributeDetails.begin();
	    iter != mAttributeDetails.end();
	    ++iter) {
	    if(iter->type->get_container() == mClassDetails.klass->get_container()) {
		// they are in the same package
		cpp_includes.insert("\"" + iter->type_path + ".h\"");
	    }
	    else {
		cpp_includes.insert("\"" + iter->type_qualified_path + ".h\"");
	    }
	}

	for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
	    iter != mReferenceDetails.end();
	    ++iter) {
	    if(iter->other_type->get_container() == mClassDetails.klass->get_container()) {
		// the are in the same package
		cpp_includes.insert("\"" + iter->other_type_path + ".h\"");
	    }
	    else {
		cpp_includes.insert("\"" + iter->other_type_qualified_path + ".h\"");
	    }
	}

	if(!cpp_includes.empty()) {
	    for(set<string>::iterator iter = cpp_includes.begin();
		iter != cpp_includes.end();
		++iter) {
		const string & include = *iter;
		cpp << "\n#include " << include;
	    }

	    cpp << "\n";
	}

	cpp << "\n#include \"" << name_to_path(mModelName) << ".h\"\n";

	cpp << "\n#include \"reflective/changes.h\"";
	cpp << "\n#include \"reflective/set_iterator.h\"";
	cpp << "\n#include \"reflective/list_iterator.h\"";
	cpp << "\n#include \"reflective/pointer_iterator.h\"";
	cpp << "\n#include \"reflective/exceptions.h\"\n";

	cpp << "\n#include \"foundation/macros.h\"";
	cpp << "\n#include \"foundation/commands.h\"";
	cpp << "\n#include \"foundation/utility.h\"";
	cpp << "\n#include \"foundation/exception.h\"";
	cpp << "\n#include \"foundation/algorithm.h\"";
	cpp << "\n#include \"foundation/functional.h\"\n";

	cpp << "\n#include <vector>"
	    << "\n#include <functional>"
	    << "\n#include <algorithm>"
	    << "\nusing namespace std;\n";
    }

    void
    Class::generate_source_statics(Common::CppStream & cpp) {
	cpp << "\nconst string " << mClassDetails.name << "::TYPE = \""
	    << mClassDetails.name << "\";\n";
    }

    void
    Class::generate_source_constructors(Common::CppStream & cpp) {
	// default constructor

	cpp << "\n" << mClassDetails.name << "::" << mClassDetails.name << "() {";

	for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
	    iter != mReferenceDetails.end();
	    ++iter) {

	    // we have to initialise pointers to null, lists and sets take care of
	    // themselves

	    if(iter->other_cardinality == UNARY) {
		cpp << "\n" << iter->other_end_member << " = 0;";
	    }
	}

	for(vector<AttributeDetails>::iterator iter = mAttributeDetails.begin();
	    iter != mAttributeDetails.end();
	    ++iter) {

	    // we have to initialise pointers to null, vectors take care of
	    // themselves

	    if(iter->cardinality == UNARY) {
		cpp << "\n" << iter->member << " = 0;";
	    }
	}

	cpp << "\n\nTRACE(9, \"Constructed: \" << static_cast<Object *>(this));";

	cpp << "\n}\n";

	// copy constructor

	cpp << "\n" << mClassDetails.name << "::" << mClassDetails.name
	    << "(const " << mClassDetails.name << " & " << mClassDetails.variable << ") {"
	    << "\nASSERT_NOT_REACHED();";
	cpp << "\n}\n";
    }

    void
    Class::generate_source_destructor(Common::CppStream & cpp) {
	cpp << "\n" << mClassDetails.name << "::~" << mClassDetails.name
	    << "() {";

	cpp << "\nTRACE(9, \"Destroyed: \" << static_cast<Object *>(this));\n";

	for(vector<AttributeDetails>::iterator iter = mAttributeDetails.begin();
	    iter != mAttributeDetails.end();
	    ++iter) {
	    if(iter->cardinality == UNARY) {
		cpp << "\nif(" << iter->member << ") {"
		    << "\n" << iter->member << "->unref();"
		    << "\n}";
	    }
	    else if(iter->cardinality == MULTI) {
		cpp << "\nfor_each(" << iter->member << ".begin(), " << iter->member << ".end(), "
		    << "mem_fun(&Object::unref));";
	    }
	    else {
		ASSERT_NOT_REACHED();
	    }
	}

	cpp << "\n}\n";
    }

    void
    Class::generate_source_body(Common::CppStream & cpp) {
	for(vector<AttributeDetails>::iterator iter = mAttributeDetails.begin();
	    iter != mAttributeDetails.end();
	    ++iter) {
	    if(iter->cardinality == UNARY) {
		generate_source_body_unary_attribute(cpp, *iter);
	    }
	    else if(iter->cardinality == MULTI) {
		generate_source_body_multi_attribute(cpp, *iter);
	    }
	    else {
		ASSERT_NOT_REACHED();
	    }
	}

	for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
	    iter != mReferenceDetails.end();
	    ++iter) {
	    if(iter->other_cardinality == UNARY) {
		if(iter->this_end_navigable) {
		    generate_source_body_reverse_navigable_pointer_role(cpp, *iter);
		}

		if(iter->other_end_navigable) {
		    generate_source_body_forward_navigable_pointer_role(cpp, *iter);
		}
	    }
	    else if(iter->other_cardinality == MULTI) {
		if(iter->other_ordering == UNORDERED) {
		    if(iter->this_end_navigable) {
			generate_source_body_reverse_navigable_set_role(cpp, *iter);
		    }

		    if(iter->other_end_navigable) {
			generate_source_body_forward_navigable_set_role(cpp, *iter);
		    }
		}
		else if(iter->other_ordering == ORDERED) {
		    if(iter->this_end_navigable) {
			generate_source_body_reverse_navigable_list_role(cpp, *iter);
		    }

		    if(iter->other_end_navigable) {
			generate_source_body_forward_navigable_list_role(cpp, *iter);
		    }
		}
		else {
		    ASSERT_NOT_REACHED();
		}
	    }
	    else {
		ASSERT_NOT_REACHED();
	    }
	}

	generate_source_body_uproot_implant(cpp);
	generate_source_body_reflective(cpp);
    }

    void
    Class::generate_source_body_unary_attribute(Common::CppStream & cpp,
						const AttributeDetails & details) {
	ASSERT(details.cardinality == UNARY);

	cpp << "\nvoid // AAA"
	    << "\n" << mClassDetails.name << "::set_" << details.role
	    << "(const " << details.type_name << " * " << details.variable << ") {"
	    << "\nbegin_transaction();"
	    << "\n"
	    << "\n//"
	    << "\n"
	    << "\n" << details.type_name << " * old_value = " << details.member << " ? "
	    << "new " << details.type_name << "(*" << details.member << ") : 0;"
	    << "\n" << details.type_name << " * new_value = " << details.variable << " ? "
	    << "new " << details.type_name << "(*" << details.variable << ") : 0;"
	    << "\n"
	    << "\n//"
	    << "\n"
	    << "\nCommand * command = command_flip_argument(this, &"
	    << mClassDetails.name << "::set_" << details.role << ","
	    << "\nnew_value, old_value);"
	    << "\nlog_command(command);"
	    << "\ncommand->unref();"
	    << "\n"
	    << "\n// make the changes"
	    << "\n"
	    << "\nif(" << details.member << ") {"
	    << "\n  " << details.member << "->unref();"
	    << "\n}"
	    << "\n"
	    << "\n" << details.member << " = new_value ? new " << details.type_name
	    << "(*new_value) : 0;"
	    << "\n"
	    << "\n// changes made"
	    << "\n"
	    << "\nChangeEvent * change = change_attribute_changed(this, \"" << details.role
	    << "\", new_value, old_value);"
	    << "\nlog_change(change);"
	    << "\nchange->unref();"
	    << "\n"
	    << "\n//"
	    << "\n"
	    << "\nif(new_value) {"
	    << "\n  new_value->unref();"
	    << "\n}"
	    << "\n"
	    << "\nif(old_value) {"
	    << "\n  old_value->unref();"
	    << "\n}"
	    << "\n"
	    << "\n//"
	    << "\n"
	    << "\nend_transaction();"
	    << "\n}\n";
    }

    void
    Class::generate_source_body_multi_attribute(Common::CppStream & cpp,
						const AttributeDetails & details) {
	ASSERT(details.cardinality == MULTI);

	cpp << "\nconst vector<const " << details.type_name << " *> &"
	    << "\n" << mClassDetails.name << "::get_" << details.role
	    << "() const {"
	    << "\nreturn death_cast<vector<const " << details.type_name << " *> >("
	    << details.member << ");"
	    << "\n}\n";

	cpp << "\nvoid // QQQ"
	    << "\n" << mClassDetails.name << "::set_" << details.role
	    << "(const vector<const " << details.type_name << " *> & " << details.variable << ") {"
	    << "\nbegin_transaction();"
	    << "\n"
	    << "\nNYI();"
	    /*
	    << "\nCommand * command = command_flip_argument(this, &"
	    << mClassDetails.name << "::set_" << details.role << ","
	    << "\n"
	    */
	    << "\n"
	    << "\nend_transaction();"
	    << "\n}\n";
    }

    void
    Class::generate_source_body_forward_navigable_list_role(Common::CppStream & cpp,
							    const ReferenceDetails & details) {
	cpp << "\nvoid // BBB"
	    << "\n" << mClassDetails.name << "::set_" << details.other_end_role << "(const list<"
	    << details.other_type_name << " *> & " << details.other_end_variable << ") {"
	    << "\n// put the items into sets so that we can call set_difference"
	    << "\n// to work out what is actually being added and removed"
	    << "\n"
	    << "\nset<" << details.other_type_name << " *> before;"
	    << "\nset<" << details.other_type_name << " *> after;"
	    << "\n"
	    << "\nbefore.insert("
	    << details.other_end_member << ".begin(), " << details.other_end_member << ".end());"
	    << "\nafter.insert("
	    << details.other_end_variable << ".begin(), " << details.other_end_variable << ".end());"
	    << "\n"
	    << "\n// check for duplicates"
	    << "\n"
	    << "\nRETURN_IF_FAIL(before.size() == " << details.other_end_member
	    << ".size());"
	    << "\nRETURN_IF_FAIL(after.size() == " << details.other_end_variable
	    << ".size());"
	    << "\n"
	    << "\n// sort the stuff being added and the stuff being removed"
	    << "\n// into separate containers"
	    << "\n"
	    << "\nvector<" << details.other_type_name << " *> added;"
	    << "\nvector<" << details.other_type_name << " *> removed;"
	    << "\n"
	    << "\nset_difference("
	    << details.other_end_variable << ".begin(), " << details.other_end_variable << ".end(),"
	    << "\n"
	    << details.other_end_member << ".begin(), " << details.other_end_member << ".end(),"
	    << "\nback_inserter(added));"
	    << "\n"
	    << "\nset_difference("
	    << details.other_end_member << ".begin(), "
	    << details.other_end_member << ".end(),"
	    << "\n"
	    << details.other_end_variable << ".begin(), "
	    << details.other_end_variable << ".end(),"
	    << "\nback_inserter(removed));"
	    << "\n";

	cpp << "\n// transaction time"
	    << "\n"
	    << "\nbegin_transaction();"		// TODO rationalise the context
	    << "\n"
	    << "\n//"
	    << "\n"
	    << "\nCommand * command = command_flip_argument(this,"
	    << "\n&" << mClassDetails.name << "::set_" << details.other_end_role << ","
	    << "\n" << details.other_end_variable << ", " << details.other_end_member << ");"
	    << "\nlog_command(command);"
	    << "\ncommand->unref();"
	    << "\n"
	    << "\n// make the changes"
	    << "\n";

	if(details.this_cardinality == UNARY) {
	    cpp << "\nfor_each(added.begin(), added.end(),"
		<< "\nbind2nd(mem_fun(&" << details.other_type_name
		<< "::internal_set_" << details.this_end_role << "), this));"
		<< "\nfor_each(removed.begin(), removed.end(),"
		<< "\nbind2nd(mem_fun(&" << details.other_type_name
		<< "::internal_set_" << details.this_end_role << "),"
		<< "\nstatic_cast<" << details.this_type_name << " *>(0)));";
	}
	else {
	    cpp << "\nfor_each(added.begin(), added.end(),"
		<< "\nbind2nd(mem_fun(&" << details.other_type_name
		<< "::internal_add_" << details.this_end_role << "), this));"
		<< "\nfor_each(removed.begin(), removed.end(),"
		<< "\nbind2nd(mem_fun(&" << details.other_type_name
		<< "::internal_remove_" << details.this_end_role << "), this));";
	}

	cpp << "\n"
	    << "\nfor_each(added.begin(), added.end(), mem_fun(&Object::ref));"
	    << "\nfor_each(removed.begin(), removed.end(), mem_fun(&Object::unref));"
	    << "\n"
	    << "\n" << details.other_end_member << " = " << details.other_end_variable << ";"
	    << "\n"
	    << "\n// changes made"
	    << "\n"
	    << "\nfor(vector<" << details.other_type_name << " *>::iterator iter = added.begin();"
	    << "\niter != added.end();"
	    << "\n++iter) {"
	    << "\n" << details.other_type_name << " * "
	    << details.other_end_variable << " = *iter;"
	    << "\nChangeEvent * change = change_role_added(this, \""
	    << details.other_end_role << "\","
	    << "\n" << details.other_end_variable << ", "
	    << (details.this_end_aggregate ? "true" : "false") << ");"
	    << "\nlog_change(change);"
	    << "\n    change->unref();"
	    << "\n}"
	    << "\n"
	    << "\nfor(vector<" << details.other_type_name << " *>::iterator iter = removed.begin();"
	    << "\niter != removed.end();"
	    << "\n++iter) {"
	    << "\n" << details.other_type_name << " * "
	    << details.other_end_variable << " = *iter;"
	    << "\nChangeEvent * change = change_role_removed(this, \""
	    << details.other_end_role << "\","
	    << "\n" << details.other_end_variable << ", "
	    << (details.this_end_aggregate ? "true" : "false") << ");"
	    << "\nlog_change(change);"
	    << "\n    change->unref();"
	    << "\n}"
	    << "\n"
	    << "\n//"
	    << "\n"
	    << "\nend_transaction();"
	    << "\n}\n";

	// add_

	cpp << "\nvoid // CCC"
	    << "\n" << mClassDetails.name << "::add_" << details.other_end_role <<
	    "(" << details.other_type_name << " * " << details.other_end_variable
	    << ", int index) {"
	    << "\n// make sure the the index isn't out of range"
	    << "\n"
	    << "\nRETURN_IF_FAIL(index <= static_cast<int>("
	    << details.other_end_member << ".size()));"
	    << "\n"
	    << "\nif(index < 0) {"
	    << "\n    index = " << details.other_end_member << ".size();"
	    << "\n}"
	    << "\n"
	    << "\nlist<" << details.other_type_name << " *>::iterator iter = "
	    << details.other_end_member << ".begin();"
	    << "\n"
	    << "\nfor(int i = 0; i < static_cast<int>("
	    << details.other_end_member << ".size()); ++i) {"
	    << "\n    RETURN_IF_FAIL(*iter != " << details.other_end_variable << ");"
	    << "\n"
	    << "\n    if(i < index) {"
	    << "\n        ++iter;"
	    << "\n    }"
	    << "\n}"
	    << "\n"
	    << "\n// transaction time"
	    << "\n"
	    << "\nbegin_transaction();"
	    << "\n"
	    << "\n//"
	    << "\n"
	    << "\nCommand * command = command_flip_method(this,"
	    << "\n&" << mClassDetails.name << "::add_" << details.other_end_role << ","
	    << "\n&" << mClassDetails.name << "::remove_" << details.other_end_role << ","
	    << "\n" << details.other_end_variable << ", index);"
	    << "\nlog_command(command);"
	    << "\ncommand->unref();"
	    << "\n"
	    << "\nChangeEvent * change = change_role_added(this, "
	    << "\"" << details.other_end_role << "\", " << details.other_end_variable << ", "
	    << (details.this_end_aggregate ? "true" : "false")
	    << ", index);"
	    << "\nlog_change(change);"
	    << "\nchange->unref();"
	    << "\n"
	    << "\n// make the changes"
	    << "\n"
	    << "\n" << details.other_end_variable << "->ref();"
	    << "\n"
	    << "\n" << details.other_end_member << ".insert(iter, "
	    << details.other_end_variable << ");";

	if(details.this_cardinality == UNARY) {
	    cpp << "\n" << details.other_end_variable << "->internal_set_"
		<< details.this_end_role
		<< "(this);";
	}
	else {
	    cpp << "\n" << details.other_end_variable << "->internal_add_"
		<< details.this_end_role
		<< "(this);";
	}

	cpp << "\n"
	    << "\n// changes made"
	    << "\n"
	    << "\nend_transaction();"
	    << "\n}\n";

	// remove_

	cpp << "\nvoid // DDD"
	    << "\n" << mClassDetails.name << "::remove_"
	    << details.other_end_role << "(" << details.other_type_name
	    << " * " << details.other_end_variable << ") {"
	    << "\n// work out the index"
	    << "\n"
	    << "\nint index = 0;"
	    << "\n"
	    << "\nlist<" << details.other_type_name << " *>::iterator iter = "
	    << details.other_end_member << ".begin();"
	    << "\n"
	    << "\nfor(; iter != " << details.other_end_member << ".end(); "
	    << "++iter, ++index) {"
	    << "\n    if(*iter == " << details.other_end_variable << ") {"
	    << "\n        break;"
	    << "\n    }"
	    << "\n}"
	    << "\n"
	    << "\nRETURN_IF_FAIL(iter != " << details.other_end_member
	    << ".end());"
	    << "\n"
	    << "\n// transaction time"
	    << "\n"
	    << "\nbegin_transaction();"
	    << "\n"
	    << "\n//"
	    << "\n"
	    << "\nCommand * command = command_flip_method(this,"
	    << "\n&" << mClassDetails.name << "::remove_" << details.other_end_role << ","
	    << "\n&" << mClassDetails.name << "::add_" << details.other_end_role << ","
	    << "\n" << details.other_end_variable << ", index);"
	    << "\nlog_command(command);"
	    << "\ncommand->unref();"
	    << "\n"
	    << "\nChangeEvent * change = change_role_removed(this, \""
	    << details.other_end_role << "\", " << details.other_end_variable << ", "
	    << (details.this_end_aggregate ? "true" : "false")
	    << ", index);"
	    << "\nlog_change(change);"
	    << "\nchange->unref();"
	    << "\n"
	    << "\n// make the changes"
	    << "\n";

	if(details.this_cardinality == UNARY) {
	    cpp << "\n" << details.other_end_variable << "->internal_set_"
		<< details.this_end_role
		<< "(static_cast<" << details.this_type_name << " *>(0));";
	}
	else {
	    cpp << "\n" << details.other_end_variable << "->internal_remove_"
		<< details.this_end_role
		<< "(this);";
	}

	cpp << "\n" << details.other_end_member << ".erase(iter);"
	    << "\n"
	    << "\n" << details.other_end_variable << "->unref();"
	    << "\n"
	    << "\n// changes made"
	    << "\n"
	    << "\nend_transaction();"
	    << "\n}\n";
    }

    void
    Class::generate_source_body_reverse_navigable_list_role(Common::CppStream & cpp,
							    const ReferenceDetails & details) {
	cpp << "\nvoid // EEE"
	    << "\n" << mClassDetails.name << "::internal_add_" << details.other_end_role
	    << "(" << details.other_type_name
	    << " * " << details.other_end_variable << ") {"
	    << "\n" << details.other_end_variable << "->ref();"
	    << "\n"
	    << "\nint index = " << details.other_end_member << ".size();"
	    << "\n"
	    << "\n" << details.other_end_member << ".push_back("
	    << details.other_end_variable << ");"
	    << "\n"
	    << "\nChangeEvent * change = change_role_added(this, \""
	    << details.other_end_role << "\"," << "\n" << details.other_end_variable << ", index, "
	    << (details.this_end_aggregate ? "true" : "false") << ");"
	    << "\nlog_change(change);"
	    << "\nchange->unref();"
	    << "\n}\n";

	// internal_remove

	// FIXME position isn't done... thinking about using another element
	// as position...

	cpp << "\nvoid // FFF"
	    << "\n" << mClassDetails.name << "::internal_remove_" << details.other_end_role
	    << "(" << details.other_type_name
	    << " * " << details.other_end_variable << ") {"
	    << "\n" << details.other_end_member << ".remove("
	    << details.other_end_variable << ");"
	    << "\n"
	    << "\nChangeEvent * change = change_role_removed(this, \""
	    << details.other_end_role << "\"," << "\n" << details.other_end_variable << ", "
	    << "-1, "
	    << (details.this_end_aggregate ? "true" : "false") << ");"
	    << "\nlog_change(change);"
	    << "\nchange->unref();"
	    << "\n"
	    << "\n" << details.other_end_variable << "->unref();"
	    << "\n}\n";
    }

    void
    Class::generate_source_body_forward_navigable_set_role(Common::CppStream & cpp,
							   const ReferenceDetails & details) {
	cpp << "\nvoid // GGG"
	    << "\n" << mClassDetails.name << "::set_" << details.other_end_role << "(const set<"
	    << details.other_type_name << " *> & " << details.other_end_variable << ") {"
	    << "\n// sort the stuff being added and the stuff being removed"
	    << "\n// into separate containers"
	    << "\n"
	    << "\nvector<" << details.other_type_name << " *> added;"
	    << "\nvector<" << details.other_type_name << " *> removed;"
	    << "\n"
	    << "\nset_difference("
	    << details.other_end_variable << ".begin(), " << details.other_end_variable << ".end(), "
	    << "\n"
	    << details.other_end_member << ".begin(), " << details.other_end_member << ".end(),"
	    << "\nback_inserter(added));"
	    << "\n"
	    << "\nset_difference("
	    << details.other_end_member << ".begin(), "
	    << details.other_end_member << ".end(), "
	    << "\n"
	    << details.other_end_variable << ".begin(), "
	    << details.other_end_variable << ".end(),"
	    << "\nback_inserter(removed));\n";

	cpp << "\n"
	    << "\n// transaction time"
	    << "\n"
	    << "\nbegin_transaction();"
	    << "\n"				// TODO rationalise the context
	    << "\n//"
	    << "\n"
	    << "\nCommand * command = command_flip_argument(this,"
	    << "\n&" << mClassDetails.name << "::set_" << details.other_end_role << ","
	    << "\n" << details.other_end_variable << ", " << details.other_end_member << ");"
	    << "\nlog_command(command);"
	    << "\ncommand->unref();"
	    << "\n"
	    << "\n// make the changes";

	if(details.this_cardinality == UNARY) {
	    cpp << "\nfor_each(added.begin(), added.end(),"
		<< "\nbind2nd(mem_fun(&" << details.other_type_name
		<< "::internal_set_" << details.this_end_role << "), this));"
		<< "\nfor_each(removed.begin(), removed.end(),"
		<< "\nbind2nd(mem_fun(&" << details.other_type_name
		<< "::internal_set_" << details.this_end_role << "),"
		<< "\nstatic_cast<" << details.this_type_name << " *>(0)));";
	}
	else if(details.this_cardinality == MULTI) {
	    if(details.this_ordering == ORDERED) {
		cpp << "\nfor_each(added.begin(), added.end(),"
		    << "\nbind2nd(mem_fun(&" << details.other_type_name
		    << "::internal_add_" << details.this_end_role << "), this));"
		    << "\nfor_each(removed.begin(), removed.end(),"
		    << "\nbind2nd(mem_fun(&" << details.other_type_name
		    << "::internal_remove_" << details.this_end_role << "), this));";
	    }
	    else if(details.this_ordering == UNORDERED) {
		cpp << "\nfor_each(added.begin(), added.end(),"
		    << "\nbind2nd(mem_fun(&" << details.other_type_name
		    << "::internal_add_" << details.this_end_role << "), this));"
		    << "\nfor_each(removed.begin(), removed.end(),"
		    << "\nbind2nd(mem_fun(&" << details.other_type_name
		    << "::internal_remove_" << details.this_end_role << "), this));";
	    }
	    else {
		ASSERT_NOT_REACHED();
	    }
	}
	else {
	    ASSERT_NOT_REACHED();
	}

	cpp << "\n"
	    << "\nfor_each(added.begin(), added.end(), mem_fun(&Object::ref));"
	    << "\nfor_each(removed.begin(), removed.end(), mem_fun(&Object::unref));"
	    << "\n"
	    << "\n" << details.other_end_member << " = " << details.other_end_variable << ";"
	    << "\n"
	    << "\n// changes made"
	    << "\n"
	    << "\nfor(vector<" << details.other_type_name << " *>::iterator iter = added.begin();"
	    << "\niter != added.end();"
	    << "\n++iter) {"
	    << "\n" << details.other_type_name << " * " << details.other_end_variable << " = *iter;"
	    << "\nChangeEvent * change = change_role_added(this, \"" << details.other_end_role
	    << "\",\n" << details.other_end_variable << ", "
	    << (details.this_end_aggregate ? "true" : "false") << ");"
	    << "\nlog_change(change);"
	    << "\nchange->unref();"
	    << "\n}"
	    << "\n"
	    << "\nfor(vector<" << details.other_type_name << " *>::iterator iter = removed.begin();"
	    << "\niter != removed.end();"
	    << "\n++iter) {"
	    << "\n" << details.other_type_name << " * " << details.other_end_variable << " = *iter;"
	    << "\nChangeEvent * change = change_role_removed(this, \"" << details.other_end_role
	    << "\",\n" << details.other_end_variable << ", "
	    << (details.this_end_aggregate ? "true" : "false") << ");"
	    << "\nlog_change(change);"
	    << "\nchange->unref();"
	    << "\n}"
	    << "\n"
	    << "\n//"
	    << "\n"
	    << "\nend_transaction();"
	    << "\n}\n";

	//
	// add
	//

	cpp << "\nvoid // HHH"
	    << "\n" << mClassDetails.name << "::add_" << details.other_end_role
	    << "(" << details.other_type_name
	    << " * " << details.other_end_variable << ") {"
	    << "\nRETURN_IF_FAIL(" << details.other_end_member
	    << ".find(" << details.other_end_variable << ") == "
	    << details.other_end_member << ".end());"
	    << "\n"
	    << "\n// transaction time"
	    << "\n"
	    << "\nbegin_transaction();"
	    << "\n"
	    << "\n//"
	    << "\n"
	    << "\nCommand * command = command_flip_method(this,"
	    << "\n&" << mClassDetails.name << "::add_" << details.other_end_role << ","
	    << "\n&" << mClassDetails.name << "::remove_" << details.other_end_role << ","
	    << "\n" << details.other_end_variable << ");"
	    << "\nlog_command(command);"
	    << "\ncommand->unref();"
	    << "\n"
	    << "\nChangeEvent * change = change_role_added(this, "
	    << "\"" << details.other_end_role << "\", " << details.other_end_variable << ", "
	    << (details.this_end_aggregate ? "true" : "false") << ");"
	    << "\nlog_change(change);"
	    << "\nchange->unref();"
	    << "\n"
	    << "\n// make the changes"
	    << "\n"
	    << "\n" << details.other_end_variable << "->ref();"
	    << "\n"
	    << "\n" << details.other_end_member << ".insert("
	    << details.other_end_variable << ");";

	if(details.this_cardinality == UNARY) {
	    cpp << "\n" << details.other_end_variable << "->internal_set_"
		<< details.this_end_role << "(this);";
	}
	else {
	    cpp << "\n" << details.other_end_variable << "->internal_add_"
		<< details.this_end_role << "(this);";
	}

	cpp << "\n"
	    << "\n// changes made"
	    << "\n"
	    << "\nend_transaction();"
	    << "\n}\n";

	//
	// remove
	//

	cpp << "\nvoid // III"
	    << "\n" << mClassDetails.name << "::remove_" << details.other_end_role
	    << "(" << details.other_type_name
	    << " * " << details.other_end_variable << ") {"
	    << "\nRETURN_IF_FAIL(" << details.other_end_member
	    << ".find(" << details.other_end_variable << ") != "
	    << details.other_end_member << ".end());"
	    << "\n"
	    << "\n// transaction time"
	    << "\n"
	    << "\nbegin_transaction();"
	    << "\n"
	    << "\n//"
	    << "\n"
	    << "\nCommand * command = command_flip_method(this,"
	    << "\n&" << mClassDetails.name << "::remove_" << details.other_end_role << ","
	    << "\n&" << mClassDetails.name << "::add_" << details.other_end_role << ","
	    << "\n" << details.other_end_variable << ");"
	    << "\nlog_command(command);"
	    << "\ncommand->unref();"
	    << "\n"
	    << "\nChangeEvent * change = change_role_removed(this, "
	    << "\"" << details.other_end_role << "\", " << details.other_end_variable << ", "
	    << (details.this_end_aggregate ? "true" : "false") << ");"
	    << "\nlog_change(change);"
	    << "\nchange->unref();"
	    << "\n"
	    << "\n// make the changes"
	    << "\n";

	if(details.this_cardinality == UNARY) {
	    cpp << "\n" << details.other_end_variable << "->internal_set_"
		<< details.this_end_role
		<< "(static_cast<" << details.this_type_name << " *>(0));";
	}
	else {
	    cpp << "\n" << details.other_end_variable << "->internal_remove_"
		<< details.this_end_role << "(this);";
	}

	cpp << "\n" << details.other_end_member << ".erase("
	    << details.other_end_variable << ");"
	    << "\n"
	    << "\n" << details.other_end_variable << "->unref();"
	    << "\n"
	    << "\n// changes made"
	    << "\n"
	    << "\nend_transaction();"
	    << "\n}\n";
    }

    void
    Class::generate_source_body_reverse_navigable_set_role(Common::CppStream & cpp,
							   const ReferenceDetails & details) {
	cpp << "\nvoid // JJJ"
	    << "\n" << mClassDetails.name << "::internal_add_" << details.other_end_role
	    << "(" << details.other_type_name
	    << " * " << details.other_end_variable << ") {"
	    << "\n" << details.other_end_member << ".insert("
	    << details.other_end_variable << ");"
	    << "\n"
	    << "\nChangeEvent * change = change_role_added(this, \""
	    << details.other_end_role << "\"," << "\n" << details.other_end_variable << ", "
	    << (details.this_end_aggregate ? "true" : "false") << ");"
	    << "\nlog_change(change);"
	    << "\nchange->unref();"
	    << "\n}\n";

	cpp << "\nvoid // KKK"
	    << "\n" << mClassDetails.name << "::internal_remove_" << details.other_end_role
	    << "(" << details.other_type_name
	    << " * " << details.other_end_variable << ") {"
	    << "\n" << details.other_end_member << ".erase("
	    << details.other_end_variable << ");"
	    << "\n"
	    << "\nChangeEvent * change = change_role_removed(this, \""
	    << details.other_end_role << "\"," << "\n" << details.other_end_variable << ", "
	    << (details.this_end_aggregate ? "true" : "false") << ");"
	    << "\nlog_change(change);"
	    << "\nchange->unref();"
	    << "\n}\n";
    }

    void
    Class::generate_source_body_forward_navigable_pointer_role(Common::CppStream & cpp,
							       const ReferenceDetails & details) {
	cpp << "\nvoid // LLL"
	    << "\n" << mClassDetails.name << "::set_" << details.other_end_role
	    << "(" << details.other_type_name
	    << " * " << details.other_end_variable << ") {"
	    << "\nif(" << details.other_end_member << " != "
	    << details.other_end_variable << ") {";

	if(details.this_end_aggregate) {
	    cpp << "\n// this end is an aggregate, so it must either have"
		<< "\n// no context or the same context as ours."
		<< "\n"
		<< "\nif(" << details.other_end_variable << ") {"
		<< "\nASSERT(!" << details.other_end_variable << "->ref_get_context() ||"
		<< "\n" << details.other_end_variable<< "->ref_get_context() == ref_get_context());"
		<< "\n}\n";
	}
	else if(details.other_end_aggregate) {
	    cpp << "\n// the other end is an aggregate, so we must either have"
		<< "\n// no context or the same context that the other end has"
		<< "\n"
		<< "\nif(" << details.other_end_variable << ") {"
		<< "\nASSERT(!ref_get_context() || "
		<< details.other_end_variable << "->ref_get_context() == ref_get_context());"
		<< "\n}\n";
	}
	else {
	    cpp << "\n// neither end is an aggregate, so we must have the same context"
		<< "\n// as each other"
		<< "\n"
		<< "\nif(" << details.other_end_variable << ") {"
		<< "\nASSERT(!" << details.other_end_variable << "->ref_get_context() ||"
		<< "\n" << details.other_end_variable <<"->ref_get_context() == ref_get_context());"
		<< "\n}\n";
	}

	cpp << "\n    begin_transaction();"
	    << "\n"
	    << "\n    Command * command = command_flip_argument(this,"
	    << "\n    &" << mClassDetails.name << "::set_" << details.other_end_role << ","
	    << "\n" << details.other_end_variable << ", "
	    << details.other_end_member << ");"
	    << "\nlog_command(command);"
	    << "\ncommand->unref();"
	    << "\n"
	    << "\n    ChangeEvent * change = change_role_changed(this,"
	    << "\n\"" << details.other_end_role << "\", " << details.other_end_variable << ","
	    << "\n" << details.other_end_member << ", "
	    << (details.this_end_aggregate ? "true" : "false") << ");"
	    << "\nlog_change(change);"
	    << "\nchange->unref();"
	    << "\n"
	    << "\nif(" << details.other_end_member << ") {";

	if(details.this_cardinality == UNARY) {
	    cpp << "\n    " << details.other_end_member << "->internal_set_"
		<< details.this_end_role
		<< "(static_cast<" << details.this_type_name << " *>(0));";
	}
	else {
	    cpp << "\n    " << details.other_end_member << "->internal_remove_"
		<< details.this_end_role << "(this);";
	}

	cpp << "\n" << details.other_end_member << "->unref();"
	    << "\n}";

	cpp << "\n"
	    << "\n" << details.other_end_member << " = "
	    << details.other_end_variable << ";";

	cpp << "\n"
	    << "\nif(" << details.other_end_member << ") {";

	if(details.this_cardinality == UNARY) {
	    cpp << "\n    " << details.other_end_member << "->internal_set_"
		<< details.this_end_role << "(this);";
	}
	else {
	    cpp << "\n    " << details.other_end_member << "->internal_add_"
		<< details.this_end_role << "(this);";
	}

	cpp << "\n" << details.other_end_member << "->ref();"
	    << "\n}"
	    << "\n"
	    << "\nend_transaction();"
	    << "\n}"
	    << "\n}\n";
    }

    void
    Class::generate_source_body_reverse_navigable_pointer_role(Common::CppStream & cpp,
							       const ReferenceDetails & details) {
	cpp << "\nvoid // MMM"
	    << "\n" << mClassDetails.name << "::internal_set_" << details.other_end_role
	    << "(" << details.other_type_name
	    << " * " << details.other_end_variable << ") {"
	    << "\nChangeEvent * change = change_role_changed(this, \""
	    << details.other_end_role << "\"," << "\n" << details.other_end_variable
	    << ", " << details.other_end_member << ", "
	    << (details.this_end_aggregate ? "true" : "false") << ");"
	    << "\nlog_change(change);"
	    << "\n"
	    << "\nchange->unref();"
	    << "\n"
	    //<< "\nif(" << details.other_end_member << " != " << details.other_end_variable << ") {"
	    << "\nif(" << details.other_end_member << ") {"
	    << "\n" << details.other_end_member << "->unref();"
	    << "\n}"
	    << "\n"
	    << "\n" << details.other_end_member << " = " << details.other_end_variable << ";"
	    << "\n"
	    << "\nif(" << details.other_end_member << ") {"
	    << "\n" << details.other_end_member << "->ref();"
	    << "\n}"
	    //<< "\n}"
	    << "\n}\n";
    }

    void
    Class::generate_source_body_uproot_implant(Common::CppStream & cpp) {
	//
	// set_context
	//

	cpp << "\nvoid"
	    << "\n" << mClassDetails.name << "::ref_set_context_impl(set<string> & visited,"
	    << "\nREF::Context * context) {"
	    << "\nif(visited.insert(TYPE).second) {";

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\n" << iter->name << "::ref_set_context_impl(visited, context);";
	}

	cpp << "\n}"
	    << "\n}\n";

	//
	// implant
	//

	cpp << "\nvoid"
	    << "\n" << mClassDetails.name << "::ref_implant_impl(set<string> & visited) throw() {"
	    << "\nif(visited.insert(TYPE).second) {"
	    << "\nREF::Context * context = ref_get_context();"
	    << "\n"
	    << "\nASSERT(context);"
	    << "\n"
	    << "\n" << mModelName << " * model = known_cast<" << mModelName << " *>(context);"
	    << "\n"
	    << "\nmodel->internal_add_" << mClassDetails.role << "(this);";

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\n"
		<< "\n" << iter->name << "::ref_implant_impl(visited);";
	}

	cpp << "\n}"
	    << "\n}\n";

	//
	// uproot
	//

	cpp << "\nvoid"
	    << "\n" << mClassDetails.name << "::ref_uproot_impl(set<string> & visited) throw() {"
	    << "\nif(visited.insert(TYPE).second) {";

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\n" << iter->name << "::ref_uproot_impl(visited);"
		<< "\n";
	}

	cpp << "\nREF::Context * context = ref_get_context();"
	    << "\n"
	    << "\nASSERT(context);"
	    << "\n"
	    << "\n" << mModelName << " * model = known_cast<" << mModelName << " *>(context);"
	    << "\n"
	    << "\nmodel->internal_add_" << mClassDetails.role << "(this);"
	    << "\n}"
	    << "\n}\n";
    }

    void
    Class::generate_source_body_reflective(Common::CppStream & cpp) {
	//
	// ref_add_role_impl
	//

	cpp << "\nbool"
	    << "\n" << mClassDetails.name
	    << "::ref_add_role_impl(set<string> & visited,"
	    << "\nconst string & role_name,"
	    << "\nREF::Class * role,"
	    << "\nREF::Class * position) {"
	    << "\nif(visited.insert(TYPE).second) {";

	for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
	    iter != mReferenceDetails.end();
	    ++iter) {
	    if(iter->other_end_navigable) {
		if(iter->other_cardinality == MULTI) {
		    if(iter->other_ordering == ORDERED) {
			cpp << "\nif(role_name == \"" << iter->other_end_role << "\") {"
			    << "\n" << iter->other_type_name << " * " << iter->other_end_variable
			    << " = dynamic_cast<" << iter->other_type_name << " *>(role);"
			    << "\n"
			    << "\nif(!" << iter->other_end_variable << ") {"
			    << "\nWARNING(\"Wrong type for role: \" + role->ref_get_type());"
			    << "\nthrow REF::BadType(\"Wrong type for role: \" "
			    << "+ role->ref_get_type() +"
			    << "\n\", in class: \" + ref_get_type());"
			    << "\n}"
			    << "\n"
			    << "\nint index = -1;"	// FIXME
			    << "\n"
			    << "\nadd_" << iter->other_end_role << "(" << iter->other_end_variable
			    << ", index);"
			    << "\nreturn true;"
			    << "\n}\n";
		    }
		    else if(iter->other_ordering == UNORDERED) {
			cpp << "\nif(role_name == \"" << iter->other_end_role << "\") {"
			    << "\n" << iter->other_type_name << " * " << iter->other_end_variable
			    << " = dynamic_cast<" << iter->other_type_name << " *>(role);"
			    << "\n"
			    << "\nif(!" << iter->other_end_variable << ") {"
			    << "\nWARNING(\"Wrong type for role: \" + role->ref_get_type() +"
			    << "\n\", in class: \" + ref_get_type());"
			    << "\nthrow REF::BadType(\"Wrong type for role: \" "
			    << "+ role->ref_get_type());"
			    << "\n}"
			    << "\n"
			    << "\nadd_" << iter->other_end_role << "("
			    << iter->other_end_variable << ");"
			    << "\nreturn true;"
			    << "\n}\n";
		    }
		    else {
			ASSERT_NOT_REACHED();
		    }
		}
	    }
	}

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\nif(" << iter->name
		<< "::ref_add_role_impl(visited, role_name, role, position)) {"
		<< "\nreturn true;"
		<< "\n}\n";
	}

	cpp << "\nreturn false;"
	    << "\n}"
	    << "\nelse {"
	    << "\nreturn false;"
	    << "\n}"
	    << "\n}\n";

	//
	// ref_remove_role_impl
	//

	cpp << "\nbool"
	    << "\n" << mClassDetails.name
	    << "::ref_remove_role_impl(set<string> & visited,"
	    << "\nconst string & role_name,"
	    << "\nREF::Class * role) {"
	    << "\nif(visited.insert(TYPE).second) {";

	for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
	    iter != mReferenceDetails.end();
	    ++iter) {
	    if(iter->other_end_navigable) {
		// because the container is unique an iterator isn't required to
		// indicate which element is being removed

		if(iter->other_cardinality == MULTI && iter->other_ordering == UNORDERED) {
		    cpp << "\nif(role_name == \"" << iter->other_end_role << "\") {"
			<< "\n" << iter->other_type_name << " * " << iter->other_end_variable
			<< " = dynamic_cast<" << iter->other_type_name << " *>(role);"
			<< "\n"
			<< "\nif(!" << iter->other_end_variable << ") {"
			<< "\nWARNING(\"Wrong type for role: \" + role->ref_get_type() +"
			<< "\n\", in class: \" + ref_get_type());"
			<< "\nthrow REF::BadType(\"Wrong type for role: \" + role->ref_get_type());"
			<< "\n}"
			<< "\n"
			<< "\nremove_" << iter->other_end_role << "("
			<< iter->other_end_variable << ");"
		        << "\nreturn true;"
			<< "\n}\n";
		}
	    }
	}

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\nif(" << iter->name
		<< "::ref_remove_role_impl(visited, role_name, role)) {"
		<< "\nreturn true;"
		<< "\n}\n";
	}

	cpp << "\nreturn false;"
	    << "\n}"
	    << "\nelse {"
	    << "\nreturn false;"
	    << "\n}"
	    << "\n}\n";

	//
	// ref_get_role
	//

	cpp << "\nbool"
	    << "\n" << mClassDetails.name
	    << "::ref_get_role_impl(set<string> & visited,"
	    << "\nREF::Class * & role,"
	    << "\nconst string & role_name) {"
	    << "\nif(visited.insert(TYPE).second) {";

	for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
	    iter != mReferenceDetails.end();
	    ++iter) {
	    if(iter->other_end_navigable) {
		if(iter->other_cardinality == UNARY) {
		    cpp << "\nif(role_name == \"" << iter->other_end_role << "\") {"
			<< "\nrole = get_" << iter->other_end_role << "();"
			<< "\nreturn true;"
			<< "\n}\n";
		}
	    }
	}

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\nif(" << iter->name
		<< "::ref_get_role_impl(visited, role, role_name)) {"
		<< "\nreturn true;"
		<< "\n}\n";
	}

	cpp << "\nreturn false;"
	    << "\n}"
	    << "\nelse {"
	    << "\nreturn false;"
	    << "\n}"
	    << "\n}\n";

	//
	// ref_set_role
	//

	cpp << "\nbool"
	    << "\n" << mClassDetails.name
	    << "::ref_set_role_impl(set<string> & visited,"
	    << "\nconst string & role_name,"
	    << "\nREF::Class * role) {"
	    << "\nif(visited.insert(TYPE).second) {";

	for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
	    iter != mReferenceDetails.end();
	    ++iter) {
	    if(iter->other_end_navigable) {
		if(iter->other_cardinality == UNARY) {
		    cpp << "\nif(role_name == \"" << iter->other_end_role << "\") {"
			<< "\n" << iter->other_type_name << " * " << iter->other_end_variable
			<< " = dynamic_cast<" << iter->other_type_name << " *>(role);"
			<< "\n"
			<< "\nif(!" << iter->other_end_variable << ") {"
			<< "\nWARNING(\"Wrong type for role: \" + role->ref_get_type() +"
			<< "\n\", in class: \" + ref_get_type());"
			<< "\nthrow REF::BadType(\"Wrong type for role: \" + role->ref_get_type());"
			<< "\n}"
			<< "\n"
			<< "\nset_" << iter->other_end_role
			<< "(" << iter->other_end_variable << ");"
			<< "\nreturn true;"
			<< "\n}\n";
		}
	    }
	}

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\nif(" << iter->name
		<< "::ref_set_role_impl(visited, role_name, role)) {"
		<< "\nreturn true;"
		<< "\n}\n";
	}

	cpp << "\nreturn false;"
	    << "\n}"
	    << "\nelse {"
	    << "\nreturn false;"
	    << "\n}"
	    << "\n}\n";

	//
	// ref_create_role_impl
	//

	cpp << "\nbool"
	    << "\n" << mClassDetails.name
	    << "::ref_create_class_impl(set<string> & visited,"
	    << "\nREF::Class * & role,"
	    << "\nconst string & role_name) const {"
	    << "\nif(visited.insert(TYPE).second) {";

	for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
	    iter != mReferenceDetails.end();
	    ++iter) {
	    if(iter->other_end_navigable) {
		cpp << "\nif(role_name == \"" << iter->other_end_role << "\") {"
		    << "\nrole = new " << iter->other_type_name << "();"
		    << "\nreturn true;"
		    << "\n}\n";
	    }
	}

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\nif(" << iter->name
		<< "::ref_create_class_impl(visited, role, role_name)) {"
		<< "\nreturn true;"
		<< "\n}\n";
	}

	cpp << "\nreturn false;"
	    << "\n}"
	    << "\nelse {"
	    << "\nreturn false;"
	    << "\n}"
	    << "\n}\n";

	//
	// ref_get_role_iterator
	//

	cpp << "\nbool"
	    << "\n" << mClassDetails.name
	    << "::ref_get_role_iterator_impl(set<string> & visited,"
	    << "\nREF::Iterator * & iterator,"
	    << "\nconst string & role_name) {"
	    << "\nif(visited.insert(TYPE).second) {";

	for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
	    iter != mReferenceDetails.end();
	    ++iter) {
	    if(iter->other_end_navigable) {
		cpp << "\nif(role_name == \"" << iter->other_end_role << "\") {"
		    << "\niterator = REF::iterator<" << iter->other_type_name
		    << ">(get_" << iter->other_end_role << "());"
		    << "\nreturn true;"
		    << "\n}\n";
	    }
	}

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\nif(" << iter->name
		<< "::ref_get_role_iterator_impl(visited, iterator, role_name)) {"
		<< "\nreturn true;"
		<< "\n}\n";
	}

	cpp << "\nreturn false;"
	    << "\n}"
	    << "\nelse {"
	    << "\nreturn false;"
	    << "\n}"
	    << "\n}\n";

	//
	// ref_get_attribute
	//

	cpp << "\nbool"
	    << "\n" << mClassDetails.name
	    << "::ref_get_attribute_impl(set<string> & visited,"
	    << "\nconst REF::Attribute * & attribute,"
	    << "\nconst string & attribute_name) const {"
	    << "\nif(visited.insert(TYPE).second) {";

	for(vector<AttributeDetails>::iterator iter = mAttributeDetails.begin();
	    iter != mAttributeDetails.end();
	    ++iter) {
	    if(iter->cardinality == UNARY) {
		cpp << "\nif(attribute_name == \"" << iter->role << "\") {"
		    << "\nattribute = get_" << iter->role << "();"
		    << "\nreturn true;"
		    << "\n}\n";
	    }
	}

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\nif(" << iter->name
		<< "::ref_get_attribute_impl(visited, attribute, attribute_name)) {"
		<< "\nreturn true;"
		<< "\n}\n";
	}

	cpp << "\nreturn false;"
	    << "\n}"
	    << "\nelse {"
	    << "\nreturn false;"
	    << "\n}"
	    << "\n}\n";

	//
	// ref_set_attribute
	//

	cpp << "\nbool"
	    << "\n" << mClassDetails.name
	    << "::ref_set_attribute_impl(set<string> & visited,"
	    << "\nconst string & attribute_name,"
	    << "\nconst REF::Attribute * attribute) {"
	    << "\nif(visited.insert(TYPE).second) {";

	for(vector<AttributeDetails>::iterator iter = mAttributeDetails.begin();
	    iter != mAttributeDetails.end();
	    ++iter) {
	    if(iter->cardinality == UNARY) {
		cpp << "\nif(attribute_name == \"" << iter->role << "\") {"
		    << "\nconst " << iter->type_name << " * " << iter->variable
		    << " = dynamic_cast<const " << iter->type_name << " *>(attribute);"
		    << "\n"
		    << "\nif(!" << iter->variable << ") {"
		    << "\nWARNING(\"Wrong type for attribute: \" + attribute->ref_get_type() +"
		    << "\n\", in class: \" + ref_get_type());"
		    << "\nthrow REF::BadType(\"Wrong type for attribute\");"
		    << "\n}"
		    << "\n"
		    << "\nset_" << iter->role << "(" << iter->variable << ");"
		    << "\nreturn true;"
		    << "\n}\n";
	    }
	}

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\nif(" << iter->name
		<< "::ref_set_attribute_impl(visited, attribute_name, attribute)) {"
		<< "\nreturn true;"
		<< "\n}\n";
	}

	cpp << "\nreturn false;"
	    << "\n}"
	    << "\nelse {"
	    << "\nreturn false;"
	    << "\n}"
	    << "\n}\n";

	//
	// ref_get_attribute_set_impl
	//

	cpp << "\nbool"
	    << "\n" << mClassDetails.name
	    << "::ref_get_attribute_set_impl(set<string> & visited,"
	    << "\nconst vector<const REF::Attribute *> * & attribute_set,"
	    << "\nconst string & attribute_name) const {"
	    << "\nif(visited.insert(TYPE).second) {";

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\nif(" << iter->name
		<< "::ref_get_attribute_set_impl(visited, attribute_set, attribute_name)) {"
		<< "\nreturn true;"
		<< "\n}\n";
	}

	cpp << "\nreturn false;"
	    << "\n}"
	    << "\nelse {"
	    << "\nreturn false;"
	    << "\n}"
	    << "\n}\n";

	//
	// ref_set_attribute_set_impl
	//

	cpp << "\nbool"
	    << "\n" << mClassDetails.name
	    << "::ref_set_attribute_set_impl(set<string> & visited,"
	    << "\nconst string & attribute_name,"
	    << "\nconst vector<const REF::Attribute *> * attribute_set) {"
	    << "\nif(visited.insert(TYPE).second) {";

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\nif(" << iter->name
		<< "::ref_set_attribute_set_impl(visited, attribute_name, attribute_set)) {"
		<< "\nreturn true;"
		<< "\n}\n";
	}

	cpp << "\nreturn false;"
	    << "\n}"
	    << "\nelse {"
	    << "\nreturn false;"
	    << "\n}"
	    << "\n}\n";

	//
	// ref_create_attribute
	//

	cpp << "\nbool"
	    << "\n" << mClassDetails.name
	    << "::ref_create_attribute_impl(set<string> & visited,"
	    << "\nREF::Attribute * & attribute,"
	    << "\nconst string & attribute_name) const {"
	    << "\nif(visited.insert(TYPE).second) {";

	for(vector<AttributeDetails>::iterator iter = mAttributeDetails.begin();
	    iter != mAttributeDetails.end();
	    ++iter) {
	    cpp << "\nif(attribute_name == \"" << iter->role << "\") {"
		<< "\nattribute = new " << iter->type_name << "();"
		<< "\nreturn true;"
		<< "\n}\n";
	}

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\nif(" << iter->name
		<< "::ref_create_attribute_impl(visited, attribute, attribute_name)) {"
		<< "\nreturn true;"
		<< "\n}\n";
	}

	cpp << "\nreturn false;"
	    << "\n}"
	    << "\nelse {"
	    << "\nreturn false;"
	    << "\n}"
	    << "\n}\n";

	//
	// ref_is_a_type_of
	//

	cpp << "\nbool"
	    << "\n" << mClassDetails.name
	    << "::ref_is_a_type_of_impl(set<string> & visited,"
	    << "\nconst string & class_name) const {"
	    << "\nif(visited.insert(TYPE).second) {";

	cpp << "\nif(class_name == \"" << mClassDetails.name << "\") {"
	    << "\nreturn true;"
	    << "\n}\n";

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\nif(" << iter->name
		<< "::ref_is_a_type_of_impl(visited, class_name)) {"
		<< "\nreturn true;"
		<< "\n}\n";
	}

	cpp << "\nreturn false;"
	    << "\n}"
	    << "\nelse {"
	    << "\nreturn false;"
	    << "\n}"
	    << "\n}\n";

	//

	cpp << "\nconst string &"
	    << "\n" << mClassDetails.name << "::ref_get_type() const {"
	    << "\nreturn TYPE;"
	    << "\n}\n";

	//

	cpp << "\nvoid"
	    << "\n" << mClassDetails.name
	    << "::ref_delete_impl(set<string> & visited) {"
	    << "\nif(visited.insert(TYPE).second) {";

	for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
	    iter != mReferenceDetails.end();
	    ++iter) {
	    if(iter->this_end_aggregate) {
		if(iter->other_cardinality == UNARY) {
		    cpp << "\nif(get_" << iter->other_end_role << "()) {"
			<< "\nget_" << iter->other_end_role << "()->ref_delete();"
			<< "\n}\n";
		}
		else if(iter->other_cardinality == MULTI) {
		    if(iter->other_ordering == ORDERED) {
			cpp << "\nlist<" << iter->other_type_name << " *> " << iter->other_end_role
			    << "_list = get_" << iter->other_end_role << "();"
			    << "\nfor_each("
			    << iter->other_end_role << "_list.begin(), "
			    << iter->other_end_role << "_list.end(),"
			    << "\nmem_fun(&REF::Class::ref_delete));\n";
		    }
		    else if(iter->other_ordering == UNORDERED) {
			cpp << "\nset<" << iter->other_type_name << " *> " << iter->other_end_role
			    << "_set = get_" << iter->other_end_role << "();"
			    << "\nfor_each("
			    << iter->other_end_role << "_set.begin(), "
			    << iter->other_end_role << "_set.end(),"
			    << "\nmem_fun(&REF::Class::ref_delete));\n";
		    }
		    else {
			ASSERT_NOT_REACHED();
		    }
		}
		else {
		    ASSERT_NOT_REACHED();
		}
	    }
	}

	for(vector<ReferenceDetails>::iterator iter = mReferenceDetails.begin();
	    iter != mReferenceDetails.end();
	    ++iter) {
	    if(iter->other_end_navigable) {
		if(iter->other_cardinality == UNARY) {
		    cpp << "\n" << iter->other_type_name << " * "
			<< iter->other_end_variable << " = 0;"
			<< "\nset_" << iter->other_end_role
			<< "(" << iter->other_end_variable << ");\n";
		}
		else if(iter->other_cardinality == MULTI) {
		    if(iter->other_ordering == ORDERED) {
			cpp << "\nlist<" << iter->other_type_name << " *> "
			    << iter->other_end_variable << ";"
			    << "\nset_" << iter->other_end_role
			    << "(" << iter->other_end_variable << ");\n";
		    }
		    else if(iter->other_ordering == UNORDERED) {
			cpp << "\nset<" << iter->other_type_name << " *> "
			    << iter->other_end_variable << ";"
			    << "\nset_" << iter->other_end_role
			    << "(" << iter->other_end_variable << ");\n";
		    }
		    else {
			ASSERT_NOT_REACHED();
		    }
		}
		else {
		    ASSERT_NOT_REACHED();
		}
	    }

	    // FIXME what to do if the other end isn't navigable
	}

	for(vector<ClassDetails>::iterator iter = mSupertypeDetails.begin();
	    iter != mSupertypeDetails.end();
	    ++iter) {
	    cpp << "\n" << iter->name << "::ref_delete_impl(visited);";
	}

	cpp << "\n}"
	    << "\n}\n";
    }

    //
    //
    //

    void
    Class::load_details() {
	ASSERT(mClass->get_name());
	ASSERT(mClass->is_abstract());

	//
	// load the details of this class
	//

	mClassDetails.klass = mClass;
	mClassDetails.name = mClass->get_name()->get_value();
	mClassDetails.role = name_to_role(mClassDetails.name);
	mClassDetails.variable = role_to_variable(mClassDetails.role);
	mClassDetails.qualified_name = mClass->get_qualified_name();
	mClassDetails.qualified_path = name_to_path(mClassDetails.qualified_name);
	mClassDetails.is_abstract = mClass->is_abstract()->get_value();
	mClassDetails.is_virtual = false;

	const list<MOF::ModelElement *> & contained_elements = mClass->get_contained_elements();

	for(list<MOF::ModelElement *>::const_iterator iter = contained_elements.begin();
	    iter != contained_elements.end();
	    ++iter) {
	    MOF::ModelElement * contained_element = *iter;
	    ASSERT(contained_element);

	    if(typeid(*contained_element) == typeid(MOF::Attribute)) {
		MOF::Attribute * attribute = known_cast<MOF::Attribute *>(contained_element);
		ASSERT(attribute->get_name());

		AttributeDetails details;

		details.attribute = attribute;
		details.name = attribute->get_name()->get_value();
		details.role = name_to_role(details.name);
		details.variable = role_to_variable(details.role);
		details.member = name_to_member(details.name);
		details.type = known_cast<MOF::DataType *>(attribute->get_type());
		ASSERT(details.type->get_name());
		details.type_name = details.type->get_name()->get_value();
		details.type_path = name_to_path(details.type_name);
		details.type_qualified_name = details.type->get_qualified_name();
		details.type_qualified_path = name_to_path(details.type_qualified_name);

		const MOF::MultiplicityType * mult = details.attribute->get_multiplicity();

		if(mult->get_upper()->get_value() == 1) {
		    details.cardinality = UNARY;
		}
		else {
		    details.cardinality = MULTI;
		}

		mAttributeDetails.push_back(details);
	    }
	    else if(typeid(*contained_element) == typeid(MOF::Reference)) {
		MOF::Reference * reference = known_cast<MOF::Reference *>(contained_element);

		ReferenceDetails details;

		details.reference = reference;

		//

		details.other_end = reference->get_referenced_end();
		ASSERT(details.other_end);
		ASSERT(details.other_end->get_name());
		details.other_end_name = details.other_end->get_name()->get_value();
		details.other_end_role = name_to_role(details.other_end_name);
		details.other_end_variable = role_to_variable(details.other_end_role);
		details.other_end_member = name_to_member(details.other_end_name);
		ASSERT(details.other_end->is_navigable());
		details.other_end_navigable =
		    details.other_end->is_navigable()->get_value();
		details.other_end_aggregate =
		    details.other_end->get_aggregation()->get_value() != "none";

		details.other_type = known_cast<MOF::Class *>(details.other_end->get_type());
		ASSERT(details.other_type->get_name());
		details.other_type_name = details.other_type->get_name()->get_value();
		details.other_type_path = name_to_path(details.other_type_name);
		details.other_type_qualified_name = details.other_type->get_qualified_name();
		details.other_type_qualified_path = name_to_path(details.other_type_qualified_name);

		const MOF::MultiplicityType * other_mult = details.other_end->get_multiplicity();
		ASSERT(other_mult);

		ASSERT(other_mult->is_unique()->get_value());		// all we can handle

		if(other_mult->get_upper()->get_value() == 1) {
		    details.other_cardinality = UNARY;
		}
		else {
		    details.other_cardinality = MULTI;
		}

		if(other_mult->get_ordered()->get_value()) {
		    details.other_ordering = ORDERED;
		}
		else {
		    //details.other_ordering = UNORDERED;		FIXME
		    details.other_ordering = ORDERED;
		}

		//

		details.this_end = details.other_end->get_other_end();
		ASSERT(details.this_end);
		ASSERT(details.this_end->get_name());
		details.this_end_name = details.this_end->get_name()->get_value();
		details.this_end_role = name_to_role(details.this_end_name);
		details.this_end_variable = role_to_variable(details.this_end_role);
		details.this_end_member = name_to_member(details.this_end_name);
		ASSERT(details.this_end->is_navigable());
		details.this_end_navigable = details.this_end->is_navigable()->get_value();
		details.this_end_aggregate =
		    details.this_end->get_aggregation()->get_value() == "composite";

		details.this_type = known_cast<MOF::Class *>(details.this_end->get_type());
		ASSERT(details.this_type == mClass);
		ASSERT(details.this_type->get_name());
		details.this_type_name = details.this_type->get_name()->get_value();
		details.this_type_path = name_to_path(details.this_type_name);
		details.this_type_qualified_name = details.this_type->get_qualified_name();
		details.this_type_qualified_path = name_to_path(details.this_type_qualified_name);

		const MOF::MultiplicityType * this_mult = details.this_end->get_multiplicity();
		ASSERT(this_mult);

		if(this_mult->get_upper()->get_value() == 1) {
		    details.this_cardinality = UNARY;
		}
		else {
		    details.this_cardinality = MULTI;
		    ASSERT(this_mult->is_unique()->get_value());		// all we can handle
		}

		if(this_mult->get_ordered()->get_value()) {
		    details.this_ordering = ORDERED;
		}
		else {
		    //details.this_ordering = UNORDERED;		FIXME
		    details.this_ordering = ORDERED;
		}

		//

		mReferenceDetails.push_back(details);
	    }
	}

	const list<MOF::GeneralizableElement *> & supertypes = mClass->get_supertypes();

	for(list<MOF::GeneralizableElement *>::const_iterator iter = supertypes.begin();
	    iter != supertypes.end();
	    ++iter) {
	    MOF::GeneralizableElement * supertype = *iter;
	    ASSERT(supertype);

	    MOF::Class * klass = known_cast<MOF::Class *>(supertype);

	    ASSERT(klass->get_name());
	    ASSERT(klass->is_abstract());

	    ClassDetails details;

	    details.klass = klass;
	    details.name = klass->get_name()->get_value();
	    details.role = name_to_role(details.name);
	    details.variable = role_to_variable(details.role);
	    details.qualified_name = klass->get_qualified_name();
	    details.qualified_path = name_to_path(details.qualified_name);
	    details.is_abstract = klass->is_abstract()->get_value();
	    details.is_virtual = false;		// FIXME

	    // FIXME hack to save analysing the class hierarchy


	    // this is for UML
	    if(mModelName == "UML") {
		if(mClassDetails.name == "Namespace" ||
		   mClassDetails.name == "GeneralizableElement" ||
		   mClassDetails.name == "EdgeElement" ||
		   mClassDetails.name == "NodeElement" ||
		   mClassDetails.name == "Relationship") {
		    details.is_virtual = true;
		}
	    }
	    else if(mModelName == "MOF2") {
		// this is for MOF
		if(mClassDetails.name == "Feature" ||
		   mClassDetails.name == "Namespace" ||
		   mClassDetails.name == "TypedElement") {
		    details.is_virtual = true;
		}
	    }

	    mSupertypeDetails.push_back(details);
	}
    }

} // namespace Generator
