// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef GENERATOR_CLASS_H
#define GENERATOR_CLASS_H

#include "foundation/object.h"

#include <string>
#include <vector>
#include <set>
using namespace std;

namespace Common {

    class CppStream;

} // namespace Common

namespace MOF {

    class Class;
    class Attribute;
    class Reference;
    class AssociationEnd;
    class DataType;
    class MultiplicityType;

} // namespace MOF

namespace Generator {

    class Class : public Object {
    public:
	Class(MOF::Class * klass, const string & model_name);

	virtual void generate(const string & output_dir, const string & include_dir);

    protected:
	~Class();

	virtual void load_details();

	//
	// header
	//

	virtual void generate_header(Common::CppStream & hpp);
	virtual void generate_header_includes(Common::CppStream & hpp);
	virtual void generate_header_forward_declarations(Common::CppStream & hpp);
	virtual void generate_header_public_members(Common::CppStream & hpp);
	virtual void generate_header_protected_members(Common::CppStream & hpp);
	virtual void generate_header_private_members(Common::CppStream & hpp);

	//
	// source
	//

	virtual void generate_source(Common::CppStream & cpp);
	virtual void generate_source_includes(Common::CppStream & cpp);
	virtual void generate_source_statics(Common::CppStream & cpp);
	virtual void generate_source_constructors(Common::CppStream & cpp);
	virtual void generate_source_destructor(Common::CppStream & cpp);
	virtual void generate_source_body(Common::CppStream & cpp);

	//

	MOF::Class * mClass;
	string mModelName;

	enum Cardinality {
	    UNARY,
	    MULTI
	};

	enum Ordering {
	    UNORDERED,
	    ORDERED
	};

	struct ClassDetails {
	    MOF::Class * klass;
	    string name;
	    string role;
	    string variable;
	    string qualified_name;
	    string qualified_path;
	    bool is_abstract;
	    bool is_virtual;
	};

	struct AttributeDetails {
	    MOF::Attribute * attribute;
	    string name;
	    string role;
	    string variable;
	    string member;
	    MOF::DataType * type;
	    string type_name;
	    string type_path;
	    string type_qualified_name;
	    string type_qualified_path;
	    Cardinality cardinality;
	};

	struct ReferenceDetails {
	    MOF::Reference * reference;

	    // this end/type

	    MOF::AssociationEnd * this_end;
	    string this_end_name;
	    string this_end_role;
	    string this_end_variable;
	    string this_end_member;
	    bool this_end_navigable;
	    bool this_end_aggregate;

	    MOF::Class * this_type;
	    string this_type_name;
	    string this_type_path;
	    string this_type_qualified_name;
	    string this_type_qualified_path;

	    Cardinality this_cardinality;
	    Ordering this_ordering;

	    // other end/type

	    MOF::AssociationEnd * other_end;
	    string other_end_name;
	    string other_end_role;
	    string other_end_variable;
	    string other_end_member;
	    bool other_end_navigable;
	    bool other_end_aggregate;

	    MOF::Class * other_type;
	    string other_type_name;
	    string other_type_path;
	    string other_type_qualified_name;
	    string other_type_qualified_path;

	    Cardinality other_cardinality;
	    Ordering other_ordering;
	};

	ClassDetails mClassDetails;

	vector<AttributeDetails> mAttributeDetails;
	vector<ReferenceDetails> mReferenceDetails;
	vector<ClassDetails> mSupertypeDetails;

    private:
	void generate_source_body_unary_attribute(Common::CppStream & cpp,
						  const AttributeDetails & details);
	void generate_source_body_multi_attribute(Common::CppStream & cpp,
						  const AttributeDetails & details);
	void generate_source_body_forward_navigable_list_role(Common::CppStream & cpp,
							      const ReferenceDetails & details);
	void generate_source_body_reverse_navigable_list_role(Common::CppStream & cpp,
							      const ReferenceDetails & details);
	void generate_source_body_forward_navigable_set_role(Common::CppStream & cpp,
							     const ReferenceDetails & details);
	void generate_source_body_reverse_navigable_set_role(Common::CppStream & cpp,
							     const ReferenceDetails & details);
	void generate_source_body_forward_navigable_pointer_role(Common::CppStream & cpp,
								 const ReferenceDetails & details);
	void generate_source_body_reverse_navigable_pointer_role(Common::CppStream & cpp,
								 const ReferenceDetails & details);
	void generate_source_body_uproot_implant(Common::CppStream & cpp);
	void generate_source_body_reflective(Common::CppStream & cpp);

	bool mDetailsLoaded;
    };

} // namespace

#endif // GENERATOR_CLASS_H
