// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "handler.h"

#include "mof/class.h"
#include "mof/package.h"
#include "mof/model.h"
#include "mof/reference.h"
#include "mof/association_end.h"
#include "mof/attribute.h"
#include "mof/data_type.h"
#include "mof/primitive_type.h"
#include "mof/structure_type.h"
#include "mof/collection_type.h"
#include "mof/alias_type.h"
#include "mof/enumeration_type.h"

#include "common/file_utilities.h"
#include "common/diff_stream.h"
#include "common/cpp_stream.h"

#include "foundation/macros.h"
#include "foundation/utility.h"

#include "utility.h"

#include <typeinfo>
#include <iostream>
using namespace std;

namespace Generator {

    Handler::Handler(MOF::Package * root_package,
		     const string & model_name) : IO(root_package, model_name) {
    }

    Handler::~Handler() {
    }

    void
    Handler::generate(const string & output_dir, const string & include_dir) {
	IO::generate(output_dir, include_dir);

	string full_path = output_dir + "/" + mHandlerDetails.qualified_path;

	Common::DiffStream hpp_stream(full_path + ".h");
	ASSERT(hpp_stream);
	Common::CppStream hpp(hpp_stream);

	Common::DiffStream cpp_stream(full_path + ".cc");
	ASSERT(cpp_stream);
	Common::CppStream cpp(cpp_stream);

	generate_header(hpp);
	generate_source(cpp);
    }

    void
    Handler::generate_header(Common::CppStream & hpp) {
	string guard = path_to_guard(mHandlerDetails.qualified_path);

	hpp << get_prologue()
	    << "\n#ifndef " << guard
	    << "\n#define " << guard << "\n";

	hpp << "\n#include \"xml/sax_handler.h\"\n";

	hpp << "\n#include \"reflective/class.h\"\n";

	//

	hpp << "\nnamespace " << mModelName << " {\n";

	//

	hpp << "\nclass Handler : public XML::SAXHandler {";

	hpp << "\npublic:";

	hpp << "\nHandler();"
	    << "\n~Handler();\n";

	//

	hpp << "\n// the sax interface"
	    << "\n"
	    << "\nvoid warning(XML::SAXParser * parser, const string & msg);"
	    << "\nvoid error(XML::SAXParser * parser, const string & msg);"
	    << "\nvoid fatal(XML::SAXParser * parser, const string & msg);"
	    << "\n"
	    << "\nvoid comment(XML::SAXParser * parser, const string & comment);"
	    << "\nvoid start_document(XML::SAXParser * parser);"
	    << "\nvoid end_document(XML::SAXParser * parser);"
	    << "\nvoid start_element(XML::SAXParser * parser, const string & name, "
	    << "const XML::Properties & properties);"
	    << "\nvoid end_element(XML::SAXParser * parser, const string & name);"
	    << "\nvoid characters(XML::SAXParser * parser, const string & data);"
	    << "\nvoid reference(XML::SAXParser * parser, const string & name);"
	    << "\nvoid cdata_block(XML::SAXParser * parser, const string & data);\n";

	//

	hpp << "\nprivate:";

	//

	string model_role = name_to_role(mModelName);

	// concrete classes

	for(list<ClassDetails>::iterator iter = mConcreteClasses.begin();
	    iter != mConcreteClasses.end();
	    ++iter) {
	    string member = model_role + "__" + iter->role;

	    hpp << "\nvoid "
		<< member << "__start(XML::SAXParser * parser, const XML::Properties & properties);"
		<< "\nvoid "
		<< member << "__end(XML::SAXParser * parser);\n";
	}

	// class roles

	for(list<ClassDetails>::iterator iter = mAllClasses.begin();
	    iter != mAllClasses.end();
	    ++iter) {
	    for(list<ReferenceDetails>::iterator iter2 = iter->references.begin();
		iter2 != iter->references.end();
		++iter2) {
		string member = model_role + "__" + iter->role + "__" + iter2->other_end_role;

		hpp << "\nvoid "
		    << member << "__start(XML::SAXParser * parser, const XML::Properties & properties);"
		    << "\nvoid "
		    << member << "__end(XML::SAXParser * parser);\n";
	    }

	    for(list<AttributeDetails>::iterator iter2 = iter->attributes.begin();
		iter2 != iter->attributes.end();
		++iter2) {
		string member = model_role + "__" + iter->role + "__" + iter2->role;

		hpp << "\nvoid "
		    << member << "__start(XML::SAXParser * parser, const XML::Properties & properties);"
		    << "\nvoid "
		    << member << "__end(XML::SAXParser * parser);\n";
	    }
	}

	hpp << "\n//"
	    << "\n//"
	    << "\n//\n";

	//

	hpp << "\nstring mCharacters;"
	    << "\n"
	    << "\ntypedef void (Handler::*StartHandler)(XML::SAXParser *, const XML::Properties &);"
	    << "\ntypedef void (Handler::*EndHandler)(XML::SAXParser *);"
	    << "\ntypedef pair<StartHandler, EndHandler> HandlerPair;"
	    << "\n"
	    << "\nhash_map<string, HandlerPair> mHandlerMap;"
	    << "\nhash_map<string, Object *> mElementMap;"
	    << "\n"
	    << "\nset<string> mUnhandledElements;"
	    << "\n"
	    << "\nvoid store(const string & id, Object * object);"
	    << "\nObject * lookup(const string & id);";

	//

	hpp << "\n};\n";

	hpp << "\n} // namespace " << mModelName << "\n";

	hpp << "\n#endif // " << guard << "\n";
    }

    void
    Handler::generate_source(Common::CppStream & cpp) {
	cpp << get_prologue()
	    << "\n#ifdef HAVE_CONFIG_H"
	    << "\n#include \"config.h\""
	    << "\n#endif // HAVE_CONFIG_H\n";

	cpp << "\n#include \"handler.h\"\n";

	// include all the headers of the classes in the model

	for(list<ClassDetails>::iterator iter = mAllClasses.begin();
	    iter != mAllClasses.end();
	    ++iter) {
	    cpp << "\n#include \"" << iter->qualified_path << ".h\"";
	}

	cpp << "\n";

	// include all the headers of the data_types in the model

	for(list<DataTypeDetails>::iterator iter = mDataTypeDetails.begin();
	    iter != mDataTypeDetails.end();
	    ++iter) {
	    cpp << "\n#include \"" << iter->qualified_path << ".h\"";
	}

	//

	cpp << "\n#include \"xml/sax_parser.h\""
	    << "\n";

	cpp << "\n#include \"common/dumb_handler.h\""
	    << "\n";

	cpp << "\n#include \"foundation/iterator.h\""
	    << "\n#include \"foundation/utility.h\""
	    << "\n#include \"foundation/macros.h\""
	    << "\n";

	cpp << "\n";

	cpp << "\n#include <iterator>"
	    << "\n#include <functional>"
	    << "\n#include <algorithm>"
	    << "\n#include <sstream>"
	    << "\nusing namespace std;"
	    << "\n";

	//

	cpp << "\nnamespace " << mModelName << " {\n";

	//

	string model_role = name_to_role(mModelName);

	//

	cpp << "\nHandler::Handler() : XML::SAXHandler() {";

	for(list<ClassDetails>::iterator iter = mConcreteClasses.begin();
	    iter != mConcreteClasses.end();
	    ++iter) {
	    string tag = mModelName + ":" + iter->name;
	    string member = model_role + "__" + iter->role;

	    cpp << "\nmHandlerMap[\"" << tag << "\"] ="
		<< "\nmake_pair(&Handler::" << member << "__start,"
		<< "\n&Handler::" << member << "__end);\n";
	}

	for(list<ClassDetails>::iterator iter = mAllClasses.begin();
	    iter != mAllClasses.end();
	    ++iter) {
	    for(list<ReferenceDetails>::iterator iter2 = iter->references.begin();
		iter2 != iter->references.end();
		++iter2) {
		if(!iter2->other_end_navigable) {
		    continue;
		}

		string tag = mModelName + ":" + iter->name + "." + iter2->other_end_name;
		string member = model_role + "__" + iter->role + "__" + iter2->other_end_role;

		cpp << "\nmHandlerMap[\"" << tag << "\"] ="
		    << "\nmake_pair(&Handler::" << member << "__start,"
		    << "\n&Handler::" << member << "__end);\n";
	    }

	    for(list<AttributeDetails>::iterator iter2 = iter->attributes.begin();
		iter2 != iter->attributes.end();
		++iter2) {
		string tag = mModelName + ":" + iter->name + "." + iter2->name;
		string member = model_role + "__" + iter->role + "__" + iter2->role;
		cpp << "\nmHandlerMap[\"" << tag << "\"] ="
		    << "\nmake_pair(&Handler::" << member << "__start,"
		    << "\n&Handler::" << member << "__end);\n";
	    }
	}

	cpp << "\n}\n";

	//

	cpp << "\nHandler::~Handler() {"
	    << "\n}\n";

	//

    cpp << "\nvoid"
	<< "\nHandler::warning(XML::SAXParser * parser, const string & msg) {"
	<< "\nWARNING(\"Unhandled warning: \" << msg);"
	<< "\n}"
	<< "\n"
	<< "\nvoid"
	<< "\nHandler::error(XML::SAXParser * parser, const string & msg) {"
	<< "\nWARNING(\"Unhandled error: \" << msg);"
	<< "\n}"
	<< "\n"
	<< "\nvoid"
	<< "\nHandler::fatal(XML::SAXParser * parser, const string & msg) {"
	<< "\nWARNING(\"Unhandled fatal: \" << msg);"
	<< "\n}"
	<< "\n"
	<< "\nvoid"
	<< "\nHandler::comment(XML::SAXParser * parser, const string & comment) {"
	<< "\n}"
	<< "\n"
	<< "\nvoid"
	<< "\nHandler::start_document(XML::SAXParser * parser) {"
	<< "\nif(parser->get_pass() == 1) {"
	<< "\n}"
	<< "\n}"
	<< "\n"
	<< "\nvoid"
	<< "\nHandler::end_document(XML::SAXParser * parser) {"
	<< "\nif(parser->get_pass() == 2) {"
	<< "\nfor(hash_map<string, Object *>::iterator iter = mElementMap.begin();"
	<< "\niter != mElementMap.end();"
	<< "\n++iter) {"
	<< "\niter->second->unref();"
	<< "\n}"
	<< "\n"
	<< "\nmElementMap.clear();"
	<< "\n}"
	<< "\n}"
	<< "\n"
	<< "\nvoid"
	<< "\nHandler::start_element(XML::SAXParser * parser, const string & name,"
	<< "\nconst XML::Properties & properties) {"
	<< "\nTRACE(5, \"Start: \" << name);"
	<< "\n"
	<< "\nmCharacters.resize(0);"
	<< "\n"
	<< "\nhash_map<string, HandlerPair>::iterator iter = mHandlerMap.find(name);"
	<< "\n"
	<< "\nif(iter == mHandlerMap.end()) {"
	<< "\nif(mUnhandledElements.insert(name).second) {"
	<< "\nWARNING(\"Unhandled element: \" << name);"
	<< "\n}"
	<< "\n"
	<< "\n// push a dumb handler so that we don't try to parse"
	<< "\n// the contents of the tag"
	<< "\n"
	<< "\nXML::SAXHandler * handler = new Common::DumbHandler();"
	<< "\nparser->use(handler);"
	<< "\nhandler->unref();"
	<< "\n}"
	<< "\nelse {"
	<< "\nHandlerPair handler_pair = iter->second;"
	<< "\n(this->*handler_pair.first)(parser, properties);"
	<< "\n}"
	<< "\n}"
	<< "\n"
	<< "\nvoid"
	<< "\nHandler::end_element(XML::SAXParser * parser, const string & name) {"
	<< "\nhash_map<string, HandlerPair>::iterator iter = mHandlerMap.find(name);"
	<< "\n"
	<< "\nif(iter == mHandlerMap.end()) {"
	<< "\n}"
	<< "\nelse {"
	<< "\nHandlerPair handler_pair = iter->second;"
	<< "\n(this->*handler_pair.second)(parser);"
	<< "\n}"
	<< "\n"
	<< "\nTRACE(5, \"End: \" << name);"
	<< "\n}"
	<< "\n"
	<< "\nvoid"
	<< "\nHandler::characters(XML::SAXParser * parser, const string & data) {"
	<< "\nmCharacters += data;"
	<< "\n}"
	<< "\n"
	<< "\nvoid"
	<< "\nHandler::reference(XML::SAXParser * parser, const string & name) {"
	<< "\nWARNING(\"Unhandled Reference: \" << name);"
	<< "\n}"
	<< "\n"
	<< "\nvoid"
	<< "\nHandler::cdata_block(XML::SAXParser * parser, const string & data) {"
	<< "\nWARNING(\"Unhandled CDATA: \" << data);"
	<< "\n}"
	<< "\n"
	<< "\nvoid"
	<< "\nHandler::store(const string & id, Object * object) {"
	<< "\nif(!mElementMap.insert(make_pair(id, object)).second) {"
	<< "\nERROR(\"Duplicate id: \" << id);"
	<< "\n}"
	<< "\n}"
	<< "\n"
	<< "\nObject *"
	<< "\nHandler::lookup(const string & id) {"
	<< "\nhash_map<string, Object *>::iterator iter = mElementMap.find(id);"
	<< "\n"
	<< "\nif(iter == mElementMap.end()) {"
	<< "\nERROR(\"Couldn't locate element with id: \" << id);"
	<< "\nthrow;	// FIXME"
	<< "\n}"
	<< "\nelse {"
	<< "\nreturn iter->second;"
	<< "\n}"
	<< "\n}\n";


	// concrete classes

	for(list<ClassDetails>::iterator iter = mConcreteClasses.begin();
	    iter != mConcreteClasses.end();
	    ++iter) {
	    string member = model_role + "__" + iter->role;

	    cpp << "\nvoid"
		<< "\nHandler::"
		<< member << "__start(XML::SAXParser * parser, const XML::Properties & properties) {"
		<< "\nif(parser->get_pass() == 1) {"
		<< "\nstring id;"
		<< "\n"
		<< "\nif(properties.get_implied_property(\"xmi.id\", id)) {"
		<< "\nstore(id, new " << iter->name << "());"
		<< "\n}"
		<< "\n}"
		<< "\nelse {"
		<< "\nstring idref;"
		<< "\nstring id;"
		<< "\n"
		<< "\nObject * " << iter->variable << ";"
		<< "\n"
		<< "\nif(properties.get_implied_property(\"xmi.idref\", idref)) {"
		<< "\n" << iter->variable << " = lookup(idref);"
		<< "\n" << iter->variable << "->ref();"
		<< "\n}"
		<< "\nelse if(properties.get_implied_property(\"xmi.id\", id)) {"
		<< "\n" << iter->variable << " = lookup(id);"
		<< "\n" << iter->variable << "->ref();"
		<< "\n}"
		<< "\nelse {"
		<< "\n" << iter->variable << " = new " << iter->name << "();"
		<< "\n}"
		<< "\n"
		<< "\nASSERT(" << iter->variable << ");"
		<< "\n"
		<< "\nparser->top().push_back(" << iter->variable << ");"
		<< "\n}"
		<< "\n}"
		<< "\n"
		<< "\nvoid"
		<< "\nHandler::"
		<< member << "__end(XML::SAXParser * parser) {"
		<< "\nif(parser->get_pass() == 1) {"
		<< "\n}"
		<< "\nelse {"
		<< "\n}"
		<< "\n}"
		<< "\n";
	}

	// class roles

	for(list<ClassDetails>::iterator iter = mAllClasses.begin();
	    iter != mAllClasses.end();
	    ++iter) {
	    for(list<ReferenceDetails>::iterator iter2 = iter->references.begin();
		iter2 != iter->references.end();
		++iter2) {
		if(!iter2->other_end_navigable) {
		    continue;
		}

		string member = model_role + "__" + iter->role + "__" + iter2->other_end_role;

		cpp << "\nvoid"
		    << "\nHandler::"
		    << member << "__start(XML::SAXParser * parser, const XML::Properties & properties) {"
		    << "\nif(parser->get_pass() == 1) {"
		    << "\n}"
		    << "\nelse {"
		    << "\nparser->push();"
		    << "\n}"
		    << "\n}"
		    << "\n"
		    << "\nvoid"
		    << "\nHandler::"
		    << member << "__end(XML::SAXParser * parser) {"
		    << "\nif(parser->get_pass() == 1) {"
		    << "\n}"
		    << "\nelse {";

		if(iter2->other_cardinality == UNARY) {
		    cpp << "\nASSERT(parser->top().size() == 1);"
			<< "\n"
			<< "\n" << iter2->other_type_name << " * " << iter2->other_end_variable
			<< " = known_cast<" << iter2->other_type_name
			<< " *>(parser->top().back());";
		}
		else if(iter2->other_cardinality == MULTI) {
		    if(iter2->other_ordering == ORDERED) {
			cpp << "\nconst vector<Object *> & objects = parser->top();"
			    << "\nlist<" << iter2->other_type_name << " *> "
			    << iter2->other_end_variable << ";"
			    << "\n"
			    << "\ntransform(objects.begin(), objects.end(), back_inserter("
			    << iter2->other_end_variable << "), Caster<Object, "
			    << iter2->other_type_name << ">());";
		    }
		    else if(iter2->other_ordering == UNORDERED) {
			cpp << "\nconst vector<Object *> & objects = parser->top();"
			    << "\nset<" << iter2->other_type_name << " *> "
			    << iter2->other_end_role << ";"
			    << "\n"
			    << "\ntransform(objects.begin(), objects.end(), associative_inserter("
			    << iter2->other_end_role << "), Caster<Object, "
			    << iter2->other_type_name << ">());";
		    }
		    else {
			ASSERT_NOT_REACHED();
		    }
		}
		else {
		    ASSERT_NOT_REACHED();
		}

		cpp << "\n"
		    << "\nparser->pop();"
		    << "\n"
		    << "\nASSERT(!parser->top().empty());"
		    << "\n"
		    << "\n" << iter->name << " * " << iter->variable
		    << " = known_cast<" << iter->name << " *>(parser->top().back());"
		    << "\n"
		    << "\n" << iter->variable << "->set_" << iter2->other_end_role
		    << "(" << iter2->other_end_variable << ");"
		    << "\n";

		if(iter2->other_cardinality == UNARY) {
		    cpp << "\n" << iter2->other_end_variable << "->unref();";
		}
		else if(iter2->other_cardinality == MULTI) {
		    if(iter2->other_ordering == ORDERED) {
		    }
		    else if(iter2->other_ordering == UNORDERED) {
		    }
		    else {
			ASSERT_NOT_REACHED();
		    }
		}
		else {
		    ASSERT_NOT_REACHED();
		}

		cpp << "\n}"
		    << "\n}"
		    << "\n";
	    }

	    for(list<AttributeDetails>::iterator iter2 = iter->attributes.begin();
		iter2 != iter->attributes.end();
		++iter2) {
		string member = model_role + "__" + iter->role + "__" + iter2->role;

		cpp << "\nvoid"
		    << "\nHandler::" << member
		    << "__start(XML::SAXParser * parser, const XML::Properties & properties) {";

		cpp << "\n}\n";

		cpp << "\nvoid"
		    << "\nHandler::" << member << "__end(XML::SAXParser * parser) {"
		    << "\nif(parser->get_pass() == 1) {"
		    << "\n}"
		    << "\nelse {";

		if(typeid(*iter2->type) == typeid(MOF::PrimitiveType)) {
		    // temp hack FIXME

		    /*
		    cpp << "\n" << iter->name << " * " << iter->variable
			<< " = known_cast<" << iter->name << " *>(parser->top().back());"
			<< "\n" << iter->variable << "->ref_set_attribute(\"" << iter2->role << "\""
			<< ", mCharacters);";
			*/
		}
		else {
		    NYI();
		}

		cpp << "\n}";

		cpp << "\n}\n";
	    }
	}

	//

	cpp << "\n} // namespace " << mModelName << "\n";
    }

    void
    Handler::load_details() {
	IO::load_details();

	mHandlerDetails.qualified_path = name_to_path(mPackage->get_qualified_name()) + "/handler";
    }

} // namespace Generator
