// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "package.h"

#include "mof/package.h"
#include "mof/class.h"
#include "mof/primitive_type.h"
#include "mof/structure_type.h"
#include "mof/enumeration_type.h"
#include "mof/collection_type.h"
#include "mof/alias_type.h"

#include "common/file_utilities.h"
#include "common/diff_stream.h"

#include "foundation/macros.h"
#include "foundation/utility.h"

#include "utility.h"
#include "class.h"
#include "primitive_type.h"
#include "structure_type.h"
#include "enumeration_type.h"
#include "collection_type.h"
#include "alias_type.h"

#include <typeinfo>
#include <iostream>
using namespace std;

namespace Generator {

    Package::Package(MOF::Package * package, const string & model_name) : Object(),
        mPackage(package),
	mModelName(model_name),
	mDetailsLoaded(false) {
	ASSERT(mPackage);
	mPackage->ref();
    }

    Package::~Package() {
	mPackage->unref();
    }

    void
    Package::generate(const string & output_dir, const string & include_dir, int depth) {
	if(!mDetailsLoaded) {
	    load_details();
	    mDetailsLoaded = true;
	}

	string full_path = output_dir + "/" + mPackageDetails.qualified_path;

	Common::ensure_directory(full_path);

	Common::DiffStream mam(full_path + "/Makefile.am");
	ASSERT(mam);
	generate_makefile(mam);

	Common::DiffStream ign(full_path + "/.cvsignore");
	ASSERT(ign);
	generate_cvsignore(ign);

	//

	cout << "    Generating classes... (" << depth << ")" << endl;

	for(vector<ClassDetails>::iterator iter = mClassDetails.begin();
	    iter != mClassDetails.end();
	    ++iter) {
	    Class * klass = new Class(iter->klass, mModelName);

	    try {
		klass->generate(output_dir, include_dir);
	    }
	    catch(...) {
		klass->unref();
		throw;
	    }

	    klass->unref();
	}

	cout << "    Generating primitive types... (" << depth << ")" << endl;

	for(vector<PrimitiveTypeDetails>::iterator iter = mPrimitiveTypeDetails.begin();
	    iter != mPrimitiveTypeDetails.end();
	    ++iter) {
	    PrimitiveType * primitive_type = new PrimitiveType(iter->primitive_type, mModelName);

	    try {
		primitive_type->generate(output_dir, include_dir);
	    }
	    catch(...) {
		primitive_type->unref();
		throw;
	    }

	    primitive_type->unref();
	}

	cout << "    Generating structure types... (" << depth << ")" << endl;

	for(vector<StructureTypeDetails>::iterator iter = mStructureTypeDetails.begin();
	    iter != mStructureTypeDetails.end();
	    ++iter) {
	    StructureType * structure_type = new StructureType(iter->structure_type, mModelName);

	    try {
		structure_type->generate(output_dir, include_dir);
	    }
	    catch(...) {
		structure_type->unref();
		throw;
	    }

	    structure_type->unref();
	}

	cout << "    Generating enumeration types... (" << depth << ")" << endl;

	for(vector<EnumerationTypeDetails>::iterator iter = mEnumerationTypeDetails.begin();
	    iter != mEnumerationTypeDetails.end();
	    ++iter) {
	    EnumerationType * enumeration_type =
		new EnumerationType(iter->enumeration_type, mModelName);

	    try {
		enumeration_type->generate(output_dir, include_dir);
	    }
	    catch(...) {
		enumeration_type->unref();
		throw;
	    }

	    enumeration_type->unref();
	}

	cout << "    Generating collection types... (" << depth << ")" << endl;

	for(vector<CollectionTypeDetails>::iterator iter = mCollectionTypeDetails.begin();
	    iter != mCollectionTypeDetails.end();
	    ++iter) {
	    CollectionType * collection_type =
		new CollectionType(iter->collection_type, mModelName);

	    try {
		collection_type->generate(output_dir, include_dir);
	    }
	    catch(...) {
		collection_type->unref();
		throw;
	    }

	    collection_type->unref();
	}

	cout << "    Generating alias types... (" << depth << ")" << endl;

	for(vector<AliasTypeDetails>::iterator iter = mAliasTypeDetails.begin();
	    iter != mAliasTypeDetails.end();
	    ++iter) {
	    AliasType * alias_type = new AliasType(iter->alias_type, mModelName);

	    try {
		alias_type->generate(output_dir, include_dir);
	    }
	    catch(...) {
		alias_type->unref();
		throw;
	    }

	    alias_type->unref();
	}

	cout << "    Generating subpackages... (" << depth << ")" << endl;

	for(vector<PackageDetails>::iterator iter = mSubPackageDetails.begin();
	    iter != mSubPackageDetails.end();
	    ++iter) {
	    Package * package = new Package(iter->package, mModelName);

	    try {
		package->generate(output_dir, include_dir, depth + 1);
	    }
	    catch(...) {
		package->unref();
		throw;
	    }

	    package->unref();
	}
    }

    void
    Package::generate_makefile(Common::DiffStream & mam) {
	//
	// includes, same for every makefile
	//

	mam << "INCLUDES += \\"
	    << "\n\t-I$(top_srcdir)/src\n";

	//
	// subdirectories
	//

	if(!mSubdirectories.empty()) {
	    mam << "\nSUBDIRS =";

	    for(set<string>::iterator iter = mSubdirectories.begin();
		iter != mSubdirectories.end();
		++iter) {
		mam << " \\\n\t" << *iter;
	    }
	}

	//
	// sources and headers
	//

	mam << "\n\nCCFILES =";

	for(set<string>::iterator iter = mBuiltSources.begin();
	    iter != mBuiltSources.end();
	    ++iter) {
	    mam << " \\\n\t" << *iter;
	}

	mam << "\n\nHFILES =";

	for(set<string>::iterator iter = mBuiltHeaders.begin();
	    iter != mBuiltHeaders.end();
	    ++iter) {
	    mam << " \\\n\t" << *iter;
	}

	//
	// common stuff
	//

	mam << "\n\nEXTRA_DIST = $(CCFILES) $(HFILES)\n";

	mam << "\nBUILT_SOURCES = $(CCFILES) $(HFILES)\n";

	mam << "\nvery-clean: clean\n"
	    << "\trm -f $(BUILT_SOURCES) $(CC_FILES:.cc=.o) $(CC_FILES:.cc=.lo)\n";
    }

    void
    Package::generate_cvsignore(Common::DiffStream & ign) {
	ign << "Makefile\n"
	    << "Makefile.in\n"
	    << ".dirstamp\n"
	    << ".deps\n"
	    << ".libs\n"
	    << "*.lo\n"
	    << "*.la\n";

	for(set<string>::iterator iter = mBuiltSources.begin();
	    iter != mBuiltSources.end();
	    ++iter) {
	    ign << *iter << "\n";
	}

	for(set<string>::iterator iter = mBuiltHeaders.begin();
	    iter != mBuiltHeaders.end();
	    ++iter) {
	    ign << *iter << "\n";
	}
    }

    void
    Package::load_details() {
	ASSERT(mPackage->get_name());

	//
	// load the details of this package
	//

	mPackageDetails.package = mPackage;
	mPackageDetails.name = mPackage->get_name()->get_value();
	mPackageDetails.path = name_to_path(mPackageDetails.name);
	mPackageDetails.qualified_name = mPackage->get_qualified_name();
	mPackageDetails.qualified_path = name_to_path(mPackageDetails.qualified_name);

	//
	// load the details of contained elements
	//

	const list<MOF::ModelElement *> & contained_elements = mPackage->get_contained_elements();

	for(list<MOF::ModelElement *>::const_iterator iter = contained_elements.begin();
	    iter != contained_elements.end();
	    ++iter) {
	    MOF::ModelElement * contained_element = *iter;

	    if(typeid(*contained_element) == typeid(MOF::Package)) {
		MOF::Package * sub_package = known_cast<MOF::Package *>(contained_element);

		ASSERT(sub_package->get_name());

		PackageDetails details;

		details.package = sub_package;
		details.name = sub_package->get_name()->get_value();
		details.path = name_to_path(details.name);
		details.qualified_name = sub_package->get_qualified_name();
		details.qualified_path = name_to_path(details.qualified_name);

		mSubPackageDetails.push_back(details);
	    }
	    else if(typeid(*contained_element) == typeid(MOF::Class)) {
		MOF::Class * klass = known_cast<MOF::Class *>(contained_element);

		ASSERT(klass->get_name());
		ASSERT(klass->is_abstract());

		ClassDetails details;

		details.klass = klass;
		details.name = klass->get_name()->get_value();
		details.path = name_to_path(details.name);
		details.qualified_name = klass->get_qualified_name();
		details.qualified_path = name_to_path(details.qualified_name);
		details.is_abstract = klass->is_abstract()->get_value();

		mClassDetails.push_back(details);
	    }
	    else if(typeid(*contained_element) == typeid(MOF::PrimitiveType)) {
		MOF::PrimitiveType * primitive_type =
		    known_cast<MOF::PrimitiveType *>(contained_element);

		ASSERT(primitive_type->get_name());

		PrimitiveTypeDetails details;

		details.primitive_type = primitive_type;
		details.name = primitive_type->get_name()->get_value();
		details.path = name_to_path(details.name);
		details.qualified_name = primitive_type->get_qualified_name();
		details.qualified_path = name_to_path(details.qualified_name);

		mPrimitiveTypeDetails.push_back(details);
	    }
	    else if(typeid(*contained_element) == typeid(MOF::StructureType)) {
		MOF::StructureType * structure_type =
		    known_cast<MOF::StructureType *>(contained_element);

		ASSERT(structure_type->get_name());

		StructureTypeDetails details;

		details.structure_type = structure_type;
		details.name = structure_type->get_name()->get_value();
		details.path = name_to_path(details.name);
		details.qualified_name = structure_type->get_qualified_name();
		details.qualified_path = name_to_path(details.qualified_name);

		mStructureTypeDetails.push_back(details);
	    }
	    else if(typeid(*contained_element) == typeid(MOF::EnumerationType)) {
		MOF::EnumerationType * enumeration_type =
		    known_cast<MOF::EnumerationType *>(contained_element);

		ASSERT(enumeration_type->get_name());

		EnumerationTypeDetails details;

		details.enumeration_type = enumeration_type;
		details.name = enumeration_type->get_name()->get_value();
		details.path = name_to_path(details.name);
		details.qualified_name = enumeration_type->get_qualified_name();
		details.qualified_path = name_to_path(details.qualified_name);

		mEnumerationTypeDetails.push_back(details);
	    }
	    else if(typeid(*contained_element) == typeid(MOF::CollectionType)) {
		MOF::CollectionType * collection_type =
		    known_cast<MOF::CollectionType *>(contained_element);

		ASSERT(collection_type->get_name());

		CollectionTypeDetails details;

		details.collection_type = collection_type;
		details.name = collection_type->get_name()->get_value();
		details.path = name_to_path(details.name);
		details.qualified_name = collection_type->get_qualified_name();
		details.qualified_path = name_to_path(details.qualified_name);

		mCollectionTypeDetails.push_back(details);
	    }
	    else if(typeid(*contained_element) == typeid(MOF::AliasType)) {
		MOF::AliasType * alias_type =
		    known_cast<MOF::AliasType *>(contained_element);

		ASSERT(alias_type->get_name());

		AliasTypeDetails details;

		details.alias_type = alias_type;
		details.name = alias_type->get_name()->get_value();
		details.path = name_to_path(details.name);
		details.qualified_name = alias_type->get_qualified_name();
		details.qualified_path = name_to_path(details.qualified_name);

		mAliasTypeDetails.push_back(details);
	    }
	}

	// load the built sources and headers

	for(vector<ClassDetails>::iterator iter = mClassDetails.begin();
	    iter != mClassDetails.end();
	    ++iter) {
	    mBuiltSources.insert(iter->path + ".cc");
	    mBuiltHeaders.insert(iter->path + ".h");
	}

	for(vector<PackageDetails>::iterator iter = mSubPackageDetails.begin();
	    iter != mSubPackageDetails.end();
	    ++iter) {
	    mSubdirectories.insert(iter->path);
	}

	for(vector<PrimitiveTypeDetails>::iterator iter = mPrimitiveTypeDetails.begin();
	    iter != mPrimitiveTypeDetails.end();
	    ++iter) {
	    mBuiltSources.insert(iter->path + ".cc");
	    mBuiltHeaders.insert(iter->path + ".h");
	}

	for(vector<StructureTypeDetails>::iterator iter = mStructureTypeDetails.begin();
	    iter != mStructureTypeDetails.end();
	    ++iter) {
	    mBuiltSources.insert(iter->path + ".cc");
	    mBuiltHeaders.insert(iter->path + ".h");
	}

	for(vector<EnumerationTypeDetails>::iterator iter = mEnumerationTypeDetails.begin();
	    iter != mEnumerationTypeDetails.end();
	    ++iter) {
	    mBuiltSources.insert(iter->path + ".cc");
	    mBuiltHeaders.insert(iter->path + ".h");
	}

	for(vector<CollectionTypeDetails>::iterator iter = mCollectionTypeDetails.begin();
	    iter != mCollectionTypeDetails.end();
	    ++iter) {
	    mBuiltSources.insert(iter->path + ".cc");
	    mBuiltHeaders.insert(iter->path + ".h");
	}

	for(vector<AliasTypeDetails>::iterator iter = mAliasTypeDetails.begin();
	    iter != mAliasTypeDetails.end();
	    ++iter) {
	    mBuiltSources.insert(iter->path + ".cc");
	    mBuiltHeaders.insert(iter->path + ".h");
	}
    }

} // namespace Generator
