// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "structure_type.h"

#include "mof/structure_type.h"
#include "mof/structure_field.h"
#include "mof/data_type.h"

#include "foundation/macros.h"
#include "foundation/utility.h"

#include "utility.h"

#include "common/cpp_stream.h"
#include "common/diff_stream.h"

#include <iostream>
#include <typeinfo>
using namespace std;

namespace Generator {

    StructureType::StructureType(MOF::StructureType * structure_type,
				 const string & model_name) : Object(),
        mStructureType(structure_type),
	mModelName(model_name),
	mDetailsLoaded(false) {
    }

    StructureType::~StructureType() {
    }

    void
    StructureType::generate(const string & output_dir, const string & include_dir) {
	if(!mDetailsLoaded) {
	    load_details();
	    mDetailsLoaded = true;
	}

	string full_path = output_dir + "/" + mStructureTypeDetails.qualified_path;

	Common::DiffStream hpp_stream(full_path + ".h");
	ASSERT(hpp_stream);
	Common::CppStream hpp(hpp_stream);

	Common::DiffStream cpp_stream(full_path + ".cc");
	ASSERT(cpp_stream);
	Common::CppStream cpp(cpp_stream);

	generate_header(hpp);
	generate_source(cpp);
    }

    void
    StructureType::generate_header(Common::CppStream & hpp) {
	string guard = path_to_guard(mStructureTypeDetails.qualified_path);

	hpp << get_prologue()
	    << "\n#ifndef " << guard
	    << "\n#define " << guard << "\n";

	generate_header_includes(hpp);

	hpp << "\nnamespace " << mModelName << " {\n";

	hpp << "\nclass " << mStructureTypeDetails.name << " : public REF::Attribute {";

	hpp << "\npublic:";

	generate_header_public_members(hpp);

	hpp << "\nprivate:";

	generate_header_private_members(hpp);

	hpp << "\n};\n";

	hpp << "\n} // namespace " << mModelName << "\n";

	hpp << "\n#endif // " << guard << "\n";
    }

    void
    StructureType::generate_source(Common::CppStream & cpp) {
	cpp << get_prologue()
	    << "\n#ifdef HAVE_CONFIG_H"
	    << "\n#include \"config.h\""
	    << "\n#endif // HAVE_CONFIG_H\n";

	generate_source_includes(cpp);

	cpp << "\nnamespace " << mModelName << " {\n";

	generate_source_statics(cpp);
	generate_source_body(cpp);

	cpp << "\n} // namespace " << mModelName << "\n";
    }

    void
    StructureType::generate_header_includes(Common::CppStream & hpp) {
	hpp << "\n#include \"reflective/attribute.h\"\n";

	if(mStructureTypeDetails.name == "Name") {
	    hpp << "\n#include <string>"
		<< "\nusing namespace std;\n";
	}

	for(set<string>::iterator iter = mStructureTypeDetails.additional_includes.begin();
	    iter != mStructureTypeDetails.additional_includes.end();
	    ++iter) {
	    hpp << "\n#include \"" << *iter << ".h\"";
	}

	hpp << "\n";
    }

    void
    StructureType::generate_header_public_members(Common::CppStream & hpp) {
	hpp << "\n" << mStructureTypeDetails.name << "();"
	    << "\n~" << mStructureTypeDetails.name << "();"
	    << "\n"
	    << "\nvoid ref_set_value(const string & value);"
	    << "\nvoid ref_get_value(string & value) const;"
	    << "\n"
	    << "\nconst string & ref_get_type() const;\n";

	hpp << "\nconst string & get_type() const;\n";
    }

    void
    StructureType::generate_header_private_members(Common::CppStream & hpp) {
	for(list<StructureFieldDetails>::iterator iter = mStructureTypeDetails.fields.begin();
	    iter != mStructureTypeDetails.fields.end();
	    ++iter) {
	    hpp << "\n" << iter->type_name << " * " << iter->member << ";";
	}

	hpp << "\nstatic const string TYPE;";
    }

    //

    void
    StructureType::generate_source_includes(Common::CppStream & cpp) {
	cpp << "\n#include \"" << mStructureTypeDetails.role << ".h\"\n"
	    << "\n#include \"reflective/exceptions.h\""
	    << "\n"
	    << "\n#include <sstream>"
	    << "\nusing namespace std;\n";
    }

    void
    StructureType::generate_source_statics(Common::CppStream & cpp) {
	cpp << "\nconst string " << mStructureTypeDetails.name << "::TYPE = \""
	    << mStructureTypeDetails.name << "\";\n";
    }

    void
    StructureType::generate_source_body(Common::CppStream & cpp) {
	cpp << "\n" << mStructureTypeDetails.name << "::" << mStructureTypeDetails.name << "() : "
	    << "REF::Attribute()";

	for(list<StructureFieldDetails>::iterator iter = mStructureTypeDetails.fields.begin();
	    iter != mStructureTypeDetails.fields.end();
	    ++iter) {
	    cpp << ", " << iter->member << "(new " << iter->type_name << "())";
	}
	    
	cpp << " {"
	    << "\n}\n";

	//
	//
	//

	cpp << "\n" << mStructureTypeDetails.name << "::~" << mStructureTypeDetails.name << "() {";

	for(list<StructureFieldDetails>::iterator iter = mStructureTypeDetails.fields.begin();
	    iter != mStructureTypeDetails.fields.end();
	    ++iter) {
	    cpp << "\n" << iter->member << "->unref();";
	}
	    
	cpp << "\n}\n";

	//
	//
	//

	cpp << "\nvoid"
	    << "\n" << mStructureTypeDetails.name << "::ref_set_value(const string & value) {"
	    << "\n}\n";

	cpp << "\nvoid"
	    << "\n" << mStructureTypeDetails.name << "::ref_get_value(string & value) const {"
	    << "\nostringstream ost;"
	    << "\nost << \"{ \";"
	    << "\nost << \" }\";"
	    << "\nvalue += ost.str();"
	    << "\n}\n";

	cpp << "\nconst string &"
	    << "\n" << mStructureTypeDetails.name << "::ref_get_type() const {"
	    << "\n  return TYPE;"
	    << "\n}\n";
    }

    void
    StructureType::load_details() {
	ASSERT(mStructureType->get_name());

	//
	// load the details of this structure type
	//

	mStructureTypeDetails.structure_type = mStructureType;
	mStructureTypeDetails.name = mStructureType->get_name()->get_value();
	mStructureTypeDetails.role = name_to_role(mStructureTypeDetails.name);
	mStructureTypeDetails.variable = role_to_variable(mStructureTypeDetails.role);
	mStructureTypeDetails.qualified_name = mStructureType->get_qualified_name();
	mStructureTypeDetails.qualified_path = name_to_path(mStructureTypeDetails.qualified_name);

	const list<MOF::ModelElement *> & contained_elements =
	    mStructureType->get_contained_elements();

	for(list<MOF::ModelElement *>::const_iterator iter = contained_elements.begin();
	    iter != contained_elements.end();
	    ++iter) {
	    MOF::ModelElement * contained_element = *iter;
	    MOF::StructureField * field = known_cast<MOF::StructureField *>(contained_element);
	    MOF::Classifier * type = field->get_type();
	    MOF::DataType * data_type = known_cast<MOF::DataType *>(type);

	    ASSERT(field->get_name());
	    ASSERT(data_type->get_name());

	    StructureFieldDetails details;

	    details.structure_field = field;
	    details.name = field->get_name()->get_value();
	    details.role = name_to_role(details.name);
	    details.member = name_to_member(details.name);
	    details.type = data_type;
	    details.type_name = data_type->get_name()->get_value();
	    details.type_path = name_to_path(details.type_name);
	    details.type_qualified_name = data_type->get_qualified_name();
	    details.type_qualified_path = name_to_path(details.type_qualified_name);

	    mStructureTypeDetails.fields.push_back(details);

	    if(details.type->get_container() == mStructureType->get_container()) {
		mStructureTypeDetails.additional_includes.insert(details.type_path);
	    }
	    else {
		mStructureTypeDetails.additional_includes.insert(details.type_qualified_path);
	    }
	}
    }

} // namespace Generator
