// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "settings.h"

#include "foundation/macros.h"

namespace GTK {

    const string Settings::KEY_PREFIX = "/apps/dachshund";

    Settings::Settings() : Toolkit::Settings(),
	mConnection(0) {
	mClient = gconf_client_get_default();

	GError * error = 0;

	gconf_client_add_dir(mClient, KEY_PREFIX.c_str(), GCONF_CLIENT_PRELOAD_NONE, &error);

	if(error) {
	    FATAL("Couldn't initialise gconf: " << error->message);
	    g_clear_error(&error);
	}

	//

	mConnection = gconf_client_notify_add(mClient, KEY_PREFIX.c_str(),
					    &Settings::static_client_notify, this, 0, &error);

	if(error) {
	    FATAL("Couldn't add client notification handler: " << error->message);
	    g_clear_error(&error);
	}
    }

    Settings::~Settings() {
	if(mConnection) {
	    ASSERT(mClient);

	    gconf_client_notify_remove(mClient, mConnection);
	}

	if(mClient) {
	    g_object_unref(G_OBJECT(mClient));
	}
    }

    //

    void
    Settings::static_client_notify(GConfClient * client, guint connection_id,
					GConfEntry * entry, gpointer user_data) {
	Settings * gconf_settings = static_cast<Settings *>(user_data);
	gconf_settings->client_notify(client, connection_id, entry);
    }

    void
    Settings::client_notify(GConfClient * client, guint connection_id, GConfEntry * entry) {
	ASSERT(mClient == client);

	string key = gconf_entry_get_key(entry);

	MESSAGE("VALUE CHANGED: " << key);

	value_changed(key);
    }

    //

    bool
    Settings::get_value(const string & key, string & value) const {
	GError * error = 0;

	gchar * result = gconf_client_get_string(mClient, (KEY_PREFIX + "/" + key).c_str(), &error);

	if(error) {
	    g_clear_error(&error);
	    return false;
	}
	else {
	    ASSERT(result);
	    value = result;
	    g_free(result);
	    return true;
	}
    }

    bool
    Settings::get_value(const string & key, int & value) const {
	GError * error = 0;

	gint result = gconf_client_get_int(mClient, (KEY_PREFIX + "/" + key).c_str(), &error);

	if(error) {
	    g_clear_error(&error);
	    return false;
	}
	else {
	    ASSERT(result);
	    value = result;
	    return true;
	}
    }

    bool
    Settings::get_value(const string & key, bool & value) const {
	GError * error = 0;

	gboolean result = gconf_client_get_bool(mClient, (KEY_PREFIX + "/" + key).c_str(), &error);

	if(error) {
	    g_clear_error(&error);
	    return false;
	}
	else {
	    value = result;
	    return true;
	}
    }

    bool
    Settings::get_value(const string & key, double & value) const {
	GError * error = 0;

	gdouble result = gconf_client_get_bool(mClient, (KEY_PREFIX + "/" + key).c_str(), &error);

	if(error) {
	    g_clear_error(&error);
	    return false;
	}
	else {
	    value = result;
	    return true;
	}
    }


    void
    Settings::set_value(const string & key, const string & value) {
	GError * error = 0;

	gconf_client_set_string(mClient, (KEY_PREFIX + "/" + key).c_str(), value.c_str(), &error);

	if(error) {
	    ERROR("Error setting string value for key: " << key);
	    g_clear_error(&error);
	}
    }

    void
    Settings::set_value(const string & key, int value) {
	GError * error = 0;

	gconf_client_set_int(mClient, (KEY_PREFIX + "/" + key).c_str(), value, &error);

	if(error) {
	    ERROR("Error setting int value for key: " << key);
	    g_clear_error(&error);
	}
    }

    void
    Settings::set_value(const string & key, bool value) {
	GError * error = 0;

	gconf_client_set_bool(mClient, (KEY_PREFIX + "/" + key).c_str(), value, &error);

	if(error) {
	    ERROR("Error setting bool value for key: " << key);
	    g_clear_error(&error);
	}
    }

    void
    Settings::set_value(const string & key, double value) {
	GError * error = 0;

	gconf_client_set_float(mClient, (KEY_PREFIX + "/" + key).c_str(), value, &error);

	if(error) {
	    ERROR("Error setting double value for key: " << key);
	    g_clear_error(&error);
	}
    }

} // namespace GTK
