/// Dachshund
// Copyright (C) 2002 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

//
// Code generator module
//

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include <set>
#include <vector>
#include <string>
#include <typeinfo>

#include <unistd.h>

#include "common/file_utilities.h"
#include "uml/foundation/data_types/name.h"
#include "uml/model_management/model.h"
#include "uml/model_management/package.h"
#include "uml/foundation/core/model_element.h"

#include "generator.h"

#include "foundation/macros.h"

using namespace std;
using namespace UML;
using namespace Common;

namespace DMCG {

#ifdef MCG_REQUIRED

	string FileModule::mTargetDirectory;

	Generator::Generator () {
		FileModule::set_target_directory("testdir/src/");
		mLanguage = DEFAULT_MCG_LANGUAGE;
		mOperatingSystem = DEFAULT_MCG_OPERATING_SYSTEM;

		// retrieve any mcg preferences...
	}

	bool
	Generator::code_generation_on (void) {
		return true;
	}

	//
	// where it all starts for a package (top-level possibly but not essential).
	// The model content is established in terms of packages.
	//
	int
	Generator::generate_code (Package *package) {
		Package *model;
		pid_t dmcg_child;

		ASSERT(package != 0);	// must be a valid 'model' pointer

		TRACE(5,"Dachshund's UML Model Code Generator.");

		// DMCG runs as a separate thread
		dmcg_child = fork();
		if (dmcg_child == 0) {
			mChosenElement = dynamic_cast<ModelElement *>(package);

			model = establish_packages(package);

			exit(0);
		} else
			return 0;
	}

	//
	// where it all starts for a class
	//
	int
	Generator::generate_code (UML::Class *klass) {
		Package *package, *model=0;
		pid_t dmcg_child;

		ASSERT(klass != 0);	// must be a valid 'model' pointer

		TRACE(5,"Dachshund's UML Model Code Generator.");

		// DMCG runs as a separate thread
		dmcg_child = fork();
		if (dmcg_child == 0) {
			mChosenElement = dynamic_cast<ModelElement *>(klass);

			package = dynamic_cast<Package *>(klass->get_namespace());
			if (package)
				model = establish_packages(package);

			exit(0);
		} else
			return 0;
	}

	//
	// suss package arrangement
	//
	Package *
	Generator::establish_packages (Package *package) {
		Package *model;

		// first, find the top level package (the model root)
		model = establish_root_package(package);

		PackageModule * new_package_module = new PackageModule();
		new_package_module->set_package (model);
		add_package_module (new_package_module);  // first 'package' is the overall model

		if (mChosenElement == model)
			mChosenModule = new_package_module;

		// for now, use default/user-pref lingo for model
		// Could be overwritten by individual package settings 
		new_package_module->set_language(mLanguage);
		new_package_module->set_operating_system(mOperatingSystem);

		// suss entire package hierarchy
		establish_package_hierarchy();

		// establish base source directory structure
		ensure_directory (FileModule::get_target_directory());

		// for each PackageModule, generate stuff
		for (vector<PackageModule *>::iterator package_ptr = mAllPackageModules.begin();
			 package_ptr < mAllPackageModules.end();
			 package_ptr++) {
			generate_package(*package_ptr);
		}
	
		return model;
	}

	//
	// traverse up the model hierarchy until we have the root package
	//
	Package *
	Generator::establish_root_package (ModelElement *model_element) {
		Package *model;
		Namespace *my_namespace, *tmp_namespace;

		my_namespace = model_element->get_namespace();

		if (!my_namespace)
			model = dynamic_cast<Package *>(model_element);
		else while (my_namespace) {
			tmp_namespace = my_namespace;
			my_namespace = tmp_namespace->get_namespace();
			// if the namespace of this element is NULL then found 'top of hierarchy'
			if (!my_namespace) {
				model = dynamic_cast<Package *>(tmp_namespace);
				break;
			}
		}

 
		return model;
	}

	//
	// then traverse down through the packages
	//
	void
	Generator::establish_package_hierarchy (void) {
		int package_nest_start = 0, package_nest_end = 0, package_nest, package_nest_count;
		bool more_to_do = true, element_is_package;
		Package *package;
		PackageModule *package_module, *new_package_module;
		set<ModelElement *> owned_elements;

		while (more_to_do) {
			more_to_do = false;
			package_nest_count = 0;
		
			// loop through outstanding model packages to establish a set of all PackageModules
			// in the user's model.
			for (package_nest = package_nest_start; package_nest <= package_nest_end; package_nest++) {

				package_module = mAllPackageModules[package_nest];
				package = package_module->get_package();
				owned_elements = package->get_owned_element();

				if (!owned_elements.empty())
					for (set<ModelElement *>::const_iterator element = owned_elements.begin();
						 element != owned_elements.end();
				   		 element++) {
				
						// only proceed if its a package or derivative
						element_is_package = ((*element)->ref_get_type() == "Package" ||
											  (*element)->ref_get_type() == "Subsystem" ||
											  (*element)->ref_get_type() == "Model");
						if (element_is_package) {
							new_package_module = new PackageModule();
							new_package_module->set_package(dynamic_cast<Package *>(*element));
							add_package_module (new_package_module);

							new_package_module->set_parent_package_module(package_module);

							if (mChosenElement == *element)
								mChosenModule = new_package_module;

							package_nest_count++;
							more_to_do = true;
						}
					}
			}
			package_nest_start = package_nest_end+1;
			package_nest_end += package_nest_count;
		}
	}

	//
	// take a package module and analyse it and its classes
	//
	int
	Generator::generate_package (PackageModule * package_module) {
		Package * package;
		PackageModule * parent_package_module;
		set<ModelElement *> owned_elements;
		MCG_Language parents_language;
		MCG_Operating_System parents_os;
		ClassModule * new_class_module;

		package = package_module->get_package();
	
		TRACE(5, "Got package \"" + package->get_name()->ref_get_value() + "\"");

		parent_package_module = package_module->get_parent_package_module();
		// copy parents settings for lingo/os
		if (parent_package_module) {
			parents_language = parent_package_module->get_language();
			parents_os = parent_package_module->get_operating_system();
			package_module->set_language(parents_language);
			package_module->set_operating_system(parents_os);
		}

		owned_elements = package->get_owned_element();

		if (!owned_elements.empty())
			for (set<ModelElement *>::const_iterator element = owned_elements.begin();
				 element != owned_elements.end();
		   		 element++) {

				// only proceed if its a class
				if ((*element)->ref_get_type() == "Class") {
					new_class_module = new ClassModule();
					new_class_module->set_class(dynamic_cast<UML::Class *>(*element));
					new_class_module->set_package_module(package_module);
					package_module->add_class(new_class_module);

					if (mChosenElement == *element)
						mChosenModule = new_class_module;

					// do the class
					generate_class(new_class_module);
				}
			}

		return 0;
	}

	//
	// take a class module and analyse it in more detail
	//
	int
	Generator::generate_class (ClassModule * class_module) {
		UML::Class * klass;
		PackageModule * package_module;
		Package * package;
		MCG_Language package_language;
		MCG_Operating_System package_os;

		klass = class_module->get_class();
		package_module = class_module->get_package_module();
		package = package_module->get_package();	// only used fro name-trace...

		TRACE(5, "Got class \"" + 
				package->get_name()->ref_get_value() + "::" +
				klass->get_name()->ref_get_value() + "\""); 

		// adopt packages build settings - maybe overridden by tagged settings
		if (package_module) {
			package_language = package_module->get_language();
			package_os = package_module->get_operating_system();
			class_module->set_language(package_language);
			class_module->set_operating_system(package_os);
		}

		return 0;
	}

#else	// MCG_REQUIRED

	// Theses func's empty if model code generator not set:
	// just provides an interface to satisfy front end
	// of Dachshund to make the various code-gen modules
	// truly modular. Ie code gen shared libs can be dropped
	// in and out as required.

	Generator::Generator () {
	}

	bool
	Generator::code_generation_on (void) {
		return false;
	}

	int
	Generator::generate_code (Package *package) {
		return -1;
	}

	int
	Generator::generate_code (UML::Class *klass) {
		return -1;
	}

#endif	// MCG_REQUIRED

}	// namespace DMCG

