// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "association_end.h"

#include "namespace.h"
#include "reference.h"
#include "model.h"

#include "foundation/macros.h"
#include "foundation/commands.h"
#include "reflective/changes.h"

#include <iostream>
#include <functional>
#include <algorithm>
using namespace std;

namespace MOF {

    const string AssociationEnd::TYPE = "AssociationEnd";

    AssociationEnd::AssociationEnd() : TypedElement(),
	mMultiplicity(0),
        mAggregation(0),
	mNavigable(0),
	mChangeable(0) {
    }

    AssociationEnd::~AssociationEnd() {
    }

    void
    AssociationEnd::set_other_end(AssociationEnd * other_end) {
	NYI();
	ASSERT_NOT_REACHED();
    }

    AssociationEnd *
    AssociationEnd::get_other_end() {
	Namespace * container = get_container();
	const list<ModelElement *> & contents = container->get_contained_elements();

	RETURN_VAL_IF_FAIL(contents.size() == 2, static_cast<AssociationEnd *>(0));

	list<ModelElement *>::const_iterator iter = contents.begin();

	AssociationEnd * other_end =
	    dynamic_cast<AssociationEnd *>(*iter == this ? *++iter : *iter);

	ASSERT(other_end);

	return other_end;
    }

    //
    //

    void
    AssociationEnd::set_navigable(const Boolean * navigable) {
	begin_transaction();

	//

	Boolean * old_value = mNavigable ? new Boolean(*mNavigable) : 0;
	Boolean * new_value = navigable ? new Boolean(*navigable) : 0;

	//

	Command * command = command_flip_argument(this, &AssociationEnd::set_navigable,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mNavigable) {
	    mNavigable->unref();
	}

	mNavigable = new_value ? new Boolean(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "navigable", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    AssociationEnd::set_changeable(const Boolean * changeable) {
	begin_transaction();

	//

	Boolean * old_value = mChangeable ? new Boolean(*mChangeable) : 0;
	Boolean * new_value = changeable ? new Boolean(*changeable) : 0;

	//

	Command * command = command_flip_argument(this, &AssociationEnd::set_changeable,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mChangeable) {
	    mChangeable->unref();
	}

	mChangeable = new_value ? new Boolean(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "changeable", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    AssociationEnd::set_multiplicity(const MultiplicityType * multiplicity) {
	begin_transaction();

	//

	MultiplicityType * old_value = mMultiplicity ? new MultiplicityType(*mMultiplicity) : 0;
	MultiplicityType * new_value = multiplicity ? new MultiplicityType(*multiplicity) : 0;

	//

	Command * command = command_flip_argument(this, &AssociationEnd::set_multiplicity,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mMultiplicity) {
	    mMultiplicity->unref();
	}

	mMultiplicity = new_value ? new MultiplicityType(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "multiplicity", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    AssociationEnd::set_aggregation(const AggregationKind * aggregation) {
	begin_transaction();

	//

	AggregationKind * old_value = mAggregation ? new AggregationKind(*mAggregation) : 0;
	AggregationKind * new_value = aggregation ? new AggregationKind(*aggregation) : 0;

	//

	Command * command = command_flip_argument(this, &AssociationEnd::set_aggregation,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mAggregation) {
	    mAggregation->unref();
	}

	mAggregation = new_value ? new AggregationKind(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "aggregation", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    //
    //
    //

    void
    AssociationEnd::internal_add_referent(Reference * referent) throw() {
	referent->ref();
	mReferents.insert(referent);
    }

    void
    AssociationEnd::internal_remove_referent(Reference * referent) throw() {
	mReferents.erase(referent);
	referent->unref();
    }

    void
    AssociationEnd::internal_add_referrer(Reference * referrer) throw() {
	referrer->ref();
	mReferrers.insert(referrer);
    }

    void
    AssociationEnd::internal_remove_referrer(Reference * referrer) throw() {
	mReferrers.erase(referrer);
	referrer->unref();
    }

    void
    AssociationEnd::ref_delete_impl(set<string> & visited) {
	if(visited.insert(TYPE).second) {
	    set<Reference *> referents_copy = mReferents;
	    set<Reference *> referrers_copy = mReferrers;
	    AssociationEnd * null_association_end = 0;

	    for_each(referents_copy.begin(), referents_copy.end(),
		     bind2nd(mem_fun(&Reference::set_referenced_end), null_association_end));

	    for_each(referrers_copy.begin(), referrers_copy.end(),
		     bind2nd(mem_fun(&Reference::set_exposed_end), null_association_end));

	    TypedElement::ref_delete_impl(visited);
	}
    }

    bool
    AssociationEnd::ref_is_a_type_of_impl(set<string> & visited,
					      const string & class_name) const {
	if(visited.insert(TYPE).second) {
	    return class_name == TYPE || TypedElement::ref_is_a_type_of(class_name);
	}
	else {
	    return false;
	}
    }

    const string &
    AssociationEnd::ref_get_type() const {
	return TYPE;
    }

    void
    AssociationEnd::ref_implant_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_add_association_end(this);

	    TypedElement::ref_implant_impl(visited);
	}
    }

    void
    AssociationEnd::ref_uproot_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_remove_association_end(this);

	    TypedElement::ref_uproot_impl(visited);
	}
    }

} // namespace MOF
