// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "constraint.h"

#include "model.h"

#include "foundation/functional.h"
#include "foundation/macros.h"
#include "foundation/commands.h"
#include "reflective/changes.h"

#include <algorithm>
#include <vector>
using namespace std;

namespace MOF {

    const string Constraint::TYPE = "Constraint";

    Constraint::Constraint() : ModelElement(),
        mExpression(0),
	mLanguage(0),
	mEvaluationPolicy(0) {
    }

    Constraint::~Constraint() {
    }

    void
    Constraint::set_expression(Object * expression) {
	begin_transaction();

	Command * command = command_flip_argument(this, &Constraint::set_expression,
						  expression, mExpression);
	log_command(command);
	command->unref();

	// make the changes

	if(mExpression) {
	    mExpression->unref();
	}

	mExpression = expression;

	if(mExpression) {
	    mExpression->ref();
	}

	//

#if 0
	ChangeEvent * change = new ChangeRoleChanged(this, "expression",
						     expression, mExpression);
	log_change(change);
	change->unref();
#endif

	end_transaction();
    }

    void
    Constraint::set_language(const Name * language) {
	begin_transaction();

	//

	Name * old_value = mLanguage ? new Name(*mLanguage) : 0;
	Name * new_value = language ? new Name(*language) : 0;

	//

	Command * command = command_flip_argument(this, &Constraint::set_language,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mLanguage) {
	    mLanguage->unref();
	}

	mLanguage = language ? new Name(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "language", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    Constraint::set_evaluation_policy(const EvaluationKind * evaluation_policy) {
	begin_transaction();

	//

	EvaluationKind * old_value = mEvaluationPolicy ? new EvaluationKind(*mEvaluationPolicy) : 0;
	EvaluationKind * new_value = evaluation_policy ? new EvaluationKind(*evaluation_policy) : 0;

	//

	Command * command = command_flip_argument(this,
						  &Constraint::set_evaluation_policy,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mEvaluationPolicy) {
	    mEvaluationPolicy->unref();
	}

	mEvaluationPolicy = new_value ? new EvaluationKind(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "visible", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    Constraint::set_constrained_elements(const set<ModelElement *> & constrained_elements) {
	begin_transaction();

	Command * command = command_flip_argument(this,
						  &Constraint::set_constrained_elements,
						  constrained_elements, mConstrainedElements);
	log_command(command);
	command->unref();

	// make the changes

	vector<ModelElement *> added;
	vector<ModelElement *> removed;

	set_difference(constrained_elements.begin(), constrained_elements.end(), 
		       mConstrainedElements.begin(), mConstrainedElements.end(),
		       back_inserter(added));

	set_difference(mConstrainedElements.begin(), mConstrainedElements.end(), 
		       constrained_elements.begin(), constrained_elements.end(),
		       back_inserter(removed));

	for_each(added.begin(), added.end(),
		 bind2nd(mem_fun(&ModelElement::internal_add_constraint), this));
	for_each(removed.begin(), removed.end(),
		 bind2nd(mem_fun(&ModelElement::internal_remove_constraint), this));

	for_each(added.begin(), added.end(), mem_fun(&Object::ref));
	for_each(removed.begin(), removed.end(), mem_fun(&Object::unref));

	mConstrainedElements = constrained_elements;

	// changes made

	for(vector<ModelElement *>::iterator iter = added.begin();
	    iter != added.end();
	    ++iter) {
	    ModelElement * constrained_element = *iter;
	    ChangeEvent * change = change_role_added(this, "constrained_element",
						     constrained_element, false);
	    log_change(change);
	    change->unref();
	}

	for(vector<ModelElement *>::iterator iter = removed.begin();
	    iter != removed.end();
	    ++iter) {
	    ModelElement * constrained_element = *iter;
	    ChangeEvent * change = change_role_removed(this, "constrained_element",
						       constrained_element, false);
	    log_change(change);
	    change->unref();
	}

	//

	end_transaction();
    }

    void
    Constraint::add_constrained_element(ModelElement * constrained_element) {
	RETURN_IF_FAIL(mConstrainedElements.find(constrained_element) ==
		       mConstrainedElements.end());

	begin_transaction();

	Command * command = command_flip_method(this,
						&Constraint::add_constrained_element,
						&Constraint::remove_constrained_element,
						constrained_element);
	log_command(command);
	command->unref();

	// make the changes

	constrained_element->ref();

	mConstrainedElements.insert(constrained_element);
	constrained_element->internal_add_constraint(this);

	// changes made

	ChangeEvent * change = change_role_added(this, "constrained_element", constrained_element,
						 false);
	log_change(change);
	log_change(change);

	//

	end_transaction();
    }

    void
    Constraint::remove_constrained_element(ModelElement * constrained_element) {
	RETURN_IF_FAIL(mConstrainedElements.find(constrained_element) !=
		       mConstrainedElements.end());

	begin_transaction();

	Command * command = command_flip_method(this,
						&Constraint::remove_constrained_element,
						&Constraint::add_constrained_element,
						constrained_element);
	log_command(command);
	command->unref();

	// make the changes

	constrained_element->internal_remove_constraint(this);
	mConstrainedElements.erase(constrained_element);

	constrained_element->unref();

	// changes made

	ChangeEvent * change = change_role_removed(this, "constrained_element",
						     constrained_element, false);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    Constraint::modify_constrained_element(ModelElement * old_constrained_element,
					   ModelElement * new_constrained_element) {
	set<ModelElement *>::iterator iter = mConstrainedElements.find(new_constrained_element);
	RETURN_IF_FAIL(iter == mConstrainedElements.end());

	iter = mConstrainedElements.find(old_constrained_element);
	RETURN_IF_FAIL(iter != mConstrainedElements.end());

	begin_transaction();

	Command * command = command_reverse_argument(this,
						     &Constraint::modify_constrained_element,
						     old_constrained_element,
						     new_constrained_element);
	log_command(command);
	command->unref();

	// make the changes

	old_constrained_element->internal_add_constraint(this);
	
	mConstrainedElements.erase(iter);
	mConstrainedElements.insert(new_constrained_element);

	new_constrained_element->internal_add_constraint(this);

	// changes made

	ChangeEvent * change = change_role_changed(this, "constrained_element",
						   old_constrained_element,
						   new_constrained_element, false);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    //
    //
    //

    void
    Constraint::internal_add_constrained_element(ModelElement * constrained_element) throw() {
	constrained_element->ref();
	mConstrainedElements.insert(constrained_element);
    }

    void
    Constraint::internal_remove_constrained_element(ModelElement * constrained_element) throw() {
	mConstrainedElements.erase(constrained_element);
	constrained_element->unref();
    }

    void
    Constraint::ref_delete_impl(set<string> & visited) {
	if(visited.insert(TYPE).second) {
	    set<ModelElement *> empty_model_element_set;

	    set_constrained_elements(empty_model_element_set);

	    ModelElement::ref_delete_impl(visited);
	}
    }

    bool
    Constraint::ref_is_a_type_of_impl(set<string> & visited, const string & class_name) const {
	if(visited.insert(TYPE).second) {
	    return class_name == TYPE || ModelElement::ref_is_a_type_of(class_name);
	}
	else {
	    return false;
	}
    }

    const string &
    Constraint::ref_get_type() const {
	return TYPE;
    }

    void
    Constraint::ref_implant_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_add_constraint(this);

	    ModelElement::ref_implant_impl(visited);
	}
    }

    void
    Constraint::ref_uproot_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_remove_constraint(this);

	    ModelElement::ref_uproot_impl(visited);
	}
    }

    const vector<const Name *> &
    Constraint::get_test() const {
	return death_cast<vector<const Name *> >(mTest);
    }

    void
    Constraint::set_test(const vector<const Name *> & test) {
	begin_transaction();

	command_flip_argument(this, &Constraint::set_test,
			      test, death_cast<vector<const Name *> >(mTest));



	for_each(mTest.begin(), mTest.end(), mem_fun(&Object::unref));

	mTest.resize(test.size());

	for(unsigned int i = 0; i < test.size(); ++i) {
	    mTest[i] = new Name(*test[i]);
	}


	end_transaction();
    }

} // namespace MOF
