// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "feature.h"

#include "model.h"

#include "foundation/macros.h"
#include "foundation/commands.h"
#include "reflective/changes.h"

namespace MOF {

    const string Feature::TYPE = "Feature";

    //

    Feature::Feature() : ModelElement(),
       mVisibility(0),
       mScope(0) {
    }

    Feature::~Feature() {
    }

    void
    Feature::set_visibility(const VisibilityKind * visibility) {
	begin_transaction();

	//

	VisibilityKind * old_value = mVisibility ? new VisibilityKind(*mVisibility): 0;
	VisibilityKind * new_value = visibility ? new VisibilityKind(*visibility) : 0;

	//

	Command * command = command_flip_argument(this, &Feature::set_visibility,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mVisibility) {
	    mVisibility->unref();
	}

	mVisibility = new_value ? new VisibilityKind(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "visibility", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    Feature::set_scope(const ScopeKind * scope) {
	begin_transaction();

	//

	ScopeKind * old_value = mScope ? new ScopeKind(*mScope) : 0;
	ScopeKind * new_value = scope ? new ScopeKind(*scope) : 0;

	//

	Command * command = command_flip_argument(this, &Feature::set_scope, new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mScope) {
	    mScope->unref();
	}

	mScope = new_value ? new ScopeKind(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "scope", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    bool
    Feature::ref_is_a_type_of_impl(set<string> & visited, const string & class_name) const {
	if(visited.insert(TYPE).second) {
	    return class_name == TYPE || ModelElement::ref_is_a_type_of(class_name);
	}
	else {
	    return false;
	}
    }

    const string &
    Feature::ref_get_type() const {
	return TYPE;
    }

    void
    Feature::ref_implant_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_add_feature(this);

	    ModelElement::ref_implant_impl(visited);
	}
    }

    void
    Feature::ref_uproot_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_remove_feature(this);

	    ModelElement::ref_uproot_impl(visited);
	}
    }

} // namespace MOF
