// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "import.h"

#include "namespace.h"
#include "model.h"

#include "foundation/commands.h"
#include "reflective/changes.h"

namespace MOF {

    const string Import::TYPE = "Import";

    //

    Import::Import() : ModelElement(),
        mVisibility(0),
	mClustered(0),
        mImported(0) {
    }

    Import::~Import() {
    }

    void
    Import::set_visibility(const VisibilityKind * visibility) {
	begin_transaction();

	//

	VisibilityKind * old_value = mVisibility ? new VisibilityKind(*mVisibility) : 0;
	VisibilityKind * new_value = visibility ? new VisibilityKind(*visibility) : 0;

	//

	Command * command = command_flip_argument(this, &Import::set_visibility,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mVisibility) {
	    mVisibility->unref();
	}

	mVisibility = new_value ? new VisibilityKind(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "visibility", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    Import::set_clustered(const Boolean * clustered) {
	begin_transaction();

	//

	Boolean * old_value = mClustered ? new Boolean(*mClustered) : 0;
	Boolean * new_value = clustered ? new Boolean(*clustered) : 0;

	//

	Command * command = command_flip_argument(this, &Import::set_clustered,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mClustered) {
	    mClustered->unref();
	}

	mClustered = new_value ? new Boolean(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "clustered", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    Import::set_imported(Namespace * imported) {
	begin_transaction();

	//

	Command * command = command_flip_argument(this, &Import::set_imported,
						  imported, mImported);
	log_command(command);
	command->unref();

	// make the changes

	if(mImported) {
	    mImported->internal_remove_importer(this);
	    mImported->unref();
	}

	mImported = imported;

	if(mImported) {
	    mImported->ref();
	    mImported->internal_add_importer(this);
	}

	// changes made

	ChangeEvent * change = change_role_changed(this, "imported", imported, mImported, false);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    Import::ref_delete_impl(set<string> & visited) {
	if(visited.insert(TYPE).second) {
	    set_imported(static_cast<Namespace *>(0));

	    ModelElement::ref_delete_impl(visited);
	}
    }

    bool
    Import::ref_is_a_type_of_impl(set<string> & visited, const string & class_name) const {
	if(visited.insert(TYPE).second) {
	    return class_name == TYPE || Import::ref_is_a_type_of(class_name);
	}
	else {
	    return false;
	}
    }

    const string &
    Import::ref_get_type() const {
	return TYPE;
    }

    void
    Import::ref_implant_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_add_import(this);

	    ModelElement::ref_implant_impl(visited);
	}
    }

    void
    Import::ref_uproot_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_remove_import(this);

	    ModelElement::ref_uproot_impl(visited);
	}
    }

} // namespace MOF
