// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "model_element.h"

#include "namespace.h"
#include "constraint.h"
#include "tag.h"
#include "model.h"

#include "name.h"
#include "boolean.h"

#include "foundation/macros.h"
#include "foundation/functional.h"
#include "foundation/commands.h"
#include "reflective/changes.h"

#include <algorithm>
#include <functional>
using namespace std;

namespace MOF {

    const string ModelElement::TYPE = "ModelElement";

    //

    ModelElement::ModelElement() : REF::Class(),
	mName(0),
	mAnnotation(0),
	mFrozen(0),
	mVisible(0),
	mContainer(0) {
    }

    ModelElement::~ModelElement() {
    }

    void
    ModelElement::set_name(const Name * name) {
	begin_transaction();

	//

	Name * old_value = mName ? new Name(*mName) : 0;
	Name * new_value = name ? new Name(*name) : 0;

	//

	Command * command = command_flip_argument(this, &ModelElement::set_name,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mName) {
	    mName->unref();
	}

	mName = new_value ? new Name(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "name", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    ModelElement::set_annotation(const Name * annotation) {
	begin_transaction();

	//

	Name * old_value = mAnnotation ? new Name(*mAnnotation) : 0;
	Name * new_value = annotation ? new Name(*annotation) : 0;

	//

	Command * command = command_flip_argument(this, &ModelElement::set_annotation,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mAnnotation) {
	    mAnnotation->unref();
	}

	mAnnotation = new_value ? new Name(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "annotation", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    ModelElement::add_provider(ModelElement * provider) {
	RETURN_IF_FAIL(mProviders.find(provider) == mProviders.end());

	//

	begin_transaction();

	//

	Command * command = command_flip_method(this,
						&ModelElement::add_provider,
						&ModelElement::remove_provider,
						provider);
	log_command(command);
	command->unref();

	// make the changes

	provider->ref();

	mProviders.insert(provider);
	provider->internal_add_dependent(this);

	// changes made

	ChangeEvent * change = change_role_added(this, "provider", provider, false);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    ModelElement::remove_provider(ModelElement * provider) {
	RETURN_IF_FAIL(mProviders.find(provider) != mProviders.end());

	//

	begin_transaction();

	//

	Command * command = command_flip_method(this,
						&ModelElement::remove_provider,
						&ModelElement::add_provider,
						provider);
	log_command(command);
	command->unref();

	// make the changes

	provider->internal_remove_dependent(this);
	mProviders.erase(provider);

	provider->unref();

	// changes made

	ChangeEvent * change = change_role_removed(this, "provider", provider, false);
	log_change(change);
	change->unref();

	end_transaction();
    }

    void
    ModelElement::set_providers(const set<ModelElement *> & providers) {
	begin_transaction();

	//

	Command * command = command_flip_argument(this,
						  &ModelElement::set_providers,
						  providers,
						  mProviders);
	log_command(command);
	command->unref();

	// make the changes

	vector<ModelElement *> added;
	vector<ModelElement *> removed;

	set_difference(providers.begin(), providers.end(),
		       mProviders.begin(), mProviders.end(), back_inserter(added));

	set_difference(mProviders.begin(), mProviders.end(), 
		       providers.begin(), providers.end(), back_inserter(removed));

	for_each(added.begin(), added.end(),
		 bind2nd(mem_fun(&ModelElement::internal_add_dependent), this));
	for_each(removed.begin(), removed.end(),
		 bind2nd(mem_fun(&ModelElement::internal_remove_dependent), this));

	for_each(added.begin(), added.end(), mem_fun(&Object::ref));
	for_each(removed.begin(), removed.end(), mem_fun(&Object::unref));

	mProviders = providers;

	// changes made

	for(vector<ModelElement *>::iterator iter = added.begin();
	    iter != added.end();
	    ++iter) {
	    ModelElement * provider = *iter;
	    ChangeEvent * change = change_role_added(this, "provider", provider, false);
	    log_change(change);
	    change->unref();
	}

	for(vector<ModelElement *>::iterator iter = removed.begin();
	    iter != removed.end();
	    ++iter) {
	    ModelElement * provider = *iter;
	    ChangeEvent * change = change_role_removed(this, "provider", provider, false);
	    log_change(change);
	    change->unref();
	}

	//

	end_transaction();
    }

    void
    ModelElement::add_tag(Tag * tag, int index) {
	RETURN_IF_FAIL(index <= static_cast<int>(mTags.size()));

	if(index < 0) {
	    index = mTags.size();
	}

	list<Tag *>::iterator iter = mTags.begin();

	for(int i = 0; i < static_cast<int>(mTags.size()); ++i) {
	    RETURN_IF_FAIL(*iter != tag);

	    if(i < index) {
		++iter;
	    }
	}

	//

	begin_transaction();

	//

	Command * command = command_flip_method(this,
						&ModelElement::add_tag,
						&ModelElement::remove_tag,
						tag, index);
	log_command(command);
	command->unref();

	//

	// make the changes

	tag->ref();

	mTags.insert(iter, tag);
	tag->internal_add_model(this);

	// changes made

	ChangeEvent * change = change_role_added(this, "tag", tag, false, index);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    ModelElement::remove_tag(Tag * tag) {
	int index = 0;

	list<Tag *>::iterator iter = mTags.begin();

	for(; iter != mTags.end(); ++iter, ++index) {
	    if(*iter == tag) {
		break;
	    }
	}

	RETURN_IF_FAIL(iter != mTags.end());

	//

	begin_transaction();

	//

	Command * command = command_flip_method(this,
						&ModelElement::remove_tag,
						&ModelElement::add_tag,
						tag, index);
	log_command(command);
	command->unref();

	// make the changes

	tag->internal_remove_model(this);
	mTags.erase(iter);

	tag->unref();

	// changes made

	ChangeEvent * change = change_role_removed(this, "tag", tag, false, index);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    ModelElement::set_tags(const list<Tag *> & tags) {
	set<Tag *> before;
	set<Tag *> after;

	before.insert(mTags.begin(), mTags.end());
	after.insert(tags.begin(), tags.end());

	// check for duplicates
	RETURN_IF_FAIL(before.size() == mTags.size());
	RETURN_IF_FAIL(after.size() == tags.size());

	// work out what is being added and what is being removed

	vector<Tag *> added;
	vector<Tag *> removed;

	set_difference(after.begin(), after.end(),
		       before.begin(), before.end(), back_inserter(added));

	set_difference(before.begin(), before.end(), 
		       after.begin(), after.end(), back_inserter(removed));

	// transaction time

	begin_transaction();

	//

	Command * command = command_flip_argument(this, &ModelElement::set_tags, tags, mTags);
	log_command(command);
	command->unref();

	// make the changes

	for_each(added.begin(), added.end(),
		 bind2nd(mem_fun(&Tag::internal_add_model), this));
	for_each(removed.begin(), removed.end(),
		 bind2nd(mem_fun(&Tag::internal_remove_model), this));

	for_each(added.begin(), added.end(), mem_fun(&Object::ref));
	for_each(removed.begin(), removed.end(), mem_fun(&Object::unref));

	mTags = tags;

	// changes made

	for(vector<Tag *>::iterator iter = added.begin();
	    iter != added.end();
	    ++iter) {
	    Tag * tag = *iter;
	    ChangeEvent * change = change_role_added(this, "tag", tag, false);
	    log_change(change);
	    change->unref();
	}

	for(vector<Tag *>::iterator iter = removed.end();
	    iter != removed.end();
	    ++iter) {
	    Tag * tag = *iter;
	    ChangeEvent * change = change_role_removed(this, "tag", tag, false);
	    log_change(change);
	    change->unref();
	}

	//

	end_transaction();
    }

    void
    ModelElement::add_constraint(Constraint * constraint) {
	RETURN_IF_FAIL(mConstraints.find(constraint) == mConstraints.end());

	//

	begin_transaction();

	//

	Command * command = command_flip_method(this,
						&ModelElement::add_constraint,
						&ModelElement::remove_constraint,
						constraint);
	log_command(command);
	command->unref();

	// make the changes

	constraint->ref();

	mConstraints.insert(constraint);
	constraint->internal_add_constrained_element(this);

	// changes made

	ChangeEvent * change = change_role_added(this, "constraint", constraint, false);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    ModelElement::remove_constraint(Constraint * constraint) {
	RETURN_IF_FAIL(mConstraints.find(constraint) != mConstraints.end());

	//

	begin_transaction();

	//

	Command * command = command_flip_method(this,
						&ModelElement::remove_constraint,
						&ModelElement::add_constraint,
						constraint);
	log_command(command);
	command->unref();

	// make the changes

	constraint->internal_remove_constrained_element(this);
	mConstraints.erase(constraint);

	constraint->unref();

	// changes made

	ChangeEvent * change = change_role_removed(this, "constraint", constraint, false);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    ModelElement::set_constraints(const set<Constraint *> & constraints) {
	begin_transaction();

	//

	Command * command = command_flip_argument(this,
						  &Namespace::set_constraints,
						  constraints,
						  mConstraints);
	log_command(command);
	command->unref();

	// make the changes

	vector<Constraint *> added;
	vector<Constraint *> removed;

	set_difference(constraints.begin(), constraints.end(),
		       mConstraints.begin(), mConstraints.end(), back_inserter(added));

	set_difference(mConstraints.begin(), mConstraints.end(),
		       constraints.begin(), constraints.end(), back_inserter(removed));

	for_each(added.begin(), added.end(),
		 bind2nd(mem_fun(&Constraint::internal_add_constrained_element), this));
	for_each(removed.begin(), removed.end(),
		 bind2nd(mem_fun(&Constraint::internal_remove_constrained_element), this));

	for_each(added.begin(), added.end(), mem_fun(&Object::ref));
	for_each(removed.begin(), removed.end(), mem_fun(&Object::unref));

	mConstraints = constraints;

	// changes made

	for(vector<Constraint *>::iterator iter = added.begin();
	    iter != added.end();
	    ++iter) {
	    Constraint * constraint = *iter;
	    ChangeEvent * change = change_role_added(this, "constraint", constraint, false);
	    log_change(change);
	    change->unref();
	}

	for(vector<Constraint *>::iterator iter = removed.begin();
	    iter != removed.end();
	    ++iter) {
	    Constraint * constraint = *iter;
	    ChangeEvent * change = change_role_removed(this, "constraint", constraint, false);
	    log_change(change);
	    change->unref();
	}

	//

	end_transaction();
    }

    //

    string
    ModelElement::get_qualified_name() const {
	if(mContainer) {
	    return mContainer->get_qualified_name() + "::" + mName->get_value();
	}
	else {
	    return mName->get_value();
	}
    }

    //
    // reflective interface
    //

    void
    ModelElement::ref_set_context_impl(set<string> & visited, REF::Context * context) {
	if(visited.insert(TYPE).second) {
	}
    }

    void
    ModelElement::ref_implant_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_add_model_element(this);
	}
    }

    void
    ModelElement::ref_uproot_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_remove_model_element(this);
	}
    }

    //

    const string &
    ModelElement::ref_get_type() const {
	return TYPE;
    }

    bool
    ModelElement::ref_is_a_type_of_impl(set<string> & visited,
				   const string & class_name) const {
	return class_name == TYPE;
    }

    //

    bool
    ModelElement::ref_add_role_impl(set<string> & visited,
					const string & role_name,
					REF::Class * role,
					REF::Class * position) {
	NYI();

	return false;
    }

    bool
    ModelElement::ref_remove_role_impl(set<string> & visited,
					   const string & role_name,
					   REF::Class * role) {
	NYI();

	return false;
    }

    bool
    ModelElement::ref_get_role_impl(set<string> & visited,
					REF::Class * & klass,
					const string & role_name) {
	NYI();

	return false;
    }

    bool
    ModelElement::ref_set_role_impl(set<string> & visited,
					const string & role_name, REF::Class * role) {
	NYI();

	return false;
    }

    bool
    ModelElement::ref_get_role_iterator_impl(set<string> & visited,
						 REF::Iterator * & iterator,
						 const string & role_name) {
	NYI();

	return false;
    }

    bool
    ModelElement::ref_get_attribute_impl(set<string> & visited,
					 const REF::Attribute * & attribute,
					 const string & attribute_name) const {
	NYI();

	return false;
    }

    bool
    ModelElement::ref_set_attribute_impl(set<string> & visited,
					     const string & attribute_name,
					     const REF::Attribute * attribute) {
	NYI();

	return false;
    }

    bool
    ModelElement::ref_get_attribute_set_impl(set<string> & visited,
					     const vector<const REF::Attribute *> * & attribute_set,
					     const string & attribute_name) const {
	NYI();

	return false;
    }

    bool
    ModelElement::ref_set_attribute_set_impl(set<string> & visited,
					     const string & attribute_name,
					     const vector<const REF::Attribute *> * attribute_set) {
	NYI();

	return false;
    }

    bool
    ModelElement::ref_create_attribute_impl(set<string> & visited,
						REF::Attribute * & attribute,
						const string & attribute_name) const {
	NYI();

	return false;
    }

    bool
    ModelElement::ref_create_class_impl(set<string> & visitd,
					REF::Class * & role,
					const string & role_name) const {
	NYI();

	return false;
    }

    void
    ModelElement::ref_delete_impl(set<string> & visited) {
	if(visited.insert(TYPE).second) {
	    Namespace * null_namespace = 0;
	    set<ModelElement *> empty_model_element_set;
	    set<Constraint *> empty_constraint_set;
	    list<Tag *> empty_tag_list;

	    set_container(null_namespace);
	    set_constraints(empty_constraint_set);
	    set_providers(empty_model_element_set);
	    set_tags(empty_tag_list);
	}
    }

    //
    //
    //

    void
    ModelElement::set_container(Namespace * container) {
	// transaction time

	begin_transaction();

	// rationalise the contexts

	if(container) {
	    internal_ref_set_context(container->ref_get_context());
	}

	//

	Command * command = command_flip_argument(this, &ModelElement::set_container,
						  container, mContainer);
	log_command(command);
	command->unref();

	// make the changes

	if(mContainer) {
	    mContainer->internal_remove_contained_element(this);
	    mContainer->unref();
	}

	mContainer = container;

	if(mContainer) {
	    mContainer->ref();
	    mContainer->internal_add_contained_element(this);
	}

	// changes made

	ChangeEvent * change = change_role_changed(this, "container", container, mContainer, true);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    ModelElement::set_frozen(const Boolean * frozen) {
	begin_transaction();

	//

	Boolean * old_value = mFrozen ? new Boolean(*mFrozen) : 0;
	Boolean * new_value = frozen ? new Boolean(*frozen) : 0;

	//

	Command * command = command_flip_argument(this, &ModelElement::set_frozen,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mFrozen) {
	    mFrozen->unref();
	}

	mFrozen = new_value ? new Boolean(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "frozen", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    ModelElement::set_visible(const Boolean * visible) {
	begin_transaction();

	//

	Boolean * old_value = mVisible ? new Boolean(*mVisible) : 0;
	Boolean * new_value = visible ? new Boolean(*visible) : 0;

	//

	Command * command = command_flip_argument(this, &ModelElement::set_visible,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mVisible) {
	    mVisible->unref();
	}

	mVisible = new_value ? new Boolean(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "name", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    //
    // internal functions, these should not be called by user code
    //

    void
    ModelElement::internal_set_container(Namespace * container) throw() {
	if(mContainer != container) {
	    if(mContainer) {
		mContainer->unref();
	    }

	    mContainer = container;

	    if(mContainer) {
		mContainer->ref();
	    }
	}
    }

    void
    ModelElement::internal_add_dependent(ModelElement * dependent) throw() {
	dependent->ref();
	mDependents.insert(dependent);
    }

    void
    ModelElement::internal_remove_dependent(ModelElement * dependent) throw() {
	mDependents.erase(dependent);
	dependent->unref();
    }

    void
    ModelElement::internal_add_tag(Tag * tag) throw() {
	tag->ref();
	mTags.push_back(tag);
    }

    void
    ModelElement::internal_remove_tag(Tag * tag) throw() {
	mTags.remove(tag);
	tag->unref();
    }

    void
    ModelElement::internal_add_constraint(Constraint * constraint) throw() {
	constraint->ref();
	mConstraints.insert(constraint);
    }

    void
    ModelElement::internal_remove_constraint(Constraint * constraint) throw() {
	mConstraints.erase(constraint);
	constraint->unref();
    }

} // namespace MOF
