// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "namespace.h"

#include "import.h"
#include "model.h"

#include "reflective/changes.h"
#include "reflective/list_iterator.h"
#include "reflective/set_iterator.h"

#include "foundation/functional.h"
#include "foundation/macros.h"
#include "foundation/commands.h"

#include <algorithm>
#include <functional>
#include <vector>
#include <iterator>
using namespace std;

namespace MOF {

    const string Namespace::TYPE = "Namespace";

    //

    Namespace::Namespace() : ModelElement() {
    }

    Namespace::~Namespace() {
    }

    void
    Namespace::add_contained_element(ModelElement * contained_element, int index) {
	// if the index is negative then we add the item to the end of the list

	if(index < 0) {
	    index = mContainedElements.size();
	}

	// check that the index is valid

	if(index > static_cast<int>(mContainedElements.size())) {
	    ERROR("index out of range");
	    throw;	// FIXME
	}

	// traverse the list to find the iterator and make sure that we aren't
	// introducing a duplicate

	list<ModelElement *>::iterator iter = mContainedElements.begin();

	for(int i = 0; i < static_cast<int>(mContainedElements.size()); ++i) {
	    if(*iter == contained_element) {
		ERROR("The element is already in the container");
		throw;	// FIXME
	    }

	    if(i < index) {
		++iter;
	    }
	}

	// check that the contained element doesn't have a context already

	if(contained_element->ref_get_context()) {
	    ERROR("Forming aggregation with element that already has a context");
	    throw;	// FIXME
	}

	// transaction time

	begin_transaction();

	// rationalise the context

	if(ref_get_context()) {
	    contained_element->internal_ref_set_context(ref_get_context());
	}

	//

	Command * command = command_flip_method(this,
						&Namespace::add_contained_element,
						&Namespace::remove_contained_element,
						contained_element, index);
	log_command(command);
	command->unref();

	// make the changes

	contained_element->ref();

	mContainedElements.insert(iter, contained_element);
	contained_element->internal_set_container(this);

	// changes made

	ChangeEvent * change = change_role_added(this, "contained_element",
						 contained_element, index);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    Namespace::remove_contained_element(ModelElement * contained_element) {
	// get an iterator to the element in our list

	int index = 0;

	list<ModelElement *>::iterator iter = mContainedElements.begin();

	for(; iter != mContainedElements.end(); ++iter, ++index) {
	    if(*iter == contained_element) {
		break;
	    }
	}

	if(iter == mContainedElements.end()) {
	    ERROR("Couldn't find element in list");
	    throw;	// FIXME
	}

	// transaction time

	begin_transaction();

	//

	Command * command = command_flip_method(this,
						&Namespace::remove_contained_element,
						&Namespace::add_contained_element,
						contained_element, index);
	log_command(command);
	command->unref();

	// make the changes

	contained_element->internal_set_container(0);
	mContainedElements.erase(iter);

	contained_element->unref();

	// delete the contained element

	contained_element->ref_delete();

	// rationalise the context

	contained_element->internal_ref_set_context(0);

	// changes made

	ChangeEvent * change = change_role_removed(this, "contained_element",
						   contained_element, index);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    Namespace::set_contained_elements(const list<ModelElement *> & contained_elements) {
	// put the items into sets so that we can call set_difference

	set<ModelElement *> before;
	set<ModelElement *> after;

	before.insert(mContainedElements.begin(), mContainedElements.end());
	after.insert(contained_elements.begin(), contained_elements.end());

	// check for duplicates

	RETURN_IF_FAIL(before.size() == mContainedElements.size());
	RETURN_IF_FAIL(after.size() == contained_elements.size());

	// work out what is being added and what is being removed

	vector<ModelElement *> added;
	vector<ModelElement *> removed;

	set_difference(after.begin(), after.end(), 
		       before.begin(), before.end(), back_inserter(added));

	set_difference(before.begin(), before.end(), 
		       after.begin(), after.end(), back_inserter(removed));

	// check that the contained elements don't have a context already

	if(!after.empty() && find_if(added.begin(), added.end(),
		mem_fun(&ModelElement::ref_get_context)) != added.end()) {
	    ERROR("Forming aggregation with element that already has a context");
	    throw;
	}

	// transaction time

	begin_transaction();

	// rationalise the context

	if(ref_get_context()) {
	    for_each(added.begin(), added.end(),
		     bind2nd(mem_fun(&ModelElement::internal_ref_set_context),
			     ref_get_context()));

	    for_each(removed.begin(), removed.end(),
		     bind2nd(mem_fun(&ModelElement::internal_ref_set_context),
			     static_cast<REF::Context *>(0)));
	}

	//

	Command * command = command_flip_argument(this,
						  &Namespace::set_contained_elements,
						  contained_elements,
						  mContainedElements);
	log_command(command);
	command->unref();

	// make the changes

	for_each(added.begin(), added.end(),
		 bind2nd(mem_fun(&ModelElement::internal_set_container), this));
	for_each(removed.begin(), removed.end(),
		 bind2nd(mem_fun(&ModelElement::internal_set_container),
			 static_cast<Namespace *>(0)));

	for_each(added.begin(), added.end(), mem_fun(&Object::ref));
	for_each(removed.begin(), removed.end(), mem_fun(&Object::unref));

	mContainedElements = contained_elements;

	// changes made

	for(vector<ModelElement *>::iterator iter = added.begin();
	    iter != added.end();
	    ++iter) {
	    ModelElement * contained_element = *iter;
	    ChangeEvent * change = change_role_added(this, "contained_element",
						     contained_element, true);
	    log_change(change);
	    change->unref();
	}

	for(vector<ModelElement *>::iterator iter = removed.begin();
	    iter != removed.end();
	    ++iter) {
	    ModelElement * contained_element = *iter;
	    ChangeEvent * change = change_role_removed(this, "contained_element",
						       contained_element, true);
	    log_change(change);
	    change->unref();
	}

	//

	end_transaction();
    }

    void
    Namespace::modify_contained_element(ModelElement * old_contained_element,
					ModelElement * new_contained_element) {
	// check that the elements are not the same

	if(old_contained_element == new_contained_element) {
	    ERROR("The elements are the same");
	    throw;	// FIXME
	}

	// get an iterator to the element in our list

	int index = 0;

	list<ModelElement *>::iterator iter = mContainedElements.begin();

	for(; iter != mContainedElements.end(); ++iter, ++index) {
	    if(*iter == old_contained_element) {
		break;
	    }
	}

	if(iter == mContainedElements.end()) {
	    ERROR("Couldn't find element in list");
	    throw;	// FIXME
	}

	// check that the element doesn't have a context already

	if(new_contained_element->ref_get_context()) {
	    ERROR("Forming aggregation with element that already has a context");
	    throw;	// FIXME
	}

	// transaction time

	begin_transaction();

	// rationalise the context

	if(ref_get_context()) {
	    new_contained_element->internal_ref_set_context(ref_get_context());
	    old_contained_element->internal_ref_set_context(0);
	}

	//

	Command * command = command_reverse_argument(this,
						     &Namespace::modify_contained_element,
						     old_contained_element,
						     new_contained_element);
	log_command(command);
	command->unref();

	// make the changes

	old_contained_element->internal_set_container(0);
	*iter = new_contained_element;
	new_contained_element->internal_set_container(this);

	new_contained_element->ref();
	old_contained_element->unref();

	// changes made

	ChangeEvent * change = change_role_changed(this, "contained_element",
						   old_contained_element,
						   new_contained_element, true);
	log_change(change);
	change->unref();

	end_transaction();
    }

    void
    Namespace::internal_add_contained_element(ModelElement * contained_element) throw() {
	contained_element->ref();
	mContainedElements.push_back(contained_element);
    }

    void
    Namespace::internal_remove_contained_element(ModelElement * contained_element) throw() {
	mContainedElements.remove(contained_element);
	contained_element->unref();
    }

    void
    Namespace::internal_add_importer(Import * importer) throw() {
	importer->ref();
	mImporters.insert(importer);
    }

    void
    Namespace::internal_remove_importer(Import * importer) throw() {
	mImporters.erase(importer);
	importer->unref();
    }

#if 0
    ModelElement *
    Namespace::lookup(const string & name) {
	list<ModelElement *>::iterator iter =
	    find_if(mContainedElements.begin(), mContainedElements.end(),
		    bind2nd_vr(mem_fun(&ModelElement::has_name), name));

	if(iter == mContainedElements.end()) {
	    ERROR("Couldn't find named element");
	    //throw NameNotFound();
	    throw;	// FIXME
	}
	else {
	    return *iter;
	}
    }
#endif

#if 0
    ModelElement *
    Namespace::resolve_qualified_name(const vector<string> & names) {
	if(names.empty()) {
	    return this;
	}
	else {
	    Namespace * ns = this;

	    for(unsigned int i = 0; i < names.size() - 1; ++i) {
		ModelElement * model_element = ns->lookup(names[i]);

		ns = dynamic_cast<Namespace *>(model_element);

		if(!ns) {
		    ERROR("Couldn't find named element");
		    //throw NameNotFound();
		    throw;	// FIXME
		}
	    }

	    return ns->lookup(names.back());
	}
    }
#endif

#if 0
    bool
    Namespace::name_is_valid(const string & proposed_name) const {
	return find_if(mContainedElements.begin(), mContainedElements.end(),
		       bind2nd_vr(mem_fun(&ModelElement::has_name),
				  proposed_name)) == mContainedElements.end();
    }
#endif

#if 0
    void
    Namespace::find_elements_by_type(set<ModelElement *> & model_element, const type_info & type) {
    }
#endif

    //
    //
    //

    void
    Namespace::ref_set_context_impl(set<string> & visited, REF::Context * context) {
	if(visited.insert(TYPE).second) {
	    ModelElement::ref_set_context_impl(visited, context);

	    for_each(mContainedElements.begin(), mContainedElements.end(),
		     bind2nd(mem_fun(&ModelElement::internal_ref_set_context), ref_get_context()));
	}
    }

    //
    //
    //

    void
    Namespace::ref_delete_impl(set<string> & visited) {
	if(visited.insert(TYPE).second) {
	    // note that we must copy the container, because it will change during
	    // the calls to remove

	    list<ModelElement *> empty_model_element_set;
	    set<Import *> importers_copy = mImporters;		// have to make a copy
	    Namespace * null_namespace = 0;

	    set_contained_elements(empty_model_element_set);
	    for_each(importers_copy.begin(), importers_copy.end(),
		     bind2nd(mem_fun(&Import::set_imported), null_namespace));

#if 0
	    // FIXME
	    list<ModelElement *> contained_element = mContainedElements;
	    for_each(contained_element.begin(), contained_element.end(),
		     mem_fun(&ModelElement::internal_ref_delete));
#endif

	    ModelElement::ref_delete_impl(visited);
	}
    }

    bool
    Namespace::ref_is_a_type_of_impl(set<string> & visited, const string & class_name) const {
	if(visited.insert(TYPE).second) {
	    return class_name == TYPE || ModelElement::ref_is_a_type_of(class_name);
	}
	else {
	    return false;
	}
    }

    const string &
    Namespace::ref_get_type() const {
	return TYPE;
    }

    void
    Namespace::ref_implant_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_add_namespace(this);

	    ModelElement::ref_implant_impl(visited);
	}
    }

    void
    Namespace::ref_uproot_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_remove_namespace(this);

	    ModelElement::ref_uproot_impl(visited);
	}
    }

} // namespace MOF
