// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // CONFIG_H

#include "parameter.h"

#include "model.h"

#include "foundation/macros.h"
#include "foundation/commands.h"
#include "reflective/changes.h"

namespace MOF {

    const string Parameter::TYPE = "Parameter";

    //

    Parameter::Parameter() : TypedElement(),
        mDirection(0),
	mMultiplicity(0) {
    }

    Parameter::~Parameter() {
    }

    void
    Parameter::set_direction(const DirectionKind * direction) {
	begin_transaction();

	//

	DirectionKind * old_value = mDirection ? new DirectionKind(*mDirection) : 0;
	DirectionKind * new_value = direction ? new DirectionKind(*direction) : 0;

	//

	Command * command = command_flip_argument(this, &Parameter::set_direction,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mDirection) {
	    mDirection->unref();
	}

	mDirection = new_value ? new DirectionKind(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "direction", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    Parameter::set_multiplicity(const MultiplicityType * multiplicity) {
	begin_transaction();

	//

	MultiplicityType * old_value = mMultiplicity ? new MultiplicityType(*mMultiplicity) : 0;
	MultiplicityType * new_value = multiplicity ? new MultiplicityType(*multiplicity) : 0;

	//

	Command * command = command_flip_argument(this, &Parameter::set_multiplicity,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mMultiplicity) {
	    mMultiplicity->unref();
	}

	mMultiplicity = new_value ? new MultiplicityType(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "multiplicity", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    //

    bool
    Parameter::ref_is_a_type_of_impl(set<string> & visited, const string & class_name) const {
	if(visited.insert(TYPE).second) {
	    return class_name == TYPE || TypedElement::ref_is_a_type_of(class_name);
	}
	else {
	    return false;
	}
    }

    const string &
    Parameter::ref_get_type() const {
	return TYPE;
    }

    void
    Parameter::ref_implant_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_add_parameter(this);

	    TypedElement::ref_implant_impl(visited);
	}
    }

    void
    Parameter::ref_uproot_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_remove_parameter(this);

	    TypedElement::ref_uproot_impl(visited);
	}
    }

} // namespace MOF
