// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "structural_feature.h"

#include "model.h"

#include "foundation/commands.h"
#include "reflective/changes.h"

namespace MOF {

    const string StructuralFeature::TYPE = "StructuralFeature";

    //

    StructuralFeature::StructuralFeature() : Feature(),
					     TypedElement(),
	mMultiplicity(0),
        mChangeable(0) {
    }

    StructuralFeature::~StructuralFeature() {
    }

    void
    StructuralFeature::set_changeable(const Boolean * changeable) {
	begin_transaction();

	//

	Boolean * old_value = mChangeable ? new Boolean(*mChangeable) : 0;
	Boolean * new_value = changeable ? new Boolean(*changeable) : 0;

	//

	Command * command = command_flip_argument(this, &StructuralFeature::set_changeable,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mChangeable) {
	    mChangeable->unref();
	}

	mChangeable = new_value ? new Boolean(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "changeable",
							new_value, old_value);

	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    StructuralFeature::set_multiplicity(const MultiplicityType * multiplicity) {
	begin_transaction();

	//

	MultiplicityType * old_value = mMultiplicity ? new MultiplicityType(*mMultiplicity) : 0;
	MultiplicityType * new_value = multiplicity ? new MultiplicityType(*multiplicity) : 0;

	//

	Command * command = command_flip_argument(this, &StructuralFeature::set_multiplicity,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mMultiplicity) {
	    mMultiplicity->unref();
	}

	mMultiplicity = new_value ? new MultiplicityType(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "multiplicity",
							new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    //

    void
    StructuralFeature::ref_delete_impl(set<string> & visited) {
	if(visited.insert(TYPE).second) {
	    Feature::ref_delete_impl(visited);
	    TypedElement::ref_delete_impl(visited);
	}
    }

    //

    bool
    StructuralFeature::ref_is_a_type_of_impl(set<string> & visited,
						 const string & class_name) const {
	if(visited.insert(TYPE).second) {
	    return class_name == TYPE ||
		Feature::ref_is_a_type_of(class_name) ||
		TypedElement::ref_is_a_type_of(class_name);
	}
	else {
	    return false;
	}
    }

    const string &
    StructuralFeature::ref_get_type() const {
	return TYPE;
    }

    void
    StructuralFeature::ref_implant_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_add_structural_feature(this);

	    Feature::ref_implant_impl(visited);
	    TypedElement::ref_implant_impl(visited);
	}
    }

    void
    StructuralFeature::ref_uproot_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_remove_structural_feature(this);

	    Feature::ref_uproot_impl(visited);
	    TypedElement::ref_uproot_impl(visited);
	}
    }

} // namespace MOF
