// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "tag.h"

#include "model.h"

#include "foundation/macros.h"
#include "foundation/functional.h"
#include "foundation/commands.h"
#include "reflective/changes.h"

#include <vector>
#include <algorithm>
using namespace std;

namespace MOF {

    const string Tag::TYPE = "Tag";

    //

    Tag::Tag() : ModelElement(),
        mTagId(0) {
    }

    Tag::~Tag() {
    }

    void
    Tag::set_tag_id(const Name * tag_id) {
	begin_transaction();

	//

	Name * old_value = mTagId ? new Name(*mTagId) : 0;
	Name * new_value = tag_id ? new Name(*tag_id) : 0;

	//

	Command * command = command_flip_argument(this, &Tag::set_tag_id,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mTagId) {
	    mTagId->unref();
	}

	mTagId = tag_id ? new Name(*tag_id) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "tag_id", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    Tag::add_model(ModelElement * model) {
	RETURN_IF_FAIL(mModels.find(model) == mModels.end());

	begin_transaction();

	//

	Command * command = command_flip_method(this, &Tag::add_model, &Tag::remove_model, model);
	log_command(command);
	command->unref();

	// make the changes

	model->ref();

	mModels.insert(model);
	model->internal_add_tag(this);

	// changes made

	ChangeEvent * change = change_role_added(this, "model", model, false);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    Tag::remove_model(ModelElement * model) {
	RETURN_IF_FAIL(mModels.find(model) != mModels.end());

	begin_transaction();

	//

	Command * command = command_flip_method(this, &Tag::remove_model, &Tag::add_model, model);
	log_command(command);
	command->unref();

	// make the changes

	model->internal_remove_tag(this);
	mModels.erase(model);

	model->unref();

	// changes made

	ChangeEvent * change = change_role_removed(this, "model", model, false);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    Tag::set_models(const set<ModelElement *> & models) {
	begin_transaction();

	Command * command = command_flip_argument(this, &Tag::set_models, models, mModels);
	log_command(command);
	command->unref();

	// make the changes

	vector<ModelElement *> added;
	vector<ModelElement *> removed;

	set_difference(models.begin(), models.end(),
		       mModels.begin(), mModels.end(), back_inserter(added));

	set_difference(mModels.begin(), mModels.end(),
		       models.begin(), models.end(), back_inserter(removed));

	for_each(added.begin(), added.end(),
		 bind2nd(mem_fun(&ModelElement::internal_add_tag), this));
	for_each(removed.begin(), removed.end(),
		 bind2nd(mem_fun(&ModelElement::internal_remove_tag), this));

	for_each(added.begin(), added.end(), mem_fun(&Object::ref));
	for_each(removed.begin(), removed.end(), mem_fun(&Object::unref));

	mModels = models;

	// changes made

	for(vector<ModelElement *>::iterator iter = added.begin();
	    iter != added.end();
	    ++iter) {
	    ModelElement * model = *iter;
	    ChangeEvent * change = change_role_added(this, "model", model, false);
	    log_change(change);
	    change->unref();
	}

	for(vector<ModelElement *>::iterator iter = removed.begin();
	    iter != removed.end();
	    ++iter) {
	    ModelElement * model = *iter;
	    ChangeEvent * change = change_role_removed(this, "model", model, false);
	    log_change(change);
	    change->unref();
	}

	//

	end_transaction();
    }

    //

    void
    Tag::internal_add_model(ModelElement * model) throw() {
	model->ref();
	mModels.insert(model);
    }

    void
    Tag::internal_remove_model(ModelElement * model) throw() {
	mModels.erase(model);
	model->unref();
    }

    void
    Tag::ref_delete_impl(set<string> & visited) {
	if(visited.insert(TYPE).second) {
	    set<ModelElement *> empty_model_element_set;

	    set_models(empty_model_element_set);

	    ModelElement::ref_delete_impl(visited);
	}
    }

    //

    bool
    Tag::ref_is_a_type_of_impl(set<string> & visited, const string & class_name) const {
	if(visited.insert(TYPE).second) {
	    return class_name == TYPE ||
		ModelElement::ref_is_a_type_of_impl(visited, class_name);
	}
	else {
	    return false;
	}
    }

    const string &
    Tag::ref_get_type() const {
	return TYPE;
    }

    void
    Tag::ref_implant_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_add_tag(this);

	    ModelElement::ref_implant_impl(visited);
	}
    }

    void
    Tag::ref_uproot_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_remove_tag(this);

	    ModelElement::ref_uproot_impl(visited);
	}
    }

} // namespace MOF
