#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "parser.h"

#include <iostream>
#include <iterator>
#include <algorithm>
using namespace std;

#include "foundation/macros.h"

namespace OCL {

    namespace {

	string
	indent_string(int columns) {
	    ASSERT(columns >= 0);

	    const int TAB_WIDTH = 8;

	    string ret;

	    fill_n(back_inserter(ret), columns / TAB_WIDTH, '\t');
	    fill_n(back_inserter(ret), columns % TAB_WIDTH, ' ');

	    return ret;
	}

    } // namespace {anonymous}

    ParserException::ParserException(const string & details) : Exception(details) {
    }

    ParserException::~ParserException() {
    }

    //
    //
    //

    SyntacticException::SyntacticException(const string & details) :
    ParserException(details) {
    }

    SyntacticException::~SyntacticException() {
    }

    //
    //
    //

    struct Incrementer {
	Incrementer(int & depth) : mDepth(depth) { ++mDepth; }
	~Incrementer() { --mDepth; }

    private:
	int & mDepth;
    };

    //
    //
    //

    Parser::Parser(istream & input) : Lexer(input), mGuessing(0), mDepth(-1) {
    }

    Parser::~Parser() {
	//ASSERT(!mGuessing);
	if(mGuessing) {
	    cerr << "Warning, we are still guessing..." << endl;
	}
    }

    void
    Parser::parse() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering parse()" << endl;
	constraint();
	cerr << indent_string(mDepth * 4) << "Leaving parse()" << endl;
    }

    void
    Parser::constraint() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering constraint()" << endl;
	context_declaration();

	bool had_something = false;

	while(lookahead(1) == TOKEN_LITERAL_DEF ||
	      lookahead(1) == TOKEN_LITERAL_PRE ||
	      lookahead(1) == TOKEN_LITERAL_POST ||
	      lookahead(1) == TOKEN_LITERAL_INV) {
	    had_something = true;
	    switch(lookahead(1)) {
	    case TOKEN_LITERAL_DEF:
		match(TOKEN_LITERAL_DEF);

		switch(lookahead(1)) {
		case TOKEN_NAME:
		    match(TOKEN_NAME);
		    break;
		case TOKEN_COLON:
		    break;
		default:
		    throw SyntacticException("Syntactic Exception");
		}

		match(TOKEN_COLON);

		while(lookahead(1) == TOKEN_LITERAL_LET) {
		    let_expression();
		}
		break;

	    case TOKEN_LITERAL_PRE:
	    case TOKEN_LITERAL_POST:
	    case TOKEN_LITERAL_INV:
		stereotype();

		switch(lookahead(1)) {
		case TOKEN_NAME:
		    match(TOKEN_NAME);
		    break;
		case TOKEN_COLON:
		    break;
		default:
		    throw SyntacticException("Syntactic Exception");
		}

		match(TOKEN_COLON);
		ocl_expression();
		break;
	    default:
		throw SyntacticException("Syntactic Exception");
	    }
	}

	if(!had_something) {
	    throw SyntacticException("Syntactic Exception");
	}

	cerr << indent_string(mDepth * 4) << "Leaving constraint()" << endl;
    }

    void
    Parser::context_declaration() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering context_declaration()" << endl;
	match(TOKEN_LITERAL_CONTEXT);

	if(lookahead(1) == TOKEN_NAME && lookahead(2) == TOKEN_DOUBLE_COLON) {
	    operation_context();
	}
	else if(lookahead(1) == TOKEN_NAME &&
		(lookahead(2) == TOKEN_LITERAL_DEF ||
		 lookahead(2) == TOKEN_COLON ||
		 lookahead(2) == TOKEN_LITERAL_PRE ||
		 lookahead(2) == TOKEN_LITERAL_POST ||
		 lookahead(2) == TOKEN_LITERAL_INV)) {
	    classifier_context();
	}
	else {
	    throw SyntacticException("Syntactic Exception");
	}

	cerr << indent_string(mDepth * 4) << "Leaving context_declaration()" << endl;
    }

    void
    Parser::let_expression() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering let_expression()" << endl;
	match(TOKEN_LITERAL_LET);
	match(TOKEN_NAME);

	switch(lookahead(1)) {
	case TOKEN_LEFT_ROUND_BRACKET:
	    match(TOKEN_LEFT_ROUND_BRACKET);
	    formal_parameter_list();
	    match(TOKEN_RIGHT_ROUND_BRACKET);
	    break;
	case TOKEN_COLON:
	case TOKEN_EQUAL:
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}

	switch(lookahead(1)) {
	case TOKEN_COLON:
	    match(TOKEN_COLON);
	    type_specifier();
	    break;
	case TOKEN_EQUAL:
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}

	match(TOKEN_EQUAL);
	ocl_expression();

	cerr << indent_string(mDepth * 4) << "Leaving let_expression()" << endl;
    }

    void
    Parser::stereotype() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering stereotype()" << endl;
	switch(lookahead(1)) {
	case TOKEN_LITERAL_PRE:
	    match(TOKEN_LITERAL_PRE);
	    break;
	case TOKEN_LITERAL_POST:
	    match(TOKEN_LITERAL_POST);
	    break;
	case TOKEN_LITERAL_INV:
	    match(TOKEN_LITERAL_INV);
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving stereotype()" << endl;
    }

    void
    Parser::ocl_expression() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering ocl_expression()" << endl;
	switch(lookahead(1)) {
	case TOKEN_LITERAL_IN:
	case TOKEN_LITERAL_LET:
	    while(lookahead(1) == TOKEN_LITERAL_LET) {
		let_expression();
	    }

	    match(TOKEN_LITERAL_IN);
	    break;

	case TOKEN_NAME:
	case TOKEN_LEFT_ROUND_BRACKET:
	case TOKEN_MINUS:
	case TOKEN_LITERAL_NOT:
	case TOKEN_LITERAL_IF:
	case TOKEN_STRING:
	case TOKEN_NUMBER:
	case TOKEN_LITERAL_SET:
	case TOKEN_LITERAL_BAG:
	case TOKEN_LITERAL_SEQUENCE:
	case TOKEN_LITERAL_COLLECTION:
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}

	expression();
	cerr << indent_string(mDepth * 4) << "Leaving ocl_expression()" << endl;
    }

    void
    Parser::operation_context() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering operation_context()" << endl;
	match(TOKEN_NAME);
	match(TOKEN_DOUBLE_COLON);

	operation_name();

	match(TOKEN_LEFT_ROUND_BRACKET);

	formal_parameter_list();

	match(TOKEN_RIGHT_ROUND_BRACKET);

	switch(lookahead(1)) {
	case TOKEN_COLON:
	    match(TOKEN_COLON);
	    return_type();
	    break;
	case TOKEN_LITERAL_DEF:
	case TOKEN_LITERAL_PRE:
	case TOKEN_LITERAL_POST:
	case TOKEN_LITERAL_INV:
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving operation_context()" << endl;
    }

    void
    Parser::classifier_context() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering classifier_context()" << endl;
	if(lookahead(1) == TOKEN_NAME && lookahead(2) == TOKEN_COLON) {
	    match(TOKEN_NAME);
	    match(TOKEN_COLON);
	    match(TOKEN_NAME);
	}
	else if(lookahead(1) == TOKEN_NAME &&
		(lookahead(2) == TOKEN_LITERAL_DEF ||
		 lookahead(2) == TOKEN_LITERAL_PRE ||
		 lookahead(2) == TOKEN_LITERAL_POST ||
		 lookahead(2) == TOKEN_LITERAL_INV)) {
	    match(TOKEN_NAME);
	}
	else {
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving classifier_context()" << endl;
    }

    void
    Parser::operation_name() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering operation_name()" << endl;
	switch(lookahead(1)) {
	case TOKEN_NAME:
	    match(TOKEN_NAME);
	    break;
	case TOKEN_EQUAL:
	    match(TOKEN_EQUAL);
	    break;
	case TOKEN_PLUS:
	    match(TOKEN_PLUS);
	    break;
	case TOKEN_MINUS:
	    match(TOKEN_MINUS);
	    break;
	case TOKEN_LESS_THAN:
	    match(TOKEN_LESS_THAN);
	    break;
	case TOKEN_GREATER_THAN:
	    match(TOKEN_GREATER_THAN);
	    break;
	case TOKEN_LESS_EQUAL:
	    match(TOKEN_LESS_EQUAL);
	    break;
	case TOKEN_GREATER_EQUAL:
	    match(TOKEN_GREATER_EQUAL);
	    break;
	case TOKEN_DIVIDE:
	    match(TOKEN_DIVIDE);
	    break;
	case TOKEN_MULTIPLY:
	    match(TOKEN_MULTIPLY);
	    break;
	case TOKEN_NOT_EQUAL:
	    match(TOKEN_NOT_EQUAL);
	    break;
	case TOKEN_LITERAL_IMPLIES:
	    match(TOKEN_LITERAL_IMPLIES);
	    break;
	case TOKEN_LITERAL_NOT:
	    match(TOKEN_LITERAL_NOT);
	    break;
	case TOKEN_LITERAL_OR:
	    match(TOKEN_LITERAL_OR);
	    break;
	case TOKEN_LITERAL_XOR:
	    match(TOKEN_LITERAL_XOR);
	    break;
	case TOKEN_LITERAL_AND:
	    match(TOKEN_LITERAL_AND);
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving operation_name()" << endl;
    }

    void
    Parser::formal_parameter_list() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering formal_parameter_list()" << endl;
	switch(lookahead(1)) {
	case TOKEN_NAME:
	    match(TOKEN_NAME);
	    match(TOKEN_COLON);
	    type_specifier();

	    while(lookahead(1) == TOKEN_COMMA) {
		match(TOKEN_COMMA);
		match(TOKEN_NAME);
		match(TOKEN_COLON);
		type_specifier();
	    }
	    break;
	case TOKEN_RIGHT_ROUND_BRACKET:
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving formal_parameter_list()" << endl;
    }

    void
    Parser::return_type() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering return_type()" << endl;
	type_specifier();
	cerr << indent_string(mDepth * 4) << "Leaving return_type()" << endl;
    }

    void
    Parser::type_specifier() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering type_specifier()" << endl;
	switch(lookahead(1)) {
	case TOKEN_NAME:
	    simple_type_specifier();
	    break;
	case TOKEN_LITERAL_SET:
	case TOKEN_LITERAL_BAG:
	case TOKEN_LITERAL_SEQUENCE:
	case TOKEN_LITERAL_COLLECTION:
	    collection_type();
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving type_specifier()" << endl;
    }

    void
    Parser::simple_type_specifier() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering simple_type_specifier()" << endl;
	path_name();
	cerr << indent_string(mDepth * 4) << "Leaving simple_type_specifier()" << endl;
    }

    void
    Parser::collection_type() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering collection_type()" << endl;
	collection_kind();
	match(TOKEN_LEFT_ROUND_BRACKET);
	simple_type_specifier();
	match(TOKEN_RIGHT_ROUND_BRACKET);
	cerr << indent_string(mDepth * 4) << "Leaving collection_type()" << endl;
    }

    void
    Parser::collection_kind() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering collection_kind()" << endl;
	switch(lookahead(1)) {
	case TOKEN_LITERAL_SET:
	    match(TOKEN_LITERAL_SET);
	    break;
	case TOKEN_LITERAL_BAG:
	    match(TOKEN_LITERAL_BAG);
	    break;
	case TOKEN_LITERAL_SEQUENCE:
	    match(TOKEN_LITERAL_SEQUENCE);
	    break;
	case TOKEN_LITERAL_COLLECTION:
	    match(TOKEN_LITERAL_COLLECTION);
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving collection_kind()" << endl;
    }

    void
    Parser::expression() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering expression()" << endl;
	logical_expression();
	cerr << indent_string(mDepth * 4) << "Leaving expression()" << endl;
    }

    void
    Parser::logical_expression() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering logical_expression()" << endl;
	relational_expression();
	while(lookahead(1) == TOKEN_LITERAL_IMPLIES ||
	      lookahead(1) == TOKEN_LITERAL_OR ||
	      lookahead(1) == TOKEN_LITERAL_XOR ||
	      lookahead(1) == TOKEN_LITERAL_AND) {
	    logical_operator();
	    relational_expression();
	}
	cerr << indent_string(mDepth * 4) << "Leaving logical_expression()" << endl;
    }

    void
    Parser::if_expression() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering if_expression()" << endl;
	match(TOKEN_LITERAL_IF);
	expression();
	match(TOKEN_LITERAL_THEN);
	expression();
	match(TOKEN_LITERAL_ELSE);
	expression();
	match(TOKEN_LITERAL_ENDIF);
	cerr << indent_string(mDepth * 4) << "Leaving if_expression()" << endl;
    }

    void
    Parser::relational_expression() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering relational_expression()" << endl;
	additive_expression();

	switch(lookahead(1)) {
	case TOKEN_EQUAL:
	case TOKEN_LESS_THAN:
	case TOKEN_LESS_EQUAL:
	case TOKEN_GREATER_EQUAL:
	case TOKEN_GREATER_THAN:
	case TOKEN_NOT_EQUAL:
	    relational_operator();
	    additive_expression();
	    break;
	case TOKEN_EOF:
	case TOKEN_LITERAL_DEF:
	case TOKEN_RIGHT_ROUND_BRACKET:
	case TOKEN_LITERAL_PRE:
	case TOKEN_LITERAL_POST:
	case TOKEN_LITERAL_INV:
	case TOKEN_LITERAL_IMPLIES:
	case TOKEN_LITERAL_OR:
	case TOKEN_LITERAL_XOR:
	case TOKEN_LITERAL_AND:
	case TOKEN_COMMA:
	case TOKEN_LITERAL_IN:
	case TOKEN_LITERAL_LET:
	case TOKEN_LITERAL_THEN:
	case TOKEN_LITERAL_ELSE:
	case TOKEN_LITERAL_ENDIF:
	case TOKEN_BAR:
	case TOKEN_RIGHT_CURLY_BRACKET:
	case TOKEN_DOT_DOT:
	case TOKEN_RIGHT_SQUARE_BRACKET:
	    break;
	default:
	    cerr << "But it is: " << lookahead(1) << endl;
	    throw SyntacticException("Syntactic Exception");
	}

	cerr << indent_string(mDepth * 4) << "Leaving relational_expression()" << endl;
    }

    void
    Parser::logical_operator() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering logical_operator()" << endl;
	switch(lookahead(1)) {
	case TOKEN_LITERAL_AND:
	    match(TOKEN_LITERAL_AND);
	    break;
	case TOKEN_LITERAL_OR:
	    match(TOKEN_LITERAL_OR);
	    break;
	case TOKEN_LITERAL_XOR:
	    match(TOKEN_LITERAL_XOR);
	    break;
	case TOKEN_LITERAL_IMPLIES:
	    match(TOKEN_LITERAL_IMPLIES);
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving logical_operator()" << endl;
    }

    void
    Parser::additive_expression() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering additive_expression()" << endl;
	multiplicative_expression();

	while(lookahead(1) == TOKEN_PLUS || lookahead(1) == TOKEN_MINUS) {
	    add_operator();
	    multiplicative_expression();
	}
	cerr << indent_string(mDepth * 4) << "Leaving additive_expression()" << endl;
    }

    void
    Parser::relational_operator() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering relational_operator()" << endl;
	switch(lookahead(1)) {
	case TOKEN_EQUAL:
	    match(TOKEN_EQUAL);
	    break;
	case TOKEN_GREATER_THAN:
	    match(TOKEN_GREATER_THAN);
	    break;
	case TOKEN_LESS_THAN:
	    match(TOKEN_LESS_THAN);
	    break;
	case TOKEN_GREATER_EQUAL:
	    match(TOKEN_GREATER_EQUAL);
	    break;
	case TOKEN_LESS_EQUAL:
	    match(TOKEN_LESS_EQUAL);
	    break;
	case TOKEN_NOT_EQUAL:
	    match(TOKEN_NOT_EQUAL);
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving relational_operator()" << endl;
    }

    void
    Parser::multiplicative_expression() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering multiplicative_expression()" << endl;
	unary_expression();

	while(lookahead(1) == TOKEN_DIVIDE ||
	      lookahead(1) == TOKEN_MULTIPLY) {
	    multiply_operator();
	    unary_expression();
	}
	cerr << indent_string(mDepth * 4) << "Leaving multiplicative_expression()" << endl;
    }

    void
    Parser::add_operator() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering add_operator()" << endl;

	switch(lookahead(1)) {
	case TOKEN_PLUS:
	    match(TOKEN_PLUS);
	    break;
	case TOKEN_MINUS:
	    match(TOKEN_MINUS);
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving add_operator()" << endl;
    }

    void
    Parser::unary_expression() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering unary_expression()" << endl;

	switch(lookahead(1)) {
	case TOKEN_MINUS:
	case TOKEN_LITERAL_NOT:
	    unary_operator();
	    postfix_expression();
	    break;
	case TOKEN_NAME:
	case TOKEN_LEFT_ROUND_BRACKET:
	case TOKEN_LITERAL_IF:
	case TOKEN_STRING:
	case TOKEN_NUMBER:
	case TOKEN_LITERAL_SET:
	case TOKEN_LITERAL_BAG:
	case TOKEN_LITERAL_SEQUENCE:
	case TOKEN_LITERAL_COLLECTION:
	    postfix_expression();
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving unary_expression()" << endl;
    }

    void
    Parser::multiply_operator() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering multiply_operator()" << endl;

	switch(lookahead(1)) {
	case TOKEN_MULTIPLY:
	    match(TOKEN_MULTIPLY);
	    break;
	case TOKEN_DIVIDE:
	    match(TOKEN_DIVIDE);
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving multiply_operator()" << endl;
    }

    void
    Parser::unary_operator() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering unary_operator()" << endl;

	switch(lookahead(1)) {
	case TOKEN_MINUS:
	    match(TOKEN_MINUS);
	    break;
	case TOKEN_LITERAL_NOT:
	    match(TOKEN_LITERAL_NOT);
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving unary_operator()" << endl;
    }

    void
    Parser::postfix_expression() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering postfix_expression()" << endl;

	primary_expression();

	while(lookahead(1) == TOKEN_DOT || lookahead(1) == TOKEN_RIGHT_ARROW) {
	    switch(lookahead(1)) {
	    case TOKEN_DOT:
		match(TOKEN_DOT);
		break;
	    case TOKEN_RIGHT_ARROW:
		match(TOKEN_RIGHT_ARROW);
		break;
	    default:
		throw SyntacticException("Syntactic Exception");
	    }

	    property_call();
	}
	cerr << indent_string(mDepth * 4) << "Leaving postfix_expression()" << endl;
    }

    void
    Parser::primary_expression() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering primary_expression()" << endl;

	switch(lookahead(1)) {
	case TOKEN_LITERAL_SET:
	case TOKEN_LITERAL_BAG:
	case TOKEN_LITERAL_SEQUENCE:
	case TOKEN_LITERAL_COLLECTION:
	    literal_collection();
	    break;
	case TOKEN_LEFT_ROUND_BRACKET:
	    match(TOKEN_LEFT_ROUND_BRACKET);
	    expression();
	    match(TOKEN_RIGHT_ROUND_BRACKET);
	    break;
	case TOKEN_LITERAL_IF:
	    if_expression();
	    break;
	default:
	    bool predicate_matched = false;
	    if((lookahead(1) == TOKEN_NAME ||
		lookahead(1) == TOKEN_STRING ||
		lookahead(1) == TOKEN_NUMBER) &&
	       (lookahead(2) == TOKEN_EOF ||
		lookahead(2) == TOKEN_LITERAL_DEF ||
		lookahead(2) == TOKEN_DOUBLE_COLON ||
		lookahead(2) == TOKEN_RIGHT_ROUND_BRACKET ||
		lookahead(2) == TOKEN_LITERAL_PRE ||
		lookahead(2) == TOKEN_LITERAL_POST ||
		lookahead(2) == TOKEN_LITERAL_INV ||
		lookahead(2) == TOKEN_EQUAL ||
		lookahead(2) == TOKEN_PLUS ||
		lookahead(2) == TOKEN_MINUS ||
		lookahead(2) == TOKEN_LESS_THAN ||
		lookahead(2) == TOKEN_LESS_EQUAL ||
		lookahead(2) == TOKEN_GREATER_THAN ||
		lookahead(2) == TOKEN_GREATER_EQUAL ||
		lookahead(2) == TOKEN_DIVIDE ||
		lookahead(2) == TOKEN_MULTIPLY ||
		lookahead(2) == TOKEN_NOT_EQUAL ||
		lookahead(2) == TOKEN_LITERAL_IMPLIES ||
		lookahead(2) == TOKEN_LITERAL_OR ||
		lookahead(2) == TOKEN_LITERAL_XOR ||
		lookahead(2) == TOKEN_LITERAL_AND ||
		lookahead(2) == TOKEN_COMMA ||
		lookahead(2) == TOKEN_LITERAL_IN ||
		lookahead(2) == TOKEN_LITERAL_LET ||
		lookahead(2) == TOKEN_LITERAL_THEN ||
		lookahead(2) == TOKEN_LITERAL_ELSE ||
		lookahead(2) == TOKEN_LITERAL_ENDIF ||
		lookahead(2) == TOKEN_DOT ||
		lookahead(2) == TOKEN_RIGHT_ARROW ||
		lookahead(2) == TOKEN_BAR ||
		lookahead(2) == TOKEN_RIGHT_CURLY_BRACKET ||
		lookahead(2) == TOKEN_DOT_DOT ||
		lookahead(2) == TOKEN_RIGHT_SQUARE_BRACKET)) {
		push();
		predicate_matched = true;

		try {
		    Incrementer i(mGuessing);
		    switch (lookahead(1)) {
		    case TOKEN_STRING:
			match(TOKEN_STRING);
			break;
		    case TOKEN_NUMBER:
			match(TOKEN_NUMBER);
			break;
		    case TOKEN_NAME:
			match(TOKEN_NAME);
			match(TOKEN_DOUBLE_COLON);
			match(TOKEN_NAME);

			for(;;) {
			    if((lookahead(1) == TOKEN_DOUBLE_COLON)) {
				match(TOKEN_DOUBLE_COLON);
				match(TOKEN_NAME);
			    }
			    else {
				break;
			    }
			}
			break;
		    default:
			throw SyntacticException("Syntactic Exception");
		    }
		}
		catch(ExpectedToken &) {
		    predicate_matched = false;
		}
		catch(SyntacticException & ex) {
		    predicate_matched = false;
		}
		catch(...) {
		    cerr <<" BLAH" << endl;
		    pop();
		    throw;
		}

		restore();
	    }

	    if(predicate_matched) {
		literal();
	    }
	    else if(lookahead(1) == TOKEN_NAME &&
		    (lookahead(2) == TOKEN_EOF ||
		     lookahead(2) == TOKEN_LITERAL_DEF ||
		     lookahead(2) == TOKEN_DOUBLE_COLON ||
		     lookahead(2) == TOKEN_LEFT_ROUND_BRACKET ||
		     lookahead(2) == TOKEN_RIGHT_ROUND_BRACKET ||
		     lookahead(2) == TOKEN_LITERAL_PRE ||
		     lookahead(2) == TOKEN_LITERAL_POST ||
		     lookahead(2) == TOKEN_LITERAL_INV ||
		     lookahead(2) == TOKEN_EQUAL ||
		     lookahead(2) == TOKEN_PLUS ||
		     lookahead(2) == TOKEN_MINUS ||
		     lookahead(2) == TOKEN_LESS_THAN ||
		     lookahead(2) == TOKEN_LESS_EQUAL ||
		     lookahead(2) == TOKEN_GREATER_THAN ||
		     lookahead(2) == TOKEN_GREATER_EQUAL ||
		     lookahead(2) == TOKEN_DIVIDE ||
		     lookahead(2) == TOKEN_MULTIPLY ||
		     lookahead(2) == TOKEN_NOT_EQUAL ||
		     lookahead(2) == TOKEN_LITERAL_IMPLIES ||
		     lookahead(2) == TOKEN_LITERAL_AND ||
		     lookahead(2) == TOKEN_LITERAL_OR ||
		     lookahead(2) == TOKEN_LITERAL_XOR ||
		     lookahead(2) == TOKEN_COMMA ||
		     lookahead(2) == TOKEN_LITERAL_IN ||
		     lookahead(2) == TOKEN_LITERAL_LET ||
		     lookahead(2) == TOKEN_LITERAL_THEN ||
		     lookahead(2) == TOKEN_LITERAL_ELSE ||
		     lookahead(2) == TOKEN_LITERAL_ENDIF ||
		     lookahead(2) == TOKEN_DOT ||
		     lookahead(2) == TOKEN_RIGHT_ARROW ||
		     lookahead(2) == TOKEN_BAR ||
		     lookahead(2) == TOKEN_RIGHT_CURLY_BRACKET ||
		     lookahead(2) == TOKEN_DOT_DOT ||
		     lookahead(2) == TOKEN_LEFT_SQUARE_BRACKET ||
		     lookahead(2) == TOKEN_RIGHT_SQUARE_BRACKET ||
		     lookahead(2) == TOKEN_AT)) {
		property_call();
	    }
	    else {
		throw SyntacticException("Syntactic Exception");
	    }
	}
	cerr << indent_string(mDepth * 4) << "Leaving primary_expression()" << endl;
    }

    void
    Parser::property_call() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering property_call()" << endl;

	path_name();

	switch(lookahead(1)) {
	case TOKEN_AT:
	    time_expression();
	    break;
	case TOKEN_EOF:
	case TOKEN_LITERAL_DEF:
	case TOKEN_LEFT_ROUND_BRACKET:
	case TOKEN_RIGHT_ROUND_BRACKET:
	case TOKEN_LITERAL_PRE:
	case TOKEN_LITERAL_POST:
	case TOKEN_LITERAL_INV:
	case TOKEN_EQUAL:
	case TOKEN_PLUS:
	case TOKEN_MINUS:
	case TOKEN_LESS_THAN:
	case TOKEN_LESS_EQUAL:
	case TOKEN_GREATER_EQUAL:
	case TOKEN_GREATER_THAN:
	case TOKEN_DIVIDE:
	case TOKEN_MULTIPLY:
	case TOKEN_NOT_EQUAL:
	case TOKEN_LITERAL_IMPLIES:
	case TOKEN_LITERAL_OR:
	case TOKEN_LITERAL_XOR:
	case TOKEN_LITERAL_AND:
	case TOKEN_COMMA:
	case TOKEN_LITERAL_IN:
	case TOKEN_LITERAL_LET:
	case TOKEN_LITERAL_THEN:
	case TOKEN_LITERAL_ELSE:
	case TOKEN_LITERAL_ENDIF:
	case TOKEN_DOT:
	case TOKEN_RIGHT_ARROW:
	case TOKEN_BAR:
	case TOKEN_RIGHT_CURLY_BRACKET:
	case TOKEN_DOT_DOT:
	case TOKEN_LEFT_SQUARE_BRACKET:
	case TOKEN_RIGHT_SQUARE_BRACKET:
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}

	switch(lookahead(1)) {
	case TOKEN_LEFT_SQUARE_BRACKET:
	    qualifiers();
	    break;
	case TOKEN_EOF:
	case TOKEN_LITERAL_DEF:
	case TOKEN_LEFT_ROUND_BRACKET:
	case TOKEN_RIGHT_ROUND_BRACKET:
	case TOKEN_LITERAL_PRE:
	case TOKEN_LITERAL_POST:
	case TOKEN_LITERAL_INV:
	case TOKEN_EQUAL:
	case TOKEN_PLUS:
	case TOKEN_MINUS:
	case TOKEN_LESS_THAN:
	case TOKEN_LESS_EQUAL:
	case TOKEN_GREATER_EQUAL:
	case TOKEN_GREATER_THAN:
	case TOKEN_DIVIDE:
	case TOKEN_MULTIPLY:
	case TOKEN_NOT_EQUAL:
	case TOKEN_LITERAL_IMPLIES:
	case TOKEN_LITERAL_OR:
	case TOKEN_LITERAL_XOR:
	case TOKEN_LITERAL_AND:
	case TOKEN_COMMA:
	case TOKEN_LITERAL_IN:
	case TOKEN_LITERAL_LET:
	case TOKEN_LITERAL_THEN:
	case TOKEN_LITERAL_ELSE:
	case TOKEN_LITERAL_ENDIF:
	case TOKEN_DOT:
	case TOKEN_RIGHT_ARROW:
	case TOKEN_BAR:
	case TOKEN_RIGHT_CURLY_BRACKET:
	case TOKEN_DOT_DOT:
	case TOKEN_RIGHT_SQUARE_BRACKET:
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}

	switch (lookahead(1)) {
	case TOKEN_LEFT_ROUND_BRACKET:
	    property_call_parameters();
	    break;
	case TOKEN_EOF:
	case TOKEN_LITERAL_DEF:
	case TOKEN_RIGHT_ROUND_BRACKET:
	case TOKEN_LITERAL_PRE:
	case TOKEN_LITERAL_POST:
	case TOKEN_LITERAL_INV:
	case TOKEN_EQUAL:
	case TOKEN_PLUS:
	case TOKEN_MINUS:
	case TOKEN_LESS_THAN:
	case TOKEN_LESS_EQUAL:
	case TOKEN_GREATER_EQUAL:
	case TOKEN_GREATER_THAN:
	case TOKEN_DIVIDE:
	case TOKEN_MULTIPLY:
	case TOKEN_NOT_EQUAL:
	case TOKEN_LITERAL_IMPLIES:
	case TOKEN_LITERAL_OR:
	case TOKEN_LITERAL_XOR:
	case TOKEN_LITERAL_AND:
	case TOKEN_COMMA:
	case TOKEN_LITERAL_IN:
	case TOKEN_LITERAL_LET:
	case TOKEN_LITERAL_THEN:
	case TOKEN_LITERAL_ELSE:
	case TOKEN_LITERAL_ENDIF:
	case TOKEN_DOT:
	case TOKEN_RIGHT_ARROW:
	case TOKEN_BAR:
	case TOKEN_RIGHT_CURLY_BRACKET:
	case TOKEN_DOT_DOT:
	case TOKEN_RIGHT_SQUARE_BRACKET:
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving property_call()" << endl;
    }

    void
    Parser::literal_collection() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering literal_collection()" << endl;

	collection_kind();

	match(TOKEN_LEFT_CURLY_BRACKET);

	switch(lookahead(1)) {
	case TOKEN_NAME:
	case TOKEN_LEFT_ROUND_BRACKET:
	case TOKEN_MINUS:
	case TOKEN_LITERAL_NOT:
	case TOKEN_LITERAL_IF:
	case TOKEN_STRING:
	case TOKEN_NUMBER:
	case TOKEN_LITERAL_SET:
	case TOKEN_LITERAL_BAG:
	case TOKEN_LITERAL_SEQUENCE:
	case TOKEN_LITERAL_COLLECTION:
	    collection_item();
	    while(lookahead(1) == TOKEN_COMMA) {
		match(TOKEN_COMMA);
		collection_item();
	    }
	    break;
	case TOKEN_RIGHT_CURLY_BRACKET:
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving literal_collection()" << endl;
    }

    void
    Parser::literal() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering literal()" << endl;

	switch(lookahead(1)) {
	case TOKEN_STRING:
	    match(TOKEN_STRING);
	    break;
	case TOKEN_NUMBER:
	    match(TOKEN_NUMBER);
	    break;
	case TOKEN_NAME:
	    enum_literal();
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving literal()" << endl;
    }

    void
    Parser::property_call_parameters() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering property_call_parameters()" << endl;

	if(lookahead(1) == TOKEN_LEFT_ROUND_BRACKET && lookahead(2) == TOKEN_NAME) {
	    push();
	    bool predicate_matched = true;

	    try {
		Incrementer i(mGuessing);
		match(TOKEN_LEFT_ROUND_BRACKET);
		match(TOKEN_NAME);

		while(lookahead(1) == TOKEN_COMMA) {
		    match(TOKEN_COMMA);
		    match(TOKEN_NAME);
		}

		switch(lookahead(1)) {
		case TOKEN_COLON:
		    match(TOKEN_COLON);
		    simple_type_specifier();
		    break;
		case TOKEN_SEMICOLON:
		case TOKEN_BAR:
		    break;
		default:
		    throw SyntacticException("Syntactic Exception");
		}

		switch(lookahead(1)) {
		case TOKEN_SEMICOLON:
		    match(TOKEN_SEMICOLON);
		    match(TOKEN_NAME);
		    match(TOKEN_COLON);
		    type_specifier();
		    match(TOKEN_EQUAL);
		    expression();
		    break;
		case TOKEN_BAR:
		    break;
		default:
		    throw SyntacticException("Syntactic Exception");
		}

		match(TOKEN_BAR);
	    }
	    catch(ExpectedToken &) {
		predicate_matched = false;
	    }
	    catch(SyntacticException & ex) {
		predicate_matched = false;
	    }
	    catch(...) {
		pop();
		throw;
	    }

	    restore();

	    if(predicate_matched) {
		match(TOKEN_LEFT_ROUND_BRACKET);
		declarator();

		switch(lookahead(1)) {
		case TOKEN_NAME:
		case TOKEN_LEFT_ROUND_BRACKET:
		case TOKEN_MINUS:
		case TOKEN_LITERAL_NOT:
		case TOKEN_LITERAL_IF:
		case TOKEN_STRING:
		case TOKEN_NUMBER:
		case TOKEN_LITERAL_SET:
		case TOKEN_LITERAL_BAG:
		case TOKEN_LITERAL_SEQUENCE:
		case TOKEN_LITERAL_COLLECTION:
		    actual_parameter_list();
		    break;
		case TOKEN_RIGHT_ROUND_BRACKET:
		    break;
		default:
		    throw SyntacticException("Syntactic Exception");
		}

		match(TOKEN_RIGHT_ROUND_BRACKET);
	    }
	    else if(lookahead(1) == TOKEN_LEFT_ROUND_BRACKET &&
		    (lookahead(2) == TOKEN_NAME ||
		     lookahead(2) == TOKEN_LEFT_ROUND_BRACKET ||
		     lookahead(2) == TOKEN_RIGHT_ROUND_BRACKET ||
		     lookahead(2) == TOKEN_MINUS ||
		     lookahead(2) == TOKEN_LITERAL_NOT ||
		     lookahead(2) == TOKEN_LITERAL_IF ||
		     lookahead(2) == TOKEN_STRING ||
		     lookahead(2) == TOKEN_NUMBER ||
		     lookahead(2) == TOKEN_LITERAL_SET ||
		     lookahead(2) == TOKEN_LITERAL_BAG ||
		     lookahead(2) == TOKEN_LITERAL_SEQUENCE ||
		     lookahead(2) == TOKEN_LITERAL_COLLECTION)) {
		match(TOKEN_LEFT_ROUND_BRACKET);

		switch(lookahead(1)) {
		case TOKEN_NAME:
		case TOKEN_LEFT_ROUND_BRACKET:
		case TOKEN_MINUS:
		case TOKEN_LITERAL_NOT:
		case TOKEN_LITERAL_IF:
		case TOKEN_STRING:
		case TOKEN_NUMBER:
		case TOKEN_LITERAL_SET:
		case TOKEN_LITERAL_BAG:
		case TOKEN_LITERAL_SEQUENCE:
		case TOKEN_LITERAL_COLLECTION:
		    actual_parameter_list();
		    break;
		case TOKEN_RIGHT_ROUND_BRACKET:
		    break;
		default:
		    throw SyntacticException("Syntactic Exception");
		}

		match(TOKEN_RIGHT_ROUND_BRACKET);
	    }
	    else {
		throw SyntacticException("Syntactic Exception");
	    }
	}
	cerr << indent_string(mDepth * 4) << "Leaving property_call_parameters()" << endl;
    }

    void
    Parser::declarator() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering declarator()" << endl;

	match(TOKEN_NAME);

	while(lookahead(1) == TOKEN_COMMA) {
	    match(TOKEN_COMMA);
	    match(TOKEN_NAME);
	}

	switch(lookahead(1)) {
	case TOKEN_COLON:
	    match(TOKEN_COLON);
	    simple_type_specifier();
	    break;
	case TOKEN_SEMICOLON:
	case TOKEN_BAR:
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}

	switch(lookahead(1)) {
	case TOKEN_SEMICOLON:
	    match(TOKEN_SEMICOLON);
	    match(TOKEN_NAME);
	    match(TOKEN_COLON);
	    type_specifier();
	    match(TOKEN_EQUAL);
	    expression();
	    break;
	case TOKEN_BAR:
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	match(TOKEN_BAR);
	cerr << indent_string(mDepth * 4) << "Leaving declarator()" << endl;
    }

    void
    Parser::actual_parameter_list() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering actual_parameter_list()" << endl;

	expression();

	while(lookahead(1) == TOKEN_COMMA) {
	    match(TOKEN_COMMA);
	    expression();
	}
	cerr << indent_string(mDepth * 4) << "Leaving actual_parameter_list()" << endl;
    }

    void
    Parser::enum_literal() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering enum_literal()" << endl;

	match(TOKEN_NAME);
	match(TOKEN_DOUBLE_COLON);
	match(TOKEN_NAME);

	while(lookahead(1) == TOKEN_DOUBLE_COLON) {
	    match(TOKEN_DOUBLE_COLON);
	    match(TOKEN_NAME);
	}
	cerr << indent_string(mDepth * 4) << "Leaving enum_literal()" << endl;
    }

    void
    Parser::path_name() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering path_name()" << endl;
	match(TOKEN_NAME);

	while(lookahead(1) == TOKEN_DOUBLE_COLON) {
	    match(TOKEN_DOUBLE_COLON);
	    match(TOKEN_NAME);
	}
	cerr << indent_string(mDepth * 4) << "Leaving path_name()" << endl;
    }

    void
    Parser::collection_item() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering collection_item()" << endl;

	expression();

	switch(lookahead(1)) {
	case TOKEN_DOT_DOT:
	    match(TOKEN_DOT_DOT);
	    expression();
	    break;
	case TOKEN_COMMA:
	case TOKEN_RIGHT_CURLY_BRACKET:
	    break;
	default:
	    throw SyntacticException("Syntactic Exception");
	}
	cerr << indent_string(mDepth * 4) << "Leaving collection_item()" << endl;
    }

    void
    Parser::time_expression() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering time_expression()" << endl;

	match(TOKEN_AT);
	match(TOKEN_LITERAL_PRE);

	cerr << indent_string(mDepth * 4) << "Leaving time_expression()" << endl;
    }

    void
    Parser::qualifiers() {
	Incrementer i(mDepth);
	cerr << indent_string(mDepth * 4) << "Entering qualifiers()" << endl;

	match(TOKEN_LEFT_SQUARE_BRACKET);
	actual_parameter_list();
	match(TOKEN_RIGHT_SQUARE_BRACKET);

	cerr << indent_string(mDepth * 4) << "Leaving qualifiers()" << endl;
    }

} // namespace OCL
