// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "foundation/object.h"
#include "foundation/macros.h"
#include "foundation/exception.h"

#include "drawing/point.h"
#include "drawing/image.h"
#include "drawing/factory.h"

// Python interface

#include "python/drawing/image.h"

#include <map>
using namespace std;

//================================================================================

namespace {
  using namespace Python;

  //typedef map<Object *, PyObject *> ReferenceMap;
  //ReferenceMap reference_table;


  //================================================================================

  PyObject *
  wrapped_image_get_w(PyObject * self, PyObject * args) {
    //virtual int get_w()
    assert(PyObject_TypeCheck(self, & drawingImageType));
    ImageReference * rp = (ImageReference *) self;

    // Extract the arguments
    if (!PyArg_ParseTuple(args, ":Drawing.get_w"))
      return NULL;
    try {

      // Finally call the method
      return Py_BuildValue("i", rp->mImage->get_w());
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }

  //--------------------------------------------------------------------------------

  PyObject *
  wrapped_image_get_h(PyObject * self, PyObject * args) {
    //virtual int get_h()
    assert(PyObject_TypeCheck(self, & drawingImageType));
    ImageReference * rp = (ImageReference *) self;

    // Extract the arguments
    if (!PyArg_ParseTuple(args, ":Drawing.get_h"))
      return NULL;
    try {

      // Finally call the method
      return Py_BuildValue("i", rp->mImage->get_h());
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }


  //--------------------------------------------------------------------------------

  struct PyMethodDef drawing_image_methods[ ] = {
    { "get_w",			wrapped_image_get_w,	    METH_VARARGS },
    { "get_h",			wrapped_image_get_h,        METH_VARARGS },
    
    { NULL, NULL }
  };


  PyObject *
  drawingImage_getattr(PyObject * self, char * name) {

    // Look for a method
    return Py_FindMethod(drawing_image_methods, self, name);
  }


  int
  drawingImage_setattr(PyObject * self, char * name, PyObject * value) {
    PyErr_SetString(PyExc_NameError, "No attributes to be set");
    return -1;
  }


  void
  drawingImage_dealloc(PyObject * self) {
    if (PyObject_TypeCheck(self, & drawingImageType)) {
      ImageReference * rp = (ImageReference *) self;
      if (rp->mImage) {
	//reference_table.erase(rp->mImage);
	rp->mImage->unref();
      }
    }
    PyMem_DEL(self);
  }

}

//================================================================================

namespace Python {

  
  PyTypeObject drawingImageType = {
    PyObject_HEAD_INIT(&PyType_Type)		// Initialisation
    0,						// ob_size	: Usually 0
    "drawing.Image",				// tp_name	: Type Name
    sizeof(ImageReference),			// tp_basicsize : Object size
    0,						// tp_itemsize  : Usually 0
  
    // Standard methods
    (destructor) drawingImage_dealloc,		// tp_dealloc	: refcount = 0
    (printfunc) 0,				// tp_print	: print x
    (getattrfunc) drawingImage_getattr,		// tp_getattr	: x.attr
    (setattrfunc) drawingImage_setattr,		// tp_setattr	: a.attr = v
    (cmpfunc) 0,				// tp_compare	: x > y
    (reprfunc) 0,				// tp_repr	: repr(x)

    // Type categories
    0,						// tp_as_number
    0,						// tp_as_sequence
    0,						// tp_as_mapping

    (hashfunc) 0,				// tp_hash
    (ternaryfunc) 0,				// tp_call
    (reprfunc) 0,				// tp_str
    (getattrofunc) 0,				// tp_getatto
    (setattrofunc) 0,				// tp_setatto

    0,						// tp_as_buffer
    0,						// tp_flags
    0,						// tp_doc
  };


  //--------------------------------------------------------------------------------


  PyObject * 
  image_reference_new(Drawing::Image * image) {

    PyObject * result = 0;
    //ReferenceMap::iterator iter = reference_table.find(image);
    //if (iter == reference_table.end()) {

    // Create a new reference
    ImageReference * self = PyObject_NEW(ImageReference, & drawingImageType);
    if (! self)
      return NULL;
    
    image->ref();		 // take joint ownership of the object
    self->mImage = image;
    result = (PyObject *) self;
      //reference_table[image] = result;
      //}
      //else {
      //TRACE(2, "Reusing an existing reference");
      // Use an existing reference
      //result = iter->second;
      //Py_INCREF(result);
      //}
    return result;
  }

  PyObject *
  image_reference_new(PyObject * self, PyObject * args) {
    //TRACE(1, "Create a new Drawing.Image");
    const char * file_name = 0;
    if (!PyArg_ParseTuple(args, "s", & file_name))
      return NULL;

    try {
      //MESSAGE("\tFile name '" << file_name << "'");
      Drawing::Factory * factory = Drawing::Factory::instance();
      Drawing::Image * image = factory->create_image(file_name);
      PyObject * result = image_reference_new(image);
      image->unref(); // release our copy
      return result;
    }
    catch (...) {
      PyErr_SetString(PyExc_AttributeError, "Unknown C++ exception");
      return NULL;
    }
  }


}
