// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H


//================================================================================


#include <unistd.h>
#include <Python.h>

#include "foundation/macros.h"

#include "python.h"

// Python interface to Foundation package
#include "python/foundation/module.h"

// Python interface to Reflective package
#include "python/reflective/module.h"

// Python interface to Drawing package
#include "python/toolkit/module.h"

// Python interface to Drawing package
#include "python/drawing/module.h"

// Python interface to Diagram package
#include "python/diagram/module.h"


namespace Python {

  //--------------------------------------------------------------------------------

  // Static data members
  Interpreter * Interpreter::mInstance = 0;


  Interpreter::Interpreter() {
    TRACE(1, "Initialise Python...");
    Py_Initialize();

    // Insert path for the python imports
    Python::Module module_sys("sys");
    PyObject * module_sys_path = module_sys.get_item("path");
    // FIXUP some error checking would be nice

    // Add the standard places (like the thing was installed or something)
    PyList_Insert(module_sys_path, 0, PyString_FromString(PYTHON_SEARCH_PATH));

    // If we are in development directory, add the path to the development source
    // I know this is just a hack, but it is sooo convenient....
    // It saves having to install the python source to try it out,
    // you can run it in place in the source tree.
    char buffer[4096];
    char * wd = getcwd(buffer, 4096);
    if (wd) {

      // Make sure we have the current directory
      PyList_Insert(module_sys_path, 0, PyString_FromString(buffer));
      if ((wd = strstr(buffer, "dachshund2/src/"))) {

	strcpy(wd, "dachshund2/src/python");
	PyList_Insert(module_sys_path, 0, PyString_FromString(buffer));
	strcpy(wd, "dachshund2/src/app/python");
	PyList_Insert(module_sys_path, 0, PyString_FromString(buffer));
	WARNING("Executing python stuff from development directory tree");
      }
    }

    // Initialise the interface packages
    foundation_initialise();
    drawing_initialise();
    reflective_initialise();
    toolkit_initialise();
    diagram_initialise();
  }


  Interpreter *
  Interpreter::instance() {
    if (! mInstance) 
      mInstance = new Interpreter();
    return mInstance;
  }


  //--------------------------------------------------------------------------------

  Module::Module(const char * name) : mName(name) {

    // Import the module
    TRACE(1, "Importing module " << name);
    mObject = PyImport_ImportModule(const_cast<char *>(name));
    if (! mObject) {
      if (PyErr_Occurred())
	PyErr_Print();
      FATAL("Error importing Python module " << name);
      // maybe an exception would be nicer
    }

    // Import the dictionary
    mDictionary = PyModule_GetDict(mObject);
    assert(mDictionary);
  }


  Module::~Module() {
    // FIXUP should release the references
  }


  // ----------------------------------------

  PyObject *
  Module::get_item(const char * name) const {
    TRACE(2, "Looking up item " << name << " in module " << mName);
    PyObject * item = PyDict_GetItemString(mDictionary, const_cast<char *>(name));
    if (! item) {
      if (PyErr_Occurred())
	PyErr_Print();
      FATAL("Error looking up item " << name);
      // maybe an exception would be nicer
    }
    return item;
  }


  // ----------------------------------------


  Callable::Callable(const Module & object, const char * name) : mName(name) {

    mFunction = object.get_item(name);
    if (! PyCallable_Check(mFunction)) {
      if (PyErr_Occurred())
	PyErr_Print();
      FATAL("Function item " << name << " not callable");
      // maybe an exception would be nicer
    }
  }

  Callable::~Callable() {
    //
  }

  // ----------------------------------------

  PyObject *
  Callable::call(PyObject * args) {

    PyObject * result = PyObject_CallObject(mFunction, args);
    if (! result) {
      if (PyErr_Occurred())
	PyErr_Print();
      FATAL("Function item " << mName << " not callable");
      // maybe an exception would be nicer
    }
    return result;
  }


}


