// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

// C++ Implementation

#include "foundation/object.h"
#include "foundation/macros.h"
#include "foundation/change_listener.h"
#include "foundation/undo_manager.h"

#include "uml/foundation/extension_mechanisms/tagged_value.h"

#include "reflective/class.h"
#include "reflective/attribute.h"
#include "reflective/iterator.h"
#include "reflective/context.h"

// Python interfaces

#include "python/foundation/object.h"
#include "python/reflective/class.h"
#include "python/reflective/context.h"

#include <map>
using namespace std;

//================================================================================


namespace {
  using namespace Python;

  // The following functions are defined by a table for the ReferenceElement python type

  // Exception objects used by this module
  //static PyObject * AttributeError;
  //static PyObject * TypeError;


  typedef map<Object *, PyObject *> ReferenceMap;
  ReferenceMap reference_table;


  //================================================================================


  PyObject *
  wrapped_context_create_class(PyObject * self, PyObject * args) {

    const char * type_name = 0;
    if (!PyArg_ParseTuple(args, "s:Context.create_class", & type_name))
      return NULL;

    assert(PyObject_TypeCheck(self, & reflectiveContextType));
    ContextReference * mrp = (ContextReference *) self;
    try {
      // Get the role, returns NULL if not set
      REF::Class * klass = mrp->mContext->ref_create_class(type_name);
      PyObject * result = class_reference_new(klass);
      klass->unref(); // release our copy
      return result;
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      ERROR("Exception during create_class for " << type_name);
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }


  PyObject *
  wrapped_context_set_root(PyObject * self, PyObject * args) {
    assert(PyObject_TypeCheck(self, & reflectiveContextType));
    ContextReference * mrp = (ContextReference *) self;

    ClassReference * ref_class = 0;
    if (!PyArg_ParseTuple(args, ":Context.set_root", & reflectiveClassType, & ref_class))
      return NULL;

    try {
      // Set the root
      mrp->mContext->ref_set_root(ref_class->get());
      return Py_BuildValue("");
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }


  PyObject *
  wrapped_context_get_root(PyObject * self, PyObject * args) {
    assert(PyObject_TypeCheck(self, & reflectiveContextType));
    ContextReference * mrp = (ContextReference *) self;

    if (!PyArg_ParseTuple(args, ":Context.get_root"))
      return NULL;

    try {
      // Get the role, returns NULL if not set
      REF::Class * klass = mrp->mContext->ref_get_root();
      return class_reference_new(klass);
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }


  struct PyMethodDef context_methods[ ] = {
    { "create_class", wrapped_context_create_class,	METH_VARARGS },
    { "set_root",     wrapped_context_set_root,		METH_VARARGS },
    { "get_root",     wrapped_context_get_root,		METH_VARARGS },
    { NULL, NULL }
  };


  //--------------------------------------------------------------------------------


  PyObject *
  context_getattr(PyObject * self, char * name) {
    // Look for a method instead
    return Py_FindMethod(context_methods, self, name);
  }


  int
  context_setattr(PyObject * self, char * name, PyObject * value) {
    //MESSAGE("Set attribute " << name);
    // Unknown attribute name
    PyErr_SetString(PyExc_NameError, "no settable attributes");
    return -1;
  }


  void
  context_dealloc(PyObject * self) {
    // Release our ownership of the object
    if (PyObject_TypeCheck(self, & reflectiveContextType)) {
      ContextReference * rp = (ContextReference *) self;
      if (rp->mContext) {
	reference_table.erase(rp->mContext);
	rp->mContext->unref();
      }
    }
    PyMem_DEL(self);
  }

}


//================================================================================


namespace Python {


  PyTypeObject reflectiveContextType = {
    PyObject_HEAD_INIT(&PyType_Type)		// Initialisation
    0,						// ob_size	: Usually 0
    "Reflective.Context",			// tp_name	: Type Name
    sizeof(ContextReference),			// tp_basicsize : Object size
    0,						// tp_itemsize  : Usually 0
  
    // Standard methods
    (destructor) context_dealloc,		// tp_dealloc	: refcount = 0
    (printfunc) 0,				// tp_print	: print x
    (getattrfunc) context_getattr,		// tp_getattr	: x.attr
    (setattrfunc) context_setattr,		// tp_setattr	: a.attr = v
    (cmpfunc) 0,				// tp_compare	: x > y
    (reprfunc) 0,				// tp_repr	: repr(x)

    // Type categories
    0,						// tp_as_number
    0,						// tp_as_sequence
    0,						// tp_as_mapping

    (hashfunc) 0,				// tp_hash
    (ternaryfunc) 0,				// tp_call
    (reprfunc) 0,				// tp_str
    (getattrofunc) 0,				// tp_getatto
    (setattrofunc) 0,				// tp_setatto

    0,						// tp_as_buffer
    0,						// tp_flags
    0,						// tp_doc  
  };

  //--------------------------------------------------------------------------------

  PyObject *
  context_reference_new(REF::Context * context) {


    // Check for NULL
    if (context == 0)
      return Py_BuildValue("");


    PyObject * result = 0;
    ReferenceMap::iterator iter = reference_table.find(context);
    if (iter == reference_table.end()) {

      // Create a new reference
      ContextReference * self = PyObject_NEW(ContextReference, & reflectiveContextType);
      if (! self)
	return NULL;
    
      context->ref();		 // take joint ownership of the object
      self->mContext = context;
      result = (PyObject *) self;
      reference_table[context] = result;
    }
    else {
      TRACE(2, "Reusing an existing reference");
      // Use an existing reference
      result = iter->second;
      Py_INCREF(result);
    }
    return result;
  }


  PyObject *
  context_reference_new(PyObject * self, PyObject * args) {
    MESSAGE("Create a new ContextReference.Context");

    NYI();
    return NULL;
  }


}
