// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include <map>
#include <list>

#include "foundation/object.h"
#include "foundation/macros.h"
#include "foundation/exception.h"
#include "foundation/undo_manager.h"
#include "foundation/change_listener.h"

#include "toolkit/widget.h"
#include "toolkit/document.h"
#include "toolkit/document_maintainer.h"
#include "toolkit/tag.h"
#include "toolkit/mark.h"
#include "toolkit/mouse_listener.h"
#include "toolkit/mouse_event.h"

#include "drawing/point.h"
#include "drawing/image.h"

// Python interfaces

#include "python/common/convert.h"
#include "python/foundation/object.h"
#include "python/drawing/image.h"
#include "python/toolkit/widget.h"
#include "python/toolkit/document.h"
#include "python/toolkit/tag.h"
#include "python/toolkit/mark.h"


//================================================================================

namespace {
  using namespace Python;


  typedef map<Object *, PyObject *> ReferenceMap;
  ReferenceMap reference_table;



/*
  PyObject *
  wrapped_document_begin_transaction(PyObject * self, PyObject * args) {
    //virtual void begin_transaction()
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * wrp = (DocumentReference *) self;
    try {
      // Extract the arguments
      if (!PyArg_ParseTuple(args, ":Document.begin_transaction"))
	return NULL;

      // Finally call the method
      wrp->mDocument->begin_transaction();
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }

  PyObject *
  wrapped_document_end_transaction(PyObject * self, PyObject * args) {
    //virtual void end_transaction()
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * wrp = (DocumentReference *) self;
    try {
      // Extract the arguments
      if (!PyArg_ParseTuple(args, ":Document.end_transaction"))
	return NULL;

      // Finally call the method
      wrp->mDocument->end_transaction();
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }
*/

  //--------------------------------------------------------------------------------

  PyObject *
  wrapped_document_get_insert_mark(PyObject * self, PyObject * args) {
    //virtual void	get_insert_mark();
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      if (!PyArg_ParseTuple(args, ":Document.get_insert_mark"))
	return NULL;
      
      // Finally call the method
      Toolkit::Mark * mark = rp->mDocument->get_insert_mark();
      PyObject * ref_mark = mark_reference_new(mark);
      mark->unref();
      return Py_BuildValue("O", ref_mark);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }


  PyObject *
  wrapped_document_get_change_mark(PyObject * self, PyObject * args) {
    //virtual void	get_change_mark();
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      if (!PyArg_ParseTuple(args, ":Document.get_change_mark"))
	return NULL;
      
      // Finally call the method
      Toolkit::Mark * mark = rp->mDocument->get_change_mark();
      PyObject * ref_mark = mark_reference_new(mark);
      mark->unref();
      return Py_BuildValue("O", ref_mark);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }


  PyObject *
  wrapped_document_get_selection_mark(PyObject * self, PyObject * args) {
    //virtual void	get_selection_mark();
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      if (!PyArg_ParseTuple(args, ":Document.get_selection_mark"))
	return NULL;
      
      // Finally call the method
      Toolkit::Mark * mark = rp->mDocument->get_selection_mark();
      PyObject * ref_mark = mark_reference_new(mark);
      mark->unref();
      return Py_BuildValue("O", ref_mark);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }




  //--------------------------------------------------------------------------------


  PyObject *
  wrapped_document_insert_block_begin(PyObject * self, PyObject * args) {
    //virtual void	insert_block_begin(Mark * location, Tag * block) = 0;
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_location = 0;
      PyObject * ref_block_mark = 0;
      PyObject * ref_block = 0;
      if (!PyArg_ParseTuple(args, "OOO:Document.insert_block_begin", 
			    & ref_location, & ref_block_mark, & ref_block))
	return NULL;
      
      Toolkit::Mark * location = 0;
      Toolkit::Mark * block_mark = 0;
      Toolkit::Tag * block = 0;
      if (!convert_to_mark(location, ref_location) ||
	  !convert_to_mark(block_mark, ref_block_mark) ||
	  !convert_to_tag(block, ref_block))
	return NULL;

      // Finally call the method
      rp->mDocument->insert_block_begin(location, block_mark, block);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }

  PyObject *
  wrapped_document_insert_block_end(PyObject * self, PyObject * args) {
    //virtual void	insert_block_end(Mark * location, Tag * block) = 0;
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_location = 0;
      PyObject * ref_block = 0;
      if (!PyArg_ParseTuple(args, "OO:Document.insert_block_end", & ref_location, & ref_block))
	return NULL;

      Toolkit::Mark * location = 0;
      Toolkit::Tag * block = 0;
      if (!convert_to_mark(location, ref_location) ||
	  !convert_to_tag(block, ref_block))
	return NULL;

      // Finally call the method
      rp->mDocument->insert_block_end(location, block);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }

  //--------------------------------------------------------------------------------

  PyObject *
  wrapped_document_insert_field_begin(PyObject * self, PyObject * args) {
    //virtual void	insert_field_begin(Mark * location, Tag * field) = 0;
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_location = 0;
      PyObject * ref_field = 0;
      if (!PyArg_ParseTuple(args, "OO:Document.insert_field_begin", & ref_location, & ref_field))
	return NULL;

      Toolkit::Mark * location = 0;
      Toolkit::Tag * field = 0;
      if (!convert_to_mark(location, ref_location) ||
	  !convert_to_tag(field, ref_field))
	return NULL;

      // Finally call the method
      rp->mDocument->insert_field_begin(location, field);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }

  PyObject *
  wrapped_document_insert_field_end(PyObject * self, PyObject * args) {
    //virtual void	insert_field_end(Mark * location, Tag * field) = 0;
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_location = 0;
      PyObject * ref_field = 0;
      if (!PyArg_ParseTuple(args, "OO:Document.insert_field_end", & ref_location, & ref_field))
	return NULL;

      Toolkit::Mark * location = 0;
      Toolkit::Tag * field = 0;
      if (!convert_to_mark(location, ref_location) ||
	  !convert_to_tag(field, ref_field))
	return NULL;

      // Finally call the method
      rp->mDocument->insert_field_end(location, field);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }

  PyObject *
  wrapped_document_insert_field_text(PyObject * self, PyObject * args) {
    //virtual void	insert_field_text(Mark * location, Tag * field, const char * text) = 0;
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_location = 0;
      PyObject * ref_field = 0;
      const char * text = 0;
      if (!PyArg_ParseTuple(args, "OOs:Document.insert_field_text", 
			    & ref_location, & ref_field, & text))
	return NULL;

      Toolkit::Mark * location = 0;
      Toolkit::Tag * field = 0;
      if (!convert_to_mark(location, ref_location) ||
	  !convert_to_tag(field, ref_field))
	return NULL;

      // Finally call the method
      rp->mDocument->insert_field_text(location, field, text);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }
  
  PyObject *
  wrapped_document_insert_content(PyObject * self, PyObject * args) {
    // combination of text, image, and widget inserts
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_location = 0;
      PyObject * ref_content = 0;
      if (!PyArg_ParseTuple(args, "OO:Document.insert_field_content", 
			    & ref_location, & ref_content))
	return NULL;

      Toolkit::Mark * location = 0;
      if (!convert_to_mark(location, ref_location))
	return NULL;

      // Check for string
      if (PyString_Check(ref_content)) {

	// Content is a string
	//rp->mDocument->insert_field_begin(location, field);
	rp->mDocument->insert_text(location, PyString_AsString(ref_content));
	//rp->mDocument->insert_field_end(location, field);
	return Py_BuildValue("");
      }
      if (ref_content->ob_type == & drawingImageType) {
	Drawing::Image * image = 0;
	if (!convert_to_image(image, ref_content))
	  return NULL;
	
	// Content is an image
	//rp->mDocument->insert_field_begin(location, field);
	rp->mDocument->insert_image(location, image);
	//rp->mDocument->insert_field_end(location, field);
	return Py_BuildValue("");
      }
      if (ref_content->ob_type == & toolkitWidgetType) {
	Toolkit::Widget * widget = 0;
	if (!convert_to_widget(widget, ref_content))
	  return NULL;
      
	//rp->mDocument->insert_field_begin(location, field);
	rp->mDocument->insert_widget(location, widget);
	//rp->mDocument->insert_field_end(location, field);
	return Py_BuildValue("");
      }

      PyErr_SetString(PyExc_TypeError, "No suitable type available");
      return NULL;
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
  }
  
  //--------------------------------------------------------------------------------

  PyObject *
  wrapped_document_clear(PyObject * self, PyObject * args) {
    //virtual void clear() = 0;
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      if (!PyArg_ParseTuple(args, ":Document.clear"))
	return NULL;

      // Finally call the method
      rp->mDocument->clear();
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }

  //--------------------------------------------------------------------------------

  PyObject *
  wrapped_document_insert_text(PyObject * self, PyObject * args) {
    //virtual void	insert_text(Mark * location, const char * text) = 0;
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_location = 0;
      const char * text = 0;
      if (!PyArg_ParseTuple(args, "Os:Document.insert_text", & ref_location, & text))
	return NULL;

      // Convert the arguments
      Toolkit::Mark * location = 0;
      if (!convert_to_mark(location, ref_location))
	return NULL;

      // Finally call the method
      rp->mDocument->insert_text(location, text);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }
  

  PyObject *
  wrapped_document_insert_widget(PyObject * self, PyObject * args) {
    //virtual void	insert_widget(Mark * location, const char * text) = 0;
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_location = 0;
      PyObject * ref_widget = 0;
      if (!PyArg_ParseTuple(args, "OO:Document.insert_text", & ref_location, & ref_widget))
	return NULL;

      // Convert the type
      Toolkit::Mark * location = 0;
      Toolkit::Widget * widget = 0;
      if (!convert_to_mark(location, ref_location) ||
	  !convert_to_widget(widget, ref_widget))
	return NULL;

      // Finally call the method
      //TRACE(1, "Insert widget " << (void *) widget);
      rp->mDocument->insert_widget(location, widget);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }

  //--------------------------------------------------------------------------------

  PyObject *
  wrapped_document_get_field_text(PyObject * self, PyObject * args) {
    //virtual void get_field_text(Mark * location, Tag * field, string & result) = 0;
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_location = 0;
      PyObject * ref_field = 0;
      if (!PyArg_ParseTuple(args, "OO:Document.get_field_text", & ref_location, & ref_field))
	return NULL;

      // Convert the type
      Toolkit::Mark * location = 0;
      Toolkit::Tag * field = 0;
      if (!convert_to_mark(location, ref_location) ||
	  !convert_to_tag(field, ref_field))
	return NULL;

      // Finally call the method
      string result;
      rp->mDocument->get_field_text(location, field, result);
      return Py_BuildValue("s", result.c_str());
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }
    
  PyObject *
  wrapped_document_replace_field_text(PyObject * self, PyObject * args) {
    //virtual void replace_field_text(Mark * location, Tag * field, const char * new_text) = 0;
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_location = 0;
      PyObject * ref_field = 0;
      const char * text = 0;
      if (!PyArg_ParseTuple(args, "OOs:Document.replace_field_text", 
			    & ref_location, & ref_field, & text))
	return NULL;

      Toolkit::Mark * location = 0;
      Toolkit::Tag * field = 0;
      if (!convert_to_mark(location, ref_location) ||
	  !convert_to_tag(field, ref_field))
	return NULL;

      // Finally call the method
      rp->mDocument->replace_field_text(location, field, text);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }
  
  //virtual void replace_field_image(Mark * location, Tag * field, Drawing::Image * image) = 0;
  //virtual void replace_field_widget(Mark * location, Tag * field, Toolkit::Widget * widget) = 0;
  //virtual void clear() = 0;

  PyObject *
  wrapped_document_delete_field(PyObject * self, PyObject * args) {
    //virtual void delete_field(Mark * location, Tag * field) = 0;
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_location = 0;
      PyObject * ref_field = 0;
      if (!PyArg_ParseTuple(args, "OO:Document.delete_field",  & ref_location, & ref_field))
	return NULL;

      Toolkit::Mark * location = 0;
      Toolkit::Tag * field = 0;
      if (!convert_to_mark(location, ref_location) ||
	  !convert_to_tag(field, ref_field))
	return NULL;

      // Finally call the method
      rp->mDocument->delete_field(location, field);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }
  
  PyObject *
  wrapped_document_delete_block(PyObject * self, PyObject * args) {
    //virtual void delete_block(Mark * location, Tag * block) = 0;
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_location = 0;
      PyObject * ref_block = 0;
      if (!PyArg_ParseTuple(args, "OO:Document.delete_block", & ref_location, & ref_block))
	return NULL;

      Toolkit::Mark * location = 0;
      Toolkit::Tag * block = 0;
      if (!convert_to_mark(location, ref_location) ||
	  !convert_to_tag(block, ref_block))
	return NULL;

      // Finally call the method
      rp->mDocument->delete_block(location, block);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }
  

  //--------------------------------------------------------------------------------
  //--------------------------------------------------------------------------------

  class ProxyDocumentMaintainer: public Toolkit::DocumentMaintainer {
    PyObject * mObject;
  public:
    // This method is called if a change occurs within any Tag sensitive to changes
    void changed(Object * object,
		 Toolkit::Mark * block_mark, Toolkit::Tag * block, 
		 Toolkit::Mark * field_mark_begin, Toolkit::Mark * field_mark_end, 
		 Toolkit::Tag * field) {

      PyObject * ref_object = object_reference_new(object);
      PyObject * ref_block_mark =       mark_reference_new(block_mark);
      PyObject * ref_field_mark_begin = mark_reference_new(field_mark_begin);
      PyObject * ref_field_mark_end =   mark_reference_new(field_mark_end);
      PyObject * ref_block = tag_reference_new(block);
      PyObject * ref_field = tag_reference_new(field);

      PyObject * ref_result = PyObject_CallMethod(mObject, "changed", "(OOOOOO)", 
						  ref_object, ref_block_mark, ref_block,
						  ref_field_mark_begin, ref_field_mark_end,
						  ref_field);
      Py_DECREF(ref_object);
      Py_DECREF(ref_block_mark);
      Py_DECREF(ref_block);
      Py_DECREF(ref_field_mark_begin);
      Py_DECREF(ref_field_mark_end);
      Py_DECREF(ref_field);

      if (ref_result == NULL) {
	ERROR("Error returned by 'changed'");
	if (PyErr_Occurred())
	  PyErr_Print();
      }
      else {
	Py_DECREF(ref_result);
      }
    }


    // This method is called to sync changes made in document with model
    // It is called just once to apply the changes when cursor moves out of changed area
    void apply_change(Object * object,
		      Toolkit::Mark * block_mark, Toolkit::Tag * block, 
		      Toolkit::Mark * field_mark_begin, Toolkit::Mark * field_mark_end,
		      Toolkit::Tag * field) {

      PyObject * ref_object = object_reference_new(object);
      PyObject * ref_block_mark =       mark_reference_new(block_mark);
      PyObject * ref_field_mark_begin = mark_reference_new(field_mark_begin);
      PyObject * ref_field_mark_end =   mark_reference_new(field_mark_end);
      PyObject * ref_block = tag_reference_new(block);
      PyObject * ref_field = tag_reference_new(field);

      PyObject * ref_result = PyObject_CallMethod(mObject, "apply_change", "(OOOOOO)", 
						  ref_object, ref_block_mark, ref_block,
						  ref_field_mark_begin, ref_field_mark_end,
						  ref_field);
      Py_DECREF(ref_object);
      Py_DECREF(ref_block_mark);
      Py_DECREF(ref_block);
      Py_DECREF(ref_field_mark_begin);
      Py_DECREF(ref_field_mark_end);
      Py_DECREF(ref_field);

      if (ref_result == NULL) {
	ERROR("Error returned by 'apply_change'");
	if (PyErr_Occurred())
	  PyErr_Print();
      }
      else {
	Py_DECREF(ref_result);
      }
    }


    ProxyDocumentMaintainer(PyObject * object) : mObject(object) { 
      Py_INCREF(mObject);
    }
    virtual ~ProxyDocumentMaintainer() {
      Py_DECREF(mObject);
    }
  };


  PyObject *
  wrapped_document_set_maintainer(PyObject * self, PyObject * args) {
    //void set_maintainer(DocumentMaintainer * maintainer);
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * object = 0;
      if (!PyArg_ParseTuple(args, "O:Document.set_maintainer", & object))
	return NULL;
      // Error checking?

      // Remove any existing proxy
      delete(rp->mDocument->get_maintainer());

      // Add the new maintainer proxy
      if (object == Py_None)
	rp->mDocument->set_maintainer(0);
      else {
	ProxyDocumentMaintainer * proxy = new ProxyDocumentMaintainer(object);
	rp->mDocument->set_maintainer(proxy);
      }
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }


  
  //--------------------------------------------------------------------------------

  class ProxyMouseListener : public Toolkit::MouseListener {
    PyObject * mObject;

    PyObject * convert(const Toolkit::MouseEvent * me) {

      // Convert the event type to a string
      const char * event_type = "unknown";
      switch (me->get_type()) {
      case Toolkit::MET_ENTERED:  
	event_type = "entered";  
	break;
      case Toolkit::MET_EXITED:	 
	event_type = "exited";	  
	break;
      case Toolkit::MET_MOVED:	 
	event_type = "moved";	  
	break;
      case Toolkit::MET_DRAGGED:	 
	event_type = "dragged";  
	break;
      case Toolkit::MET_PRESSED:	 
	event_type = "pressed";  
	break;
      case Toolkit::MET_RELEASED: 
	event_type = "released"; 
	break;
      case Toolkit::MET_LAST: break;
      }

      // Build a tupple with the event details with string and numbers:
      // ("event_type", (x_pos, y_pos), button, click_count)
      return Py_BuildValue("(s(ii)ii)", event_type, 
			   me->get_point().get_x(), me->get_point().get_y(),
			   me->get_button(), me->get_click_count());
    }
	
  public:

    virtual bool mouse_entered(const Toolkit::MouseEvent * mouse_event) {
      PyObject * ref_event = convert(mouse_event);
      PyObject * ref_result = PyObject_CallMethod(mObject, "mouse_entered", "(O)", ref_event);
      bool result = false;
      Py_DECREF(ref_event);
      if (ref_result == NULL) {
	ERROR("Error returned by 'entered'");
	if (PyErr_Occurred())
	  PyErr_Print();
      }
      else {
	if (PyObject_IsTrue(ref_result))
	  result = true;
	Py_DECREF(ref_result);
      }
      return result;
    }

    virtual bool mouse_exited(const Toolkit::MouseEvent * mouse_event) {
      PyObject * ref_event = convert(mouse_event);
      PyObject * ref_result = PyObject_CallMethod(mObject, "mouse_exited", "(O)", ref_event);
      bool result = false;
      Py_DECREF(ref_event);
      if (ref_result == NULL) {
	ERROR("Error returned by 'exited'");
	if (PyErr_Occurred())
	  PyErr_Print();
      }
      else {
	if (PyObject_IsTrue(ref_result))
	  result = true;
	Py_DECREF(ref_result);
      }
      return result;
    }

    virtual bool mouse_pressed(const Toolkit::MouseEvent * mouse_event) {
      PyObject * ref_event = convert(mouse_event);
      PyObject * ref_result = PyObject_CallMethod(mObject, "mouse_pressed", "(O)", ref_event);
      bool result = false;
      Py_DECREF(ref_event);
      if (ref_result == NULL) {
	ERROR("Error returned by 'pressed'");
	if (PyErr_Occurred())
	  PyErr_Print();
      }
      else {
	if (PyObject_IsTrue(ref_result))
	  result = true;
	Py_DECREF(ref_result);
      }
      return result;
    }

    virtual bool mouse_released(const Toolkit::MouseEvent * mouse_event) {
      PyObject * ref_event = convert(mouse_event);
      PyObject * ref_result = PyObject_CallMethod(mObject, "released", "(O)", ref_event);
      bool result = false;
      Py_DECREF(ref_event);
      if (ref_result == NULL) {
	ERROR("Error returned by 'released'");
	if (PyErr_Occurred())
	  PyErr_Print();
      }
      else {
	if (PyObject_IsTrue(ref_result))
	  result = true;
	Py_DECREF(ref_result);
      }
      return result;
    }
    virtual bool mouse_moved(const Toolkit::MouseEvent * mouse_event) {
      PyObject * ref_event = convert(mouse_event);
      PyObject * ref_result = PyObject_CallMethod(mObject, "mouse_moved", "(O)", ref_event);
      bool result = false;
      Py_DECREF(ref_event);
      if (ref_result == NULL) {
	ERROR("Error returned by 'moved'");
	if (PyErr_Occurred())
	  PyErr_Print();
      }
      else {
	if (PyObject_IsTrue(ref_result))
	  result = true;
	Py_DECREF(ref_result);
      }
      return result;

    }
    virtual bool mouse_dragged(const Toolkit::MouseEvent * mouse_event) {
      PyObject * ref_event = convert(mouse_event);
      PyObject * ref_result = PyObject_CallMethod(mObject, "mouse_dragged", "(O)", ref_event);
      bool result = false;
      Py_DECREF(ref_event);
      if (ref_result == NULL) {
	ERROR("Error returned by 'dragged'");
	if (PyErr_Occurred())
	  PyErr_Print();
      }
      else {
	if (PyObject_IsTrue(ref_result))
	  result = true;
	Py_DECREF(ref_result);
      }
      return result;
    }

    PyObject * get() { return mObject; }
    ProxyMouseListener(PyObject * object) : mObject(object) { 
      Py_INCREF(mObject);

      // Check each of the functions to see if it is callable?      
    }
    virtual ~ProxyMouseListener() { 
      Py_DECREF(mObject); 
    }
  };
  
  //--------------------------------------------------------------------------------
  
  // List of maintainers registered
  list<ProxyMouseListener *> listener_list;

  
  PyObject *
  wrapped_document_listener_add(PyObject * self, PyObject * args) {
    //void add(MouseListener * mouse_listener);
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * object = 0;
      if (!PyArg_ParseTuple(args, "O:Document.listener_add", & object))
	return NULL;
      // type check?

      ProxyMouseListener * proxy = new ProxyMouseListener(object);
      listener_list.push_back(proxy);
      //rp->mDocument->add(proxy);
      NYI();
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }
  
  PyObject *
  wrapped_document_listener_remove(PyObject * self, PyObject * args) {
    //void remove(const MouseListener * mouse_listener);
    assert(PyObject_TypeCheck(self, & toolkitDocumentType));
    DocumentReference * rp = (DocumentReference *) self;
    try {
      // Extract the arguments
      PyObject * object = 0;
      if (!PyArg_ParseTuple(args, "O:Document.listener_remove", & object))
	return NULL;

      // Lookup the proxy to remove
      for (list<ProxyMouseListener *>::iterator iter = listener_list.begin();
	   iter != listener_list.end(); iter++) {
	if (object == (* iter)->get()) {

	  // Unregister the listener, delete the object, remove from list
	  NYI();
	  //rp->mDocument->remove(* iter);
	  delete(* iter);
	  listener_list.erase(iter);
	  break;
	}
      }
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }

  //--------------------------------------------------------------------------------

  struct PyMethodDef toolkit_document_methods[ ] = {
/*
    { "begin_transaction",	wrapped_document_begin_transaction, METH_VARARGS },
    { "end_transaction",	wrapped_document_end_transaction, METH_VARARGS },
*/
    { "get_insert_mark",	wrapped_document_get_insert_mark,	METH_VARARGS },
    { "get_change_mark",	wrapped_document_get_change_mark,	METH_VARARGS },
    { "get_selection_mark",	wrapped_document_get_selection_mark,	METH_VARARGS },

    { "insert_block_begin",	wrapped_document_insert_block_begin,	METH_VARARGS },
    { "insert_block_end",	wrapped_document_insert_block_end,	METH_VARARGS },

    { "insert_field_begin",	wrapped_document_insert_field_begin,	METH_VARARGS },
    { "insert_field_end",	wrapped_document_insert_field_end,	METH_VARARGS },
    { "insert_field_text",	wrapped_document_insert_field_text,	METH_VARARGS },
    { "insert_content",		wrapped_document_insert_content,	METH_VARARGS },

    { "clear",			wrapped_document_clear,			METH_VARARGS },

    { "insert_text",		wrapped_document_insert_text,		METH_VARARGS },
  //{ "insert_image",		wrapped_document_insert_image,		METH_VARARGS },
    { "insert_widget",		wrapped_document_insert_widget,		METH_VARARGS },

    { "get_field_text",		wrapped_document_get_field_text,	METH_VARARGS },
    { "replace_field_text",	wrapped_document_replace_field_text,	METH_VARARGS },
  //{ "replace_field_image",	wrapped_document_replace_field_image,	METH_VARARGS },
  //{ "replace_field_widget",	wrapped_document_replace_field_widget,	METH_VARARGS },
    { "delete_field",		wrapped_document_delete_field,		METH_VARARGS },
    { "delete_block",		wrapped_document_delete_block,		METH_VARARGS },


    { "set_maintainer",         wrapped_document_set_maintainer,  METH_VARARGS },

    { "listener_add",		wrapped_document_listener_add,    METH_VARARGS },
    { "listener_remove",	wrapped_document_listener_remove, METH_VARARGS },
    
    { NULL, NULL }
  };



  PyObject *
  document_getattr(PyObject * self, char * name) {

    // Look for a method
    return Py_FindMethod(toolkit_document_methods, self, name);
  }


  int
  document_setattr(PyObject * self, char * name, PyObject * value) {
    PyErr_SetString(PyExc_NameError, "No attributes to be set");
    return -1;
  }


  void
  document_dealloc(PyObject * self) {
    if (PyObject_TypeCheck(self, & toolkitDocumentType)) {
      DocumentReference * rp = (DocumentReference *) self;
      if (rp->mDocument) {
	reference_table.erase(rp->mDocument);
	rp->mDocument->unref();
      }
    }
    PyMem_DEL(self);
  }

}

//================================================================================

namespace Python {


  PyTypeObject toolkitDocumentType = {
    PyObject_HEAD_INIT(&PyType_Type)		// Initialisation
    0,						// ob_size	: Usually 0
    "toolkit.Document",				// tp_name	: Type Name
    sizeof(DocumentReference),			// tp_basicsize : Object size
    0,						// tp_itemsize  : Usually 0
    
    // Standard methods
    (destructor) document_dealloc,		// tp_dealloc	: refcount = 0
    (printfunc) 0,				// tp_print	: print x
    (getattrfunc) document_getattr,		// tp_getattr	: x.attr
    (setattrfunc) document_setattr,		// tp_setattr	: a.attr = v
    (cmpfunc) 0,				// tp_compare	: x > y
    (reprfunc) 0,				// tp_repr	: repr(x)

    // Type categories
    0,						// tp_as_number
    0,						// tp_as_sequence
    0,						// tp_as_mapping

    (hashfunc) 0,				// tp_hash
    (ternaryfunc) 0,				// tp_call
    (reprfunc) 0,				// tp_str
    (getattrofunc) 0,				// tp_getatto
    (setattrofunc) 0,				// tp_setatto
    
    0,						// tp_as_buffer
    0,						// tp_flags
    0,						// tp_doc
    
    (traverseproc) 0,				// tp_traverse
    (inquiry) 0,				// tp_clear
    (richcmpfunc) 0,				// tp_richcompare
    0,						// tp_weaklistoffset
    (getiterfunc) 0,				// tp_iter
    (iternextfunc) 0,				// tp_iternext
    0, 						// tp_methods
    0, 						// tp_members
    0,						// tp_getset
    0,						// tp_base
  };


  //--------------------------------------------------------------------------------


  PyObject *
  document_reference_new(Toolkit::Document * document) {

    PyObject * result = 0;
    ReferenceMap::iterator iter = reference_table.find(document);
    if (iter == reference_table.end()) {

      // Create a new reference
      DocumentReference * self = PyObject_NEW(DocumentReference, & toolkitDocumentType);
      if (! self)
	return NULL;
    
      document->ref();		 // take joint ownership of the object
      self->mDocument = document;
      result = (PyObject *) self;
      reference_table[document] = result;
    }
    else {
      //TRACE(2, "Reusing an existing reference");
      // Use an existing reference
      result = iter->second;
      Py_INCREF(result);
    }
    return result;
  }

  PyObject *
  document_reference_new(PyObject * self, PyObject * args) {
    //TRACE(1, "Create a new Toolkit.Document");
    PyErr_SetString(PyExc_TypeError, "don't know how to construct");
    return NULL;
  }

}

