// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "foundation/object.h"
#include "foundation/macros.h"
#include "foundation/exception.h"

#include "toolkit/factory.h"
#include "toolkit/mark.h"

// Python interfaces

#include "python/common/convert.h"
#include "python/foundation/object.h"
#include "python/toolkit/mark.h"

#include <map>
using namespace std;

//================================================================================

namespace {
  using namespace Python;

  typedef map<Object *, PyObject *> ReferenceMap;
  ReferenceMap reference_table;


  PyObject *
  wrapped_mark_get_name(PyObject * self, PyObject * args) {
    // const string & get_name() const;
    assert(PyObject_TypeCheck(self, & toolkitMarkType));
    MarkReference * rp = (MarkReference *) self;
    try {
      // Extract the arguments
      if (!PyArg_ParseTuple(args, ":Mark.get_name"))
	return NULL;

      // Finally call the method
      const string & name = rp->mMark->get_name();
      return Py_BuildValue("s", name.c_str());
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }


  //--------------------------------------------------------------------------------



  PyObject *
  wrapped_mark_relocate_to_mark(PyObject * self, PyObject * args) {
    // virtual void	relocate_to_mark(Mark * mark) = 0;
    assert(PyObject_TypeCheck(self, & toolkitMarkType));
    MarkReference * rp = (MarkReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_mark = 0;
      if (!PyArg_ParseTuple(args, "O:Mark.relocate_to_mark", & ref_mark))
	return NULL;

      // Convert the types
      Toolkit::Mark * mark = 0;
      if (! convert_to_mark(mark, ref_mark))
	return NULL;

      // Finally call the method
      rp->mMark->relocate_to_mark(mark);
      return Py_BuildValue("");
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }


  PyObject *
  wrapped_mark_relocate_to_tag_end(PyObject * self, PyObject * args) {
    // virtual void	relocate_to_tag_end(Toolkit::Tag * tag)
    assert(PyObject_TypeCheck(self, & toolkitMarkType));
    MarkReference * rp = (MarkReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_tag = 0;
      if (!PyArg_ParseTuple(args, "O:Mark.relocate_to_tag_end", & ref_tag))
	return NULL;

      // Convert the types
      Toolkit::Tag * tag = 0;
      if (! convert_to_tag(tag, ref_tag))
	return NULL;

      // Finally call the method
      rp->mMark->relocate_to_tag_end(tag);
      return Py_BuildValue("");
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }

  PyObject *
  wrapped_mark_at_tag_start(PyObject * self, PyObject * args) {
    // virtual void	at_tag_start(Toolkit::Tag * tag)
    assert(PyObject_TypeCheck(self, & toolkitMarkType));
    MarkReference * rp = (MarkReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_tag = 0;
      if (!PyArg_ParseTuple(args, "O:Mark.at_tag_start", & ref_tag))
	return NULL;

      // Convert the types
      Toolkit::Tag * tag = 0;
      if (! convert_to_tag(tag, ref_tag))
	return NULL;

      // Finally call the method
      int result = rp->mMark->at_tag_start(tag);
      return Py_BuildValue("i", result);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }


  PyObject *
  wrapped_mark_at_tag_end(PyObject * self, PyObject * args) {
    // virtual void	at_tag_end(Toolkit::Tag * tag)
    assert(PyObject_TypeCheck(self, & toolkitMarkType));
    MarkReference * rp = (MarkReference *) self;
    try {
      // Extract the arguments
      PyObject * ref_tag = 0;
      if (!PyArg_ParseTuple(args, "O:Mark.at_tag_end", & ref_tag))
	return NULL;

      // Convert the types
      Toolkit::Tag * tag = 0;
      if (! convert_to_tag(tag, ref_tag))
	return NULL;

      // Finally call the method
      int result = rp->mMark->at_tag_end(tag);
      return Py_BuildValue("i", result);
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }



  //--------------------------------------------------------------------------------


  struct PyMethodDef toolkit_mark_methods[ ] = {

    { "get_name",		wrapped_mark_get_name,  METH_VARARGS },

    { "relocate_to_mark",	wrapped_mark_relocate_to_mark, METH_VARARGS },
    { "relocate_to_tag_end",	wrapped_mark_relocate_to_tag_end, METH_VARARGS },
    { "at_tag_start",		wrapped_mark_at_tag_start, METH_VARARGS },
    { "at_tag_end",		wrapped_mark_at_tag_end, METH_VARARGS },

    { NULL, NULL }
  };



  PyObject *
  mark_getattr(PyObject * self, char * name) {

    // Look for a method
    return Py_FindMethod(toolkit_mark_methods, self, name);
  }


  int
  mark_setattr(PyObject * self, char * name, PyObject * value) {
    PyErr_SetString(PyExc_NameError, "No attributes to be set");
    return -1;
  }


  void
  mark_dealloc(PyObject * self) {
    if (PyObject_TypeCheck(self, & toolkitMarkType)) {
      MarkReference * rp = (MarkReference *) self;
      if (rp->mMark) {
	reference_table.erase(rp->mMark);
	rp->mMark->unref();
      }
    }
    PyMem_DEL(self);
  }

}

//================================================================================

namespace Python {

  PyTypeObject toolkitMarkType = {
    PyObject_HEAD_INIT(&PyType_Type)		// Initialisation
    0,						// ob_size	: Usually 0
    "toolkit.Mark",				// tp_name	: Type Name
    sizeof(MarkReference),			// tp_basicsize : Object size
    0,						// tp_itemsize  : Usually 0
    
    // Standard methods
    (destructor) mark_dealloc,			// tp_dealloc	: refcount = 0
    (printfunc) 0,				// tp_print	: print x
    (getattrfunc) mark_getattr,			// tp_getattr	: x.attr
    (setattrfunc) mark_setattr,			// tp_setattr	: a.attr = v
    (cmpfunc) 0,				// tp_compare	: x > y
    (reprfunc) 0,				// tp_repr	: repr(x)

    // Type categories
    0,						// tp_as_number
    0,						// tp_as_sequence
    0,						// tp_as_mapping

    (hashfunc) 0,				// tp_hash
    (ternaryfunc) 0,				// tp_call
    (reprfunc) 0,				// tp_str
    (getattrofunc) 0,				// tp_getatto
    (setattrofunc) 0,				// tp_setatto
    
    0,						// tp_as_buffer
    0,						// tp_flags
    0,						// tp_doc
    
    (traverseproc) 0,				// tp_traverse
    (inquiry) 0,				// tp_clear
    (richcmpfunc) 0,				// tp_richcompare
    0,						// tp_weaklistoffset
    (getiterfunc) 0,				// tp_iter
    (iternextfunc) 0,				// tp_iternext
    0, 						// tp_methods
    0, 						// tp_members
    0,						// tp_getset
    0,						// tp_base

  };


  //--------------------------------------------------------------------------------


  PyObject *
  mark_reference_new(Toolkit::Mark * mark) {

    PyObject * result = 0;
    ReferenceMap::iterator iter = reference_table.find(mark);
    if (iter == reference_table.end()) {

      // Create a new reference
      MarkReference * self = PyObject_NEW(MarkReference, & toolkitMarkType);
      if (! self)
	return NULL;
    
      mark->ref();		 // take joint ownership of the object
      self->mMark = mark;
      result = (PyObject *) self;
      reference_table[mark] = result;
    }
    else {
      //TRACE(2, "Reusing an existing reference");
      // Use an existing reference
      result = iter->second;
      Py_INCREF(result);
    }
    return result;
  }


  PyObject *
  mark_reference_new_mark(PyObject * self, PyObject * args) {
    //TRACE(1, "Create a new Toolkit.Mark");
    try {
      // Extract the arguments
      const char * name = 0;
      PyObject * ref_location = 0;
      if (!PyArg_ParseTuple(args, "sO:Mark", & name, & ref_location))
	return NULL;

      Toolkit::Mark * location = 0;
      if (!convert_to_mark(location, ref_location))
	return NULL;

      // Finally call the method
      Toolkit::Factory * factory = Toolkit::Factory::instance();
      Toolkit::Mark * mark = factory->create_mark(name, location);
      PyObject * result = mark_reference_new(mark);
      mark->unref(); // release our copy
      return result;
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }

  PyObject *
  mark_reference_new_block_mark(PyObject * self, PyObject * args) {
    //TRACE(1, "Create a new Toolkit.BlockMark");
    try {
      // Extract the arguments
      const char * name = 0;
      PyObject * ref_location = 0;
      PyObject * ref_block = 0;
      int level = 0;
      PyObject * ref_object = 0;
      if (!PyArg_ParseTuple(args, "sO|OiO:BlockMark", 
			    & name, & ref_location, & ref_block, & level, & ref_object))
	return NULL;

      Toolkit::Mark * location = 0;      
      Toolkit::Tag * block = 0;      
      Object * object = 0;      
      if (!convert_to_mark(location, ref_location) ||
	  !convert_to_tag(block, ref_block) ||
	  !convert_to_object(object, ref_object))
	return NULL;

      // Finally call the method
      Toolkit::Factory * factory = Toolkit::Factory::instance();
      Toolkit::Mark * mark = factory->create_block_mark(name, location, block, level, object);
      PyObject * result = mark_reference_new(mark);
      mark->unref(); // release our copy
      return result;
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }

}

