// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "foundation/object.h"
#include "foundation/macros.h"
#include "foundation/exception.h"
#include "foundation/command.h"
#include "foundation/command_callback.h"

#include "toolkit/widget.h"
#include "toolkit/menu.h"
#include "toolkit/factory.h"


// Python interfaces

#include "python/common/convert.h"
#include "python/foundation/object.h"
#include "python/toolkit/menu.h"

#include <map>
#include <functional>
#include <algorithm>
using namespace std;

//================================================================================

namespace {
    using namespace Python;

    typedef map<Object *, PyObject *> ReferenceMap;
    ReferenceMap reference_table;


    PyObject *
    wrapped_menu_show(PyObject * self, PyObject * args) {
        // Widget methods
        //virtual void show()
        assert(PyObject_TypeCheck(self, & toolkitMenuType));
        MenuReference * rp = (MenuReference *) self;
        try {
            // Extract the arguments
            if (!PyArg_ParseTuple(args, ":Menu.show"))
                return NULL;

            // Finally call the method
            rp->mMenu->show();
        }
        catch (Exception & e) {
            string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
            PyErr_SetString(PyExc_RuntimeError, error.c_str());
            return NULL;
        }
        catch (...) {
            PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
            return NULL;      
        }
        // Return None
        return Py_BuildValue("");
    }


    PyObject *
    wrapped_menu_hide(PyObject * self, PyObject * args) {
        // Widget methods
        //virtual void hide()
        assert(PyObject_TypeCheck(self, & toolkitMenuType));
        MenuReference * rp = (MenuReference *) self;
        try {
            // Extract the arguments
            if (!PyArg_ParseTuple(args, ":Menu.hide"))
                return NULL;

            // Finally call the method
            rp->mMenu->hide();
        }
        catch (Exception & e) {
            string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
            PyErr_SetString(PyExc_RuntimeError, error.c_str());
            return NULL;
        }
        catch (...) {
            PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
            return NULL;      
        }
        // Return None
        return Py_BuildValue("");
    }


    //--------------------------------------------------------------------------------

    PyObject *
    wrapped_menu_popup(PyObject * self, PyObject * args) {
        //virtual void popup(const MouseEvent * mouse_event) = 0;
        assert(PyObject_TypeCheck(self, & toolkitMenuType));
        MenuReference * rp = (MenuReference *) self;
        try {
            // Extract the arguments
            PyObject * ref_event;
            if (!PyArg_ParseTuple(args, "O:Menu.popup", &ref_event))
                return NULL;

            // Convert the tuple into a mouse event again
            const Toolkit::MouseEvent * event = 0;
            if (! convert_to_mouse_event(event, ref_event))
                return NULL;

            // Finally call the method
            rp->mMenu->popup(event);
        }
        catch (Exception & e) {
            string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
            PyErr_SetString(PyExc_RuntimeError, error.c_str());
            return NULL;
        }
        catch (...) {
            PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
            return NULL;      
        }
        // Return None
        return Py_BuildValue("");
    }

    /*
       FIXUP The interface to the menu needs a little more though
       I may have to make the command a real python object
       and/or I may have to have a list of commands to cross-reference

       virtual void add_radio(Command * command, const string & label, Command * group,
       bool sensitive = true) = 0;
       virtual void add_submenu(Menu * submenu, const string & label) = 0;

       virtual void set_sensitive(Command * command, bool sensitive) = 0;
       virtual void set_state(Command * command, bool state) = 0;	// for radio/check buttons
     */
    //--------------------------------------------------------------------------------


    class ProxyCommandCallback : public Object {
        PyObject * mCallable;

    public:
        ProxyCommandCallback(PyObject * callable) : mCallable(callable) {
            Py_INCREF(mCallable);
        }

        ~ProxyCommandCallback() {
            Py_DECREF(mCallable);
        }

        void callback(Object * node) {
            //PyObject * ref_node = object_reference_new(node);
            PyObject * ref_node = object_reference_new(node);
            PyObject * ref_args = Py_BuildValue("(O)", ref_node);
            PyObject * ref_result = PyObject_CallObject(mCallable, ref_args);
            Py_DECREF(ref_node);
            Py_DECREF(ref_args);
            if (ref_result == NULL) {
                ERROR("Error returned by callback function");
                if (PyErr_Occurred())
                    PyErr_Print();
            }
            else {
                Py_DECREF(ref_result);
            }      
        }
    };

    //--------------------------------------------------------------------------------

    PyObject *
    wrapped_menu_add_push(PyObject * self, PyObject * args) {
        //virtual void add_push(Command * command, const string & label, bool sensitive = true) = 0;
        assert(PyObject_TypeCheck(self, & toolkitMenuType));
        MenuReference * rp = (MenuReference *) self;
        try {
            // Extract the arguments
            PyObject * ref_command = 0;
            PyObject * ref_arg = 0;
            const char * label = 0;
            int sensitive = 1;
            if (!PyArg_ParseTuple(args, "OOs|i:Menu.add", & ref_command, & ref_arg, 
                                    & label, & sensitive))
                return NULL;

            // Create a proxy to accept the callback
            ProxyCommandCallback * proxy = new ProxyCommandCallback(ref_command);

            // Convert the argument to a REF::Class
            Object * arg = 0;
            if (!convert_to_object(arg, ref_arg))
                return NULL;

            // Finally add the callback
            rp->mMenu->add_push(command_callback(proxy, arg, &ProxyCommandCallback::callback),
                                label, sensitive);
        }
        catch (Exception & e) {
            string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
            PyErr_SetString(PyExc_RuntimeError, error.c_str());
            return NULL;
        }
        catch (...) {
            PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
            return NULL;      
        }
        // Return None
        return Py_BuildValue("");
    }

    PyObject *
    wrapped_menu_add_check(PyObject * self, PyObject * args) {
        //virtual void add_check(Command * command, const string & label, bool sensitive = true) = 0;
        assert(PyObject_TypeCheck(self, & toolkitMenuType));
        MenuReference * rp = (MenuReference *) self;
        try {
            // Extract the arguments
            PyObject * ref_command = 0;
            PyObject * ref_arg = 0;
            const char * label = 0;
            int sensitive = 1;
            if (!PyArg_ParseTuple(args, "OOs|i:Menu.add", & ref_command, & ref_arg, 
                                    & label, & sensitive))
                return NULL;

            // Create a proxy to accept the callback
            ProxyCommandCallback * proxy = new ProxyCommandCallback(ref_command);

            // Convert the argument to an Object
            Object * arg = 0;
            if (!convert_to_object(arg, ref_arg))
                return NULL;

            // Finally add the callback
            rp->mMenu->add_check(command_callback(proxy, arg, &ProxyCommandCallback::callback),
                                    label, sensitive);
        }
        catch (Exception & e) {
            string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
            PyErr_SetString(PyExc_RuntimeError, error.c_str());
            return NULL;
        }
        catch (...) {
            PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
            return NULL;      
        }
        // Return None
        return Py_BuildValue("");
    }

    PyObject *
    wrapped_menu_add_separator(PyObject * self, PyObject * args) {
        //virtual void add_separator() = 0;
        assert(PyObject_TypeCheck(self, & toolkitMenuType));
        MenuReference * rp = (MenuReference *) self;
        try {
            // Extract the arguments
            if (!PyArg_ParseTuple(args, ":Menu.add_separator"))
                return NULL;

            // Finally call the method
            rp->mMenu->add_separator();
        }
        catch (Exception & e) {
            string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
            PyErr_SetString(PyExc_RuntimeError, error.c_str());
            return NULL;
        }
        catch (...) {
            PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
            return NULL;      
        }
        // Return None
        return Py_BuildValue("");
    }

    //--------------------------------------------------------------------------------

    struct PyMethodDef menu_methods[ ] = {

        { "show",			wrapped_menu_show,	METH_VARARGS },
        { "hide",			wrapped_menu_hide,	METH_VARARGS },

        { "popup",			wrapped_menu_popup,	METH_VARARGS },
        { "add_push",		wrapped_menu_add_push,	METH_VARARGS },
        { "add_check",		wrapped_menu_add_check,	METH_VARARGS },
        { "add_separator",		wrapped_menu_add_separator, METH_VARARGS },

        { NULL, NULL }
    };



    PyObject *
    menu_getattr(PyObject * self, char * name) {

        // Look for a method
        return Py_FindMethod(menu_methods, self, name);
    }


    int
    menu_setattr(PyObject * self, char * name, PyObject * value) {
        PyErr_SetString(PyExc_NameError, "No attributes to be set");
        return -1;
    }


    void
    menu_dealloc(PyObject * self) {
        if (PyObject_TypeCheck(self, & toolkitMenuType)) {
            MenuReference * rp = (MenuReference *) self;
            if (rp->mMenu) {
                reference_table.erase(rp->mMenu);
                //FIXUP - this destroys the menu --- review ownership protocol for menus
                //rp->mMenu->unref();
            }
        }
        PyMem_DEL(self);
    }


}

//================================================================================

namespace Python {


  PyTypeObject toolkitMenuType = {
    PyObject_HEAD_INIT(&PyType_Type)		// Initialisation
    0,						// ob_size	: Usually 0
    "toolkit.Menu",				// tp_name	: Type Name
    sizeof(MenuReference),			// tp_basicsize : Object size
    0,						// tp_itemsize  : Usually 0
    
    // Standard methods
    (destructor) menu_dealloc,			// tp_dealloc	: refcount = 0
    (printfunc) 0,				// tp_print	: print x
    (getattrfunc) menu_getattr,			// tp_getattr	: x.attr
    (setattrfunc) menu_setattr,			// tp_setattr	: a.attr = v
    (cmpfunc) 0,				// tp_compare	: x > y
    (reprfunc) 0,				// tp_repr	: repr(x)

    // Type categories
    0,						// tp_as_number
    0,						// tp_as_sequence
    0,						// tp_as_mapping

    (hashfunc) 0,				// tp_hash
    (ternaryfunc) 0,				// tp_call
    (reprfunc) 0,				// tp_str
    (getattrofunc) 0,				// tp_getatto
    (setattrofunc) 0,				// tp_setatto
    
    0,						// tp_as_buffer
    0,						// tp_flags
    0,						// tp_doc
    
    (traverseproc) 0,				// tp_traverse
    (inquiry) 0,				// tp_clear
    (richcmpfunc) 0,				// tp_richcompare
    0,						// tp_weaklistoffset
    (getiterfunc) 0,				// tp_iter
    (iternextfunc) 0,				// tp_iternext
    0, 						// tp_methods
    0, 						// tp_members
    0,						// tp_getset
    0,						// tp_base

  };


  //--------------------------------------------------------------------------------


  PyObject *
  menu_reference_new(Toolkit::Menu * menu) {
    PyObject * result = 0;
    ReferenceMap::iterator iter = reference_table.find(menu);
    if (iter == reference_table.end()) {

      // Create a new reference
      MenuReference * self = PyObject_NEW(MenuReference, & toolkitMenuType);
      if (! self)
	return NULL;
    
      menu->ref();		 // take joint ownership of the object
      self->mMenu = menu;
      result = (PyObject *) self;
      reference_table[menu] = result;
    }
    else {
      //TRACE(2, "Reusing an existing reference");
      // Use an existing reference
      result = iter->second;
      Py_INCREF(result);
    }
    return result;
  }


  PyObject *
  menu_reference_new(PyObject * self, PyObject * args) {
    //TRACE(1, "Create a new Toolkit.Menu");
    try {
      // Extract the arguments
      if (!PyArg_ParseTuple(args, ":Menu"))
	return NULL;

      // Finally call the method
      Toolkit::Factory * factory = Toolkit::Factory::instance();
      Toolkit::Menu * menu = factory->create_menu();
      PyObject * result = menu_reference_new(menu);
      menu->unref(); // release our copy
      return result;
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }

} // namespace Python
