// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "foundation/object.h"
#include "foundation/macros.h"
#include "foundation/exception.h"

#include "toolkit/widget.h"
#include "toolkit/mouse_event.h"

#include "drawing/point.h"

// Python interfaces

#include "python/foundation/object.h"
#include "python/reflective/class.h"
#include "python/toolkit/mouse_event.h"

#include <list>
using namespace std;

//================================================================================

namespace {
    using namespace Python;


    PyObject *
    wrapped_mouse_event_get_type(PyObject * self, PyObject * args) {
        //MouseEventType get_type()
        assert(PyObject_TypeCheck(self, & toolkitMouseEventType));
        MouseEventReference * rp = (MouseEventReference *) self;
        try {
            // Extract the arguments
            if (!PyArg_ParseTuple(args, ":MouseEvent.get_type"))
                return NULL;

            // Finally call the method
            Toolkit::MouseEventType type = rp->mMouseEvent->get_type();

            // Convert the event type to a string
            const char * event_type = "unknown";
            switch (type) {
            case Toolkit::MET_ENTERED:  
                event_type = "entered";  
                break;
            case Toolkit::MET_EXITED:	 
                event_type = "exited";	  
                break;
            case Toolkit::MET_MOVED:	 
                event_type = "moved";	  
                break;
            case Toolkit::MET_DRAGGED:	 
                event_type = "dragged";  
                break;
            case Toolkit::MET_PRESSED:	 
                event_type = "pressed";  
                break;
            case Toolkit::MET_RELEASED: 
                event_type = "released"; 
                break;
            case Toolkit::MET_LAST: break;
            }
            return Py_BuildValue("s", event_type);
        }
        catch (Exception & e) {
            string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
            PyErr_SetString(PyExc_RuntimeError, error.c_str());
            return NULL;
        }
        catch (...) {
            PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
            return NULL;      
        }
    }


    PyObject *
    wrapped_mouse_event_get_point(PyObject * self, PyObject * args) {
        //Drawing::Point get_point()
        assert(PyObject_TypeCheck(self, & toolkitMouseEventType));
        MouseEventReference * rp = (MouseEventReference *) self;
        try {
            // Extract the arguments
            if (!PyArg_ParseTuple(args, ":MouseEvent.get_point"))
                return NULL;

            // Finally call the method
            const Drawing::Point & point = rp->mMouseEvent->get_point();
            return Py_BuildValue("(ii)", point.get_x(), point.get_y());
        }
        catch (Exception & e) {
            string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
            PyErr_SetString(PyExc_RuntimeError, error.c_str());
            return NULL;
        }
        catch (...) {
            PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
            return NULL;      
        }
    }

    PyObject *
    wrapped_mouse_event_get_button(PyObject * self, PyObject * args) {
        //int get_button()
        assert(PyObject_TypeCheck(self, & toolkitMouseEventType));
        MouseEventReference * rp = (MouseEventReference *) self;
        try {
            // Extract the arguments
            if (!PyArg_ParseTuple(args, ":MouseEvent.get_button"))
                return NULL;

            // Finally call the method
            int button = rp->mMouseEvent->get_button();
            return Py_BuildValue("i", button);
        }
        catch (Exception & e) {
            string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
            PyErr_SetString(PyExc_RuntimeError, error.c_str());
            return NULL;
        }
        catch (...) {
            PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
            return NULL;      
        }
    }


    PyObject *
    wrapped_mouse_event_get_click_count(PyObject * self, PyObject * args) {
        //int get_click_count()
        assert(PyObject_TypeCheck(self, & toolkitMouseEventType));
        MouseEventReference * rp = (MouseEventReference *) self;
        try {
            // Extract the arguments
            if (!PyArg_ParseTuple(args, ":MouseEvent.get_click_count"))
                return NULL;

            // Finally call the method
            int click_count = rp->mMouseEvent->get_click_count();
            return Py_BuildValue("i", click_count);
        }
        catch (Exception & e) {
            string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
            PyErr_SetString(PyExc_RuntimeError, error.c_str());
            return NULL;
        }
        catch (...) {
            PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
            return NULL;      
        }
    }

    //--------------------------------------------------------------------------------

    struct PyMethodDef mouse_event_methods[ ] = {
        { "get_type",		wrapped_mouse_event_get_type, METH_VARARGS },
        { "get_point",		wrapped_mouse_event_get_point, METH_VARARGS },
        { "get_button",		wrapped_mouse_event_get_button, METH_VARARGS },
        { "get_click_count",	wrapped_mouse_event_get_click_count, METH_VARARGS },

        { NULL, NULL }
    };



    PyObject *
    mouse_event_getattr(PyObject * self, char * name) {

        // Look for a method
        return Py_FindMethod(mouse_event_methods, self, name);
    }


    int
    mouse_event_setattr(PyObject * self, char * name, PyObject * value) {
        PyErr_SetString(PyExc_NameError, "No attributes to be set");
        return -1;
    }


    void
    mouse_event_dealloc(PyObject * self) {
        assert(PyObject_TypeCheck(self, & toolkitMouseEventType));
        PyMem_DEL(self);
    }


}

//================================================================================

namespace Python {

    PyTypeObject toolkitMouseEventType = {
        PyObject_HEAD_INIT(&PyType_Type)		// Initialisation
            0,						// ob_size	: Usually 0
        "toolkit.MouseEvent",			// tp_name	: Type Name
        sizeof(MouseEventReference),		// tp_basicsize : Object size
        0,						// tp_itemsize  : Usually 0

        // Standard methods
        (destructor) mouse_event_dealloc,		// tp_dealloc	: refcount = 0
        (printfunc) 0,				// tp_print	: print x
        (getattrfunc) mouse_event_getattr,		// tp_getattr	: x.attr
        (setattrfunc) mouse_event_setattr,		// tp_setattr	: a.attr = v
        (cmpfunc) 0,				// tp_compare	: x > y
        (reprfunc) 0,				// tp_repr	: repr(x)

        // Type categories
        0,						// tp_as_number
        0,						// tp_as_sequence
        0,						// tp_as_mapping

        (hashfunc) 0,				// tp_hash
        (ternaryfunc) 0,				// tp_call
        (reprfunc) 0,				// tp_str
        (getattrofunc) 0,				// tp_getatto
        (setattrofunc) 0,				// tp_setatto

        0,						// tp_as_buffer
        0,						// tp_flags
        0,						// tp_doc

        (traverseproc) 0,				// tp_traverse
        (inquiry) 0,				// tp_clear
        (richcmpfunc) 0,				// tp_richcompare
        0,						// tp_weaklistoffset
        (getiterfunc) 0,				// tp_iter
        (iternextfunc) 0,				// tp_iternext
        0, 						// tp_methods
        0, 						// tp_members
        0,						// tp_getset
        0,						// tp_base

    };


    //--------------------------------------------------------------------------------


    PyObject *
    mouse_event_reference_new(const Toolkit::MouseEvent * mouse_event) {
        MouseEventReference * self = PyObject_NEW(MouseEventReference, & toolkitMouseEventType);
        if (! self)
            return NULL;
        self->mMouseEvent = mouse_event;
        return (PyObject *) self;
    }

    PyObject *
    mouse_event_reference_new(PyObject * self, PyObject * args) {
        TRACE(1, "Create a new Toolkit.MouseEvent");
        PyErr_SetString(PyExc_TypeError, "don't know how to construct");
        return NULL;
    }

} // namespace {anonymous}
