// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "canvas.h"

#include "mouse_listener.h"
#include "canvas_listener.h"
#include "key_listener.h"
#include "canvas_event.h"

#include "drawing/vector.h"

#include <algorithm>
#include <functional>
#include <iostream>
using namespace std;

namespace Toolkit {

    Canvas::Canvas() : Widget(), Drawing::Drawable() {
    }

    Canvas::~Canvas() {
    }

    void
    Canvas::add(MouseListener * mouse_listener) {
	mMouseListeners.push_front(mouse_listener);
    }

    void
    Canvas::remove(const MouseListener * mouse_listener) {
	list<MouseListener *>::iterator iter = find(mMouseListeners.begin(), mMouseListeners.end(),
						    mouse_listener);

	if(iter == mMouseListeners.end()) {
	    cerr << "Tried to remove mouse listener but couldn't find it" << endl;
	}
	else {
	    mMouseListeners.erase(iter);
	}
    }

    void
    Canvas::add(CanvasListener * canvas_listener) {
	mCanvasListeners.push_front(canvas_listener);
    }

    void
    Canvas::remove(const CanvasListener * canvas_listener) {
	list<CanvasListener *>::iterator iter =
	    find(mCanvasListeners.begin(), mCanvasListeners.end(), canvas_listener);

	if(iter == mCanvasListeners.end()) {
	    cerr << "Tried to remove canvas listener but couldn't find it" << endl;
	}
	else {
	    mCanvasListeners.erase(iter);
	}
    }

    void
    Canvas::add(KeyListener * key_listener) {
	mKeyListeners.push_front(key_listener);
    }

    void
    Canvas::remove(const KeyListener * key_listener) {
	list<KeyListener *>::iterator iter = find(mKeyListeners.begin(), mKeyListeners.end(),
						    key_listener);

	if(iter == mKeyListeners.end()) {
	    cerr << "Tried to remove key listener but couldn't find it" << endl;
	}
	else {
	    mKeyListeners.erase(iter);
	}
    }

    bool
    Canvas::mouse_entered(const MouseEvent * mouse_event) {
	list<MouseListener *>::const_iterator iter =
	    find_if(mMouseListeners.begin(), mMouseListeners.end(),
		    bind2nd(mem_fun(&MouseListener::mouse_entered), mouse_event));

	return iter != mMouseListeners.end();
    }

    bool
    Canvas::mouse_exited(const MouseEvent * mouse_event) {
	list<MouseListener *>::const_iterator iter =
	    find_if(mMouseListeners.begin(), mMouseListeners.end(),
		    bind2nd(mem_fun(&MouseListener::mouse_exited), mouse_event));

	return iter != mMouseListeners.end();
    }

    bool
    Canvas::mouse_pressed(const MouseEvent * mouse_event) {
	list<MouseListener *>::const_iterator iter =
	    find_if(mMouseListeners.begin(), mMouseListeners.end(),
		    bind2nd(mem_fun(&MouseListener::mouse_pressed), mouse_event));

	return iter != mMouseListeners.end();
    }

    bool
    Canvas::mouse_released(const MouseEvent * mouse_event) {
	list<MouseListener *>::const_iterator iter =
	    find_if(mMouseListeners.begin(), mMouseListeners.end(),
		    bind2nd(mem_fun(&MouseListener::mouse_released), mouse_event));

	return iter != mMouseListeners.end();
    }

    bool
    Canvas::mouse_moved(const MouseEvent * mouse_event) {
	list<MouseListener *>::const_iterator iter =
	    find_if(mMouseListeners.begin(), mMouseListeners.end(),
		    bind2nd(mem_fun(&MouseListener::mouse_moved), mouse_event));

	return iter != mMouseListeners.end();
    }

    bool
    Canvas::mouse_dragged(const MouseEvent * mouse_event) {
	list<MouseListener *>::const_iterator iter =
	    find_if(mMouseListeners.begin(), mMouseListeners.end(),
		    bind2nd(mem_fun(&MouseListener::mouse_dragged), mouse_event));

	return iter != mMouseListeners.end();
    }

    bool
    Canvas::key_pressed(const KeyEvent * key_event) {
	list<KeyListener *>::const_iterator iter =
	    find_if(mKeyListeners.begin(), mKeyListeners.end(),
		    bind2nd(mem_fun(&KeyListener::key_pressed), key_event));

	return iter != mKeyListeners.end();
    }

    bool
    Canvas::key_released(const KeyEvent * key_event) {
	list<KeyListener *>::const_iterator iter =
	    find_if(mKeyListeners.begin(), mKeyListeners.end(),
		    bind2nd(mem_fun(&KeyListener::key_released), key_event));

	return iter != mKeyListeners.end();
    }

    bool
    Canvas::exposed(const CanvasEvent * canvas_event) {
	list<CanvasListener *>::const_iterator iter;

	const Drawing::Vector & size = get_size();

	if(canvas_event->get_area().get_w() < size.get_x() ||
	canvas_event->get_area().get_h() < size.get_y()) {
	    push_clip(canvas_event->get_area());

	    iter = find_if(mCanvasListeners.begin(),
			   mCanvasListeners.end(),
			   bind2nd(mem_fun(&CanvasListener::canvas_exposed), canvas_event));

	    pop_clip();
	}
	else {
	    iter = find_if(mCanvasListeners.begin(),
			   mCanvasListeners.end(),
			   bind2nd(mem_fun(&CanvasListener::canvas_exposed), canvas_event));
	}

	return iter != mCanvasListeners.end();
    }

} // namespace Toolkit
