// Dachshund,
// Copyright (C) 2002 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "class_diagram_model.h"

#include "reflective/changes.h"

#include "uml/foundation/data_types/name.h"
#include "uml/foundation/core/class.h"
#include "uml/foundation/core/association.h"
#include "uml/foundation/core/association_end.h"
#include "uml/foundation/core/generalization.h"
#include "uml/foundation/core/dependency.h"
#include "uml/model_management/package.h"

#include "uml/diagram/diagram.h"
#include "uml/diagram/class_presentation.h"
#include "uml/diagram/package_presentation.h"
#include "uml/diagram/association_presentation.h"
#include "uml/diagram/generalization_presentation.h"
#include "uml/diagram/dependency_presentation.h"

#include "foundation/macros.h"
#include "foundation/utility.h"
#include "foundation/undo_manager.h"

#include <algorithm>
#include <utility>
using namespace std;

// TODO
//
// add a system that can be used to get all of the edges and nodes from
// the graph model.
//
// the perspective layer would use this to render all of the figs when
// the diagram is opened

namespace UMLExt {

    ClassDiagramModel::ClassDiagramModel(UML::Diagram * diagram) : Graph::Model(),
        mDiagram(diagram) {
	ASSERT(diagram);
	mDiagram->ref();

	UndoManager * undo_manager = UndoManager::instance();
	undo_manager->add_change_listener(static_cast<ChangeListener *>(this));
    }

    ClassDiagramModel::~ClassDiagramModel() {
	UndoManager * undo_manager = UndoManager::instance();
	undo_manager->remove_change_listener(static_cast<ChangeListener *>(this));

	mDiagram->unref();
    }

    void
    ClassDiagramModel::add_node(Object * node, Object * parent) throw(Graph::Exception) {
	RETURN_IF_FAIL(node);

	UML::PresentationElement * pe = dynamic_cast<UML::PresentationElement *>(node);

	if(!pe) {
	    throw Graph::Exception("Attempting to add an illegal object type from the "
                                   "class-diagram model");
	}
	else if(pe->ref_get_type() == "ClassPresentation") {
	    begin_transaction();

	    if(pe->get_subject()) {
		// the subject exists, so we see if there are any edges
		// that can be added

		UML::Class * c = known_cast<UML::Class *>(pe->get_subject());

		const list<UML::AssociationEnd *> & associations = c->get_association();

		for(list<UML::AssociationEnd *>::const_iterator iter = associations.begin();
		    iter != associations.end();
		    ++iter) {
		    UML::AssociationEnd * ae = *iter;

		    // TODO:
		    //
		    //  get the other end, get the class on the other end,
		    //  if that class is in the diagram then create a presentation
		    //  element for the association
		}
	    }
	    else {
		// the presentation element has no subject, so we automatically
		// create one for it

		UML::Class * c = new UML::Class();

		UML::Name * name = new UML::Name();
		name->ref_set_value("NewClass");
		c->set_name(name);
		name->unref();

		pe->set_subject(c);

		UML::Namespace * n = mDiagram->get_namespace();

		n->add_owned_element(c);

		c->unref();
	    }

	    mDiagram->add_element(pe);

	    end_transaction();
	}
	else if(pe->ref_get_type() == "PackagePresentation") {
	    begin_transaction();

	    if(pe->get_subject()) {
		UML::Package * p = known_cast<UML::Package *>(pe->get_subject());

		// TODO added edges that can be added...
	    }
	    else {
		UML::Package * p = new UML::Package();

		UML::Name * name = new UML::Name();
		name->ref_set_value("NewPackage");
		p->set_name(name);
		name->unref();

		pe->set_subject(p);

		UML::Namespace * n = mDiagram->get_namespace();

		n->add_owned_element(p);

		p->unref();
	    }

	    mDiagram->add_element(pe);

	    end_transaction();
	}
	else {
	    throw Graph::Exception("Node must be a class or a package");
	}
    }

    void
    ClassDiagramModel::remove_node(Object * node) throw(Graph::Exception) {
	RETURN_IF_FAIL(node);

	UML::PresentationElement * p = dynamic_cast<UML::PresentationElement *>(node);

	if(!p) {
	    throw Graph::Exception("Attempting to remove an illegal object type from the "
                                   "class-diagram model");
	}
	else if(p->ref_get_type() == "ClassPresentation") {
	    UML::ClassPresentation * cp = known_cast<UML::ClassPresentation *>(p);

	    begin_transaction();

            // remove the associations that this class is a source for

	    const list<UML::AssociationPresentation *> & apss = cp->get_association_source();

	    for(list<UML::AssociationPresentation *>::const_iterator iter = apss.begin();
		iter != apss.end();
		++iter) {
		remove_edge(*iter);
	    }

            // remove the associations that this class is a destination for

	    const list<UML::AssociationPresentation *> & apds = cp->get_association_destination();

	    for(list<UML::AssociationPresentation *>::const_iterator iter = apds.begin();
		iter != apds.end();
		++iter) {
		remove_edge(*iter);
	    }

            // remove the generalizations that this class is a subtype for

	    const list<UML::GeneralizationPresentation *> & gps = cp->get_generalization();

	    for(list<UML::GeneralizationPresentation *>::const_iterator iter = gps.begin();
		iter != gps.end();
		++iter) {
		remove_edge(*iter);
	    }

            // remove the generalizations that this class is a supertype for

	    const list<UML::GeneralizationPresentation *> & sps = cp->get_specialization();

	    for(list<UML::GeneralizationPresentation *>::const_iterator iter = sps.begin();
		iter != sps.end();
		++iter) {
		remove_edge(*iter);
	    }

            //

	    cp->ref_delete();

	    end_transaction();
	}
	else {
	    throw Graph::Exception("Node must be a Class or a Package");
	}
    }

    //

    void
    ClassDiagramModel::add_edge(Object * edge,
				Object * node1, Object * node2) throw(Graph::Exception) {
	RETURN_IF_FAIL(edge);
        RETURN_IF_FAIL(node1);
        RETURN_IF_FAIL(node2);

	UML::PresentationElement * p = dynamic_cast<UML::PresentationElement *>(edge);

	if(!p) {
	    throw Graph::Exception("Attempting to add an illegal object type from the "
                                   "class-diagram model");
	}
	else if(p->ref_get_type() == "AssociationPresentation") {
	    UML::AssociationPresentation * ap = known_cast<UML::AssociationPresentation *>(p);
	    UML::ClassPresentation * cp1 = dynamic_cast<UML::ClassPresentation *>(node1);
	    UML::ClassPresentation * cp2 = dynamic_cast<UML::ClassPresentation *>(node2);

	    if(!cp1 || !cp2) {
		throw Graph::Exception("An association can only be made between two Classifiers");
	    }

	    begin_transaction();

	    if(!p->get_subject()) {
		UML::Name * name = new UML::Name();

		// create the association

		UML::Association * a = new UML::Association();

		name->ref_set_value("NewAssociation");
		a->set_name(name);

		// create the ends

		UML::AssociationEnd * ae1 = new UML::AssociationEnd();

		name->ref_set_value("End1");
		ae1->set_name(name);

		UML::AssociationEnd * ae2 = new UML::AssociationEnd();

		name->ref_set_value("End2");
		ae2->set_name(name);

		// add the association to the namespace

		UML::Namespace * n = mDiagram->get_namespace();

		n->add_owned_element(a);

		// add the ends to the association

		a->add_connection(ae1);
		a->add_connection(ae2);

		// tell the presentation what its subject is

		ap->set_subject(a);

		// unref stuff

		ae2->unref();
		ae1->unref();
		a->unref();
		name->unref();
	    }

	    // connect the presentation of the association to the classes

	    ap->set_source(cp1);
	    ap->set_destination(cp2);

	    // add the presentation to the diagram

	    mDiagram->add_element(ap);

	    end_transaction();
	}
        else if(p->ref_get_type() == "GeneralizationPresentation") {
            UML::GeneralizationPresentation * gp = known_cast<UML::GeneralizationPresentation *>(p);
            UML::ClassPresentation * cp1 = dynamic_cast<UML::ClassPresentation *>(node1);
            UML::ClassPresentation * cp2 = dynamic_cast<UML::ClassPresentation *>(node2);

            if(!cp1 && cp2) {
                throw Graph::Exception("A generalization can only be made between two "
                                       "generalizable-elements of the same type");
            }

            begin_transaction();

            if(!gp->get_subject()) {
                UML::Name * name = new UML::Name();

                // create the generalization

                UML::Generalization * g = new UML::Generalization();

                name->ref_set_value("NewGeneralization");
                g->set_name(name);

                // add the generalization to the package

                UML::Namespace * ns = mDiagram->get_namespace();

                ns->add_owned_element(g);

                // tell the presentation what its subject is

                gp->set_subject(g);

                g->unref();
                name->unref();
            }

            // connect the presentation of the generalization to the classes

            gp->set_subtype(cp1);
            gp->set_supertype(cp2);

            // add the presentation to the diagram

            mDiagram->add_element(gp);

            end_transaction();
        }
        else if(p->ref_get_type() == "DependencyPresentation") {
            UML::DependencyPresentation * dp = known_cast<UML::DependencyPresentation *>(p);
            UML::NodeElement * n1 = dynamic_cast<UML::NodeElement *>(node1);
            UML::NodeElement * n2 = dynamic_cast<UML::NodeElement *>(node2);

            if(!n1 && !n2) {
                throw Graph::Exception("A dependency can only be formed between two nodes. "
                                       "(This will be fixed...)");
            }

            begin_transaction();

            if(!dp->get_subject()) {
                UML::Name * name = new UML::Name();

                // create the dependency

                UML::Dependency * d = new UML::Dependency();

                name->ref_set_value("NewDependency");
                d->set_name(name);

                // add the dependency to the package

                UML::Namespace * ns = mDiagram->get_namespace();

                ns->add_owned_element(d);

                // tell the presentation what is subject is

                dp->set_subject(d);

                d->unref();
                name->unref();
            }

            // connect the presentation of the dependency to the nodes

            dp->set_client(n1);
            dp->set_supplier(n2);

            // add the dependency to the diagram

            mDiagram->add_element(dp);

            end_transaction();
        }
	else {
            throw Graph::Exception("Edge must be a generalization or an association");
	}
    }

    void
    ClassDiagramModel::remove_edge(Object * edge) throw(Graph::Exception) {
	RETURN_IF_FAIL(edge);

	UML::PresentationElement * p = dynamic_cast<UML::PresentationElement *>(edge);

	if(!p) {
	    throw Graph::Exception("Attempting to remove an illegal object type from the "
                                   "class-diagram model");
	}
	else if(p->ref_get_type() == "AssociationPresentation") {
	    UML::AssociationPresentation * ap = known_cast<UML::AssociationPresentation *>(p);

	    UML::Association * a = known_cast<UML::Association *>(ap->get_subject());

	    ap->ref_delete();
	    a->ref_delete();

	    edge_removed(edge);
	}
	else {
	    throw Graph::Exception("Node must be an Association");
	}
    }

    //
    //
    //

    void
    ClassDiagramModel::changes_occurred(const list<ChangeEvent *> & changes) throw() {
	for_each(changes.begin(), changes.end(),
		 bind1st(mem_fun(&ClassDiagramModel::process_change), this));
    }

    void
    ClassDiagramModel::process_change(ChangeEvent * change) {
	ASSERT(change);

	if(typeid(*change) == typeid(REF::ChangeAttributeChanged)) {
	    REF::ChangeAttributeChanged * cac = known_cast<REF::ChangeAttributeChanged *>(change);

	    REF::Class * object = cac->get_object();
	    ASSERT(object);

	    if(object->ref_is_a_type_of("PresentationElement") &&
	       object->ref_get_role("diagram") == mDiagram) {
		// a presentation in our diagram has changed

		node_changed(object);
	    }
	    else if(object->ref_is_a_type_of("ModelElement")) {
		// if this is a presentation elemtn of ours is
		// representing this model element then need
		// to emit a node/edge_changed signal

		UML::ModelElement * me = known_cast<UML::ModelElement *>(object);

		const list<UML::PresentationElement *> & presentations = me->get_presentation();

		for(list<UML::PresentationElement *>::const_iterator iter = presentations.begin();
		    iter != presentations.end();
		    ++iter) {
		    UML::PresentationElement * pe = *iter;
		    ASSERT(pe);

		    if(pe->get_diagram() == mDiagram) {
			node_changed(pe);
			break;
		    }
		}
	    }
	}
	else if(typeid(*change) == typeid(REF::ChangeRoleAdded)) {
	    // we are interested if:
	    //     - a presentation element has been added to the diagram
	    //
	    //     - an edge has been added to the syntax model that 
	    //       we can link in the presentation model

	    REF::ChangeRoleAdded * cra = known_cast<REF::ChangeRoleAdded *>(change);

	    REF::Class * object = cra->get_object();
	    ASSERT(object);

	    REF::Class * value = cra->get_value();
	    ASSERT(value);

	    if(object->ref_is_a_type_of("Diagram") &&
	       cra->get_role_name() == "element" && object == mDiagram) {
		if(value->ref_is_a_type_of("EdgeElement")) {
		    edge_added(value);
		}
		else if(value->ref_is_a_type_of("NodeElement")) {
		    node_added(value);
		}
		else {
		    ERROR("Unknown type added to diagram");
		}
	    }
	}
	else if(typeid(*change) == typeid(REF::ChangeRoleRemoved)) {
	    // we are interested if:
	    //     - a presentation element has been removed from the diagram
	    //     - an object that a presentation elements represents has been
	    //       removed from the syntax model

	    REF::ChangeRoleRemoved * crr = known_cast<REF::ChangeRoleRemoved *>(change);

	    REF::Class * object = crr->get_object();
	    ASSERT(object);

	    REF::Class * value = crr->get_value();
	    ASSERT(value);

	    if(object->ref_is_a_type_of("Diagram") &&
	       crr->get_role_name() == "element" && object == mDiagram) {
		if(value->ref_is_a_type_of("EdgeElement")) {
		    edge_removed(value);
		}
		else if(value->ref_is_a_type_of("NodeElement")) {
		    node_removed(value);
		}
		else {
		    ERROR("Unknown type removed from diagram");
		}
	    }
	}
    }

} // namespace UMLExt
