// Dachshund,
// Copyright (C) 2002 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "class_diagram_renderer.h"

#include "uml/diagram/class_presentation.h"
#include "uml/diagram/association_presentation.h"
#include "uml/diagram/generalization_presentation.h"
#include "uml/diagram/dependency_presentation.h"

#include "diagram/layer.h"

#include "foundation/macros.h"
#include "foundation/utility.h"

#include "fig_class.h"
#include "fig_package.h"
#include "fig_association.h"
#include "fig_generalization.h"
#include "fig_dependency.h"

namespace UMLExt {

    ClassDiagramRenderer::ClassDiagramRenderer() : Renderer() {
    }

    ClassDiagramRenderer::~ClassDiagramRenderer() {
    }

    FigNode *
    ClassDiagramRenderer::create_fig_node_for(Diagram::Layer * layer, Object * node) {
	UML::PresentationElement * klass = known_cast<UML::PresentationElement *>(node);

	if(klass->ref_get_type() == "ClassPresentation") {
	    UML::ClassPresentation * cp = known_cast<UML::ClassPresentation *>(klass);

	    return new FigClass(cp);
	}
        else if(klass->ref_get_type() == "PackagePresentation") {
	    UML::PackagePresentation * pp = known_cast<UML::PackagePresentation *>(klass);

	    return new FigPackage(pp);
	}
	else {
	    ASSERT_NOT_REACHED();
	}
    }

    FigEdge *
    ClassDiagramRenderer::create_fig_edge_for(Diagram::Layer * layer, Object * edge) {
	UML::PresentationElement * klass = known_cast<UML::PresentationElement *>(edge);

	if(klass->ref_get_type() == "AssociationPresentation") {
	    UML::AssociationPresentation * ap = known_cast<UML::AssociationPresentation *>(klass);

            // make sure the association connects two classes

	    ASSERT(ap->get_source());
	    ASSERT(ap->get_destination());

	    //

	    Diagram::Fig * source = layer->get_presentation_for(ap->get_source());
	    Diagram::Fig * destination = layer->get_presentation_for(ap->get_destination());

	    ASSERT(source);
	    ASSERT(destination);

	    FigClass * source_class = known_cast<FigClass *>(source);
	    FigClass * destination_class = known_cast<FigClass *>(destination);

	    return new FigAssociation(ap, source_class, destination_class);
	}
        else if(klass->ref_get_type() == "GeneralizationPresentation") {
            UML::GeneralizationPresentation * gp =
                known_cast<UML::GeneralizationPresentation *>(klass);

            // make sure the generalization has a subtype and a supertype

            ASSERT(gp->get_subtype());
            ASSERT(gp->get_supertype());

            //

            Diagram::Fig * subtype = layer->get_presentation_for(gp->get_subtype());
            Diagram::Fig * supertype = layer->get_presentation_for(gp->get_supertype());

            ASSERT(subtype);
            ASSERT(supertype);

            FigClass * subtype_class = known_cast<FigClass *>(subtype);
            FigClass * supertype_class = known_cast<FigClass *>(supertype);

            return new FigGeneralization(gp, subtype_class, supertype_class);
        }
        else if(klass->ref_get_type() == "DependencyPresentation") {
            UML::DependencyPresentation * dp =
                known_cast<UML::DependencyPresentation *>(klass);

            // make sure the dependency has a client and a supplier

            ASSERT(dp->get_client());
            ASSERT(dp->get_supplier());

            //

            Diagram::Fig * client = layer->get_presentation_for(dp->get_client());
            Diagram::Fig * supplier = layer->get_presentation_for(dp->get_supplier());

            ASSERT(client);
            ASSERT(supplier);

            FigNode * client_node = known_cast<FigNode *>(client);
            FigNode * supplier_node = known_cast<FigNode *>(supplier);

            return new FigDependency(dp, client_node, supplier_node);
        }
	else {
	    ASSERT_NOT_REACHED();
	}
    }

} // namespace UMLExt
