// Dachshund,
// Copyright (C) 2002 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "fig_edge.h"

#include "fig_node.h"
#include "fig_gimlet.h"
#include "colors.h"

#include "uml/foundation/data_types/name.h"
#include "uml/diagram/edge_element.h"

#include "drawing/drawable.h"

#include "foundation/utility.h"
#include "foundation/macros.h"

#include <sstream>
using namespace std;

namespace UMLExt {

    FigEdge::FigEdge(UML::PresentationElement * pe,
		     FigNode * source,
		     FigNode * destination) : Diagram::FigEdge(pe, source, destination) {
	set_line_color(EDGE_LINE_COLOR);
	set_fill_color(EDGE_FILL_COLOR);

	mSourceGimlet = new FigGimlet();
	mDestinationGimlet = new FigGimlet();
    }

    FigEdge::~FigEdge() {
	mSourceGimlet->unref();
	mDestinationGimlet->unref();
    }

    void
    FigEdge::paint(Drawing::Drawable * drawable) const {

        //              ^
	//            / |
	//           /__|
	//           /
	//          /
	//         /
	//        /
	//       /
	//      /
	//

	ASSERT(mPoints.size() >= 2);

	if(is_filled()) {
	    drawable->set_color(get_fill_color());
	    drawable->draw_polygon(mPoints, true);
	}
	else {
	    drawable->set_color(get_line_color());
	    drawable->set_line_style(get_line_style());
	    drawable->draw_lines(mPoints);
	}

	mSourceGimlet->paint(drawable);
	mDestinationGimlet->paint(drawable);
    }

    void
    FigEdge::cleanup() {
	Diagram::FigEdge::cleanup();
    }

    void
    FigEdge::begin_interaction() {
        Diagram::FigEdge::begin_interaction();
    }

    void
    FigEdge::end_interaction() {
	begin_transaction();

	UML::EdgeElement * ee = known_cast<UML::EdgeElement *>(get_owner());

	ostringstream ost;

	for(unsigned int i = 1; i < mPoints.size() - 1; ++i) {
	    ost << mPoints[i];
	}

	UML::Name * geometry = new UML::Name();
	geometry->ref_set_value(ost.str());
	ee->set_geometry(geometry);
	geometry->unref();

        Diagram::FigEdge::end_interaction();

	end_transaction();
    }

    void
    FigEdge::update_to_owner() {
        Diagram::FigEdge::update_to_owner();

	UML::EdgeElement * ee = known_cast<UML::EdgeElement *>(get_owner());

	const UML::Name * geometry = ee->get_geometry();

	if(geometry) {
	    vector<Drawing::Point> points;
	    istringstream ist(geometry->get_value());

	    Drawing::Point source_center;
	    mSourceFig->get_center(source_center);
	    points.push_back(source_center);

	    for(;;) {
		Drawing::Point point;
		ist >> point;

		if(ist.good()) {
		    points.push_back(point);
		}
		else {
		    break;
		}
	    }

	    Drawing::Point destination_center;
	    mDestinationFig->get_center(destination_center);
	    points.push_back(destination_center);

	    set_points(points);

	    update_edge_to_nodes();
	}
	else {
	    ERROR("No geometry in edge element");
	}
    }

    void
    FigEdge::set_layer(Diagram::Layer * layer) {
	Diagram::FigEdge::set_layer(layer);

	mSourceGimlet->set_layer(layer);
	mDestinationGimlet->set_layer(layer);
    }

    void
    FigEdge::update_bounds_to_points() {
	Diagram::FigEdge::update_bounds_to_points();

	update_gimlets();

	mBounds |= mSourceGimlet->get_bounds();
	mBounds |= mDestinationGimlet->get_bounds();
    }

    void
    FigEdge::update_gimlets() {
	RETURN_IF_FAIL(mPoints.size() >= 2);

	mSourceGimlet->damage();
	mSourceGimlet->set_control_points(mPoints[1], mPoints[0]);
	mSourceGimlet->damage();

	mDestinationGimlet->damage();
	mDestinationGimlet->set_control_points(mPoints[mPoints.size() - 2],
					       mPoints[mPoints.size() - 1]);
	mDestinationGimlet->damage();
    }

} // namespace UMLExt
