// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "node.h"

#include "property.h"

#include <algorithm>
#include <vector>
#include <set>
#include <iterator>
#include <iostream>
using namespace std;

#include "foundation/functional.h"
#include "foundation/macros.h"
#include "foundation/utility.h"
#include "foundation/iterator.h"
#if 0
#include "foundation/changes.h"
#endif
#include "foundation/commands.h"

namespace XML {

    Node::Node() : Object() {
    }

    Node::~Node() {
	for_each(mChildren.begin(), mChildren.end(), mem_fun(&Object::unref));
	for_each(mProperties.begin(), mProperties.end(), mem_fun(&Object::unref));
    }

    void
    Node::set_name(const string & name) {
	if(mName != name) {
	    begin_transaction();

	    Command * command = command_flip_argument(this, &Node::set_name, name, mName);

#if 0
	    ChangeEvent * change = new ChangeAttributeChanged(this, "name", name, mName);
#endif

	    // make the changes

	    mName = name;

	    // changes made

	    log_command(command);
#if 0
	    log_change(change);
#endif

#if 0
	    change->unref();
#endif
	    command->unref();

	    end_transaction();
	}
    }

    void
    Node::set_content(const string & content) {
	if(mContent != content) {
	    begin_transaction();

	    Command * command = command_flip_argument(this, &Node::set_content, content, mContent);
#if 0
	    ChangeEvent * change = new ChangeAttributeChanged(this, "content", content, mContent);
#endif

	    // make the changes

	    mContent = content;

	    // changes made

	    log_command(command);
#if 0
	    log_change(change);
#endif

#if 0
	    change->unref();
#endif
	    command->unref();

	    end_transaction();
	}
    }

    void
    Node::add_property(Property * property, int index) {
	RETURN_IF_FAIL(index <= static_cast<int>(mProperties.size()));

	if(index < 0) {
	    index = mProperties.size();
	}

	list<Property *>::iterator iter = mProperties.begin();

	for(int i = 0; i < static_cast<int>(mProperties.size()); ++i) {
	    RETURN_IF_FAIL(*iter != property);

	    if(i < index) {
		++iter;
	    }
	}

	begin_transaction();

	Command * command = command_flip_method(this,
						&Node::add_property,
						&Node::remove_property,
						property, index);

#if 0
	ChangeEvent * change = new ChangeRoleAdded(this, "property", property, index);
#endif

	// make the changes

	property->ref();

	mProperties.insert(iter, property);

	// changes made

	log_command(command);
#if 0
	log_change(change);
#endif

#if 0
	change->unref();
#endif
	command->unref();

	end_transaction();
    }

    void
    Node::remove_property(Property * property) {
	int index = 0;

	list<Property *>::iterator iter = mProperties.begin();

	for(; iter != mProperties.end(); ++iter, ++index) {
	    if(*iter == property) {
		break;
	    }
	}

	RETURN_IF_FAIL(iter != mProperties.end());

	begin_transaction();

	Command * command = command_flip_method(this,
						&Node::remove_property,
						&Node::add_property,
						property, index);

#if 0
	ChangeEvent * change = new ChangeRoleRemoved(this, "property", property, index);
#endif

	// make the changes

	mProperties.erase(iter);

	property->unref();

	// changes made

	log_command(command);
#if 0
	log_change(change);
#endif

#if 0
	change->unref();
#endif
	command->unref();

	end_transaction();
    }

    void
    Node::set_properties(const list<Property *> & properties) {
	set<Property *> before;
	set<Property *> after;

	before.insert(mProperties.begin(), mProperties.end());
	after.insert(properties.begin(), properties.end());

	// check for duplicates
	RETURN_IF_FAIL(before.size() == mProperties.size());
	RETURN_IF_FAIL(after.size() == properties.size());

	begin_transaction();

	Command * command = command_flip_argument(this, &Node::set_properties,
						  properties, mProperties);

#if 0
	ChangeEvent * change = new ChangeAttributeChanged(this, "properties",
							  properties, mProperties);
#endif

	// make the changes

	vector<Property *> added;
	vector<Property *> removed;

	set_difference(after.begin(), after.end(), 
		       before.begin(), before.end(), back_inserter(added));

	set_difference(before.begin(), before.end(), 
		       after.begin(), after.end(), back_inserter(removed));

	for_each(added.begin(), added.end(), mem_fun(&Object::ref));
	for_each(removed.begin(), removed.end(), mem_fun(&Object::unref));

	mProperties = properties;

	// changes made

	log_command(command);
#if 0
	log_change(change);
#endif

#if 0
	change->unref();
#endif
	command->unref();

	end_transaction();
    }

    void
    Node::add_child(Node * child, int index) {
	RETURN_IF_FAIL(index <= static_cast<int>(mChildren.size()));

	if(index < 0) {
	    index = mChildren.size();
	}

	list<Node *>::iterator iter = mChildren.begin();

	for(int i = 0; i < static_cast<int>(mChildren.size()); ++i) {
	    RETURN_IF_FAIL(*iter != child);

	    if(i < index) {
		++iter;
	    }
	}

	begin_transaction();

	Command * command = command_flip_method(this,
						&Node::add_child,
						&Node::remove_child,
						child, index);

#if 0
	ChangeEvent * change = new ChangeRoleAdded(this, "child", child, index);
#endif

	// make the changes

	child->ref();

	mChildren.insert(iter, child);

	// changes made

	log_command(command);
#if 0
	log_change(change);
#endif

#if 0
	change->unref();
#endif
	command->unref();

	end_transaction();
    }

    void
    Node::remove_child(Node * child) {
	int index = 0;

	list<Node *>::iterator iter = mChildren.begin();

	for(; iter != mChildren.end(); ++iter, ++index) {
	    if(*iter == child) {
		break;
	    }
	}

	RETURN_IF_FAIL(iter != mChildren.end());

	begin_transaction();

	Command * command = command_flip_method(this,
						&Node::remove_child,
						&Node::add_child,
						child, index);

#if 0
	ChangeEvent * change = new ChangeRoleRemoved(this, "child", child, index);
#endif

	// make the changes

	mChildren.erase(iter);

	child->unref();

	// changes made

	log_command(command);
#if 0
	log_change(change);
#endif

#if 0
	change->unref();
#endif
	command->unref();

	end_transaction();
    }

    void
    Node::set_children(const list<Node *> & children) {
	set<Node *> before;
	set<Node *> after;

	before.insert(mChildren.begin(), mChildren.end());
	after.insert(children.begin(), children.end());

	// check for duplicates
	RETURN_IF_FAIL(before.size() == mChildren.size());
	RETURN_IF_FAIL(after.size() == children.size());

	begin_transaction();

	Command * command = command_flip_argument(this, &Node::set_children,
						  children, mChildren);

#if 0
	ChangeEvent * change = new ChangeAttributeChanged(this, "children", children, mChildren);
#endif

	// make the changes

	vector<Node *> added;
	vector<Node *> removed;

	set_difference(after.begin(), after.end(), 
		       before.begin(), before.end(), back_inserter(added));

	set_difference(before.begin(), before.end(), 
		       after.begin(), after.end(), back_inserter(removed));

	for_each(added.begin(), added.end(), mem_fun(&Object::ref));
	for_each(removed.begin(), removed.end(), mem_fun(&Object::unref));

	mChildren = children;

	// changes made

	log_command(command);
#if 0
	log_change(change);
#endif

#if 0
	change->unref();
#endif
	command->unref();

	end_transaction();
    }

    ostream &
    operator << (ostream & ost, const Node & node) {
	static bool initialised = false;
	static int i = 0;

	if(!initialised) {
	    i = ios::xalloc();
	    initialised = true;
	}

	long & indent_depth = ost.iword(i);

	fill_n(ostream_iterator<const char *>(ost), indent_depth * 4, " ");

	ost << "<" << node.get_name();

	const list<Property *> & properties = node.mProperties;

	transform(properties.begin(), properties.end(), ostream_iterator2<Property>(ost, " "),
		  Dereferencer<Property>());

	const list<Node *> & children = node.mChildren;

	if(children.empty()) {
	    ost << "/>" << endl;
	}
	else {
	    ost << ">" << endl;

	    ++indent_depth;

	    transform(children.begin(), children.end(), ostream_iterator<Node>(ost),
		      Dereferencer<Node>());

	    --indent_depth;

	    fill_n(ostream_iterator<const char *>(ost), indent_depth * 4, " ");

	    ost << "</" << node.get_name() << ">" << endl;
	}

	return ost;
    }

} // namespace XML
