#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "sax_parser.h"

#include "foundation/macros.h"

#include "sax_handler.h"

#include <libxml/tree.h>

#include <cstdarg>
using namespace std;

namespace XML {

    namespace {

	template<void (SAXParser::*method)(const string &)>
	void
	var_args_handler(void * context, const char * msg, ...) {
	    SAXParser * parser = static_cast<SAXParser *>(context);

	    va_list arg;
	    const int BUFFER_SIZE = 1024;
	    char buffer[BUFFER_SIZE];

	    va_start(arg, msg);
	    vsnprintf(buffer, BUFFER_SIZE, msg, arg);
	    va_end(arg);

	    (parser->*method)(buffer);
	}

	template<void (SAXParser::*method)()>
	void
	void_handler(void * context) {
	    SAXParser * parser = static_cast<SAXParser *>(context);

	    (parser->*method)();
	}

	template<void (SAXParser::*method)(const string &)>
	void
	string_handler(void * context, const xmlChar * name) {
	    SAXParser * parser = static_cast<SAXParser *>(context);

	    (parser->*method)(string(reinterpret_cast<const char *>(name)));
	}

	template<void (SAXParser::*method)(const string &)>
	void
	data_handler(void * context, const xmlChar * data, int length) {
	    SAXParser * parser = static_cast<SAXParser *>(context);

	    (parser->*method)(string(reinterpret_cast<const char *>(data), length));
	}

	// special cases that we can't generalise

	xmlEntityPtr
	get_entity(void * context, const xmlChar * name) {
	    return xmlGetPredefinedEntity(name);
	}

	template<void (SAXParser::*method)(const string &, const Properties &)>
	void
	_start_element(void * context, const xmlChar * name,
		      const xmlChar ** attributes) {
	    SAXParser * parser = static_cast<SAXParser *>(context);

	    Properties properties;

	    if(attributes) {
		for(; *attributes; ++attributes) {
		    string name = *reinterpret_cast<const char **>(attributes);
		    string value = *reinterpret_cast<const char **>(++attributes);

#ifdef DEBUG
		    ASSERT(properties.insert(make_pair(name, value)).second);
#else
		    properties[name] = value;
#endif
		}
	    }

	    (parser->*method)(string(reinterpret_cast<const char *>(name)), properties);
	}

	template<void (SAXParser::*method)(const string &)>
	void
	_end_element(void * context, const xmlChar * name) {
	    SAXParser * parser = static_cast<SAXParser *>(context);

	    (parser->*method)(string(reinterpret_cast<const char *>(name)));
	}

    } // namespace {anonymous}

    //
    //
    //

    SAXParser::SAXParser(SAXHandler * root_handler) : Object(),
	mPass(0),
        mTagDepth(0) {
	ASSERT(root_handler);

	// NOTE that we push the root_handler onto the stack with a
	// depth of -1 so that it never gets taken off the stack.

	mHandlerStack.push(make_pair(root_handler, -1));
	root_handler->ref();
    }

    SAXParser::~SAXParser() {
        if(mHandlerStack.size() != 1) {
            ERROR("The handler stack should be of size 1 when the parser is finished");
        }

        while(!mHandlerStack.empty()) {
            mHandlerStack.top().first->unref();
            mHandlerStack.pop();
        }
    }

    void
    SAXParser::parse(istream & is, int passes) {
	vector<char> buffer;

	ios::fmtflags old_flags = is.flags();
	is.flags(old_flags & ~ios::skipws);

	try {
	    copy(istream_iterator<char>(is), istream_iterator<char>(), back_inserter(buffer));
	}
	catch(...) {
	    is.flags(old_flags);
	    throw;
	}

	is.flags(old_flags);	// restore the flags

	//
	//
	//

	xmlSAXHandler legacy_handler = {
	    0
	};

	legacy_handler.warning = &var_args_handler<&SAXParser::warning>;
	legacy_handler.error = &var_args_handler<&SAXParser::error>;
	legacy_handler.fatalError = &var_args_handler<&SAXParser::fatal>;

	legacy_handler.getEntity = &get_entity;

	legacy_handler.comment = &string_handler<&SAXParser::comment>;
	legacy_handler.startDocument = &void_handler<&SAXParser::start_document>;
	legacy_handler.endDocument = &void_handler<&SAXParser::end_document>;
	legacy_handler.startElement = &_start_element<&SAXParser::start_element>;
	legacy_handler.endElement = &_end_element<&SAXParser::end_element>;
	legacy_handler.characters = &data_handler<&SAXParser::characters>;
	legacy_handler.cdataBlock = &data_handler<&SAXParser::cdata_block>;

	mPass = 0;
	mTagDepth = 0;

	while(passes-- > 0) {
	    ++mPass;
	    xmlSAXUserParseMemory(&legacy_handler,
				  static_cast<void *>(this),
				  &buffer.front(), buffer.size());
	}

	mPass = 0;
	ASSERT(mTagDepth == 0);
    }

    void
    SAXParser::use(SAXHandler * handler) {
	ASSERT(handler);

	TRACE(9, "Pushing: " << static_cast<void *>(handler) << " on at depth: " << mTagDepth);

	mHandlerStack.push(make_pair(handler, mTagDepth));
	handler->ref();

	handler->start_document(this);
    }

    void
    SAXParser::push() {
        mProduceStack.push(vector<Object *>());
    }

    void
    SAXParser::pop() {
        ASSERT(!mProduceStack.empty());
        mProduceStack.pop();
    }

    //

    void
    SAXParser::warning(const string & msg) {
	ASSERT(!mProduceStack.empty());

	SAXHandler * handler = mHandlerStack.top().first;

	handler->warning(this, msg);
    }

    void
    SAXParser::error(const string & msg) {
	ASSERT(!mProduceStack.empty());

	SAXHandler * handler = mHandlerStack.top().first;

	handler->error(this, msg);
    }

    void
    SAXParser::fatal(const string & msg) {
	ASSERT(!mHandlerStack.empty());

	SAXHandler * handler = mHandlerStack.top().first;

	handler->fatal(this, msg);
    }

    //

    void
    SAXParser::comment(const string & comment) {
	ASSERT(!mHandlerStack.empty());

	SAXHandler * handler = mHandlerStack.top().first;

	handler->comment(this, comment);
    }

    void
    SAXParser::start_document() {
	ASSERT(!mHandlerStack.empty());

	SAXHandler * handler = mHandlerStack.top().first;

	handler->start_document(this);
    }

    void
    SAXParser::end_document() {
	ASSERT(!mHandlerStack.empty());

	SAXHandler * handler = mHandlerStack.top().first;

	handler->end_document(this);
    }

    void
    SAXParser::start_element(const string & name, const Properties & properties) {
	ASSERT(!mHandlerStack.empty());

	SAXHandler * handler = mHandlerStack.top().first;

	TRACE(10, static_cast<void *>(handler) << " is handling: <" << name << ">");

	handler->start_element(this, name, properties);

	++mTagDepth;
    }

    void
    SAXParser::end_element(const string & name) {
	--mTagDepth;

	ASSERT(!mHandlerStack.empty());

	if(mTagDepth == mHandlerStack.top().second) {
	    SAXHandler * handler = mHandlerStack.top().first;

	    TRACE(9, "Popping: " << static_cast<void *>(handler) << " on at depth: " << mTagDepth);

	    handler->end_document(this);

	    handler->unref();
	    mHandlerStack.pop();
	}

	SAXHandler * handler = mHandlerStack.top().first;

	TRACE(10, static_cast<void *>(handler) << " is handling: </" << name << ">");

	handler->end_element(this, name);
    }

    void
    SAXParser::characters(const string & data) {
	ASSERT(!mHandlerStack.empty());

	SAXHandler * handler = mHandlerStack.top().first;

	handler->characters(this, data);
    }

    void
    SAXParser::reference(const string & name) {
	ASSERT(!mHandlerStack.empty());

	SAXHandler * handler = mHandlerStack.top().first;

	handler->reference(this, name);
    }

    void
    SAXParser::cdata_block(const string & data) {
	ASSERT(!mHandlerStack.empty());

	SAXHandler * handler = mHandlerStack.top().first;

	handler->cdata_block(this, data);
    }

} // namespace XML
